unit CompileDlg;

// 13-NOV-99 ml: structural rework
// 03-JUL-99 ml: implementation so far
//----------------------------------------------------------------------------------------------------------------------
// This unit contains a dialog which looks very much like the Delphi compiler progress dialog
// and has been designed to do also the same as its counterpart in the IDE.
// There are, though, two additonal tasks to be done by this dialog: One is the generation of
// lexer and/or parser and the second is to compile an arbitrary Delphi project. The generation
// part is done by creating the lexer and parser generator classes and let them do their work.
// The compilation (always as last step in this process chain) calls DCC32 and redirects
// its output to DCG. This gives us the opportunity to provide the user with
// exactly the same error messages, hints and warnings as (s)he is used from the IDE.

interface

uses
  Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, Buttons, ExtCtrls,
  LexerGenerator, ParserGenerator, Project, DCGTools;

type
  TCompileDialog = class(TForm)
    CloseButton: TButton;
    Bevel1: TBevel;
    Bevel2: TBevel;
    Label1: TLabel;
    ProjectLabel: TLabel;
    Bevel3: TBevel;
    Bevel4: TBevel;
    Bevel5: TBevel;
    Label6: TLabel;
    Label3: TLabel;
    SourceLineLabel: TLabel;
    TotalCountLabel: TLabel;
    GenerateStateLabel2: TLabel;
    Bevel7: TBevel;
    Label5: TLabel;
    TemplateLineLabel: TLabel;
    Bevel6: TBevel;
    Bevel8: TBevel;
    Bevel9: TBevel;
    Label2: TLabel;
    Label4: TLabel;
    DelphiSourceLineLabel: TLabel;
    DelphiTotalCountLabel: TLabel;
    CompileStateLabel1: TLabel;
    CompileStateLabel2: TLabel;
    Bevel10: TBevel;
    Bevel11: TBevel;
    Bevel12: TBevel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    HintCountLabel: TLabel;
    WarningCountLabel: TLabel;
    ErrorCountLabel: TLabel;
    GenerateStateLabel1: TLabel;
    procedure CloseButtonClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    FLexerGenerator: TLexerGenerator;
    FParserGenerator: TParserGenerator;
    FTotalCount: Cardinal;
    FProject: TProjectForm;
    procedure OnMessage(Sender: TObject; Info: TMessageInfo);
    procedure OnProcessSourceLine(Sender: TObject; const Line: String; Index: Cardinal);
    procedure OnProcessTemplateLine(Sender: TObject; const Line: String; Index: Cardinal);
    procedure DoCompileDelphiProject(Run: Boolean);
    function MessageInfo(Code: TMessageType; const Msg, Source: String; Line, Position: Cardinal): TMessageInfo; overload;
    function MessageInfo(const Msg: String): TMessageInfo; overload;
    // the third overloaded version could also be implemented using default parameter in the first
    // version but this is discouraged as Delphi has problems with overloaded routines using  default parameters
    function MessageInfo(Code: TMessageType; const Msg: String): TMessageInfo; overload;
    procedure UpdateCountLabels;
  public
    procedure DoGenerate(Project: TProjectForm);

    property LexerGenerator: TLexerGenerator read FLexerGenerator;
    property ParserGenerator: TParserGenerator read FParserGenerator;
  end;

var
  CompileDialog: TCompileDialog;

//----------------------------------------------------------------------------------------------------------------------

implementation

uses
  DCGStrings, DCGData, FileCtrl;

{$R *.DFM}

//----------------------------------------------------------------------------------------------------------------------

procedure TCompileDialog.DoGenerate(Project: TProjectForm);

var
  LexerSourceTemp,
  LexerCodeTemp,
  ParserSourceTemp,
  ParserCodeTemp: TStringList;
  BasePath: String;

begin
  FProject := Project;
  FProject.ClearMessages;
  CloseButton.Caption := 'Cancel';
  // temporary string lists to hold the lines if the corresponding edit is not loaded
  // in the editor
  LexerSourceTemp := nil;  
  LexerCodeTemp := nil;
  ParserSourceTemp := nil;
  ParserCodeTemp := nil;
  try
    try
      with Project, ProjectSettings do
      begin
        try
          BasePath := ExtractFilePath(WorkingFile);
          // initialize dialog
          ProjectLabel.Caption := MinimizeName(WorkingFile, Self.Canvas, ProjectLabel.Width);
          if GenerateLexer or GenerateParser then
          begin
            GenerateStateLabel1.Font.Style := [];
            GenerateStateLabel1.Caption := 'generating:';
            GenerateStateLabel2.visible := True;
            GenerateStateLabel2.Font.Style := [];
            GenerateStateLabel2.Caption := '';
          end
          else
          begin
            GenerateStateLabel1.Font.Style := [];
            GenerateStateLabel1.Caption := 'nothing to generate';
            GenerateStateLabel2.visible := False;
          end;

          if CompileDelphiProject then
          begin
            CompileStateLabel2.Font.Style := [];
            CompileStateLabel2.Caption := 'compiling:';
            CompileStateLabel2.visible := True;
            CompileStateLabel2.Font.Style := [];
            CompileStateLabel2.Caption := '';
          end
          else
          begin
            CompileStateLabel2.Font.Style := [];
            CompileStateLabel2.Caption := 'nothing to compile';
            CompileStateLabel2.visible := False;
          end;

          SourceLineLabel.Caption := '0';
          TemplateLineLabel.Caption := '0';
          TotalCountLabel.Caption := '0';
          DelphiSourceLineLabel.Caption := '0';
          DelphiTotalCountLabel.Caption := '0';

          // start generation
          if GenerateLexer then
          begin
            FProject.AddMessage(MessageInfo(mtLog, 'Generating lexical analyzer...'));
            FTotalCount := 0;
            LexerTargetEdit.Lines.Clear;

            // generate default target file name
            if LexerTarget = '' then
              LexerTarget := ChangeFileExt(LexerSource, '.pas');

            GenerateStateLabel2.Caption := 'lexer (' + ExtractFileName(LexerTarget) + ')';

            if FLexerGenerator = nil then FLexerGenerator := TLexerGenerator.Create;
            LexerSourceTemp := TStringList.Create;
            FLexerGenerator.Source := LexerSourceTemp;
            if LexerSourceEdit = nil then
            begin
              if FileExists(BasePath + LexerTemplate) then LexerSourceTemp.LoadFromFile(BasePath + LexerSource);
            end
            else LexerSourceEdit.Lines.AssignTo(FLexerGenerator.Source);
            if FLexerGenerator.Source.Count = 0 then FProject.AddMessage(MessageInfo(mtWarning, SLexerWarning));

            LexerCodeTemp := TStringList.Create;
            FLexerGenerator.Template := LexerCodeTemp;
            if LexerTemplateEdit = nil then
            begin
              if FileExists(BasePath + LexerTemplate) then LexerCodeTemp.LoadFromFile(BasePath + LexerTemplate);
            end
            else LexerTemplateEdit.Lines.AssignTo(FLexerGenerator.Template);
            if FLexerGenerator.Template.Count = 0 then FProject.AddMessage(MessageInfo(mtWarning, SLexerTemplateWarning));
            FLexerGenerator.OnSourceLine := OnProcessSourceLine;
            FLexerGenerator.OnTemplateLine := OnProcessTemplateLine;
            FLexerGenerator.OnMessage := OnMessage;
            Application.ProcessMessages;
            FLexergenerator.Execute;
            LexerTargetEdit.Lines.AddStrings(FLexergenerator.Target);
            NeedSave := True;
          end;

          if (LastErrorCount = 0) and GenerateParser then
          begin
            FProject.AddMessage(MessageInfo(mtLog, 'Generating parser...'));
            FTotalCount := 0;
            ParserTargetEdit.Lines.Clear;

            // generate default target file name
            if ParserTarget = '' then
              ParserTarget := ChangeFileExt(ParserSource, '.pas');

            GenerateStateLabel2.Caption := 'parser (' + ExtractFileName(ParserTarget) + ')';

            if FParserGenerator = nil then FParserGenerator := TParserGenerator.Create;
            ParserSourceTemp := TStringList.Create;
            FParserGenerator.Source := ParserSourceTemp;
            if ParserSourceEdit = nil then
            begin
              if FileExists(BasePath + ParserSource) then ParserSourceTemp.LoadFromFile(BasePath + ParserSource);
            end
            else ParserSourceEdit.Lines.AssignTo(FParserGenerator.Source);
            if FParserGenerator.Source.Count = 0 then FProject.AddMessage(MessageInfo(mtWarning, SParserWarning));

            ParserCodeTemp := TStringList.Create;
            FParserGenerator.Template := ParserCodeTemp;
            if ParserTemplateEdit = nil then
            begin
              if FileExists(BasePath + ParserTemplate) then ParserCodeTemp.LoadFromFile(BasePath + ParserTemplate);
            end
            else ParserTemplateEdit.Lines.AssignTo(FParserGenerator.Template);
            if FParserGenerator.Template.Count = 0 then FProject.AddMessage(MessageInfo(mtWarning, SParserTemplateWarning));

            FParserGenerator.Report := False;
            FParserGenerator.OnSourceLine := OnProcessSourceLine;
            FParserGenerator.OnTemplateLine := OnProcessTemplateLine;
            FParserGenerator.OnMessage := OnMessage;
            Application.ProcessMessages;
            FParserGenerator.Execute;
            ParserTargetEdit.Lines.AddStrings(FParserGenerator.Target);
            NeedSave := True;
          end;

        finally
          if GenerateLexer or GenerateParser then
          begin
            GenerateStateLabel1.Font.Style := [fsBold];
            GenerateStateLabel1.Caption := 'Done:';
            GenerateStateLabel2.Font.Style := [fsBold];
            if LastErrorCount > 0 then GenerateStateLabel2.Caption := 'There are errors.'
                                  else
              if LastWarningCount > 0 then GenerateStateLabel2.Caption := 'There are warnings.'
                                    else
                if LastHintCount > 0 then GenerateStateLabel2.Caption := 'There are hints.'
                                    else GenerateStateLabel2.Caption := 'Generated.';
          end;
        end;

        if (LastErrorCount = 0) and CompileDelphiProject then
        begin
          try
            DoCompileDelphiProject(False);
          finally
            CompileStateLabel1.Font.Style := [fsBold];
            CompileStateLabel1.Caption := 'Done:';
            CompileStateLabel2.Font.Style := [fsBold];
            if LastErrorCount > 0 then CompileStateLabel2.Caption := 'There are errors.'
                                  else
              if LastWarningCount > 0 then CompileStateLabel2.Caption := 'There are warnings.'
                                    else
                if LastHintCount > 0 then CompileStateLabel2.Caption := 'There are hints.'
                                     else CompileStateLabel2.Caption := 'Compiled.';
          end;
        end;
      end;
    except
      // catch all exceptions to allow reenabling task windows in project form
      FProject.AddMessage(MessageInfo(mtError, 'Internal error! Contact vendor..'));
    end;
  finally
    LexerSourceTemp.Free;
    LexerCodeTemp.Free;
    ParserSourceTemp.Free;
    ParserCodeTemp.Free;
    CloseButton.Caption := 'OK';
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCompileDialog.CloseButtonClick(Sender: TObject);

begin
  FProject.GeneratingDone;
  Close;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCompileDialog.OnProcessSourceLine(Sender: TObject; const Line: String; Index: Cardinal);

begin
  SourceLineLabel.Caption := IntToStr(Index);
  Inc(FTotalCount);
  TotalCountLabel.Caption := IntToStr(FTotalCount);
  if (Index mod 20) = 0 then Update;
end;

//-----------------------------------------------------------------------------

procedure TCompileDialog.OnProcessTemplateLine(Sender: TObject; const Line: String; Index: Cardinal);

begin
  TemplateLineLabel.Caption := IntToStr(Index);
  Inc(FTotalCount);
  TotalCountLabel.Caption := IntToStr(FTotalCount);
  if (Index mod 20) = 0 then Update;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCompileDialog.FormDestroy(Sender: TObject);

begin
  FLexerGenerator.Free;
  FParserGenerator.Free;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCompileDialog.DoCompileDelphiProject(Run: Boolean);

// compiles the given Delphi project using the command line compiler dcc32
// The execution of dcc is hidden and its output redirected to DCG using an
// anonymous pipe. Most of the work here is to parse the result strings received
// from dcc32.
// Note: The Run paramater is not used currently, but will later determine whether
//       to run the just compiled project or not.

var
  SA: TSecurityAttributes;
  SI: TStartupInfo;
  PI: TProcessInformation;
  StdOutPipeRead,
  StdOutPipeWrite: THandle;
  WasOK: Boolean;
  Buffer: array[0..255] of Char;
  I, J,
  BytesRead: Cardinal;
  FirstLine: Boolean;
  DCCPath,
  CommandLine,
  WorkDir,
  Item,
  Message,
  Line: String;
  TotalCount,
  OP, CP: Cardinal;
  FileList: TStringList;
  Index: Integer;
  MI: TMessageInfo;

begin
  DCCPath := ReplaceMacroByString(SDelphiPathMacro, GlobalSettings.DCC32Path, DelphiRoot);
  if not FileExists(DCCPath) then
  begin
    FProject.AddMessage(MessageInfo(mtFatalError, SDCCNotFound));
    UpdateCountLabels;
  end
  else
  begin
    Application.ProcessMessages;
    with SA do
    begin
      nLength := SizeOf(SA);
      bInheritHandle := True;
      lpSecurityDescriptor := nil;
    end;
    // create pipe for standard output redirection
    CreatePipe(StdOutPipeRead,  // read handle
               StdOutPipeWrite, // write handle
               @SA,             // security attributes
               0                // number of bytes reserved for pipe - 0 default
               );
    try
      // Make child process use StdOutPipeWrite as standard out,
      // and make sure it does not show on screen.
      with SI do
      begin
        FillChar(SI, SizeOf(SI), 0);
        cb := SizeOf(SI);
        dwFlags := STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
        wShowWindow := SW_HIDE;
        hStdInput := GetStdHandle(STD_INPUT_HANDLE); // don't redirect std input
        hStdOutput := StdOutPipeWrite;
        hStdError := StdOutPipeWrite;
      end;

      // launch the command line compiler
      with FProject, ProjectSettings do
      begin
        // Note: The project config file (ProjectName.cfg) must already been written
        //       and must contain search pathes to all files needed. The default Lib path of Delphi
        //       is used, though.
        WorkDir := ExtractFilePath(WorkingFile);
        CommandLine := DCCPath + ' "' + WorkDir + DelphiProject + '" -h -w -u"' + ExtractFilePath(DCCPath) + '..\lib"';
        WorkDir := ExtractFilePath(WorkDir + DelphiProject);
      end;
      WasOK := CreateProcess(nil, PChar(CommandLine), nil, nil, True, 0, nil, PChar(WorkDir), SI, PI);

      // Now that the handle has been inherited, close it to be safe.
      // We don't want to read or write to it accidentally.
      CloseHandle(StdOutPipeWrite);
      // if process could be created then handle its output
      if not WasOK then
      begin
        FProject.AddMessage(MessageInfo(mtFatalError, SysErrorMessage(GetLastError)));
        UpdateCountLabels;
      end
      else
      try
        // get all output until dcc finishes
        Line := '';
        FirstLine := True;
        FileList := TStringList.Create;
        FileList.Duplicates := dupIgnore;
        FileList.Sorted := True;
        repeat
          // read block of characters (might contain carriage returns and line feeds)
          WasOK := ReadFile(StdOutPipeRead, Buffer, 255, BytesRead, nil);

          // has anything been read?
          if BytesRead > 0 then
          begin
            // finish buffer to PChar
            Buffer[BytesRead] := #0;
            // combine the buffer with the rest of the last run
            Line := Line + Buffer;
            Item := '';
            // parse entire line
            for I := 1 to Length(Line) do
            begin
              // a line feed doesn't mean anything to us
              if Line[I] = LF then Continue
                                else
                // a line has been completed, update display
                if Line[I] = CR then
                begin
                  // prepare message variables
                  Message := '';
                  MI.Line := 0;
                  MI.Position := 0;
                  MI.Source := '';

                  // ignore copyright information
                  if FirstLine then FirstLine := False
                               else
                  begin
                    // one item has been read, now split the parts,
                    // find line number (which is enclosed in parenthesis)
                    OP := Pos('(', Item);
                    // normal line information?
                    if OP > 0 then
                    begin
                      CP := Pos(')', Item);

                      // extract filename
                      MI.Source := Copy(Item, 1, OP - 1);
                      // files in the working path aren't listed with path information
                      // we need to add it here therefor
                      if Pos('\', MI.Source) = 0 then MI.Source := WorkDir + MI.Source;
                      CompileStateLabel2.Caption := ExtractFileName(MI.Source);

                      // add this filename to our list (if not already there) and determine its index
                      if not FileList.Find(CompileStateLabel2.Caption, Index) then Index := FileList.Add(CompileStateLabel2.Caption);

                      // extract line number
                      DelphiSourceLineLabel.Caption := Copy(Item, OP + 1, CP - OP - 1);
                      MI.Line := StrToInt(DelphiSourceLineLabel.Caption);
                      FileList.Objects[Index] := Pointer(MI.Line);

                      // update total lines
                      TotalCount := 0;
                      for J := 0 to FileList.Count -1 do Inc(TotalCount, Cardinal(FileList.Objects[J]));
                      DelphiTotalCountLabel.Caption := IntToStr(TotalCount);
                      DelphiTotalCountLabel.Update;  // update display
                      // format rest of line (might be a hint or warning)
                      Message := Copy(Item, CP + 1, Length(Item));
                    end
                    else
                    begin
                      // this isn't a normal line but an error (-> reformat it as message) or
                      // a status message (get info and swallow the line)
                      if Pos('lines', Item) > 0 then
                      begin
                        DelphiTotalCountLabel.Caption := Copy(Item, 1, Pos('lines', Item) - 1);
                        DelphiTotalCountLabel.Update;
                      end
                      else Message := Item;
                    end;
                    Message := Trim(Message);
                    if Message <> '' then
                    begin
                      MI := MessageInfo(Message);
                      FProject.AddMessage(MI);
                      UpdateCountLabels;
                    end;
                  end;
                  Item := '';
                end
                else Item := Item + Line[I];
            end;
            // keep not processed characters until next run
            Line := Item;
          end;
        until not WasOK or (BytesRead = 0);

        FileList.Free;
        // wait for dcc32 to finish (should be already at this point)
        WaitForSingleObject(PI.hProcess, INFINITE);
      finally
        // Close all remaining handles
        CloseHandle(PI.hThread);
        CloseHandle(PI.hProcess);
      end;
    finally
      CloseHandle(StdOutPipeRead);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCompileDialog.OnMessage(Sender: TObject; Info: TMessageInfo);

// event handler for messages coming in from the generators

begin
  // determine filename
  if Info.Source <> '' then
    if Sender = FLexerGenerator then
    begin
      if Info.Source = 'S' then Info.Source := FProject.ProjectSettings.LexerSource
                           else Info.Source := FProject.ProjectSettings.LexerTemplate;
    end
    else
    begin
      if Info.Source = 'S' then Info.Source := FProject.ProjectSettings.ParserSource
                           else Info.Source := FProject.ProjectSettings.ParserTemplate;
    end;
  // add message to listing
  FProject.AddMessage(Info);
  UpdateCountLabels;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCompileDialog.UpdateCountLabels;

begin
  // update labels
  with FProject do
  begin
    ErrorCountLabel.Caption := IntToStr(LastErrorCount);
    HintCountLabel.Caption := IntToStr(LastHintCount);
    WarningCountLabel.Caption := IntToStr(LastWarningCount);
  end;
  WarningCountLabel.Update;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCompileDialog.MessageInfo(Code: TMessageType; const Msg, Source: String;
                                    Line, Position: Cardinal): TMessageInfo;

// formats a TMessageInfoRecord from DCG data

begin
  Result.Code := Code;
  Result.Message := Msg;
  Result.Source := Source;
  Result.Line := Line;
  Result.Position := Position;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCompileDialog.MessageInfo(const Msg: String): TMessageInfo;

// Formats a TMessageInfoRecord from dcc32 output, which requires some parsing.
// Line number and source file have already been extracted.

var
  Colon: Integer;
  S: String;

begin
  S := UpperCase(Msg);
  if Pos('FATAL', S) = 1 then Result.Code := mtFatalError
                           else
    if Pos('ERROR', S) = 1 then Result.Code := mtError
                             else
      if Pos('WARNING', S) = 1 then Result.Code := mtWarning
                               else
        if Pos('HINT', S) = 1 then Result.Code := mtHint
                                 else Result.Code := mtLog;

  Colon := Pos(':', Msg);
  if Colon > 0 then Result.Message := Copy(Msg, Colon + 1, Length(Msg));
end;

//----------------------------------------------------------------------------------------------------------------------

function TCompileDialog.MessageInfo(Code: TMessageType; const Msg: String): TMessageInfo;

// formats a TMessageInfoRecord from internal message production

begin
  Result.Code := Code;
  Result.Message := Msg;
  Result.Source := '';
  Result.Line := 0;
  Result.Position := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

end.
