unit DCGHighlighter;

// this unit contains the implementation of a lexical analyzer
// see documentation for an extensive description

interface

uses Classes, Graphics, Highlighter;

const CR = #13;
      LF = #10;

// user defined start states and tokens

const UnknownAttrs = 2;
      AttrParm = 4;
      CommentAlt = 6;
      Script = 8;


const DOCTYPE = 257;
      FLOATNUMBER = 258;
      IDENTIFIER = 259;
      INTEGERNUMBER = 260;
      KEYWORD = 261;
      MLCOMMENT = 262;
      NAMETOKEN = 263;
      SCRIPTPART = 264;
      STRINGCONSTANT = 265;
      SYMBOL = 266;
      UNKNOWN = 267;
      UNKNOWNEND = 268;
      UNKNOWNPARM = 269;
      UNKNOWNPSEUDO = 270;
      UNKNOWNPSEUDOEND = 271;
      URLSTRING = 272;
      WHITESPACE = 273;



type
  TDCGHighlighter = class(TBaseHighlighter)
  private
    FCurrentToken: Integer;            // type of current token
    FInput: PChar;                     // input string to be tokenized
    FCurrentPos: Cardinal;             // current lexer position in FInput
    FTokenPos: Cardinal;               // current token position in FInput
    FLookahead: Char;                  // next char after current position
    FCurrentState,
    FStartState,
    FLineState: Cardinal;              // lexer states
    FLastChar: Char;                   // last matched char (#0 if no match)
    FPositions: array of Cardinal;     // for each rule the last marked position, zeroed when
                                       // rule has already been considered (allocated once)
    FMatchStack: array of Cardinal;    // stack containing matched rules (growing dynamically)
    FMatchCount: Integer;              // contains current number of matches
    FRejected: Boolean;                // current match rejected?
    FDoStop: Boolean;                  // token or eof found?
    FEOL: Boolean;
    FCanvas: TCanvas;

    // !! adjust attributes !!
    FIdentifierAttributes,
    FNumberAttributes,
    FSpaceAttributes,
    FStringAttributes,
    FCommentAttributes,
    FSymbolAttributes,
    FKeyAttributes,
    FUnknownAttributes,
    FParamAttributes,
    FURLAttributes,
    FDocTypeAttributes,
    FNameAttributes,
    FScriptAttributes: THighlightAttributes;

    procedure RuleToToken(Rule: Integer);
    function GetNextChar: Char;
    function MatchRule(var Rule: Integer): Boolean;
    function DoDefault: Boolean;
    function GetCurrentChar: Char;
    procedure SetAttribute(const Index: Integer; const Value: THighlightAttributes);
    procedure HighlightChange(Sender: TObject);
    procedure AssignAttributes(Attributes: THighlightAttributes);
  protected
    procedure Reset; virtual;
    procedure SetMatch(I: Integer);
    procedure SetState(Value: Cardinal); virtual;
    procedure SetToken(Token: Integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function GetEol: Boolean; override;
    function GetToken: String; override;
    function GetTokenInfo(var S: PChar; var Pos: Integer): Integer; override;
    procedure Next; override;
    procedure Reject;
    procedure SetCanvas(Value: TCanvas); override;
    function GetAttribute(Index: Integer): THighlightAttributes; override;
    function GetAttributeCount: Integer; override;
    function GetIdentChars: TIdentChars; override;
    function GetLanguageName: String; override;
    function GetRange: Pointer; override;
    function IsKeyword: Boolean;
    procedure ResetRange; override;
    procedure SetRange(Value: Pointer); override;
    procedure SetLine(const NewValue: String); override;

    property CurrentChar: Char read GetCurrentChar;
    property NextChar: Char read GetNextChar;
    property Lookahead: Char read FLookahead;
    property State: Cardinal read FCurrentState write SetState; // read to get current lexer state
                                                                // write to set start state for next run
    property TokenKind: Integer read FCurrentToken;
    property TokenPosition: Cardinal read FTokenPos;
  published
    // !! adjust attributes !!
    property IdentifierAttributes: THighlightAttributes index 0 read FIdentifierAttributes write SetAttribute;
    property NumberAttributes: THighlightAttributes index 1 read FNumberAttributes write SetAttribute;
    property SpaceAttributes: THighlightAttributes index 2 read FSpaceAttributes write SetAttribute;
    property StringAttributes: THighlightAttributes index 3 read FStringAttributes write SetAttribute;
    property CommentAttributes: THighlightAttributes index 4 read FCommentAttributes write SetAttribute;
    property SymbolAttributes: THighlightAttributes index 5 read FSymbolAttributes write SetAttribute;
    property KeyAttributes: THighlightAttributes index 6 read FKeyAttributes write SetAttribute;
    property UnknownAttributes: THighlightAttributes index 7 read FUnknownAttributes write SetAttribute;
    property ParamAttributes: THighlightAttributes index 8 read FParamAttributes write SetAttribute;
    property URLAttributes: THighlightAttributes index 9 read FURLAttributes write SetAttribute;
    property DocTypeAttributes: THighlightAttributes index 10 read FDocTypeAttributes write SetAttribute;
    property NameAttributes: THighlightAttributes index 11 read FNameAttributes write SetAttribute;
    property ScriptAttributes: THighlightAttributes index 12 read FScriptAttributes write SetAttribute;
  end;

//--------------------------------------------------------------------------------

implementation

uses SysUtils;

// DFA table


//--------------------------------------------------------------------------------

constructor TDCGHighlighter.Create(AOwner: TComponent);

begin
  inherited;
  // allocate positions array (length is highest possible mark position)
  SetLength(FPositions, MarksHigh[StateCount - 1] + 1);

  DefaultFilter := 'Lexical description (*.l)|*.l|Parser description (*.y)|*.y|DCG code template (*.cod)|*.cod|';

  // !! adjust attributes !!
  FIdentifierAttributes := THighlightAttributes.Create('identifier');
  FIdentifierAttributes.OnChange := HighlightChange;
  FNumberAttributes := THighlightAttributes.Create('number');
  FNumberAttributes.OnChange := HighlightChange;
  FSpaceAttributes := THighlightAttributes.Create('space');
  FSpaceAttributes.OnChange := HighlightChange;
  FStringAttributes := THighlightAttributes.Create('string');
  FStringAttributes.OnChange := HighlightChange;
  FCommentAttributes := THighlightAttributes.Create('comment');
  FCommentAttributes.OnChange := HighlightChange;
  FSymbolAttributes := THighlightAttributes.Create('symbol');
  FSymbolAttributes.OnChange := HighlightChange;
  FKeyAttributes := THighlightAttributes.Create('keyword');
  FKeyAttributes.OnChange := HighlightChange;
  FUnknownAttributes := THighlightAttributes.Create('unknown');
  FUnknownAttributes.OnChange := HighlightChange;
  FParamAttributes := THighlightAttributes.Create('parameter');
  FParamAttributes.OnChange := HighlightChange;
  FURLAttributes := THighlightAttributes.Create('URL');
  FURLAttributes.OnChange := HighlightChange;
  FDocTypeAttributes := THighlightAttributes.Create('doctype');
  FDocTypeAttributes.OnChange := HighlightChange;
  FNameAttributes := THighlightAttributes.Create('name');
  FNameAttributes.OnChange := HighlightChange;
  FScriptAttributes := THighlightAttributes.Create('script');
  FScriptAttributes.OnChange := HighlightChange;
end;

//-----------------------------------------------------------------------------

destructor TDCGHighlighter.Destroy;

begin
  FPositions := nil;
  FMatchStack := nil;

  // !! adjust attributes !!
  FIdentifierAttributes.Free;
  FNumberAttributes.Free;
  FSpaceAttributes.Free;
  FStringAttributes.Free;
  FCommentAttributes.Free;
  FSymbolAttributes.Free;
  FKeyAttributes.Free;
  FUnknownAttributes.Free;
  FParamAttributes.Free;
  FURLAttributes.Free;
  FDocTypeAttributes.Free;
  FNameAttributes.Free;
  FScriptAttributes.Free;

  inherited;
end;

//-----------------------------------------------------------------------------

procedure TDCGHighlighter.SetCanvas(Value: TCanvas);

begin
  FCanvas := Value;
end;

//--------------------------------------------------------------------------------

function TDCGHighlighter.IsKeyword: Boolean;

// checks whether the current token is a keyword,
// if so then the current token type is set accordingly

var Word: String;

begin
  SetString(Word, FInput + FTokenPos, FCurrentPos - FTokenPos);
  FCurrentToken := GetKeywordToken(Word);
  if FCurrentToken > 0 then
  begin
    Result := True;
    FDoStop := True;
  end
  else Result := False;
end;

//-----------------------------------------------------------------------------

function TDCGHighlighter.GetEol: Boolean;

begin
  Result := (FCurrentToken = -1) or FEOL;
end;

//-----------------------------------------------------------------------------

function TDCGHighlighter.GetToken: String;

var Len: Cardinal;

begin
  Len := FCurrentPos - FTokenPos;
  SetString(Result, FInput + FTokenPos, Len);
end;

//-----------------------------------------------------------------------------

procedure TDCGHighlighter.Reset;

begin
  FTokenPos := 0;
  FCurrentPos := 0;
  FCurrentToken := 0;
  FLineState := 1;
  FLastChar := #0;
  FMatchCount := 0;
  FEOL := False;
end;

//-----------------------------------------------------------------------------

procedure TDCGHighlighter.SetToken(Token: Integer);

begin
  FCurrentToken := Token;
  FDoStop := True;
end;

//-----------------------------------------------------------------------------

procedure TDCGHighlighter.RuleToToken(Rule: Integer);

// var S : String;

begin
  // SetString(S, FInput + FTokenPos, FCurrentPos - FTokenPos);
  // actions
  case Rule of
                                              State := Script;
                                              SetToken(SCRIPTPART);
                                            end;
                                              State := Comment;
                                              SetToken(MLCOMMENT);
                                            end;
                                              State := CommentAlt;
                                              SetToken(MLCOMMENT);
                                            end;
                                              State := UnknownAttrs;
                                              SetToken(KEYWORD); // SetToken(UNKNOWN);
                                            end;
                                              State := UnknownAttrs;
                                              SetToken(KEYWORD); // SetToken(UNKNOWNPSEUDO);
                                            end;
                                              State := Normal;
                                              SetToken(KEYWORD); // SetToken(UNKNOWNEND);
                                            end;
                                              State := Normal;
                                              SetToken(KEYWORD); // SetToken(UNKNOWNPSEUDOEND);
                                            end;

                                              SetToken(SYMBOL);
                                              State := AttrParm;
                                            end;
                                              State := Normal;
                                              SetToken(KEYWORD);
                                            end;

                                              SetToken(INTEGERNUMBER);
                                              State := UnknownAttrs;
                                            end;
                                                  SetToken(URLSTRING);
                                                  State := UnknownAttrs;
                                                end;
                                              State := UnknownAttrs;
                                              SetToken(STRINGCONSTANT);
                                            end;
                                              State := UnknownAttrs;
                                              SetToken(NAMETOKEN);
                                            end;

                                              SetToken(MLCOMMENT);
                                              State := Normal;
                                            end;

                                              SetToken(MLCOMMENT);
                                              State := Normal;
                                            end;
                                               State := Normal;
                                               SetToken(SCRIPTPART);
                                             end;

  end;
end;

//-----------------------------------------------------------------------------

function TDCGHighlighter.MatchRule(var Rule : Integer ) : Boolean;

// finds the last match and the corresponding marked position and adjusts
// the matched string accordingly;
// returns:
// - True if a rule has been matched, False otherwise
// - Rule: the number of the matched rule

begin
  FRejected := False;
  while (FMatchCount > 0) and (FPositions[FMatchStack[FMatchCount - 1]] = 0) do Dec(FMatchCount);

  if FMatchCount > 0 then
  begin
    Rule := FMatchStack[FMatchCount - 1];
    FCurrentPos := FTokenPos + FPositions[Rule];
    FPositions[Rule] := 0;
    if (FCurrentPos - FTokenPos) > 0 then FLastChar := FInput[FCurrentPos]
                                     else FLastChar := #0;
    Result := True;
  end
  else
  begin
    // return position to starting point
    FCurrentPos := FTokenPos;
    FLastChar := #0;
    Result := False;
  end
end;

//-----------------------------------------------------------------------------

function TDCGHighlighter.DoDefault : Boolean;

begin
  FRejected := False;
  if NextChar <> #0 then Result := True
                    else
  begin
    FLineState := 1;
    Result := False;
  end;
  if not FEOL then FLastChar := FInput[FCurrentPos];
end;

//-----------------------------------------------------------------------------

procedure TDCGHighlighter.AssignAttributes(Attributes: THighlightAttributes);

begin
 if FCanvas.Brush.Color <> Attributes.Background then FCanvas.Brush.Color:= Attributes.Background;
 if FCanvas.Font.Color <> Attributes.Foreground then FCanvas.Font.Color:= Attributes.Foreground;
 if FCanvas.Font.Style <> Attributes.Style then FCanvas.Font.Style:= Attributes.Style;
end;

//--------------------------------------------------------------------------------

procedure TDCGHighlighter.Next;

// scans for the next token and determines its kind (FCurrentToken and FTokenPos
// are set appropriately)

var I : Integer;
    Len : Cardinal;
    Matched : Boolean;

begin
  FDoStop := False;

  repeat
    // initialize run
    FTokenPos := FCurrentPos;

    if FLastChar <> #0 then
      if FLastChar = LF then FLineState := 1
                        else FLineState := 0;
    FCurrentState := FStartState + FLineState;

    repeat // character scan loop
      // mark positions and matches
      Len := FCurrentPos - FTokenPos;
      for I := MarksLow[FCurrentState] to MarksHigh[FCurrentState] do FPositions[MarkPositionTable[I]] := Len;
      for I := MatchesHigh[FCurrentState] downto MatchesLow[FCurrentState] do SetMatch(MatchTable[I]);

      if TransitionsLow[FCurrentState] > TransitionsHigh[FCurrentState] then Break;  // dead state, do action

      // determine action
      I := TransitionsLow[FCurrentState];
      while (I <= TransitionsHigh[FCurrentState]) and not (FInput[FCurrentPos] in TransitionTable[I].CharClass) do Inc(I);

      if I > TransitionsHigh[FCurrentState] then Break; // no transition on current char in this state

      // get next character
      NextChar;

      // switch to new state
      FCurrentState := TransitionTable[I].NextState;

    until False;

    repeat
      Matched := MatchRule(I);
      if Matched then RuleToToken(I)
                 else Break;
      // FRejected can be set in RuleToToken by calling Reject
    until Matched and not FRejected;

    if not (Matched or DoDefault) then
    begin
      FStartState := Normal;
      FLineState := 1;
      FLastChar := #0;
      SetToken(0);
    end;
  until FDoStop;

  // !! adjust attributes !!
  if Assigned(FCanvas) then
    case FCurrentToken of
      IDENTIFIER       : AssignAttributes(FIdentifierAttributes);
      INTEGERNUMBER,
      FLOATNUMBER      : AssignAttributes(FNumberAttributes);
      WHITESPACE       : AssignAttributes(FSpaceAttributes);
      STRINGCONSTANT   : AssignAttributes(FStringAttributes);
      MLCOMMENT        : AssignAttributes(FCommentAttributes);
      SYMBOL           : AssignAttributes(FSymbolAttributes);
      KEYWORD          : AssignAttributes(FKeyAttributes);
      UNKNOWN,
      UNKNOWNPSEUDO,
      UNKNOWNEND,
      UNKNOWNPSEUDOEND : AssignAttributes(FUnknownAttributes);
      UNKNOWNPARM      : AssignAttributes(FParamAttributes);
      URLSTRING        : AssignAttributes(FURLAttributes);
      DOCTYPE          : AssignAttributes(FDocTypeAttributes);
      NAMETOKEN        : AssignAttributes(FNameAttributes);
      SCRIPTPART       : AssignAttributes(FScriptAttributes);
    end;
end;

//-----------------------------------------------------------------------------

procedure TDCGHighlighter.SetLine(const NewValue: String);

begin
  FInput := PChar(NewValue);
  Reset;
  Next;
end;

//-----------------------------------------------------------------------------

function TDCGHighlighter.GetNextChar: Char;

begin
  if not FEOL and (FInput[FCurrentPos] = #0) then FEOL := True;
  Inc(FCurrentPos);
  if not FEOL then
  begin
    FLookahead := FInput[FCurrentPos + 1];
    Result := FInput[FCurrentPos];
  end
  else Result := #0;
end;

//-----------------------------------------------------------------------------

procedure TDCGHighlighter.SetState(Value: Cardinal);

begin
  FStartState := Value;
end;

//-----------------------------------------------------------------------------

procedure TDCGHighlighter.SetMatch(I: Integer);

begin
  if FMatchCount >= Length(FMatchStack) then SetLength(FMatchStack, FMatchCount + 1);
  FMatchStack[FMatchCount] := I;
  Inc(FMatchCount);
end;

//-----------------------------------------------------------------------------

procedure TDCGHighlighter.Reject;

begin
  FRejected := True;
end;

//-----------------------------------------------------------------------------

function TDCGHighlighter.GetTokenInfo(var S: PChar; var Pos: Integer): Integer;

// returns token string, token position relative to entire input string, and token length in one call

begin
  Pos := FTokenPos;
  if FEOL then
  begin
    S := StrEnd(FInput);
    Result := S - FInput - Pos;
  end
  else
  begin
    S := FInput + FTokenPos;
    Result := FCurrentPos - Cardinal(Pos);
  end;
end;

//-----------------------------------------------------------------------------

function TDCGHighlighter.GetCurrentChar: Char;

begin
  Result := FInput[FCurrentPos];
end;

//-----------------------------------------------------------------------------

procedure TDCGHighlighter.HighlightChange(Sender:TObject);

begin
  WindowList.Invalidate;
end;

//--------------------------------------------------------------------------------

function TDCGHighlighter.GetIdentChars: TIdentChars;

begin
  Result := ['_', '0'..'9', 'a'..'z', 'A'..'Z'];
end;

//--------------------------------------------------------------------------------

function TDCGHighlighter.GetAttributeCount: Integer;

begin
  Result := 7;
end;

//--------------------------------------------------------------------------------

function TDCGHighlighter.GetAttribute(Index: Integer): THighlightAttributes;

begin
  // !! adjust attributes !!
  case Index of
    0  : Result := FIdentifierAttributes;
    1  : Result := FNumberAttributes;
    2  : Result := FSpaceAttributes;
    3  : Result := FStringAttributes;
    4  : Result := FCommentAttributes;
    5  : Result := FSymbolAttributes;
    6  : Result := FKeyAttributes;
    7  : Result := FUnknownAttributes;
    8  : Result := FParamAttributes;
    9  : Result := FURLAttributes;
    10 : Result := FDocTypeAttributes;
    11 : Result := FNameAttributes;
    12 : Result := FScriptAttributes;
    else Result := nil;
  end;
end;

//--------------------------------------------------------------------------------

function TDCGHighlighter.GetLanguageName: String;

begin
  Result := 'DCG';
end;

//--------------------------------------------------------------------------------

procedure TDCGHighlighter.SetAttribute(const Index: Integer; const Value: THighlightAttributes);

begin
  // !! adjust attributes !!
  case Index of
    0  : FIdentifierAttributes.Assign(Value);
    1  : FNumberAttributes.Assign(Value);
    2  : FSpaceAttributes.Assign(Value);
    3  : FStringAttributes.Assign(Value);
    4  : FCommentAttributes.Assign(Value);
    5  : FSymbolAttributes.Assign(Value);
    6  : FKeyAttributes.Assign(Value);
    7  : FUnknownAttributes.Assign(Value);
    8  : FParamAttributes.Assign(Value);
    9  : FURLAttributes.Assign(Value);
    10 : FDocTypeAttributes.Assign(Value);
    11 : FNameAttributes.Assign(Value);
    12 : FScriptAttributes.Assign(Value);
  end;
end;

//--------------------------------------------------------------------------------

function TDCGHighlighter.GetRange: Pointer;

begin
  Result := Pointer(FStartState);
end;

//--------------------------------------------------------------------------------

procedure TDCGHighlighter.SetRange(Value: Pointer);

begin
  FStartState := Integer(Value);
end;

//--------------------------------------------------------------------------------

procedure TDCGHighlighter.ResetRange;

begin
  FStartState := Normal;
  // in case someone will ask for it immediately afterwards
  FCurrentState := Normal;
end;

//--------------------------------------------------------------------------------

end.


