unit DCGData;

// 05-DEC-99 ml: structural rework
// 04-JUL-99 ml: implementation so far
//----------------------------------------------------------------------------------------------------------------------
// DCGData is the data module used in DCG. This unit is responsible for initialization of
// the GlobalSettings structure and initiates loading and storing application data to/from the
// registry (which actually happens in the main form). Furthermore, there is the declaration
// for project depending settings and the source code of a base project form, which implements
// commonly used methods and properties and is the ancestor of the project form. This two-step
// way has been choosen to enable implementation of different project forms with the same
// base functionality.

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, SyntaxEditor,
  EditorKeyCommands, DCGHighlighter, Highlighter, DelphiHighlighter,
  ImgList, SystemImageList, HTMLHighlighter, CPPHighlighter, SQLHighlighter,
  Dialogs;

type
  TGlobalSettings = record
    DCC32Path: String;
    EditorCommands: TKeystrokes;
    AutoSaveEditor,
    AutoSaveDesktop,
    CreateBackups,
    DontWriteSettings: Boolean;
    EditOptions: TSyntaxEditOptions;
    ScrollbarMode: TScrollBarMode;
    BlockIndent,
    UndoLimit,
    TabSize: Cardinal;
    MultilinePageControl,
    ShowFileIcons: Boolean;
    EditorFontName: String;
    EditorFontSize: Integer;
    EditorMarkForeground,
    EditorMarkBackground: TColor;
    MinimizeToTray: Boolean;
    HighlighterList: TList;
  end;

  TProjectSettings = record
    GenerateLexer,
    GenerateParser,
    CompileDelphiProject: Boolean;
    LexerSource,
    LexerTemplate,
    LexerTarget,
    ParserSource,
    ParserTemplate,
    ParserTarget,
    DelphiProject: String;             // file names isolated from corresponding edit
    LexerSourceEdit,
    LexerTemplateEdit,
    LexerTargetEdit,
    ParserSourceEdit,
    ParserTemplateEdit,
    ParserTargetEdit: TSyntaxEdit;     // edit references
  end;
  
  TBaseProjectForm = class(TForm)
  private
    FNeedSave: Boolean;
    FLoading: Integer;
    FWorkingFile: String;
    procedure SetNeedSave(const Value: Boolean);
    procedure SetWorkingFile(const Value: String);
    function GetLoading: Boolean;
  protected
    function GetNewProject: Boolean; virtual; abstract;
  public
    procedure BeginLoading;
    function CanShow: Boolean; virtual;
    procedure CheckFileAges; virtual; abstract;
    procedure EndLoading;

    procedure Execute; virtual; abstract;
    procedure SetStatus(const Text: String); virtual; abstract;
    procedure Save; virtual; abstract;
    function SaveAll: Boolean; virtual; abstract;
    procedure SaveAs; virtual; abstract;
    function SaveProject: Boolean; virtual; abstract;
    function SaveProjectAs: Boolean; virtual; abstract;

    property Loading: Boolean read GetLoading;
    property NeedSave: Boolean read FNeedSave write SetNeedSave;
    property NewProject: Boolean read GetNewProject;
    property WorkingFile: String read FWorkingFile write SetWorkingFile;
  end;

  TApplicationData = class(TDataModule)
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    DelphiHighlighter1: TDelphiHighlighter;
    DCGHighlighter1: TDCGHighlighter;
    HTMLHighlighter1: THTMLHighlighter;
    SystemImages: TSystemImageList;
    CPPHighlighter1: TCPPHighlighter;
    SQLHighlighter1: TSQLHighlighter;
    procedure ApplicationDataCreate(Sender: TObject);
    procedure ApplicationDataDestroy(Sender: TObject);
  end;

procedure ShowDCGError(const Msg: String); overload;
procedure ShowDCGError(const Msg: String; Args: array of const); overload;

var
  ApplicationData: TApplicationData;

  // all settings relevant to the program
  GlobalSettings: TGlobalSettings;
  DelphiRoot: String; // complete path of one of the Delphi installations of the machine

//----------------------------------------------------------------------------------------------------------------------

implementation

uses
  Project, WindowList, Main, DCGStrings, DCGTools;

type
  EDCGException = class(Exception);

{$R *.DFM}

//----------------------------------------------------------------------------------------------------------------------

procedure ShowDCGError(const Msg: String);

begin
  raise EDCGException.Create(Msg);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ShowDCGError(const Msg: String; Args: array of const);

begin
  raise EDCGException.CreateFmt(Msg, Args);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TApplicationData.ApplicationDataCreate(Sender: TObject);

var
  I: Integer;
  List: TStringList;

begin
  Application.Title := SVersion;
  // set default values for global settings and read then user defined values
  FillChar(GlobalSettings, SizeOf(GlobalSettings), 0);
  ReadDelphiRoot;
  with GlobalSettings do
  begin
    EditorCommands := TKeyStrokes.Create(nil);
    EditorCommands.ResetDefaults;
    EditOptions := SyntaxEditor.DefaultOptions - [eoHideSelection] + [eoWantTabs];
    BlockIndent := 2;
    UndoLimit := 32767;
    TabSize := 8;
    MultilinePageControl := True;
    ShowFileIcons := True;
    EditorFontName := 'Courier New';
    EditorFontSize := 8;
    EditorMarkForeground := clWhite;
    EditorMarkBackground := clMaroon;
    MinimizeToTray := False;
    HighlighterList := TList.Create;
    // add all highlighters currently used in the project
    for I := 0 to ComponentCount -1 do
      if Components[I] is TSyntaxHighlighter then HighlighterList.Add(Components[I]);
  end;
  MainForm.ReadPreferences;

  // parse command line
  List := TStringList.Create;
  List.Duplicates := dupIgnore;
  for I := 1 to ParamCount do List.Add(ParamStr(I));

  MainForm.AddProjectFiles(List);
  List.Free;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TApplicationData.ApplicationDataDestroy(Sender: TObject);

begin
  MainForm.WritePreferences;
  with GlobalSettings do
  begin
    EditorCommands.Free;
    HighlighterList.Free;
  end;
end;

//----------------- TBaseProjectForm -----------------------------------------------------------------------------------

procedure TBaseProjectForm.BeginLoading;

begin
  Inc(FLoading);
end;

//----------------------------------------------------------------------------------------------------------------------

function TBaseProjectForm.CanShow: Boolean;

begin
  Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TBaseProjectForm.EndLoading;

begin
  if FLoading > 0 then Dec(FLoading);
end;

//----------------------------------------------------------------------------------------------------------------------

function TBaseProjectForm.GetLoading: Boolean;

begin
  Result := FLoading > 0;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TBaseProjectForm.SetNeedSave(const Value: Boolean);

begin
  if FLoading = 0 then
  begin
    FNeedSave := Value;
    if MainForm.ActiveProject = Self then MainForm.SaveAction.enabled := Value;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TBaseProjectForm.SetWorkingFile(const Value: String);

begin
  if FWorkingFile <> Value then
  begin
    FWorkingFile := Value;
    WindowListForm.UpdateList;
    MainForm.UpdateCaption(PureName(WorkingFile));
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

end.

