unit DCGTools;

// 05-DEC-99 ml: structural rework
// 20-JUN-99 ml: last change
//----------------------------------------------------------------------------------------------------------------------
// This unit collects the basic data types and operations used in the
// Lexer and Parser generators as well as other global types used in DCG.
// Additionally there are some utility functions used everywhere.

interface

uses
  Windows, Classes, EditorKeyCommands;

const
  // symbolic character constants
  BS   = #8;
  TAB  = #9;
  LF   = #10; // line feed
  CR   = #13;
  FF   = #12; // form feed
  CRLF = #13#10;
  NL   = #10; // one char symbol to signal a CRLF or LF

type
  // string and character class pointers
  TCClass = set of Char;

  // sorted integer sets
  TIntegerList = array of Integer;

  // Regular expressions:
  TExprType = (ntMark, ntChar, ntString, ntCClass, ntStar, ntPlus, ntOptional, ntConcat, ntAlternative);

  PRegExpr = ^TRegExpr;
  TRegExpr = record
  case ExprType: TExprType of
    ntMark:
      (Rule, Position: Integer);
    ntChar:
      (C: Char);
    ntString:
      (S: PChar);
    ntCClass:
      (CC: TCClass);
    ntStar,
    ntPlus,
    ntOptional:
      (R: PRegExpr);
    ntConcat,
    ntAlternative:
      (R1, R2: PRegExpr);
  end;

const
  Letters: TCClass = ['A'..'Z', 'a'..'z', '_'];
  Digits: TCClass = ['0'..'9'];
  Alpha: TCClass = ['A'..'Z', 'a'..'z', '_', '0'..'9'];

  EpsEpr: PRegExpr = nil;
  // characters not allowed in Delphi identifiers but probably used in keywords
  InvalidChars = ['#', '$', '&', '''', '(', ')', '*', '+', ',', '', '.', '/',
                  ':', ';', '<', '=', '>', '@', '[', ']', '^', '{', '}', '%', '!',
                  '?', '~', ''];

type
  // a special string list which can deal with case sensitive as well as
  // not case sensitive strings
  TSymbolListDuplicates = (cidIgnore, cidAccept, cidError);

  PSymbolListItem = ^TSymbolListItem;
  TSymbolListItem = record
    FString: String;
    FObject: TObject;
  end;

  // provides inforomation about the last message (log, hint, warning, error, fatal error)
  // a generator has produced
  TMessageType = (mtLog, mtHint, mtWarning, mtError, mtSyntaxError, mtSemanticError, mtFatalError);

  PMessageInfo = ^TMessageInfo;
  TMessageInfo = record
   Code: TMessageType;
   Message,                   // a human readable description of the error
   Source: String;            // source file concerning the message
                              // (might be empty, indicating a message not related to
                              //  a particular file, e.g. 'done' message or fatal errors)
   Line,                      // line number
   Position: Cardinal;        // character index in Line
  end;

  TOnProcessLineEvent = procedure(Sender: TObject; const Line: String; Index: Cardinal) of object;
  TOnMessageEvent = procedure(Sender: TObject; Info: TMessageInfo) of object;

  PSymbolItemList = ^TSymbolItemList;
  TSymbolItemList = array[0..MaxListSize div SizeOf(TSymbolListItem)] of TSymbolListItem;

  TSymbolList = class(TStrings)
  private
    FList: PSymbolItemList;
    FCount: Integer;
    FCapacity: Integer;
    FSorted: Boolean;
    FDuplicates: TSymbolListDuplicates;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    FUpdating: Boolean;
    FCaseSensitive: Boolean;
    FCompareFunction: function(const S1, S2: String): Integer;
    procedure ExchangeItems(Index1, Index2: Integer);
    procedure Grow;
    procedure QuickSort(L, R: Integer);
    procedure InsertItem(Index: Integer; const S: String);
    procedure SetSorted(Value: Boolean);
    procedure SetCaseSensitive(const Value: Boolean);
  protected
    procedure Changed; virtual;
    procedure Changing; virtual;
    function Get(Index: Integer): String; override;
    function GetCapacity: Integer; override;
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure Put(Index: Integer; const S: String); override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
    procedure SetCapacity(NewCapacity: Integer); override;
    procedure SetUpdateState(Updating: Boolean); override;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    function Add(const S: String): Integer; override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Exchange(Index1, Index2: Integer); override;
    function Find(const S: String; var Index: Integer): Boolean; virtual;
    function IndexOf(const S: String): Integer; override;
    procedure Insert(Index: Integer; const S: String); override;
    procedure Sort; virtual;

    property CaseSensitive: Boolean read FCaseSensitive write SetCaseSensitive default True;
    property Duplicates: TSymbolListDuplicates read FDuplicates write FDuplicates;
    property Sorted: Boolean read FSorted write SetSorted;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
  end;

// Tools
procedure AppendSlash(var Folder: String);
procedure EmptyRegistryKey(Key: HKEY);
function PureName(const FileName: String): String;
procedure ReadPreferenceStrings(Key: HKEY; PrefName: String; List: TStrings);
function StringToBoolean(const Str: String): Boolean;
procedure WritePreferenceStrings(Key: HKEY; PrefName: String; List: TStrings);
procedure ReadDelphiRoot;
function ReplaceMacroByString(const Macro, Source, Replacement: String): String;
function ReplaceStringByMacro(const Macro, Source, SubString: String): String;
procedure SaveKeystrokesToRegistry(Key: HKEY; const Commands: TKeyStrokes);
procedure LoadKeystrokesFromRegistry(Key: HKEY; const Commands: TKeyStrokes);
procedure SaveHighlightersToRegistry(Key: HKEY);
procedure LoadHighlightersFromRegistry(Key: HKEY);
function SelectDirectory(const Caption, InitialDir: String;  const Root: WideString; ShowStatus: Boolean;
  out Directory: String): Boolean; overload;
function MaxPoint(P1, P2: TPoint): TPoint;

// Integer sets
procedure ISSingleton(var M: TIntegerList; I: Integer);
procedure ISInclude(var M: TIntegerList; I: Integer);
procedure ISExclude(var M: TIntegerList; I: Integer);
procedure ISUnion(var M, N: TIntegerList);
procedure ISMinus(var M, N: TIntegerList);
procedure ISIntersect(var M, N: TIntegerList);
function ISMember(M: TIntegerList; I: Integer): Boolean;
function ISEqual(M, N: TIntegerList): Boolean;
function ISSubset(M, N: TIntegerList): Boolean;

// Regular expressions
function MakeMarkNode(Rule, Position: Integer): PRegExpr;
function MakeCharNode(C: Char): PRegExpr;
function MakeStringNode(const Str: String): PRegExpr;
function MakeCClassNode(CC: TCClass): PRegExpr;
function MakeStarNode(R: PRegExpr): PRegExpr;
function MakePlusNode(R: PRegExpr): PRegExpr;
function MakeOptionalNode(R: PRegExpr): PRegExpr;
function ExpandMNExpression(R: PRegExpr; M, N: Integer): PRegExpr;
function MakeConcatExpression(R1, R2: PRegExpr): PRegExpr;
function MakeAlternativeExpression(R1, R2: PRegExpr): PRegExpr;

// Unifiers for regular expressions
// The following predicates check whether the specified regular
// expression R is of the denoted type; if the predicate succeeds,
// the other arguments of the predicate are instantiated to the
// corresponding values.
function IsEpsExpr(R: PRegExpr): Boolean;
function IsMarkExpr(R: PRegExpr; var Rule, Position: Integer): Boolean;
function IsCharExpr(R: PRegExpr; var C: Char): Boolean;
function ISStrExpr(R: PRegExpr; var Str: String): Boolean;
function IsCClassExpr(R: PRegExpr; var CC: TCClass): Boolean;
function IsStarExpr(R: PRegExpr; var R1: PRegExpr): Boolean;
function IsPlusExpr(R: PRegExpr; var R1: PRegExpr): Boolean;
function IsOptExpr(R: PRegExpr; var R1: PRegExpr): Boolean;
function IsConcatExpr(R: PRegExpr; var R1, R2: PRegExpr): Boolean;
function IsAltExpr(R: PRegExpr; var R1, R2: PRegExpr): Boolean;

type
  TOrderPredicate = function (I, J: Integer): Boolean of object;
  TSwapProc = procedure (I, J: Integer) of object;
  TTableLookupProc = function(K: Integer): String of object;
  TTableEntryProc = procedure(K: Integer; Symbol: String) of object;

procedure QuickSort(Lo, Hi: Integer; Less: TOrderPredicate; Swap: TSwapProc);
function Key(Symbol: String; TableSize: Integer; Lookup: TTableLookupProc; Entry: TTableEntryProc): Integer;
function DefinedKey(Symbol: String; TableSize: Integer; Lookup: TTableLookupProc): Boolean;
function Cardinality(CC: TCClass): Integer;
function BlankStr(Str: String): String;
function IsInt(Str: String; var I: Integer): Boolean;
function CharStr(C: Char; Reserved: TCClass): String;
function SingleQuoteStr(Str: String): String;
function DoubleQuoteStr(Str: String): String;
function CClassStr(CC: TCClass): String;
function CClassOrCharStr(CC: TCClass): String;
function RegExprStr(R: PRegExpr): String;

//----------------------------------------------------------------------------------------------------------------------

implementation

uses
  Consts, DCGStrings, DCGData, Highlighter, Dialogs, Menus, Registry, SysUtils, ShlObj, ActiveX, Forms;

//----------------- TSymbolList ----------------------------------------------------------------------------------------

constructor TSymbolList.Create;

begin
  inherited;
  FCaseSensitive := True;
  FCompareFunction := CompareStr;
end;

//----------------------------------------------------------------------------------------------------------------------

destructor TSymbolList.Destroy;

begin
  FOnChange := nil;
  FOnChanging := nil;
  inherited Destroy;
  if FCount <> 0 then Finalize(FList[0], FCount);
  FCount := 0;
  SetCapacity(0);
end;

//----------------------------------------------------------------------------------------------------------------------

function TSymbolList.Add(const S: String): Integer;

begin
  if not Sorted then Result := FCount
                else
    if Find(S, Result) then
      case Duplicates of
        cidIgnore:
          Exit;
        cidError:
          Error(SDuplicateString, 0);
      end;
  InsertItem(Result, S);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.Changed;

begin
  if not FUpdating and Assigned(FOnChange) then FOnChange(Self);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.Changing;

begin
  if not FUpdating and Assigned(FOnChanging) then FOnChanging(Self);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.Clear;

begin
  if FCount <> 0 then
  begin
    Changing;
    Finalize(FList[0], FCount);
    FCount := 0;
    SetCapacity(0);
    Changed;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.Delete(Index: Integer);

begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Changing;
  Finalize(FList[Index]);
  Dec(FCount);
  if Index < FCount then System.Move(FList[Index + 1], FList[Index], (FCount - Index) * SizeOf(TSymbolListItem));
  Changed;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.Exchange(Index1, Index2: Integer);

begin
  if (Index1 < 0) or (Index1 >= FCount) then Error(SListIndexError, Index1);
  if (Index2 < 0) or (Index2 >= FCount) then Error(SListIndexError, Index2);
  Changing;
  ExchangeItems(Index1, Index2);
  Changed;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.ExchangeItems(Index1, Index2: Integer);

var
  Temp: Integer;
  Item1,
  Item2: PSymbolListItem;

begin
  Item1 := @FList[Index1];
  Item2 := @FList[Index2];
  Temp := Integer(Item1.FString);
  Integer(Item1.FString):= Integer(Item2.FString);
  Integer(Item2.FString):= Temp;
  Temp := Integer(Item1.FObject);
  Integer(Item1.FObject):= Integer(Item2.FObject);
  Integer(Item2.FObject):= Temp;
end;

//----------------------------------------------------------------------------------------------------------------------

function TSymbolList.Find(const S: String; var Index: Integer): Boolean;

var
  L, H,
  I, C: Integer;

begin
  Result := False;
  L := 0;
  H := FCount - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := FCompareFunction(FList[I].FString, S);
    if C < 0 then L := I + 1
             else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        if Duplicates <> cidAccept then L := I;
      end;
    end;
  end;
  Index := L;
end;

//----------------------------------------------------------------------------------------------------------------------

function TSymbolList.Get(Index: Integer): String;

begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Result := FList[Index].FString;
end;

//----------------------------------------------------------------------------------------------------------------------

function TSymbolList.GetCapacity: Integer;

begin
  Result := FCapacity;
end;

//----------------------------------------------------------------------------------------------------------------------

function TSymbolList.GetCount: Integer;

begin
  Result := FCount;
end;

//----------------------------------------------------------------------------------------------------------------------

function TSymbolList.GetObject(Index: Integer): TObject;

begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Result := FList[Index].FObject;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.Grow;

var
  Delta: Integer;

begin
  if FCapacity > 64 then Delta := FCapacity div 4
                    else
    if FCapacity > 8 then Delta := 16
                     else Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

//----------------------------------------------------------------------------------------------------------------------

function TSymbolList.IndexOf(const S: String): Integer;

begin
  if not Sorted then Result := inherited IndexOf(S)
                else
    if not Find(S, Result) then Result := -1;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.Insert(Index: Integer; const S: String);

begin
  if Sorted then Error(SSortedListError, 0);
  if (Index < 0) or (Index > FCount) then Error(SListIndexError, Index);
  InsertItem(Index, S);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.InsertItem(Index: Integer; const S: String);

begin
  Changing;
  if FCount = FCapacity then Grow;
  if Index < FCount then System.Move(FList[Index], FList[Index + 1], (FCount - Index) * SizeOf(TSymbolListItem));
  with FList[Index] do
  begin
    Pointer(FString):= nil;
    FObject := nil;
    FString := S;
  end;
  Inc(FCount);
  Changed;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.Put(Index: Integer; const S: String);

begin
  if Sorted then Error(SSortedListError, 0);
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Changing;
  FList[Index].FString := S;
  Changed;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.PutObject(Index: Integer; AObject: TObject);

begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Changing;
  FList[Index].FObject := AObject;
  Changed;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.QuickSort(L, R: Integer);

var
  I, J: Integer;
  P: String;

begin
  repeat
    I := L;
    J := R;
    P := FList^[(L + R) shr 1].FString;
    repeat
      while FCompareFunction(FList[I].FString, P) < 0 do Inc(I);
      while FCompareFunction(FList[J].FString, P) > 0 do Dec(J);
      if I <= J then
      begin
        ExchangeItems(I, J);
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J);
    L := I;
  until I >= R;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.SetCapacity(NewCapacity: Integer);

begin
  ReallocMem(FList, NewCapacity * SizeOf(TSymbolListItem));
  FCapacity := NewCapacity;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.SetSorted(Value: Boolean);

begin
  if FSorted <> Value then
  begin
    if Value then Sort;
    FSorted := Value;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.SetUpdateState(Updating: Boolean);

begin
  FUpdating:=Updating;
  if Updating then Changing else Changed;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.SetCaseSensitive(const Value: Boolean);

begin
  if FCaseSensitive <> Value then
  begin
    Clear;
    FCaseSensitive := Value;
    if FCaseSensitive then FCompareFunction := CompareStr
                      else FCompareFunction := CompareText;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSymbolList.Sort;

begin
  if not Sorted and (FCount > 1) then
  begin
    Changing;
    QuickSort(0, FCount - 1);
    Changed;
  end;
end;

//----------------- Tools ----------------------------------------------------------------------------------------------

procedure AppendSlash(var Folder: String);

// appends a slash at the end of Folder if there isn't one already and the
// length of Folder is > 0

begin
  if (Length(Folder) > 0) and (Folder[Length(Folder)] <> '\') then Folder := Folder + '\';
end;

//----------------------------------------------------------------------------------------------------------------------

procedure EmptyRegistryKey(Key: HKEY);

// Deletes all values and sub keys under the given registry key. This key must
// either be one of the predefined keys (e.g. HKEY_CURRENT_USER) or an previously
// opened key.
//
// BE CAREFUL! This procedure can damage your registry seriously! Using a predefined
//             key is in not recommended at all and you should check twice which
//             key you will finally clear.

const
  DefValueSize  = 512;
  DefBufferSize = 8192;

var
  Status: Integer;
  Index: Integer;
  ValueSize,
  BufferSize: Cardinal;
  KeyType: Integer;
  ValueName: String;
  Buffer: Pointer;
  SubKey: HKEY;

begin
  SetLength(ValueName, DefValueSize);
  Buffer := AllocMem(DefBufferSize);
  try
    Index := 0;
    repeat
      ValueSize := DefValueSize;
      BufferSize := DefBufferSize;
      // enumerate data values at current key
      Status := RegEnumValue(Key, Index, PChar(ValueName), ValueSize, nil, @KeyType, Buffer, @BufferSize);
      // delete old value
      if Status = ERROR_SUCCESS then RegDeleteValue(Key, PChar(ValueName));
    until Status <> ERROR_SUCCESS; // Loop until all values found

    // start over, looking for keys now instead of values
    Index:=0;
    repeat
      ValueSize := DefValueSize;
      BufferSize := DefBufferSize;
      Status := RegEnumKeyEx(Key, Index, PChar(ValueName), ValueSize, nil, Buffer, @BufferSize, nil);
      // was a valid key found?
      if Status = ERROR_SUCCESS then
      begin
        // open the key if found
        Status:=RegOpenKey(Key, PChar(ValueName), SubKey);
        if Status = ERROR_SUCCESS then
        begin
          // if that worked, recurse back here
          EmptyRegistryKey(SubKey);
          RegCloseKey(SubKey);
          RegDeleteKey(Key, PChar(ValueName));
        end;
      end;
    until Status <> ERROR_SUCCESS; // loop until key enum fails
  finally
    FreeMem(Buffer);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function PureName(const FileName: String): String;

// peels a pure file name out without path or extension

begin
  Result := ExtractFileName(FileName);
  if Pos('.', Result) > 0 then Delete(Result, Pos('.', Result), Length(Result));
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ReadPreferenceStrings(Key: HKEY; PrefName: String; List: TStrings);

// Reads a string value (given by PrefName) in a string list, considering only
// the types REG_MULTI_SZ, REG_EXPAND_SZ and REG_SZ.
// Note: REG_EXPAND_SZ and REG_SZ return only one string, which will be added as element 0.
//       For REG_MULTI_SZ the entire list will be returned.

var
  Buffer: PChar;
  Head,
  Tail: PChar;
  Size: DWORD;
  ValueType: DWORD;
  Value: String;

begin
  List.Clear;
  // first, determine size of buffer
  if RegQueryValueEx(Key, PChar(PrefName), nil, @ValueType, nil, @Size) = ERROR_SUCCESS then
  begin
    GetMem(Buffer, Size);
    try
      if RegQueryValueEx(Key, PChar(PrefName), nil, @ValueType, Pointer(Buffer), @Size) = ERROR_SUCCESS then
      begin
        case ValueType of
          REG_MULTI_SZ:
            begin
              Head := Buffer;
              while Head^ <> #0 do
              begin
                Tail := Head;
                while Tail^ <> #0 do Inc(Tail);
                SetString(Value, Head, Tail - Head);
                List.Add(Value);
                Head := Tail + 1; // next character after final null
                                  // (can again be a null, indicating the string's end)
              end;
            end;
          REG_SZ,
          REG_EXPAND_SZ:
            List.Add(Buffer);
        end;
      end;
    finally
      FreeMem(Buffer);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure WritePreferenceStrings(Key: HKEY; PrefName: String; List: TStrings);

// Writes the values of the given list in a REG_MULTI_SZ key entry (given by PrefName)

var
  Buffer: PChar;
  Walker: PChar;
  Size: DWORD;
  I, L: Integer;
  Value: String;

begin
  // determine buffer size
  Size := 0;
  for I := 0 to List.Count-1 do Inc(Size, Length(List[I]) + 1); // length of string plus final null
  // end of buffer is marked by a second 0
  Inc(Size);
  // fill buffer
  GetMem(Buffer, Size);
  Walker := Buffer;
  try
    for I := 0 to List.Count - 1 do
    begin
      Value := List[I];
      L := Length(Value);
      System.Move(PChar(Value)^, Walker^, L);
      Inc(Walker, L);
      Walker^ := #0;
      Inc(Walker);
    end;
    Walker^ := #0;
    RegSetValueEx(Key, PChar(PrefName), 0, REG_MULTI_SZ, Pointer(Buffer), Size);
  finally
    FreeMem(Buffer);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function StringToBoolean(const Str: String): Boolean;

begin
  Result:=(CompareText(Str,'True') = 0) or
          (CompareText(Str,'T') = 0) or
          (CompareText(Str,'Yes') = 0) or
          (CompareText(Str,'Y') = 0);
end;

//----------------------------------------------------------------------------------------------------------------------

function ReplaceMacroByString(const Macro, Source, Replacement: String): String;

// searches the Source string for the given Macro and replaces it by Replacement
// if found

var
  MacroPos: Cardinal;

begin
  Result := Source;
  MacroPos := Pos(Macro, Result);
  if MacroPos > 0 then
  begin
    Delete(Result, MacroPos, Length(Macro));
    Insert(Replacement, Result, MacroPos);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function ReplaceStringByMacro(const Macro, Source, SubString: String): String;

// searches the Source string for the given SubString and replaces it by Macro if found

var
  SubPos: Cardinal;

begin
  Result := Source;
  SubPos := Pos(UpperCase(SubString), UpperCase(Result));
  if SubPos > 0 then
  begin
    Delete(Result, SubPos, Length(SubString));
    Insert(Macro, Result, SubPos);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure SaveKeystrokesToRegistry(Key: HKEY; const Commands: TKeyStrokes);

// Saves the given keystrokes as list of keys into the given registry key.
// Each entry is named as the corresponding command ID. The given key must be either
// one of the predefined keys (e.g. HKEY_CURRENT_USER) or a previously opened key.

var
  I,
  NewIndex: Integer;
  SubKey,
  CommandKey: HKEY;
  Disposition: Integer;
  Name: String;
  Value: DWORD;

begin
  if RegCreateKeyEx(Key, 'Commands', 0, nil, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS,
                    nil, SubKey, @Disposition) = ERROR_SUCCESS then
  try
    // clear all command keys since there might be fewer now than before
    EmptyRegistryKey(SubKey);
    for I := 0 to Commands.Count - 1 do
    begin
      Name := EditorCommandToCodeString(Commands[I].Command);
      Value := Commands[I].ShortCut;
      // one command may have several shortcuts, which will be stored as value list
      // in a key named like the command itself
      if RegCreateKeyEx(SubKey, PChar(Name), 0, nil, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS,
                        nil, CommandKey, @Disposition) = ERROR_SUCCESS then
      begin
        // read number of values already there
        RegQueryInfoKey(CommandKey, nil, nil, nil, nil, nil, nil, @NewIndex, nil, nil, nil, nil);
        // use this number to create an unique value name
        Name := IntToStr(NewIndex);
        // set the entry to the given value
        RegSetValueEx(CommandKey, PChar(Name), 0, REG_DWORD, @Value, 4);
        RegCloseKey(CommandKey);
      end;
    end;
  finally
    RegCloseKey(SubKey);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure LoadKeystrokesFromRegistry(Key: HKEY; const Commands: TKeyStrokes);

// Loads all keystrokes from the given registry key into Commands. The given key must be either
// one of the predefined keys (e.g. HKEY_CURRENT_USER) or a previously opened key.

var
  CommandIndex,
  ShortCutIndex: Cardinal;
  SubKey, CommandKey: HKEY;
  CommandName: array[0..64] of Char;
  ValueName: array[0..5] of Char;
  ValueSize,
  BufferSize,
  Value: DWORD;
  CommandID,
  KeyStatus,
  ValueStatus: Integer;

begin
  if RegOpenKeyEx(Key, 'Commands', 0, KEY_READ, SubKey) = ERROR_SUCCESS then
  try
    // read number of subkeys (defined commands)
    RegQueryInfoKey(SubKey, nil, nil, nil, @CommandIndex, nil, nil, nil, nil, nil, nil, nil);
    if CommandIndex > 0 then
    begin
      Commands.Clear;
      CommandIndex := 0;
      repeat
        BufferSize := 64;
        // enumerate keys at commands key
        KeyStatus := RegEnumKeyEx(SubKey, CommandIndex, @CommandName, BufferSize, nil, nil, nil, nil);
        // successfull?
        if KeyStatus = ERROR_SUCCESS then
          // can we open the current command key?
          if RegOpenKeyEx(SubKey, @CommandName, 0, KEY_READ, CommandKey) = ERROR_SUCCESS then
          begin
            // for each command key read its values and create keystroke entries
            ShortCutIndex := 0;
            repeat
              ValueSize := 4;
              BufferSize := 5;
              // enumerate data values at current command key
              ValueStatus := RegEnumValue(CommandKey, ShortCutIndex, @ValueName, BufferSize, nil, nil, @Value, @ValueSize);
              // could we read the value?
              if ValueStatus = ERROR_SUCCESS then
              begin
                // convert command name into its corresponding identifier
                if IdentToEditorCommand(CommandName, CommandID) then
                  with Commands.Add do
                  begin
                    Command := CommandID;
                    try
                      ShortCut := Value;
                    except
                      on EKeyException do
                      begin
                        MessageDlg('The keystroke "' + Menus.ShortCutToText(Value) +
                                   '" for editor command "' + CommandName + '" is already assigned ' +
                                   'to another editor command.'#13#10, Dialogs.mtError,
                                   [mbOK], 0);
                        ShortCut := ecNone;
                      end
                      else raise;
                    end;
                  end;
              end;
              Inc(ShortCutIndex);
            until ValueStatus <> ERROR_SUCCESS; // Loop until all values are processed
            RegCloseKey(CommandKey);
          end;
        Inc(CommandIndex);
      until KeyStatus <> ERROR_SUCCESS; // Loop until all keys are processed
    end;
  finally
    RegCloseKey(SubKey);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure SaveHighlightersToRegistry(Key: HKEY);

// Saves all color and fontstyle values of all registered highlighters into the given registry key.
// Each entry is named as the corresponding highlighter language. The given key must be either
// one of the predefined keys (e.g. HKEY_CURRENT_USER) or a previously opened key.

var I: Integer;
    SubKey: HKEY;
    Disposition: Integer;
    HL: TSyntaxHighlighter;

begin
  if RegCreateKeyEx(Key, 'Highlighter', 0, nil, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS,
                    nil, SubKey, @Disposition) = ERROR_SUCCESS then
  try
    // clear all command keys since they might have changed due to development
    EmptyRegistryKey(SubKey);
    // go through all highlighters registered in GlobalSettings.HighlighterList
    with GlobalSettings do
      for I := 0 to HighlighterList.Count - 1 do
      begin
        HL := HighlighterList[I];
        HL.SaveToRegistry(SubKey, HL.LanguageName);
      end;
  finally
    RegCloseKey(SubKey);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure LoadHighlightersFromRegistry(Key: HKEY);

// Loads all color and fontstyle values of all registered highlighters from the given registry key.
// The given key must be either one of the predefined keys (e.g. HKEY_CURRENT_USER)
// or a previously opened key.

var
  Index: Cardinal;
  SubKey: HKEY;
  Language: array[0..64] of Char;
  BufferSize: DWORD;
  KeyStatus,
  I: Integer;
  HL: TSyntaxHighlighter;

begin
  if RegOpenKeyEx(Key, 'Highlighter', 0, KEY_READ, SubKey) = ERROR_SUCCESS then
  try
    // read number of subkeys (stored highlighters)
    RegQueryInfoKey(SubKey, nil, nil, nil, @Index, nil, nil, nil, nil, nil, nil, nil);
    if Index > 0 then
    begin
      Index := 0;
      repeat
        BufferSize := 64;
        // enumerate keys at highlighter key
        KeyStatus := RegEnumKeyEx(SubKey, Index, @Language, BufferSize, nil, nil, nil, nil);
        // successfull?
        if KeyStatus = ERROR_SUCCESS then
          with GLobalSettings do
          begin
            // find corresponding highlighter(s) and let it (them) load its (their) values
            for I := 0 to HighlighterList.Count - 1 do
            begin
              HL := HighlighterList[I];
              if CompareText(Language, HL.LanguageName) = 0 then HL.LoadFromRegistry(SubKey, Language);
            end;
          end;
        Inc(Index);
      until KeyStatus <> ERROR_SUCCESS; // Loop until all keys are processed
    end;
  finally
    RegCloseKey(SubKey);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function MaxPoint(P1, P2: TPoint): TPoint;

// helper function for editor

begin
  if P1.Y > P2.Y then Result := P1
                 else
    if P1.Y < P2.Y then Result := P2
                   else
      if P1.X > P2.X then Result := P1
                     else Result := P2;
end;

//----------------- Integer sets ---------------------------------------------------------------------------------------

procedure ISSingleton(var M: TIntegerList; I: Integer);

begin
  SetLength(M, 1);
  M[0] := I;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ISInclude(var M: TIntegerList; I: Integer);

var
  L, R, K: Integer;
  Count: Integer;

begin
  if M = nil then ISSingleton(M, I)
             else
  begin
    Count := Length(M);
    // binary search
    L := 0;
    R := Count - 1;
    K := L + (R - L) div 2;
    while (L < R) and (M[K] <> I) do
    begin
      if M[K] < I then L := Succ(K)
                  else R := Pred(K);
      K := L + (R - L) div 2
    end;

    // end of set reached or value not already there?
    if (K = Count) or (M[K] <> I) then
    begin
      // expand set
      SetLength(M, Count + 1);
      // new value to included after the current place
      if M[K] < I then
      begin
        // is the place to put the new value in the last one in the set, then move all
        // following entries one place to the right
        if (K + 1) < Count then Move(M[K + 1], M[K + 2], (Count - K - 1) * SizeOf(Integer));
        // store the new value
        M[K + 1] := I;
      end
      else
      begin
        // new value must be stored before the current value, thus move all values
        // (inlcuding the one of the current place) one place to the right
        Move(M[K], M[K + 1], (Count - K) * SizeOf(Integer));
        // store the new value
        M[K] := I;
      end;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ISExclude(var M: TIntegerList; I: Integer);

var
  L, R, K: Integer;
  Count: Integer;

begin
  if Assigned(M) then
  begin
    Count := Length(M);
    // binary search
    L := 0;
    R := Count - 1;
    K := L + (R - L) div 2;
    while (L < R) and (M[K] <> I) do
    begin
      if M[K] < I then L := Succ(K)
                  else R := Pred(K);
      K := L + (R - L) div 2
    end;

    if (K < Count) and (M[K] = I) then
    begin
      if K < Count - 1 then Move(M[K + 1], M[K], (Count - K - 1) * SizeOf(Integer));
      SetLength(M, Count - 1);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ISUnion(var M, N: TIntegerList);

// merges M and N into M

var
  Union: TIntegerList;
  I, J,
  K, L: Integer;

begin
  // merge sort
  I := 0;
  K := 0;
  L := 0;
  SetLength(Union, Length(M) + Length(N));

  while (K < Length(M)) and (L < Length(N)) do
  begin
    if M[K] < N[L] then
    begin
      Union[I] := M[K];
      Inc(K)
    end
    else
      if N[L] < M[K] then
      begin
        Union[I] := N[L];
        Inc(L)
      end
      else
      begin
        Union[I] := M[K];
        Inc(K);
        Inc(L)
      end;
    Inc(I);
  end;

  for J := K to Length(M) -1 do
  begin
    Union[I] := M[J];
    Inc(I);
  end;

  for J := L to Length(N) - 1 do
  begin
    Union[I] := N[J];
    Inc(I);
  end;

  M := nil;
  M := Copy(Union, 0, I);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ISMinus(var M, N: TIntegerList);

var
  Minus: TIntegerList;
  I, K,
  L: Integer;

begin
  I := 0;
  L := 0;
  SetLength(Minus, Length(M));
  for K := 0 to Length(M) - 1 do
  begin
    while (L < Length(N)) and (N[L] < M[K]) do Inc(L);
    if (L = Length(N)) or (N[L] > M[K]) then
    begin
      Minus[I] := M[K];
      Inc(I);
    end
    else Inc(L);
  end;

  M := nil;
  if I > 0 then M := Copy(Minus, 0, I);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ISIntersect(var M, N: TIntegerList);

var
  Intersect: TIntegerList;
  I, K, L: Integer;

begin
  I := 0;
  L := 0;
  SetLength(Intersect, Length(M));
  for K := 0 to Length(M) - 1 do
  begin
    while (L < Length(N)) and (N[L] < M[K]) do Inc(L);
    if (L < Length(N)) and (N[L] = M[K]) then
    begin
      Intersect[I] := M[K];
      Inc(I);
      Inc(L);
    end;
  end;
  M := nil;
  if I > 0 then M := Copy(Intersect, 0, I);
end;

//----------------------------------------------------------------------------------------------------------------------

function ISMember(M: TIntegerList; I: Integer): Boolean;

var
  L, R, K : Integer;

begin
  // binary search
  L := 0;
  R := Length(M) - 1;
  K := L + (R - L) div 2;
  while (L < R) and (M[K] <> I) do
  begin
    if M[K] < I then L := Succ(K)
                else R := Pred(K);
    K := L + (R - L) div 2;
  end;
  ISMember := (K < Length(M)) and (M[K] = I);
end;

//----------------------------------------------------------------------------------------------------------------------

function ISEqual(M, N: TIntegerList): Boolean;

var
  I: Integer;

begin
  Result := False;
  if Length(M) = Length(N) then
  begin
    for I := 0 to Length(M) - 1 do
      if M[I] <> N[I] then Break;
    if I = Length(M) then Result := True;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function ISSubset(M, N: TIntegerList): Boolean;

// checks whether M is a subset of N

var
  K, L: Integer;

begin
  Result := False;
  if Length(M) <= Length(N) then
  begin
    L := 0;
    for K := 0 to Length(M) - 1 do
    begin
      while (L < Length(N)) and (N[L] < M[K]) do Inc(L);
      if (L = Length(N)) or (N[L] > M[K]) then Break
                                          else Inc(L);
    end;
    if K = Length(M) then Result := True;
  end
end;

//----------------- constructors for regular expressions ---------------------------------------------------------------

function NewExpression(ExpressionType: TExprType; N: Integer): PRegExpr;

// returns new PRegExpr node (N: number of bytes to allocate)

begin
  Result := AllocMem (SizeOf(TExprType) + N);
  Result.ExprType := ExpressionType;
end;

//----------------------------------------------------------------------------------------------------------------------

function MakeMarkNode(Rule, Position: Integer): PRegExpr;

// markers are used to denote endmarkers of rules, as well as other
// special positions in rules, e.g. the position of the lookahead
// operator; they are considered nullable; by convention, we use
// the following Position numbers:
// - 0: endmarker position
// - 1: lookahead operator position

begin
  Result := NewExpression(ntMark, 2 * SizeOf(Integer));
  Result.Rule := Rule;
  Result.Position := Position;
end;

//----------------------------------------------------------------------------------------------------------------------

function MakeCharNode(C: Char): PRegExpr;

begin
  Result := NewExpression(ntChar, SizeOf(Char));
  Result.C := C;
end;

//----------------------------------------------------------------------------------------------------------------------

function MakeStringNode(const Str: String): PRegExpr;

begin
  Result := NewExpression(ntString, SizeOf(PChar));
  Result.S := StrNew(PChar (Str));
end;

//----------------------------------------------------------------------------------------------------------------------

function MakeCClassNode(CC: TCClass): PRegExpr;

begin
  Result := NewExpression(ntCClass, SizeOf(TCClass));
  Result.CC := CC;
end;

//----------------------------------------------------------------------------------------------------------------------

function MakeStarNode(R: PRegExpr): PRegExpr;

var
  Temp: PRegExpr;

begin
  Temp := NewExpression(ntStar, SizeOf(PRegExpr));
  Temp.R := R;

  Result := Temp;
end;

//----------------------------------------------------------------------------------------------------------------------

function MakePlusNode(R: PRegExpr): PRegExpr;

var
  Temp: PRegExpr;

begin
  Temp := NewExpression(ntPlus, SizeOf(PRegExpr));
  Temp.R := R;

  Result := Temp;
end;

//----------------------------------------------------------------------------------------------------------------------

function MakeOptionalNode(R: PRegExpr): PRegExpr;

var
  Temp: PRegExpr;

begin
  Temp := NewExpression(ntOptional, SizeOf(PRegExpr));
  Temp.R := R;

  Result := Temp;
end;

//----------------------------------------------------------------------------------------------------------------------

function ExpandMNExpression(R: PRegExpr; M, N: Integer): PRegExpr;

// constructor expanding expression R{m, N} to the corresponding
// alt expression R^m|...|R^N

var
  ri,
  rmn: PRegExpr;
  I: Integer;

begin
  if (M > N) or (N = 0) then Result := EpsEpr
                        else
  begin
    // construct R^m
    if M = 0 then ri := EpsEpr
             else
    begin
      ri := R;
      for I := 2 to M do ri := MakeConcatExpression (ri, R);
    end;

    // construct R{m,N}
    rmn := ri; // R{m,N} := R^m
    for I := M + 1 to N do
    begin
      if IsEpsExpr(ri) then ri := R
                         else ri := MakeConcatExpression (ri, R);
      rmn := MakeAlternativeExpression (rmn, ri)
      // R{m,N} := R{m,N} | R^I,
      // I=m+1,...,N

    end;
    Result := rmn;
  end
end;

//----------------------------------------------------------------------------------------------------------------------

function MakeConcatExpression(R1, R2: PRegExpr): PRegExpr;

var
  Temp: PRegExpr;

begin
  Temp := NewExpression(ntConcat, 2 * SizeOf(PRegExpr));
  Temp.R1 := R1;
  Temp.R2 := R2;

  Result := Temp;
end;

//----------------------------------------------------------------------------------------------------------------------

function MakeAlternativeExpression(R1, R2: PRegExpr): PRegExpr;

var
  Temp: PRegExpr;

begin
  Temp := NewExpression(ntAlternative, 2 * SizeOf(PRegExpr));
  Temp.R1 := R1;
  Temp.R2 := R2;

  Result := Temp;
end;

//----------------------------------------------------------------------------------------------------------------------

function IsEpsExpr(R: PRegExpr): Boolean;

begin
  Result := R = EpsEpr;
end;

//----------------------------------------------------------------------------------------------------------------------

function IsMarkExpr(R: PRegExpr; var Rule, Position: Integer): Boolean;

begin
  if R = EpsEpr then Result := False
                else
    if R.ExprType = ntMark then
    begin
      Result := True;
      Rule := R.Rule;
      Position := R.Position
    end
    else Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

function IsCharExpr(R: PRegExpr; var C: Char): Boolean;

begin
  if R = EpsEpr then Result := False
                else
    if R.ExprType = ntChar then
    begin
      Result := True;
      C := R.C;
    end
    else Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

function ISStrExpr (R: PRegExpr; var Str: String): Boolean;

begin
  if R = EpsEpr then Result := False
                else
    if R^.ExprType = ntString then
    begin
      Result := True;
      Str := R.S;
    end
    else
      Result := False
end;

//----------------------------------------------------------------------------------------------------------------------

function IsCClassExpr(R: PRegExpr; var CC: TCClass): Boolean;

begin
  if R = EpsEpr then Result := False
                else
    if R.ExprType = ntCClass then
    begin
      Result := True;
      CC := R.CC;
    end
    else Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

function IsStarExpr(R: PRegExpr; var R1: PRegExpr): Boolean;

begin
  if R = EpsEpr then Result := False
                else
    if R.ExprType = ntStar then
    begin
      Result := True;
      R1 := R.R;
    end
    else Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

function IsPlusExpr(R: PRegExpr; var R1: PRegExpr): Boolean;

begin
  if R = EpsEpr then Result := False
                else
    if R.ExprType = ntPlus then
    begin
      Result := True;
      R1 := R.R;
    end
    else Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

function IsOptExpr(R: PRegExpr; var R1: PRegExpr): Boolean;

begin
  if R = EpsEpr then Result := False
                else
    if R.ExprType = ntOptional then
    begin
      Result := True;
      R1 := R.R;
    end
    else Result := False
end;

//----------------------------------------------------------------------------------------------------------------------

function IsConcatExpr(R: PRegExpr; var R1, R2: PRegExpr): Boolean;

begin
  if R = EpsEpr then Result := False
                else
    if R.ExprType = ntConcat then
    begin
      Result := True;
      R1 := R.R1;
      R2 := R.R2;
    end
    else Result := False
end;

//----------------------------------------------------------------------------------------------------------------------

function IsAltExpr(R: PRegExpr; var R1, R2: PRegExpr): Boolean;

begin
  if R = EpsEpr then Result := False
                else
    if R.ExprType = ntAlternative then
    begin
      Result := True;
      R1 := R.R1;
      R2 := R.R2;
    end
    else Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure QuickSort(Lo, Hi: Integer; Less: TOrderPredicate; Swap: TSwapProc);

// General inplace sorting procedure based on the QuickSort algorithm.
// This procedure can be applied to any sequential data structure;
// only the corresponding routines Less which compares, and Swap which
// swaps two elements I, J of the target data structure, must be
// supplied as appropriate for the target data structure.
// - Lo, Hi: the lower and higher indices, indicating the elements to
//   be sorted
// - Less(I, J): should return True if element no. I 'is Less than'
//   element no. J, and False otherwise; any total quasi-ordering may
//   be supplied here (if neither Less(I, J) nor Less(J, I) then elements
//   I and J are assumed to be 'equal').
// - Swap(I, J): should swap the elements with index I and J
// Generic hash table routines (based on quadratic rehashing; hence the
// table size must be a prime number):

  //--------------- local functions -------------------------------------------

  procedure Sort(L, R: Integer);

  var
    I, J, K : Integer;

  begin
    I := L;
    J := R;
    K := (L + R) div 2;
    repeat
      while Less(I, K) do Inc(I);
      while Less(K, J) do Dec(J);
      if I <= J then
      begin
        Swap(I, J);
        if K = I then K := J // pivot element swapped!
                 else
          if K = J then K := I;
        Inc(I);
        Dec(J)
      end;
    until I > J;

    if L < J then Sort(L, J);
    if I < R then Sort(I, R);
  end;

  //--------------- end local functions ---------------------------------------

begin
  if Lo < Hi then Sort(Lo, Hi)
end;

//----------------- Generic has table routines -------------------------------------------------------------------------

function Hash(Str: String; TableSize: Integer): Integer;

// computes a hash Key for Str

var
  I, Key: Integer;

begin
  Key := 0;
  for I := 1 to Length(Str) do  Inc(Key, Ord(Str[I]));
  Result := Key mod TableSize + 1;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure NewPosition(var Position, Incr, Count: Integer; TableSize: Integer);

// computes a New position in the table (quadratic collision strategy)
// - Position: current position (+Inc)
// - Incr: current increment (+2)
// - Count: current number of collisions (+1)
// quadratic collision formula for position of str after N collisions:
// Position(str, N) = (hash(str)+N^2) mod TableSize +1
// note that N^2-(N-1)^2 = 2n-1 <=> N^2 = (N-1)^2 + (2n-1) for N>0,
// I.e. the increment Inc=2n-1 increments by two in each collision 

begin
  Inc(Count);
  Inc(Position, Incr);
  if Position > TableSize then Position := Position mod TableSize + 1;
  Inc(Incr, 2);
end;

//----------------------------------------------------------------------------------------------------------------------

function Key(Symbol: String; TableSize: Integer; Lookup: TTableLookupProc; Entry : TTableEntryProc): Integer;

// returns a hash table key for Symbol; inserts the symbol into the
// table if necessary
// - TableSize is the Symbol table size and must be a fixed prime number
// - Lookup is the table Lookup procedure which should return the string
//   at Key K in the table ('' if Entry is ClearIntSet)
// - Entry is the table Entry procedure which is assumed to store the
//   given Symbol at the given location

var
  Position,
  Incr,
  Count: Integer;

begin
  Position := Hash(Symbol, TableSize);
  Incr := 1;
  Count := 0;
  Result := 0;
  while Count <= TableSize do
    if Lookup(Position) = '' then
    begin
      Entry(Position, Symbol);
      Result := Position;
      Break;
    end
    else
      if Lookup(Position) = Symbol then
      begin
        Result := Position;
        Break;
      end
      else
      begin
        NewPosition(Position, Incr, Count, TableSize);
        Result := Position;
      end;
end;

//----------------------------------------------------------------------------------------------------------------------

function DefinedKey(Symbol: String; TableSize: Integer; Lookup: TTableLookupProc): Boolean;

// checks the table to see if symbol is in the table

var
  Position,
  Incr,
  Count: Integer;

begin
  Position := Hash(Symbol, TableSize);
  Incr := 1;
  Count := 0;
  while Count <= TableSize do
    if Lookup (Position) = '' then
    begin
      Result := False;
      Exit;
    end
    else
      if Lookup (Position) = Symbol then
      begin
        Result := True;
        Exit;
      end
      else NewPosition(Position, Incr, Count, TableSize);
  Result := False;
end;

//----------------- Utility routines -----------------------------------------------------------------------------------

function Cardinality(CC: TCClass): Integer;

// returns the cardinality (number of characters) of a character class

var
  C: Char;
  Count: Integer;

begin
  Count := 0;
  for C := #0 to #255 do
    if C in CC then Inc(Count);
  Result := Count;
end;

//----------------------------------------------------------------------------------------------------------------------

function BlankStr(Str: String): String;

// returns string of same length as str, with all non-whitespace characters
// replaced by blanks

var
  I: Integer;

begin
  for I := 1 to Length(Str) do
    if Str[I] <> TAB then Str[I] := ' ';
  Result := Str;
end;

//----------------------------------------------------------------------------------------------------------------------

function IsInt(Str: String; var I: Integer): Boolean;

// checks whether str represents an integer; if so, returns the
// value of it in I
// Utility functions for list generating routines:

var
  Error: Integer;

begin
  Val(Str, I, Error);
  Result := Error = 0;
end;

//----------------------------------------------------------------------------------------------------------------------

// Utility functions for list generating routines

function CharStr(C: Char; Reserved: TCClass): String;

// returns a print name for character C, using the standard escape
// conventions; Reserved is the class of `Reserved' special characters
// which should be quoted with \ (\ itself is always quoted)

  //--------------- local functions -------------------------------------------

  function OctStr(C: Char): String;

  // return octal string representation of character C *)

  begin
    Result := IntToStr(Ord(C) div 64) + IntToStr((Ord(C) mod 64) div 8) + IntToStr(Ord(C) mod 8);
  end;

  //--------------- end local functions ---------------------------------------

begin
  case C of
    #0 .. #7,
    #11 ,
    #14 .. #31,
    #127 .. #255:
      Result := '\' + OctStr(C);
    BS:
      Result := '\b';
    TAB:
      Result := '\t';
    LF:
      Result := '\n';
    CR:
      Result := '\c';
    FF:
      Result := '\f';
    '\':
      Result := '\\';
  else
    if C in Reserved then Result := '\' + C
                     else Result := C;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function SingleQuoteStr(Str: String): String;

// returns print name of str enclosed in single quotes, using the
// standard escape conventions

var
  I : Integer;
  S : String;

begin
  S := '';
  for I := 1 to Length(Str) do S := S + CharStr(Str[I], ['''']);
  Result := '''' + S + '''';
end;

//----------------------------------------------------------------------------------------------------------------------

function DoubleQuoteStr(Str: String): String;

// returns print name of Str enclosed in double quotes, using the
// standard escape conventions

var
  I: Integer;
  S: String;

begin
  S := '';
  for I := 1 to Length(Str) do S := S + CharStr(Str[I], ['"']);
  Result := '"' + S + '"';
end;

//----------------------------------------------------------------------------------------------------------------------

function CClassStr(CC: TCClass): String;

// returns print name of character class CC, using the standard escape
// conventions; if CC contains more than 128 elements, the complement
// notation (^) is used; if CC is the class of all (non-null) characters
// except newline, the period notation is used

const
  Reserved: TCClass = ['^', '-', ']'];

var
  c1,
  c2: Char;
  Str: String;

begin
  if CC = [#1..#255] - [LF] then Result := '.'
                            else
  begin
    Str := '';
    if Cardinality(CC) > 128 then
    begin
      Str := '^';
      CC := [#0..#255] - CC;
    end;

    c1 := #0;
    while c1 <= #255 do
      if c1 in CC then
      begin
        c2 := c1;
        while (c2 < #255) and (Succ(c2) in CC) do Inc(c2);
        if c1 = c2 then Str := Str + CharStr(c1, Reserved)
                   else
          if c2 = Succ(c1) then Str := Str + CharStr(c1, Reserved) + CharStr(c2, Reserved)
                           else Str := Str + CharStr(c1, Reserved) + '-' + CharStr(c2, Reserved);
        c1 := c2;
      end
      else Inc(c1);
    Result := '[' + Str + ']';
  end
end;

//----------------------------------------------------------------------------------------------------------------------

function CClassOrCharStr(CC: TCClass): String;

// returns a print name for character class CC (either cclassStr, or,
// if CC contains only one element, character in single quotes)

var
  Count: Integer;
  C,
  C1: Char;

begin
  Count := 0;
  C1 := #0;
  for C := #0 to #255 do
    if C in CC then
    begin
      c1 := C;
      Inc(Count);
      if Count > 1 then
      begin
        Result := CClassStr(CC);
        Exit;
      end
    end;

  if Count = 1 then Result := SingleQuoteStr(c1)
               else Result := '[]';
end;

//----------------------------------------------------------------------------------------------------------------------

function RegExprStr(R: PRegExpr): String;

// unparses a regular expression

  //--------------- local functions -------------------------------------------

  function UnparseExpr(R: PRegExpr): String;

  var
    rule_no,
    Position: Integer;
    C: Char;
    Str: String;
    CC: TCClass;
    R1,
    R2: PRegExpr;

  begin
    if IsEpsExpr(R) then Result := ''
                    else
      if IsMarkExpr(R, rule_no, Position) then Result := '#(' + IntToStr(rule_no) + ',' + IntToStr(Position) + ')'
                                          else
        if IsCharExpr(R, C) then Result := CharStr(C, ['"', '.', '^', '$', '[', ']', '*', '+', '?' , '{', '}', '|', '(', ')', '/', '<', '>'])
                            else
          if ISStrExpr (R, Str) then Result := DoubleQuoteStr(Str)
                                else
            if IsCClassExpr(R, CC) then Result := CClassStr(CC)
                                   else
              if IsStarExpr(R, R1) then Result := UnparseExpr(R1) + '*'
                                   else
                if IsPlusExpr(R, R1) then Result := UnparseExpr(R1) + '+'
                                     else
                  if IsOptExpr(R, R1) then Result := UnparseExpr(R1) + '?'
                                      else
                    if IsConcatExpr(R, R1, R2) then Result := '(' + UnparseExpr(R1) + UnparseExpr(R2) + ')'
                                               else
                      if IsAltExpr(R, R1, R2) then Result := '(' + UnparseExpr(R1) + '|' + UnparseExpr(R2) + ')'
                                              else raise Exception.Create('invalid expression');

  end;

  //--------------- end local functions ---------------------------------------

begin
  Result := UnparseExpr(R);
end;

//----------------------------------------------------------------------------------------------------------------------

function BrowseCallbackProc(hwnd: HWND; uMsg: UINT; lParam, lpData: LPARAM): Integer; stdcall;

// callback function used in SelectDirectory to set the status text and choose an initial dir

var
  Path: array[0..MAX_PATH] of Char;

begin
  case uMsg of
    BFFM_INITIALIZED:
      begin
        // Initialization has been done, now set our initial directory which is passed in lpData
        // (and set btw. the status text too).
        // Note: There's no need to cast lpData to a PChar since the following call needs a
        //       LPARAM parameter anyway.
        SendMessage(hwnd, BFFM_SETSELECTION, 1, lpData);
        SendMessage(hwnd, BFFM_SETSTATUSTEXT, 0, lpData);
      end;
    BFFM_SELCHANGED:
      begin
        // Set the status window to the currently selected path.
        if SHGetPathFromIDList(Pointer(lParam), Path) then SendMessage(hwnd, BFFM_SETSTATUSTEXT, 0, Integer(@Path));
      end;
  end;
  Result := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

function SelectDirectory(const Caption, InitialDir: String; const Root: WideString;
                         ShowStatus: Boolean; out Directory: String): Boolean;

// Another (overloaded) browse-for-folder function with the ability to select an intial directory
// (other SelectDirectory functions are in FileCtrl.pas).
// I had to make this overloading unambiguous in its parameter list so I included a flag which
// indicates whether or not to show a status text line in the dialog (which will receive the
// currently selected path if enabled).

var
  BrowseInfo: TBrowseInfo;
  Buffer: PChar;
  RootItemIDList,
  ItemIDList: PItemIDList;
  ShellMalloc: IMalloc;
  IDesktopFolder: IShellFolder;
  Eaten, Flags: LongWord;
  Windows: Pointer;
  Path: String;

begin
  Result := False;
  Directory := '';
  Path := InitialDir;
  if (Length(Path) > 0) and (Path[Length(Path)] = '\') then Delete(Path, Length(Path), 1);
  FillChar(BrowseInfo, SizeOf(BrowseInfo), 0);
  if (ShGetMalloc(ShellMalloc) = S_OK) and (ShellMalloc <> nil) then
  begin
    Buffer := ShellMalloc.Alloc(MAX_PATH);
    try
      SHGetDesktopFolder(IDesktopFolder);
      IDesktopFolder.ParseDisplayName(Application.Handle, nil, PWideChar(Root), Eaten, RootItemIDList, Flags);
      with BrowseInfo do
      begin
        hwndOwner := Application.Handle;
        pidlRoot := RootItemIDList;
        pszDisplayName := Buffer;
        lpszTitle := PChar(Caption);
        ulFlags := BIF_RETURNONLYFSDIRS;
        if ShowStatus then ulFlags := ulFlags or BIF_STATUSTEXT;
        lParam := Integer(PChar(Path));
        lpfn := BrowseCallbackProc;
      end;

      // make the browser dialog modal
      Windows := DisableTaskWindows(Application.Handle);
      try
        ItemIDList := ShBrowseForFolder(BrowseInfo);
      finally
        EnableTaskWindows(Windows);
      end;

      Result :=  ItemIDList <> nil;
      if Result then
      begin
        ShGetPathFromIDList(ItemIDList, Buffer);
        ShellMalloc.Free(ItemIDList);
        Directory := Buffer;
      end;
    finally
      ShellMalloc.Free(Buffer);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure ReadDelphiRoot;

// tries to read Delphi's installation path from registry (D4 first, then D3)

var
  Key: HKEY;
  Size: DWORD;
  Buffer: array[0..MAX_PATH] of Char;

begin
  DelphiRoot := '';
  Size := SizeOf(Buffer);
  if RegOpenKeyEx(HKEY_LOCAL_MACHINE, 'Software\Borland\Delphi\5.0', 0, KEY_READ, Key) = ERROR_SUCCESS then
  begin
    if RegQueryValueEx(Key, 'RootDir', nil, nil, @Buffer, @Size) = ERROR_SUCCESS then DelphiRoot := Buffer;
    RegCloseKey(Key);
  end
  else
  begin
    if RegOpenKeyEx(HKEY_LOCAL_MACHINE, 'Software\Borland\Delphi\4.0', 0, KEY_READ, Key) = ERROR_SUCCESS then
    begin
      if RegQueryValueEx(Key, 'RootDir', nil, nil, @Buffer, @Size) = ERROR_SUCCESS then DelphiRoot := Buffer;
      RegCloseKey(Key);
    end
    else
    begin
      // Delphi 4 seems not to be installed, so try Delphi 3
      if RegOpenKeyEx(HKEY_LOCAL_MACHINE, 'Software\Borland\Delphi\3.0', 0, KEY_READ, Key) = ERROR_SUCCESS then
      begin
        if RegQueryValueEx(Key, 'RootDir', nil, nil, @Buffer, @Size) = ERROR_SUCCESS then DelphiRoot := Buffer;
        RegCloseKey(Key);
      end
    end;
  end;
  if DelphiRoot <> '' then GlobalSettings.DCC32Path := DelphiRoot + '\bin\dcc32.exe';
  GlobalSettings.DCC32Path := ReplaceStringByMacro(SDelphiPathMacro, GlobalSettings.DCC32Path, DelphiRoot);
end;

//----------------------------------------------------------------------------------------------------------------------

end.

