unit Highlighter;

// Copyright (c) 1999, 2000 Mike Lischke (public@lischke-online.de)
//----------------------------------------------------------------------------------------------------------------------
// This unit implements the base functionality for a syntax highlighter which is used
// in TSyntaxEdit. Some of the code used here was taken form mwEdit.

interface

uses
  Windows, SysUtils, Classes, Graphics, Registry, Controls;

type
  TWindowList = class(TList)
  public
    procedure AddInstance(Control: TWinControl);
    procedure RemoveInstance(Control: TWinControl);
    procedure Invalidate;
  end;

  THighlightData = record
    Foreground,
    Background: TColor;
    Style: TFontStyles;
  end;

  TTokenData = record
    Token: PChar;
    TokenType: Integer;
    Position,
    Length: Cardinal;
    Background,
    ForeGround: TColor;
    Style: TFontStyles;
  end;

  THighlightAttributes = class(TPersistent)
  private
    FHighlightData: THighlightData;
    FOnChange: TNotifyEvent;
    FName: String;
    procedure SetBackground(Value: TColor);
    procedure SetForeground(Value: TColor);
    procedure SetStyle(Value: TFontStyles);
  public
    procedure Assign(Source: TPersistent); override;
    function LoadFromBorlandRegistry(Root: HKEY; Key, Name: String; OldStyle: Boolean): Boolean; virtual;
    function LoadFromRegistry(Reg: TRegistry): Boolean;
    function SaveToRegistry(Reg: TRegistry): Boolean;

    property Name: String read FName;
  published
    constructor Create(Name: String);

    property Background: TColor read FHighlightData.Background write SetBackground;
    property Foreground: TColor read FHighlightData.Foreground write SetForeground;
    property Style: TFontStyles read FHighlightData.Style write SetStyle;

    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  TIdentChars = set of Char;

  THighlighterCapability = (
    hcUserSettings, // supports Enum/UseUserSettings
    hcRegistry      // supports LoadFrom/SaveToRegistry
  );

  THighlighterCapabilities = set of THighlighterCapability;

  TSyntaxHighlighter = class(TComponent)
  private
    FWindowList: TWindowList;
    FDefaultFilter: String;
  protected
    function GetIdentChars: TIdentChars; virtual;
    function GetLanguageName: String; virtual; abstract;

    function GetAttributeCount: Integer; virtual; abstract;
    function GetAttribute(idx: Integer): THighlightAttributes; virtual; abstract;
    function GetCapabilities: THighlighterCapabilities; virtual;
    function GetDefaultFilter: String; virtual;
    procedure SetDefaultFilter(Value: String); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure AssignAttributesFrom(const Source: TSyntaxHighlighter); virtual;
    function EOL: Boolean; virtual; abstract;
    function GetRange: Pointer; virtual; abstract;
    function GetToken: String; virtual; abstract;
    function GetTokenInfo: TTokenData; virtual; abstract; 
    function GetTokenPos: Integer; virtual; abstract;
    procedure Next; virtual; abstract;
    procedure SetLine(const Value: String); virtual; abstract;
    procedure SetRange(Value: Pointer); virtual; abstract;
    procedure ReSetRange; virtual; abstract;
    // Index = 0..GetAttributeCount-1
    function UseUserSettings(SettingIndex: Integer): Boolean; virtual;
    procedure EnumUserSettings(Settings: TStrings); virtual;
    function LoadFromRegistry(RootKey: HKEY; Key: String): Boolean; virtual;
    function SaveToRegistry(RootKey: HKEY; Key: String): Boolean; virtual;

    property WindowList: TWindowList read FWindowList;
    property IdentChars: TIdentChars read GetIdentChars;
    property LanguageName: String read GetLanguageName;

    property AttributeCount: Integer read GetAttributeCount;
    property Attribute[Index: Integer]: THighlightAttributes read GetAttribute;
    property Capabilities: THighlighterCapabilities read GetCapabilities;
  published
    property DefaultFilter: String read GetDefaultFilter write SetDefaultFilter;
  end;

  THighlighterClass = class of TSyntaxHighlighter;
  
//----------------------------------------------------------------------------------------------------------------------

implementation

//----------------- TWindowList ----------------------------------------------------------------------------------------

procedure TWindowList.AddInstance(Control: TWinControl);

begin
  if IndexOf(Control) = -1 then Add(Control);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TWindowList.RemoveInstance(Control: TWinControl);

var Index: Integer;

begin
  Index:= IndexOf(Control);
  if Index <> -1 then Delete(Index);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TWindowList.Invalidate;

var
  I: Integer;

begin
  for I := 0 to Count - 1 do TWinControl(Items[I]).Invalidate;
end;

//----------------- THighlightAttributes -----------------------------------------

procedure THighlightAttributes.Assign(Source: TPersistent);

begin
  if Source is THighlightAttributes then
  begin
    FHighlightData := (Source as THighlightAttributes).FHighlightData;
    FName := (Source as THighlightAttributes).FName;
  end
  else inherited Assign(Source);
end;

//----------------------------------------------------------------------------------------------------------------------

constructor THighlightAttributes.Create(Name: String);

begin
  inherited Create;
  FHighlightData.Background := clWindow;
  FHighlightData.Foreground := clWindowText;
  FName := Name;
end;

//----------------------------------------------------------------------------------------------------------------------

function THighlightAttributes.LoadFromBorlandRegistry(Root: HKEY; Key, Name: String; OldStyle: Boolean): Boolean;

const
  Pal16: array [0..15] of TColor = (
    clBlack, clMaroon, clGreen, clOlive, clNavy, clPurple, clTeal, clDkGray,
    clLtGray, clRed, clLime, clYellow, clBlue, clFuchsia, clAqua, clWhite
  );

  //--------------- local functions -------------------------------------------

  function IsTrue(Value: String): Boolean;

  begin
    Result := not ((UpperCase(Value) = 'FALSE') or (Value = '0'));
  end;

  //---------------------------------------------------------------------------

  function Get(var Name: String): String;

  var
    I: Integer;

  begin
    I := Pos(',', Name);
    if I = 0 then I := Length(Name)+1;
    Result := Copy(Name, 1, I-1);
    Name := Copy(Name, I+1, Length(Name)-I);
  end;

  //---------------------------------------------------------------------------

  function LoadOldStyle(Root: HKEY; Key, Name: String): Boolean;

  var
    descript,
    fgColRGB,
    bgColRGB,
    fontStyle,
    fgDefault,
    bgDefault,
    fgIndex16,
    bgIndex16: String;
    Reg: TRegistry;

  begin
    Result := False;
    try
      reg := TRegistry.Create;
      Reg.RootKey := Root;
      try
        with Reg do
        begin
          if OpenKeyReadOnly(Key) then
          begin
            try
              if ValueExists(Name) then
              begin
                descript := ReadString(Name);
                fgColRGB := Get(descript);
                bgColRGB := Get(descript);
                fontStyle := Get(descript);
                fgDefault := Get(descript);
                bgDefault := Get(descript);
                fgIndex16 := Get(descript);
                bgIndex16 := Get(descript);
                if bgDefault = '1' then Background := clWindow
                                   else Background := Pal16[StrToInt(bgIndex16)];
                if fgDefault = '1' then Foreground := clWindowText
                                   else Foreground := Pal16[StrToInt(fgIndex16)];
                Style := [];
                if Pos('B', fontStyle) > 0 then Style := Style + [fsBold];
                if Pos('I', fontStyle) > 0 then Style := Style + [fsItalic];
                if Pos('U', fontStyle) > 0 then Style := Style + [fsUnderline];
                Result := True;
              end;
            finally
              CloseKey;
            end;
          end;
        end;
      finally
        Reg.Free;
      end;
    except
    end;
  end;

  //---------------------------------------------------------------------------

  function LoadNewStyle(Root: HKEY; Key, Name: String): Boolean;

  var
    fgIndex16,
    bgIndex16: DWORD;
    fontBold,
    fontItalic,
    fontUnderline,
    fgDefault,
    bgDefault: String;
    Reg: TRegistry;

  begin
    Result := False;
    try
      Reg := TRegistry.Create;
      Reg.RootKey := Root;
      try
        with Reg do
        begin
          if OpenKeyReadOnly(Key + '\' + Name) then
          begin
            try
              if ValueExists('Foreground Color') then fgIndex16 := ReadInteger('Foreground Color')
                                                 else fgIndex16 := 15;
              if ValueExists('Background Color') then bgIndex16 := ReadInteger('Background Color')
                                                 else bgIndex16 := 0;
              if ValueExists('Bold') then fontBold := ReadString('Bold')
                                     else Abort;
              if ValueExists('Italic') then fontItalic := ReadString('Italic')
                                       else Abort;
              if ValueExists('Underline') then fontUnderline := ReadString('Underline')
                                          else Abort;
              if ValueExists('Default Foreground') then fgDefault := ReadString('Default Foreground')
                                                   else Abort;
              if ValueExists('Default Background') then bgDefault := ReadString('Default Background')
                                                   else Abort;
              if IsTrue(bgDefault) then Background := clWindow
                                   else Background := Pal16[bgIndex16];
              if IsTrue(fgDefault) then Foreground := clWindowText
                                   else Foreground := Pal16[fgIndex16];
              Style := [];
              if IsTrue(fontBold) then Style := Style + [fsBold];
              if IsTrue(fontItalic) then Style := Style + [fsItalic];
              if IsTrue(fontUnderline) then Style := Style + [fsUnderline];
              Result := True;
            finally
              CloseKey;
            end;
          end;
        end;
      finally
        Reg.Free;
      end;
    except
    end;
  end;

  //--------------- end local functions ---------------------------------------

begin
  if OldStyle then Result := LoadOldStyle(Root, Key, Name)
              else Result := LoadNewStyle(Root, Key, Name);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure THighlightAttributes.SetBackground(Value: TColor);

begin
  with FHighlightData do
    if Background <> Value then
    begin
      Background := Value;
      if Assigned(FOnChange) then FOnChange(Self);
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure THighlightAttributes.SetForeground(Value: TColor);

begin
  with FHighlightData do
    if Foreground <> Value then
    begin
      Foreground := Value;
      if Assigned(FOnChange) then FOnChange(Self);
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure THighlightAttributes.SetStyle(Value: TFontStyles);

begin
  with FHighlightData do
    if Style <> Value then
    begin
      Style := Value;
      if Assigned(FOnChange) then FOnChange(Self);
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

function THighlightAttributes.LoadFromRegistry(Reg: TRegistry): Boolean;

var
  Key: String;

begin
  Key := Reg.CurrentPath;
  if Reg.OpenKeyReadOnly(Name) then
  begin
    if Reg.ValueExists('Attributes') then Reg.ReadBinaryData('Attributes', FHighlightData, SizeOf(FHighlightData));
    Reg.OpenKeyReadOnly('\' + Key);
    Result := True;
  end
  else Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

function THighlightAttributes.SaveToRegistry(Reg: TRegistry): Boolean;

var
  Key: String;

begin
  Key := Reg.CurrentPath;
  if Reg.OpenKey(Name, True) then
  begin
    Reg.WriteBinaryData('Attributes', FHighlightData, SizeOf(FHighlightData));
    Reg.OpenKey('\' + Key, False);
    Result := True;
  end
  else Result := False;
end;

//----------------- TSyntaxHighlighter -----------------------------------------------------------------------------------

constructor TSyntaxHighlighter.Create(AOwner: TComponent);

begin
  inherited Create(AOwner);
  FWindowList := TWindowList.Create;
  FDefaultFilter := '';
end;

//----------------------------------------------------------------------------------------------------------------------

destructor TSyntaxHighlighter.Destroy;

begin
  FWindowList.Free;
  inherited Destroy;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSyntaxHighlighter.EnumUserSettings(Settings: TStrings);

begin
  Settings.Clear;
end;

//----------------------------------------------------------------------------------------------------------------------

function TSyntaxHighlighter.UseUserSettings(SettingIndex: Integer): Boolean;

begin
  Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

function TSyntaxHighlighter.GetIdentChars: TIdentChars;

begin
  Result := [#33..#255];
end;

//----------------------------------------------------------------------------------------------------------------------

function TSyntaxHighlighter.LoadFromRegistry(RootKey: HKEY; Key: String): Boolean;

var
  Reg: TRegistry;
  I: Integer;

begin
  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;
    if Reg.OpenKeyReadOnly(Key) then
    begin
      Result := True;
      for I := 0 to AttributeCount - 1 do
        Result := Result and Attribute[i].LoadFromRegistry(Reg);
    end
    else Result := False;
  finally
    Reg.Free;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TSyntaxHighlighter.SaveToRegistry(RootKey: HKEY; Key: String): Boolean;

var
  Reg: TRegistry;
  I: Integer;

begin
  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;
    if Reg.OpenKey(Key, True) then
    begin
      Result := True;
      for I := 0 to AttributeCount-1 do
        Result := Result and Attribute[i].SaveToRegistry(Reg);
    end
    else Result := False;
  finally
    Reg.Free;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TSyntaxHighlighter.GetCapabilities: THighlighterCapabilities;

begin
  Result := [hcRegistry];
end;

//----------------------------------------------------------------------------------------------------------------------

function TSyntaxHighlighter.GetDefaultFilter: String;

begin
  Result := FDefaultFilter;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSyntaxHighlighter.SetDefaultFilter(Value: String);

begin
  if FDefaultFilter <> Value then FDefaultFilter := Value;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSyntaxHighlighter.AssignAttributesFrom(const Source: TSyntaxHighlighter);

// copies all attribute settings from Source to the local list, only attributes whose names
// do match are copied

var
  I, J: Integer;
  AttrName: String;

begin
  for I := 0 to Source.AttributeCount - 1 do
  begin
    AttrName := Source.Attribute[I].Name;
    for J := 0 to AttributeCount - 1 do
      if CompareText(AttrName, Attribute[J].Name) = 0 then Attribute[J].Assign(Source.Attribute[I]);
  end;
  FWindowList.Invalidate;
end;

//----------------------------------------------------------------------------------------------------------------------

end.
