unit Instance;

// 20-JUN-99 ml: last change
//----------------------------------------------------------------------------------------------------------------------
// This unit provides a mechanism to allow only one instance of an application to run. Any
// following start of the same executable results in a notification of the first instance
// passing all parameters given to the second instance through to the first one.
//
// usage:
//  Call (as soon as possible) the function 'CheckForOtherInstance'.
//  You can use the source of the project:
//
//  begin
//    if not CheckForOtherInstance then
//    begin
//      Application.Title := 'Blah blah';
//      Application.CreateForm(TMainForm, MainForm);
//        :
//      Application.Run;
//    end;
//  end;
//
//  That's all for this part. The application will not be terminated automatically as this might not
//  desirable, but with the example shown above not much more happens beyond the check.
//
//  If the first instance needs to know which parameters have been passed then include a hook of the
//  application window (HookMainWindow/UnhookMainWindow) somewhere it fits (place is uncritical,
//  the handler doesn't even need to be in a TWinControl). The message to be catched is WM_OTHERINSTANCE
//  (posted, not sent).
//
//  While receiving the message WM_OTHERINSTANCE the WindowHook can use "GetInstanceParams" to get
//  a string list with the passed parameters (this list is guarded with a mutex object to synchronize
//  access to it. Note: The entry 0 in the list is the file name of the executable
//  (just like in ParamCount/ParamStr).

interface

uses
  Windows, SysUtils, Classes;

const
  OtherInstanceMsgName = 'MessageOtherInstance';
  OtherInstanceMutexName = 'MutexOtherInstance';

var
  WM_OTHERINSTANCE: Cardinal;

function CheckForOtherInstance: Boolean;
procedure GetInstanceParams(const List: TStringList);

//----------------------------------------------------------------------------------------------------------------------

implementation

uses
  Dialogs, Forms;

const
  ParamSize = 4096;

type
  PInstInfo = ^TInstInfo;
  TInstInfo = record
    FirstInstance: THandle;
    Mutex: THandle; // for process synchronisation
    Params: array[0..ParamSize - 1] of Char;
  end;

var
  MappingHandle: THandle;  // handle of the filemapping object

//------------------------------------------------------------------------------

function CheckForOtherInstance: Boolean;

// Checks whether there is already a running instance of the same executable and
// ensures (if there's one) that it is activated and the own command line parameters
// are passed to it.

var
  Info: PInstInfo;
  CommandLine: PChar;

begin
  Result := False;
  // create shared memory 
  MappingHandle := CreateFileMapping(DWORD(-1), nil, PAGE_READWRITE, 0, SizeOf(TInstInfo), PChar(ExtractFileName(Application.ExeName)));
  if MappingHandle <> 0 then
  begin
    // get pointer to data
    Info := MapViewOfFile(MappingHandle, FILE_MAP_READ or FILE_MAP_WRITE, 0, 0, SizeOf(TInstInfo));
    // does the mapping already exist?
    if GetLastError = ERROR_ALREADY_EXISTS then
    begin
      // well, there's already another instance running, so pass the parameter list,
      // notify the first instanz and get out off here
      Result := True;
      // ParamCount and ParamStr are much too long winded
      CommandLine := GetCommandLine;
      StrLCopy(Info.Params, CommandLine, ParamSize);
      if IsIconic(Info.FirstInstance) then ShowWindow(Info.FirstInstance, SW_RESTORE);
      SetForegroundWindow(Info.FirstInstance);
      PostMessage(Info.FirstInstance, WM_OTHERINSTANCE, 0, 0);
    end
    else
    begin
      // hey, this is the first instance, so init the data area and store the application handle
      FillChar(Info^, SizeOf(Info^), 0);
      Info.FirstInstance := Application.Handle;
      Info.Mutex := CreateMutex(nil, False, OtherInstanceMutexName);
    end;
    // free the view
    UnmapViewOfFile(Info);
  end
  else raise Exception.Create('internal error (Instance check): file mapping failed');
end;

//------------------------------------------------------------------------------

procedure GetInstanceParams(const List : TStringList);

// fills a list with the strings contained in the info structure

var
  Info: PInstInfo;

begin
  if assigned(List) then
  begin
    Info := MapViewOfFile(MappingHandle, FILE_MAP_READ or FILE_MAP_WRITE, 0, 0, SizeOf(TInstInfo));
    // wait until mutex is free
    WaitForSingleObject(Info.Mutex, INFINITE);
    List.Clear;
    List.CommaText := Info.Params;
    // free the view
    ReleaseMutex(Info.Mutex);
    UnmapViewOfFile(Info);
  end;
end;

//------------------------------------------------------------------------------

procedure FreeInstanceInfo;

// remove all allocated data from paging file

var
  Info: PInstInfo;
  
begin
  // something to do?
  if MappingHandle <> 0 then
  begin
    // free the mutex if it has been created by this instance
    Info := MapViewOfFile(MappingHandle, FILE_MAP_READ or FILE_MAP_WRITE, 0, 0, SizeOf(TInstInfo));
    if Info.FirstInstance = Application.Handle then
    begin
      // wait until mutex is free
      WaitForSingleObject(Info.Mutex, INFINITE);
      CloseHandle(Info.Mutex);
    end;
    UnmapViewOfFile(Info);
    CloseHandle(MappingHandle);
  end;
end;

//------------------------------------------------------------------------------

initialization
  WM_OTHERINSTANCE := RegisterWindowMessage(OtherInstanceMsgName);
finalization
  FreeInstanceInfo;
end.


