unit LexerGenerator;

// 05-DEC-99 ml: structural rework
// 13-NOV-99 ml: implementation so far
//----------------------------------------------------------------------------------------------------------------------
// This unit contains a class used to create lexical scanners.

interface

uses
  Classes, DCGStrings, DCGTools;

const
  MaxSymbols = 997;       // Length of static hash symbol table (prime number!)
  MacroBufferSize = 8192; // Length of static macro stack used to parse Lexer input
  MaxItems = 255;         // Length of item buffers used by the Lexer parser

  SectionDelimiter = '%%';
      
type
  TLexerGenerator = class;

  // table data structures

  TSymbolTableEntry = record
    Name: String;
    case SymbolType: (stNone, stMacro, stStartState ) of
      stMacro:
        (Substitute: PChar); // macro substitution, String is not allowed in variant
                             // reords, hence there's a PChar
      stStartState:
        (StartState: Cardinal);
  end;

  TSymbolTable = array[1..MaxSymbols] of TSymbolTableEntry;

  TPositionTableEntry = record
    FollowPos: TIntegerList;    // set of Follow positions
    case PositionType: (ptChar, ptCClass, ptMark) of
      ptChar:
        (C: Char);              // character Position
      ptCClass:
        (CClass: TCClass);      // character class Position
      ptMark:
        (Rule,
         Position: Integer );   // mark Position
  end;

  TPosTable = array of TPositionTableEntry;

  // First positions for start states (even states
  // are entered anywhere on the line, odd states only
  // at the beginning of the line; states 0 and 1 denote
  // default, states 2..2*FStartStatesCount+1 user-defined
  // start states)
  TFirstPosTable  = array of TIntegerList;

  TStateTableEntry = record
    StatePosition: TIntegerList;   // positions covered by state
    Final: Boolean;               // Final state?
    TransitionLow,
    TransitionHigh: Integer;      // transitions
  end;

  TStateTable = array of TStateTableEntry;

  TTransTableEntry = record
    CC: TCClass;            // characters of transition
    FollowPos: TIntegerList; // Follow positions (positions of next state)
    NextState: Integer;     // next state
  end;

  TTransitionTable = array of TTransTableEntry;

  TPartitionTable = array of TIntegerList; // state partitions (DFA optimization)
  TStatePartitionTable = array of Integer; // partition number of states

  TLexerGenerator = class
  private
    FSource,
    FTemplate: TStrings;
    FTarget: TStringList;
    FCurrentSourceLine,
    FCurrentTemplateLine: String;             // current input lines without lexer comments
    FSourceIndex,                             // index into FSource
    FTemplateIndex,                           // dito for FTemplate
    FRuleCount,
    FErrors: Cardinal;
    FSymbolTable: TSymbolTable;
    FPositionTable: TPosTable;
    FFirstPositionTable: TFirstPosTable;
    FStateTable: TStateTable;
    FTransitionTable: TTransitionTable;
    FCurrentState: Integer;
    FStartStatesCount: Integer;
    FGenerating: Boolean;

    // Lex-Parser variables
    FCurrentPosition,
    FBufferPosition : Integer;                // current position in source line and input stack pointer
    FMacroBuffer : array[1..MacroBufferSize] of Char; // input buffer
    FParserStringState,
    FParserCClassState,
    FParserQuoteState : Boolean;              // state information
    FMacroCount : Integer;                    // number of macros currently on stack
    FCurrentExpression,
    FCurrentStatement : String;               // expression and statement part of rule
    FCaretFlag : Boolean;                     // caret flag
    FParserStateCount : Integer;              // number of start states in prefix
    FParserStartStates : TIntegerList;         // start states
    FParsedExpression : PRegExpr;             // parsed expression
    FItemString: String;
    FItemCount: Integer;
    FItemPos,
    FItemLength: array [1..MaxItems] of Integer;
    FLexerName: String;                       // name of target lexer class and module
    FCaseSensitive: Boolean;                  // indication whether keywords are to be consider case sensitive
    FTokens,                                  // list of defined tokens for automatic creation of token constants
    FKeywords: TSymbolList;                   // list of defined keywords for automatic creation of the keyword table
    FInsetStartLine: Integer;                 // line marker to place the final tables at
    FCurrentTokenID: Integer;                 // running number for keyword and token constants

    // DFA optimization
    FPartitionTable: TPartitionTable;
    FStatePartitionTable: TStatePartitionTable;
    FOptimizedStateTable: TStateTable;
    FOptimizedTransitionTable: TTransitionTable;

    // events for application callbacks
    FOnSourceLine,
    FOnTemplateLine: TOnProcessLineEvent;
    FOnMessage: TOnMessageEvent;
    procedure Error(ErrorType: TMessageType; Msg: String; Pos: Integer; InSource: Boolean);
    procedure FatalError(Msg: String);
    procedure DoLog(const S: String);
    procedure NextTemplateSection;
    procedure AddRule;
    procedure AuxiliaryProcs;
    procedure Code;
    procedure DefineMacro(Symbol, Replacement: String);
    procedure DefineStartState(Symbol: String; Pos: Integer);
    procedure Definitions;
    procedure GenerateTables;
    function Itemv(I: Integer): String;
    procedure Rules;
    procedure Split(Str: String; Count: Integer);
    procedure MakeDFATable;
    procedure AddCClassPos(CC: TCClass);
    procedure AddCharPos(C: Char);
    procedure AddMarkPos(Rule, Pos: Integer);
    function AddState(Pos: TIntegerList): Integer;
    procedure AddTransition(CC: TCClass; Follow: TIntegerList);
    procedure EndStateTransition;
    procedure Entry(K: Integer; Symbol: String);
    procedure Evaluate(R: PRegExpr; var First, Last: TIntegerList; var nullable: Boolean);
    function Lookup(K: Integer): String;
    procedure MergeTransition;
    function TransitionCount(I: Integer): Integer;
    function MakeNewState(Pos: TIntegerList): Integer;
    procedure SortTransition;
    procedure StartStateTransition;
    function TransitionLess(I, J: Integer): Boolean;
    function TransitionLessNextState(I, J: Integer): Boolean;
    procedure SwapTransition(I, J: Integer);
    function EquivalentStates(I, J: Integer): Boolean;
    procedure OptimizeDFATable;
    procedure AddExpression(R: PRegExpr; var First: TIntegerList);
    function CurrentParserChar(Source: Boolean): Char;
    procedure AddStartState(symbol: String);
    procedure FetchNextParserChar(Source: Boolean);
    procedure InitScanner(Source: Boolean);
    procedure MarkError(ErrorType: TMessageType; Msg: String; Offset: Integer; Source: Boolean);
    procedure ParseRule(RuleNumber: Integer);
    procedure PushString(Str: String);
    procedure ClearStructures;
    procedure FreeExpression(MemList: TList; var R: PRegExpr);
    function StripComment(const Line: String): String;
    function NextSourceLine: Boolean;
    function NextTemplateLine: Boolean;
    procedure Globals;
    procedure Keywords;
    function CheckID(const Symbol: String): Boolean;
    procedure WriteKeywordSupport;
    procedure KewordConstants;
    procedure DefineToken(Symbol: String; Pos: Integer);
    procedure TokenConstants;
    function ReplaceInvalidChars(const S: String): String;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure Execute;
    property Generating: Boolean read FGenerating;
    property Source: TStrings read FSource write FSource;
    property Target: TStringList read FTarget;
    property Template: TStrings read FTemplate write FTemplate;

    property OnMessage: TOnMessageEvent read FOnMessage write FOnMessage;
    property OnSourceLine: TOnProcessLineEvent read FOnSourceLine write FOnSourceLine;
    property OnTemplateLine: TOnProcessLineEvent read FOnTemplateLine write FOnTemplateLine;
  end;

//----------------------------------------------------------------------------------------------------------------------

implementation

uses
  SysUtils;

//----------------- TLexerGenerator ------------------------------------------------------------------------------------

constructor TLexerGenerator.Create;

begin
  FTarget := TStringList.Create;
  FTokens := TSymbolList.Create;
  FTokens.Sorted := True;
  FKeywords := TSymbolList.Create;
  FKeywords.Sorted := True;
end;

//----------------------------------------------------------------------------------------------------------------------

destructor TLexerGenerator.Destroy;

begin
  ClearStructures;
  FTokens.Free;
  FKeywords.Free;
  inherited;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.ClearStructures;

var
  I: Integer;

begin
  for I := 1 to MaxSymbols do
  begin
    FSymbolTable[I].Name := '';
    if (FSymbolTable[I].SymbolType = stMacro) and
       Assigned(FSymbolTable[I].Substitute) then StrDispose(FSymbolTable[I].Substitute);
  end;

  for I:= 0 to Length(FStateTable) - 1 do FStateTable[I].StatePosition := nil;
  FStateTable := nil;

  for I:= 0 to Length(FTransitionTable) - 1 do FTransitionTable[I].FollowPos := nil;
  FTransitionTable := nil;

  FPositionTable := nil;
  FFirstPositionTable := nil;
  FParserStartStates := nil;

  if assigned(FTokens) then FTokens.Clear;
  if assigned(FKeywords) then FKeywords.Clear;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.Execute;

// Main routine, parses source grammar and generates a string list of delphi source code
// which can then be saved to a file.

begin
  if FSource = nil then FatalError('source not assigned');
  if FTemplate = nil then FatalError('template not assigned');
  FGenerating := True;
  try
    ClearStructures; // clears symbol table too

    // initialize
    FErrors := 0;
    FSourceIndex := 0;
    FTemplateIndex := 0;
    FRuleCount := 0;
    FStartStatesCount := 0;
    FCurrentTokenID := 257;
    FLexerName := 'Lexer';
    FCaseSensitive := False;
    SetLength(FFirstPositionTable, 2);
    FFirstPositionTable[0] := nil;
    FFirstPositionTable[1] := nil;

    FTarget.Clear;

    // parse source grammar
    DoLog('parsing source grammar ... ');

    // read in global symbols and keywords
    Globals;
    Keywords;
    // find and copy first code template section
    NextTemplateSection;
    Definitions;
    KewordConstants;
    TokenConstants;
    Rules;
    if FRuleCount = 0 then Error(mtError, SEmptyGrammar, Length(TrimRight(FCurrentSourceLine)), True);
    if FErrors = 0 then
    begin
      // generate DFA table and listings and write code
      DoLog('DFA construction ... ');
      MakeDFATable;
      DoLog('DFA optimization ... ');
      OptimizeDFATable;
      DoLog('Code generation ... ');
      GenerateTables;
      WriteKeywordSupport;
      NextTemplateSection;
    end;

    AuxiliaryProcs;
    if FErrors = 0 then DoLog('Result');

    // print statistics
    if FErrors = 0 then DoLog(IntToStr(FSource.Count) + ' line(s),     ' +
                              IntToStr(FRuleCount) + ' rule(s),     ' +
                              IntToStr(FTokens.Count) + ' token(s),     ' +
                              IntToStr(FKeywords.Count) + ' keyword(s),     ' +
                              IntToStr(Length(FPositionTable)) + ' position(s),     ' +
                              IntToStr(Length(FStateTable)) + ' state(s),     ' +
                              IntToStr(Length(FTransitionTable)) + ' transition(s)');
  finally
    FGenerating := False;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.DoLog(const S: String);

var
  MI: TMessageInfo;

begin
  if assigned(FOnMessage) then
  begin
    with MI do
    begin
      Code := mtLog;
      Message := S;
      Source := '';
      Line := 0;
      Position := 0;
    end;
    FOnMessage(Self, MI);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.Error(ErrorType: TMessageType; Msg: String; Pos: Integer; InSource: Boolean);

var
  MI: TMessageInfo;

begin
  Inc(FErrors);
  if assigned(FOnMessage) then
  begin
    with MI do
    begin
      Code := ErrorType;
      Message := Msg;
      if InSource then
      begin
        Source := 'S';  // actual filenames must be inserted by the receiver
        Line := FSourceIndex;
      end
      else
      begin
        Source := 'T';
        Line := FTemplateIndex;
      end;
      Position := Pos;
    end;
    FOnMessage(Self, MI);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.FatalError(Msg: String);

var
  MI: TMessageInfo;

begin
  Inc(FErrors);
  if assigned(FOnMessage) then
  begin
    with MI do
    begin
      Code := mtFatalError;
      Message := Msg;
      Source := '';
      Line := 0;
      Position := 0;
    end;
    FOnMessage(Self, MI);
  end;
  Abort;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.NextTemplateSection;

// find next section mark (%%) in code template, copy all intermediate lines from
// template to target and replace lexer symbols appropriately

var
  TargetLine: String;
  C: Char;

begin
  while NextTemplateLine and (FCurrentTemplateLine <> SectionDelimiter) do
  begin
    if FCurrentTemplateLine <> '' then
    begin
      InitScanner(False);
      TargetLine := '';
      // have to take the long way in order to replace macros
      repeat
        C := CurrentParserChar(False);
        if C = #0 then Break;
        TargetLine := TargetLine + C;
        FetchNextParserChar(False);
      until False;
      if TargetLine <> '' then FTarget.Add(TargetLine);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

// Semantic routines: 

procedure TLexerGenerator.DefineStartState(Symbol: String; Pos: Integer);

// process start state definition

begin
  with FSymbolTable[Key(Symbol, MaxSymbols, Lookup, Entry)] do
    if SymbolType = stNone then
    begin
      // two default states are already there (not counted in FStartStateCount)
      SetLength(FFirstPositionTable, 2 * (FStartStatesCount + 3));
      FFirstPositionTable[2 * FStartStatesCount + 2] := nil;
      FFirstPositionTable[2 * FStartStatesCount + 3] := nil;
      Inc(FStartStatesCount);
      SymbolType := stStartState;
      StartState := FStartStatesCount;
      if FStartStatesCount = 1 then FTarget.Add('const ' + Symbol + ' = ' + IntToStr(2 * StartState) + ';')
                               else FTarget.Add('      ' + Symbol + ' = ' + IntToStr(2 * StartState) + ';');
    end
    else Error(mtSemanticError, SSymbolAlreadyDefined, Pos, True);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.DefineMacro(Symbol, Replacement: String);

// process macro definition

begin
  with FSymbolTable[Key(''+Symbol+'', MaxSymbols, Lookup, Entry)] do
    if SymbolType = stNone then
    begin
      SymbolType := stMacro;
      Substitute := StrNew(PChar(Replacement));
    end
    else Error(mtSemanticError, SSymbolAlreadyDefined, 1, True);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.DefineToken(Symbol: String; Pos: Integer);

var
  Index: Integer;

begin
  if FTokens.Find(Symbol, Index) then Error(mtSemanticError, SSymbolAlreadyDefined, Pos, True)
                                 else FTokens.Add(Symbol);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.FreeExpression(MemList: TList; var R: PRegExpr);

// frees allocated memory of the expression R

begin
  if Assigned(R) then
  begin
    case R.ExprType of
      ntCClass,
      ntMark,
      ntChar:
        ; // nothing do free
      ntString:
        if Assigned(R.S) then StrDispose(R.S);
      ntStar,
      ntPlus,
      ntOptional:
        FreeExpression(MemList, R.R);
      ntConcat,
      ntAlternative:
        begin
          FreeExpression(MemList, R.R1);
          FreeExpression(MemList, R.R2);
        end;
    end;
    if MemList.IndexOf(R) = -1 then MemList.Add(R);
    R := nil;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.AddRule;

// process Rule

var
  I: Integer;
  First: TIntegerList;
  MemList: TList;

begin
  AddExpression(FParsedExpression, First);
  if FParserStateCount = 0 then
    if FCaretFlag then ISUnion(FFirstPositionTable[1], First)
                  else
    begin
      ISUnion(FFirstPositionTable[0], First);
      ISUnion(FFirstPositionTable[1], First);
    end
                           else
    if FCaretFlag then
      for I := 0 to FParserStateCount - 1 do
        ISUnion(FFirstPositionTable[2 * FParserStartStates[I] + 1], First)
                  else
      for I := 0 to FParserStateCount - 1 do
      begin
        ISUnion(FFirstPositionTable[2 * FParserStartStates[I]], First);
        ISUnion(FFirstPositionTable[2 * FParserStartStates[I] + 1], First);
      end;

  // collect allocated memory and free it finally
  MemList := TList.Create;
  try
    FreeExpression(MemList, FParsedExpression);
    for I := 0 to MemList.Count - 1 do FreeMem(MemList[I]);
  finally
    MemList.Free;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.GenerateTables;

// write the DFA table to the output file
// Tables are represented as a collection of typed array constants:
//
// type TTransition = record
//                 CC : set of Char; { characters }
//                 S  : Integer;     { next state }
//               end;
//
// const
//
// { table sizes: }
//
// MarkPositionCount   = ...;
// MatchCount = ...;
// TransitionCount   = ...;
// StateCount  = ...;
//
// { Rules of mark positions for each state: }
//
// MarkPositionTable : array [1..MarkPositionCount] of Integer = ...;
//
// { Rules of matches for each state: }
//
// MatchTable : array [1..MatchCount] of Integer = ...;
//
// { transition table: }
//
// TransitionTable : array [1..TransitionCount] of TTransition = ...;
//
// { offsets into the marks, matches and transition Tables: }
//
// MarksLow, MarksHigh,
// MatchesLow, MatchesHigh,
// TransitionsLow, TransitionsHigh : array[0..StateCount-1] of Integer = ...;

var
  MarkPositionCount,
  MatchCount,
  TransitionCount,
  StateCount: Integer;
  MarksLow,
  MarksHigh,
  MatchesLow,
  MatchesHigh,
  TransitionsLow,
  TransitionsHigh: array of Integer;

  //--------------- local functions -------------------------------------------

  procedure Counters;

  // compute counters and offsets

  var
    S, I: Integer;

  begin
    StateCount := Length(FStateTable);
    TransitionCount := Length(FTransitionTable);
    MarkPositionCount := 0;
    MatchCount := 0;

    // allocate memory for local tables
    SetLength(TransitionsLow, StateCount);
    SetLength(TransitionsHigh, StateCount);
    SetLength(MarksLow, StateCount);
    SetLength(MatchesLow, StateCount);
    SetLength(MarksHigh, StateCount);
    SetLength(MatchesHigh, StateCount);

    for S := 0 to StateCount-1 do
      with FStateTable[S] do
      begin
        TransitionsLow[S] := TransitionLow;
        TransitionsHigh[S] := TransitionHigh;
        MarksLow[S] := MarkPositionCount;
        MatchesLow[S] := MatchCount;
        for I := 0 to Length(StatePosition) - 1 do
          with FPositionTable[StatePosition[I]] do
            if PositionType = ptMark then
            begin
              if Position = 0 then Inc(MatchCount)
                              else
                if Position = 1 then Inc(MarkPositionCount);
            end;
        MarksHigh[S] := MarkPositionCount - 1;
        MatchesHigh[S] := MatchCount - 1;
      end;
  end;

  //---------------------------------------------------------------------------

  function WriteCharacterClass(CC: TCClass): String;

  // print the given character class

    //-------------------------------------------------------------------------

    function CharacterToString(C: Char) : String;

    begin
      case C of
        #0..#31,     // nonprintable characters
        #127..#255:
          Result := '#' + IntToStr(Ord(C));
        '''':
          Result := '''''''''';
        else
          Result := '''' + C + '''';
      end;
    end;

    //-------------------------------------------------------------------------

  var
    Char1,
    Char2: SmallInt;
    Column: Integer;
    Tag: String;

  begin
    Result := '[';
    Column := 0;
    Char1 := 0;
    while Char1 < 256 do
    begin
      if Char(Char1) in CC then
      begin
        if Column > 0 then
        begin
          Result := Result + ', ';
          Inc(Column);

          if Column > 60 then
          begin
            // insert line break
            Result := Result + CR;
            Result := Result + StringOfChar(' ', 17);
            Column := 0;
          end;
        end;

        Char2 := Char1;
        while (Char2 < 255) and (Char(Succ(Char2)) in CC) do Inc(Char2);
        if Char1 = Char2 then Tag := CharacterToString(Char(Char1))
                         else Tag := CharacterToString(Char(Char1))+'..'+CharacterToString(Char(Char2));
        Result := Result + Tag;
        Inc(Column, Length(Tag));
        Char1 := Char2 + 1;
      end
      else Inc(Char1);
    end;
    Result := Result + ']';
  end;

  //---------------------------------------------------------------------------

  procedure PrintTables;

  var
    S, I,
    Index,
    Count: Integer;
    Line: String;

  begin
    Index := FInsetStartLine;
    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, 'type TTransition = record');
    Inc(Index);
    FTarget.Insert(Index, '       CharClass : set of Char;');
    Inc(Index);
    FTarget.Insert(Index, '       NextState : Integer;');
    Inc(Index);
    FTarget.Insert(Index, '     end;');
    Inc(Index);
    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, 'const');
    Inc(Index);

    // table sizes
    FTarget.Insert(Index, '  MarkPositionCount = ' + IntToStr(MarkPositionCount) + ';');
    Inc(Index);
    FTarget.Insert(Index, '  MatchCount        = ' + IntToStr(MatchCount) + ';');
    Inc(Index);
    FTarget.Insert(Index, '  TransitionCount   = ' + IntToStr(TransitionCount) + ';');
    Inc(Index);
    FTarget.Insert(Index, '  StateCount        = ' + IntToStr(StateCount) + ';');
    Inc(Index);

    // mark table
    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, '  MarkPositionTable : array[0..MarkPositionCount - 1] of Integer = (');
    Inc(Index);
    Count := 0;
    Line := '    ';
    for S := 0 to StateCount - 1 do
      with FStateTable[S] do
      begin
        for I := 0 to Length(StatePosition) - 1 do
          with FPositionTable[StatePosition[I]] do
            if (PositionType = ptMark) and (Position = 1) then
            begin
              Line := Line + IntToStr(Rule);
              Inc(Count);
              if Count < MarkPositionCount then Line := Line + ', ';
              if Length(Line) >= 80 then
              begin
                FTarget.Insert(Index, Line);
                Inc(Index);
                Line := '    ';
              end;
            end;
      end;
    FTarget.Insert(Index, Line + ');');
    Inc(Index);

    // match table
    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, '  MatchTable : array [0..MatchCount - 1] of Integer = (');
    Inc(Index);
    Count := 0;
    Line := '    ';
    for S := 0 to StateCount - 1 do
      with FStateTable[S] do
      begin
        for I := 0 to Length(StatePosition) - 1 do
          with FPositionTable[StatePosition[I]] do
            if (PositionType = ptMark) and (Position = 0) then
            begin
              Line := Line + IntToStr(Rule);
              Inc(Count);
              if Count < MatchCount then Line := Line + ', ';
              if Length(Line) >= 80 then
              begin
                FTarget.Insert(Index, Line);
                Inc(Index);
                Line := '    ';
              end;
            end;
    end;
    FTarget.Insert(Index, Line + ');');
    Inc(Index);

    // transition table
    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, '  TransitionTable : array [0..TransitionCount - 1] of TTransition = (');
    Inc(Index);
    Count := 0;
    for S := 0 to StateCount - 1 do
      with FStateTable[S] do
      begin
        for I := TransitionLow to TransitionHigh do
          with FTransitionTable[I] do
          begin
            Line := '    (CharClass: ' + WriteCharacterClass(CC) + '; NextState: ' + IntToStr(NextState) + ')';
            Inc(Count);
            if Count < TransitionCount then Line := Line + ',';
            // consider line breaks
            if Pos(CR, Line) = 0 then
            begin
              FTarget.Insert(Index, Line);
              Inc(Index);
            end
            else
            begin
              while Pos(CR, Line) > 0 do
              begin
                FTarget.Insert(Index, Copy(Line, 1, Pos(CR, Line) - 1));
                Delete(Line, 1, Pos(CR, Line));
                Inc(Index);
              end;
              if Line <> '' then
              begin
                FTarget.Insert(Index, Line);
                Inc(Index);
              end;
            end;
          end;
      end;
    FTarget.Insert(Index, '    );');
    Inc(Index);

    // Offset tables
    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, '  MarksLow : array [0..StateCount-1] of Integer = (');
    Inc(Index);
    Line := '    ';
    for S := 0 to StateCount - 1 do
    begin
      Line := Line + IntToStr(MarksLow[S]);
      if S < StateCount - 1 then Line := Line + ', ';
      if Length(Line) >= 80 then
      begin
        FTarget.Insert(Index, Line);
        Inc(Index);
        Line := '    ';
      end;
    end;
    FTarget.Insert(Index, Line + ');');
    Inc(Index);

    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, '  MarksHigh : array [0..StateCount-1] of Integer = (');
    Inc(Index);
    Line := '    ';
    for S := 0 to StateCount - 1 do
    begin
      Line := Line + IntToStr(MarksHigh[S]);
      if S < StateCount - 1 then Line := Line + ', ';
      if Length(Line) >= 80 then
      begin
        FTarget.Insert(Index, Line);
        Inc(Index);
        Line := '    ';
      end;
    end;
    FTarget.Insert(Index, Line + ');');
    Inc(Index);

    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, '  MatchesLow : array [0..StateCount-1] of Integer = (');
    Inc(Index);
    Line := '    ';
    for S := 0 to StateCount - 1 do
    begin
      Line := Line + IntToStr(MatchesLow[S]);
      if S < StateCount - 1 then Line := Line + ', ';
      if Length(Line) >= 80 then
      begin
        FTarget.Insert(Index, Line);
        Inc(Index);
        Line := '    ';
      end;
    end;
    FTarget.Insert(Index, Line + ');');
    Inc(Index);

    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, '  MatchesHigh : array [0..StateCount-1] of Integer = (');
    Inc(Index);
    Line := '    ';
    for S := 0 to StateCount - 1 do
    begin
      Line := Line + IntToStr(MatchesHigh[S]);
      if S < StateCount - 1 then Line := Line + ', ';
      if Length(Line) >= 80 then
      begin
        FTarget.Insert(Index, Line);
        Inc(Index);
        Line := '    ';
      end;
    end;
    FTarget.Insert(Index, Line + ');');
    Inc(Index);

    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, '  TransitionsLow : array [0..StateCount-1] of Integer = (');
    Inc(Index);
    Line := '    ';
    for S := 0 to StateCount - 1 do
    begin
      Line := Line + IntToStr(TransitionsLow[S]);
      if S < StateCount - 1 then Line := Line + ', ';
      if Length(Line) >= 80 then
      begin
        FTarget.Insert(Index, Line);
        Inc(Index);
        Line := '    ';
      end;
    end;
    FTarget.Insert(Index, Line + ');');
    Inc(Index);

    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, '  TransitionsHigh : array [0..StateCount-1] of Integer = (');
    Inc(Index);
    Line := '    ';
    for S := 0 to StateCount - 1 do
    begin
      Line := Line + IntToStr(TransitionsHigh[S]);
      if S < StateCount - 1 then Line := Line + ', ';
      if Length(Line) >= 80 then
      begin
        FTarget.Insert(Index, Line);
        Inc(Index);
        Line := '    ';
      end;
    end;
    FTarget.Insert(Index, Line + ');');
    Inc(Index);

    FTarget.Insert(Index, '');
    Inc(Index);

    // store last line index for other stuff to be inserted
    FInsetStartLine := Index;
  end;

  //--------------- end local functions ---------------------------------------

begin
  Counters;
  PrintTables;
end;

//----------------- TLexerGenerator ------------------------------------------------------------------------------------

procedure TLexerGenerator.Split(Str: String; Count: Integer);

// Split Str into at most Count whitespace-delimited items
// (result in FItemString, FItemCount, FItemPos, FItemLength)

  //--------------- local functions -------------------------------------------

  procedure Scan(var Position : Integer);

  // scan one item

  var
    L: Integer;

  begin
    while (Position <= Length(FItemString)) and
          ((FItemString[Position] = ' ') or (FItemString[Position] = TAB)) do Inc(Position);
    L := 0;
    while (Position + L <= Length(FItemString)) and
          (FItemString[Position + L] <> ' ')    and
          (FItemString[Position + L] <> TAB)    do Inc(L);
    Inc(FItemCount);
    FItemPos[FItemCount] := Position;
    FItemLength[FItemCount] := L;
    Inc(Position, L + 1);
    while (Position <= Length(FItemString)) and
          ((FItemString[Position] = ' ') or (FItemString[Position] = TAB)) do Inc(Position);
  end;

  //--------------- end local functions ---------------------------------------

var
  Position: Integer;

begin
  FItemString := Str;
  Position := 1;
  FItemCount := 0;
  while (FItemCount < Count - 1) and (Position <= Length(FItemString)) do Scan(Position);
  if Position <= Length(FItemString) then
  begin
    Inc(FItemCount);
    FItemPos[FItemCount] := Position;
    FItemLength[FItemCount] := Length(FItemString) - Position + 1;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.Itemv(I: Integer): String;

// return Ith item in splitted string (whole string for I = 0)

begin
  if I = 0 then Result := FItemString
           else
    if (I < 0) or (I > FItemCount) then Result := ''
                                   else Result := Copy(FItemString, FItemPos[I], FItemLength[I]);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.Code;

begin
  while NextSourceLine do
  begin
    if FCurrentSourceLine = '%}' then Break
                                 else FTarget.Add(FCurrentSourceLine);
  end;
  if FCurrentSourceLine <> '%}' then Error(mtSyntaxError, SUnmatchedLeftbrace, Length(TrimRight(FCurrentSourceLine)), True);
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.CheckID(const Symbol: String): Boolean;

// check whether Symbol is a valid identifier

var
  I: Integer;

begin
  if (Symbol = '') or not (Symbol[1] in Letters) then Result := False
                                                 else
  begin
    Result := True;
    for I := 2 to Length(Symbol) do
      if not (Symbol[I] in Alpha) then
      begin
        CheckID := False;
        Break;
      end;
  end
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.Definitions;

  //--------------- local functions -------------------------------------------

  procedure Definition;

  var
    I: Integer;
    Command : String;

  begin
    Split(FCurrentSourceLine, 2);
    Command := UpperCase(Itemv(1));
    if (Command = '%S') or (Command = '%START') then
    begin
      Split(FCurrentSourceLine, MaxItems);
      for I := 2 to FItemCount do
        if CheckID(Itemv(I)) then DefineStartState(Itemv(I), FItemPos[I])
                             else Error(mtSyntaxError, SIdentifierExpected, FItemPos[I], True);
    end
    else
      if (Command = '%T') or (Command = '%TOKEN') then
      begin
        Split(FCurrentSourceLine, MaxItems);
        for I := 2 to FItemCount do
          if CheckID(Itemv(I)) then DefineToken(Itemv(I), FItemPos[I])
                               else Error(mtSyntaxError, SIdentifierExpected, FItemPos[I], True);
      end
      else
        if CheckID(Itemv(1)) then DefineMacro(Itemv(1), Itemv(2))
                             else Error(mtSyntaxError, SIdentifierExpected, 0, True);
  end;

  //--------------- end local functions ---------------------------------------

begin
  while NextSourceLine do
    if FCurrentSourceLine <> '' then
      if FCurrentSourceLine = SectionDelimiter then Break
                     else
        if FCurrentSourceLine = '%{' then Code
                       else
          if (FCurrentSourceLine[1] = '%') or
             (FCurrentSourceLine[1] in Letters) then Definition
                                                else FTarget.Add(FCurrentSourceLine);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.Rules;

begin
  NextTemplateSection;
  FInsetStartLine := FTarget.Count;
  if FCurrentSourceLine = SectionDelimiter then
  begin
    while NextSourceLine do
    begin
      if FCurrentSourceLine <> '' then
        if FCurrentSourceLine = SectionDelimiter then Break
                                                 else
          if FCurrentSourceLine = '%{' then Code
                         else
            if (FCurrentSourceLine[1] <> ' ') and
               (FCurrentSourceLine[1] <> TAB) then
            begin
              if FRuleCount = 0 then NextTemplateSection;
              ParseRule(FRuleCount);
              if FErrors = 0 then
              begin
                AddRule;
                FTarget.Add('    // ' + FCurrentExpression);
                if Trim(FCurrentStatement) = '|' then FTarget.Add(Format('      %:d,', [FRuleCount]))
                                                 else FTarget.Add(Format('      %:d: %s', [FRuleCount, Trim(FCurrentStatement)]));
              end;
              Inc(FRuleCount);
            end
            else FTarget.Add('      ' + FCurrentSourceLine);
    end;
  end
  else Error(mtError, SUnexpectedEOI, Length(TrimRight(FCurrentSourceLine)), True);
  NextTemplateSection;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.AuxiliaryProcs;

begin
  FTarget.Add('');
  while NextSourceLine do
    FTarget.Add(FCurrentSourceLine);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.MakeDFATable;

var
  I: Integer;

begin
  // initialize start states
  for I := 2 to Length(FFirstPositionTable) - 1 do ISUnion(FFirstPositionTable[I], FFirstPositionTable[I mod 2]);
  for I := 0 to 2 * FStartStatesCount + 1 do MakeNewState(FFirstPositionTable[I]);
  FCurrentState := -1;
  while Succ(FCurrentState) < Length(FStateTable) do
  begin
    Inc(FCurrentState);

    // add transitions for active state
    StartStateTransition;
    for I := 0 to Length(FStateTable[FCurrentState].StatePosition) - 1 do
      with FPositionTable[FStateTable[FCurrentState].StatePosition[I]] do
        if PositionType = ptChar then AddTransition([C], FollowPos)
                                 else
          if PositionType = ptCClass then AddTransition(CClass, FollowPos)
                                     else
            if Position = 0 then FStateTable[FCurrentState].Final := True;

    // assign next states
    I := FStateTable[FCurrentState].TransitionLow;
    while I < Length(FTransitionTable) do
    begin
      with FTransitionTable[I] do NextState := AddState(FollowPos);
      Inc(I);
    end;

    // merge transitions for the same next state
    MergeTransition;
    SortTransition;
    EndStateTransition;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

// Hash table routines

function TLexerGenerator.Lookup(K: Integer) : String;

begin
  Result := FSymbolTable[K].Name;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.Entry(K: Integer; Symbol: String);

begin
  with FSymbolTable[K] do
  begin
    Name := Symbol;
    SymbolType := stNone;
  end
end;

//----------------------------------------------------------------------------------------------------------------------

// Routines to build the Position table

procedure TLexerGenerator.AddCharPos(C: Char);

var
  Count: Integer;

begin
  Count := Length(FPositionTable);
  SetLength(FPositionTable, Count + 1);
  FPositionTable[Count].FollowPos := nil;
  FPositionTable[Count].PositionType := ptChar;
  FPositionTable[Count].C := C;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.AddCClassPos(CC: TCClass);

var
  Count: Integer;

begin
  Count := Length(FPositionTable);
  SetLength(FPositionTable, Count + 1);
  FPositionTable[Count].FollowPos := nil;
  FPositionTable[Count].PositionType := ptCClass;
  FPositionTable[Count].CClass := CC;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.AddMarkPos(Rule, Pos: Integer);

var
  Count: Integer;

begin
  Count := Length(FPositionTable);
  SetLength(FPositionTable, Count + 1);
  FPositionTable[Count].FollowPos := nil;
  FPositionTable[Count].PositionType := ptMark;
  FPositionTable[Count].Rule := Rule;
  FPositionTable[Count].Position := Pos;
end;

//----------------------------------------------------------------------------------------------------------------------

// Routines to build the state table

function TLexerGenerator.MakeNewState(Pos: TIntegerList): Integer;

var
  Count: Integer;

begin
  Count := Length(FStateTable);
  SetLength(FStateTable, Count + 1);
  Result := Count;
  with FStateTable[Count] do
  begin
    StatePosition := Copy(Pos, 0, Length(Pos));
    Final := False;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.AddState(Pos: TIntegerList): Integer;

var
  I: Integer;

begin
  for I := 0 to Length(FStateTable) - 1 do
    if ISEqual(Pos, FStateTable[I].StatePosition) then
    begin
      Result := I;
      Exit;
    end;
  Result := MakeNewState(Pos);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.StartStateTransition;

begin
  FStateTable[FCurrentState].TransitionLow := Length(FTransitionTable);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.EndStateTransition;

begin
  FStateTable[FCurrentState].TransitionHigh := Length(FTransitionTable) - 1;
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.TransitionCount(I: Integer): Integer;

begin
  with FStateTable[I] do
    Result := TransitionHigh - TransitionLow + 1;
end;

//----------------------------------------------------------------------------------------------------------------------

// Construction of the transition table:
// This implementation here uses a simple optimization which tries to avoid
// the construction of different transitions for each individual character
// in large character classes by MERGING transitions whenever possible. The
// transitions, at any time, will be partitioned into transitions on disjoint
// character classes. When adding a new transition on character class CC, we
// repartition the transitions as follows:
// 1. If the current character class cc equals an existing one, we can
//    simply add the new Follow set to the existing one.
// 2. Otherwise, for some existing transition on some character class
//    cc1 with cc*cc1<>[], we replace the existing transition by a new
//    transition on CC*cc1 with Follow set = cc1's Follow set + CC's Follow
//    set, and, if necessary (I.e. if cc1-CC is nonempty), a transition on
//    cc1-cc with Follow set = cc1's Follow set. We then remove the elements
//    of cc1 from cc, and proceed again with step 1.
// We may stop this process as soon as cc becomes empty (then all characters
// in cc have been distributed among the existing partitions). If cc does
// NOT become empty, we have to construct a new transition for the remaining
// character class (which then will be disjoint from all other character
// classes in the transition table).

procedure TLexerGenerator.AddTransition(CC: TCClass; Follow: TIntegerList);

var
  I: Integer;
  cc0,
  cc1,
  cc2: TCClass;
  Count: Integer;
    
begin
  for I := FStateTable[FCurrentState].TransitionLow to Length(FTransitionTable) - 1 do
    if FTransitionTable[I].CC = CC then
    begin
      ISUnion(FTransitionTable[I].FollowPos, Follow);
      Exit;
    end
    else
    begin
      cc0 := CC * FTransitionTable[I].CC;
      if cc0 <> [] then
        begin
          cc1 := FTransitionTable[I].CC - CC;
          cc2 := CC - FTransitionTable[I].CC;
          if cc1 <> [] then
          begin
            FTransitionTable[I].CC := cc1;
            Count := Length(FTransitionTable);
            SetLength(FTransitionTable, Count + 1);
            FTransitionTable[Count].CC := cc0;
            FTransitionTable[Count].FollowPos := nil;
            FTransitionTable[Count].FollowPos := Copy(FTransitionTable[I].FollowPos, 0, Length(FTransitionTable[I].FollowPos));
            ISUnion(FTransitionTable[Count].FollowPos, Follow);
          end
          else
          begin
            FTransitionTable[I].CC := cc0;
            ISUnion(FTransitionTable[I].FollowPos, Follow);
          end;
          CC := cc2;
          if CC=[] then Exit;
        end
    end;

  Count := Length(FTransitionTable);
  SetLength(FTransitionTable, Count + 1);
  FTransitionTable[Count].CC := CC;
  FTransitionTable[Count].FollowPos := nil;
  FTransitionTable[Count].FollowPos := Copy(Follow, 0, Length(Follow));
end;

//----------------------------------------------------------------------------------------------------------------------

// comparison and swap procedures for sorting transitions

function TLexerGenerator.TransitionLessNextState(I, J: Integer): Boolean;

// compare transitions based on next states

begin
  Result := FTransitionTable[I].NextState < FTransitionTable[J].NextState;
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.TransitionLess(I, J: Integer): Boolean;

// lexical order on transitions

var
  C: Char;
  xi,
  xj: Boolean;

begin
  for C := #0 to #255 do
  begin
    xi := C in FTransitionTable[I].CC;
    xj := C in FTransitionTable[J].CC;
    if xi <> xj then
    begin
      Result := xi > xj;
      Exit;
    end;
  end;
  Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.SwapTransition(I, J: Integer);

// swap transitions I and J

var
  Entry: TTransTableEntry;

begin
  Entry := FTransitionTable[I];
  FTransitionTable[I] := FTransitionTable[J];
  FTransitionTable[J] := Entry;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.MergeTransition;

var
  I, J,
  n_deleted: Integer;

begin
  // sort transitions w.r.t. next states
  QuickSort(FStateTable[FCurrentState].TransitionLow, Length(FTransitionTable) - 1, TransitionLessNextState, SwapTransition);

  // merge transitions for the same next state
  n_deleted := 0;
  for I := FStateTable[FCurrentState].TransitionLow to Length(FTransitionTable) - 1 do
    if FTransitionTable[I].CC <> [] then
    begin
      J := Succ(I);
      while (J < Length(FTransitionTable)) and
            (FTransitionTable[I].NextState = FTransitionTable[J].NextState) do
      begin
        // merge CClasses of transitions I and J, then mark
        // transition J as deleted
        FTransitionTable[I].CC := FTransitionTable[I].CC + FTransitionTable[J].CC;
        FTransitionTable[J].CC := [];
        Inc(n_deleted);
        Inc(J);
      end;
    end;

  // remove deleted transitions
  J := FStateTable[FCurrentState].TransitionLow;
  for I := FStateTable[FCurrentState].TransitionLow to Length(FTransitionTable) - 1 do
    if FTransitionTable[I].CC <> [] then
      if I <> J then
      begin
        FTransitionTable[J] := FTransitionTable[I];
        Inc(J);
      end
      else Inc(J);

  // update transition Count
  if n_deleted > 0 then SetLength(FTransitionTable, Length(FTransitionTable) - n_deleted);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.SortTransition;

begin
  QuickSort(FStateTable[FCurrentState].TransitionLow, Length(FTransitionTable) - 1, TransitionLess, SwapTransition);
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.EquivalentStates(I, J: Integer): Boolean;

// checks whether states I and J are equivalent; two states are considered equivalent if:
// - they cover the same marker positions (/ and endmarkers of rules)
// - they have transitions on the same symbols/characters, and corresponding
//   transitions go to states in the same partition
// two different start states are never considered equivalent

var
  II, JJ, K: Integer;
  MarkerPos_I,
  MarkerPos_J: TIntegerList;

begin
  // check for start states
  if (I <= 2 * FStartStatesCount + 1) and
     (J <= 2 * FStartStatesCount + 1) and
     (I <> J) then Result := False
              else
  begin
    // check end positions
    MarkerPos_I := nil;
    with FStateTable[I] do
      for K := 0 to Length(StatePosition) - 1 do
        if FPositionTable[StatePosition[K]].PositionType = ptMark then ISInclude(MarkerPos_I, StatePosition[K]);

    MarkerPos_J := nil;
    with FStateTable[J] do
      for K := 0 to Length(StatePosition) - 1 do
        if FPositionTable[StatePosition[K]].PositionType = ptMark then ISInclude(MarkerPos_J, StatePosition[K]);

    if not ISEqual(MarkerPos_I, MarkerPos_J) then
    begin
      EquivalentStates := False;
      Exit;
    end;

    // check transitions:
    if TransitionCount(I) <> TransitionCount(J) then Result := False
                                                else
    begin
      II := FStateTable[I].TransitionLow;
      JJ := FStateTable[J].TransitionLow;
      for K := 0 to Pred(TransitionCount(I)) do
        if (FTransitionTable[II + K].CC <> FTransitionTable[JJ + K].CC) then
        begin
          Result := False;
          Exit;
        end
        else
          if FStatePartitionTable[FTransitionTable[II + K].NextState] <>
             FStatePartitionTable[FTransitionTable[JJ + K].NextState] then
          begin
            Result := False;
            Exit;
          end;
      Result := True;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.OptimizeDFATable;

var
  I, J, K: Integer;
  CurrentPartition,
  NewPartition,
  NewPartitionCount: Integer;
  OptTransitionCount,
  PartitionCount: Integer;

begin
  // Initially, create one partition containing ALL states:
  PartitionCount := 1;

  SetLength(FStatePartitionTable, Length(FStateTable));
  SetLength(FPartitionTable, PartitionCount + 1);

  FPartitionTable[0] := nil;
  for I := 0 to Length(FStateTable) - 1 do
  begin
    ISInclude(FPartitionTable[0], I);
    FStatePartitionTable[I] := 0;
  end;

  // Now, repeatedly pass over the created partitions, breaking up
  // partitions if they contain nonequivalent states, until no more
  // partitions have been added during the last pass:

  repeat
    NewPartitionCount := 0;
    CurrentPartition := 0;
    NewPartition := PartitionCount;
    FPartitionTable[NewPartition] := nil;
    while (PartitionCount < Length(FStateTable)) and (CurrentPartition < PartitionCount) do
    begin
      for I := 1 to Length(FPartitionTable[CurrentPartition]) - 1 do
        if not EquivalentStates(FPartitionTable[CurrentPartition][0], FPartitionTable[CurrentPartition][I]) then
          // add to new partition:
          ISInclude(FPartitionTable[NewPartition], FPartitionTable[CurrentPartition][I]);

      if Length(FPartitionTable[NewPartition]) <> 0 then
      begin
        // add new partition
        Inc(PartitionCount);
        Inc(NewPartitionCount);
        // remove new partition from old one
        ISMinus(FPartitionTable[CurrentPartition], FPartitionTable[NewPartition]);
        // update partition assignments
        for I := 0 to Length(FPartitionTable[NewPartition]) - 1 do
          FStatePartitionTable[FPartitionTable[NewPartition][I]] := NewPartition;
        SetLength(FPartitionTable, PartitionCount + 1);
        FPartitionTable[NewPartition + 1] := nil;
        Inc(NewPartition);
      end;
      Inc(CurrentPartition);
    end;
  until NewPartitionCount = 0;

  // build the optimized tables
  SetLength(FOptimizedStateTable, PartitionCount);
  OptTransitionCount := 0;
  for I := 0 to PartitionCount - 1 do
  begin
    J := FPartitionTable[I][0];
    FOptimizedStateTable[I] := FStateTable[J];
    with FOptimizedStateTable[I] do
    begin
      TransitionLow := OptTransitionCount;
      TransitionHigh := OptTransitionCount + FStateTable[J].TransitionHigh - FStateTable[J].TransitionLow;
      for K := 1 to Length(FPartitionTable[I]) - 1 do
        ISUnion(StatePosition, FStateTable[FPartitionTable[I][K]].StatePosition);
    end;

    SetLength(FOptimizedTransitionTable, OptTransitionCount + FStateTable[J].TransitionHigh - FStateTable[J].TransitionLow + 1);
    for K := FStateTable[J].TransitionLow to FStateTable[J].TransitionHigh do
    begin
      FOptimizedTransitionTable[OptTransitionCount] := FTransitionTable[K];
      with FOptimizedTransitionTable[OptTransitionCount] do NextState := FStatePartitionTable[NextState];
      Inc(OptTransitionCount);
    end;
  end;

  // update state table
  FStateTable := Copy(FOptimizedStateTable, 0, PartitionCount);
  FTransitionTable := Copy(FOptimizedTransitionTable, 0, OptTransitionCount);

  // free optimization tables
  for I := 0 to Length(FPartitionTable) - 1 do FPartitionTable[I] := nil;
  FPartitionTable := nil;
  for I := 0 to Length(FOptimizedStateTable) - 1 do FOptimizedStateTable[I].StatePosition := nil;
  FOptimizedStateTable := nil;

  for I := 0 to Length(FOptimizedTransitionTable) - 1 do FOptimizedTransitionTable[I].FollowPos := nil;
  FOptimizedTransitionTable := nil;

  FStatePartitionTable := nil;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.Evaluate(R: PRegExpr; var First, Last: TIntegerList; var nullable: Boolean);

// Evaluates the expression R, adding the positions in R to the position
// table and assigning First, Last and Follow sets accordingly
// Returns:
// - First: the set of First positions in R
// - Last: the set of Last positions in R
// - nullable: denotes whether the R is nullable (i.e. is matched by the empty string).

var
  C : Char;
  Str: String;
  CC: TCClass;
  Rule,
  Pos: Integer;
  R1, R2: PRegExpr;
  First1,
  Last1: TIntegerList;
  nullable1: Boolean;
  I: Integer;

begin
  if IsEpsExpr(R) then
  begin
    First := nil;
    Last := nil;
    nullable := True;
  end
  else
    if IsMarkExpr(R, Rule, Pos) then
    begin
      AddMarkPos(Rule, Pos);
      ISSingleton(First, Length(FPositionTable) - 1);
      ISSingleton(Last, Length(FPositionTable) - 1);
      nullable := True;
    end
    else
      if IsCharExpr(R, c) then
      begin
        AddCharPos(c);
        ISSingleton(First, Length(FPositionTable) - 1);
        ISSingleton(Last, Length(FPositionTable) - 1);
        nullable := False
      end
      else
        if IsStrExpr(R, Str) then
          if Length(Str) = 0 then
          begin
            // empty string is treated as empty expression
            First := nil;
            Last := nil;
            nullable := True;
          end
          else
          begin
            AddCharPos(Str[1]);
            ISSingleton(First, Length(FPositionTable) - 1);
            for I := 2 to Length(Str) do
            begin
              AddCharPos(Str[I]);
              ISSingleton(FPositionTable[Length(FPositionTable) - 2].FollowPos, Length(FPositionTable) - 1);
            end;
            ISSingleton(Last, Length(FPositionTable) - 1);
            nullable := False;
          end
                             else
          if IsCClassExpr(R, CC) then
          begin
            AddCClassPos(CC);
            ISSingleton(First, Length(FPositionTable) - 1);
            ISSingleton(Last, Length(FPositionTable) - 1);
            nullable := False;
          end
          else
            if IsStarExpr(R, R1) then
            begin
              Evaluate(R1, First, Last, nullable);
              for I := 0 to Length(Last) - 1 do ISUnion(FPositionTable[Last[I]].FollowPos, First);
              nullable := True;
            end
            else
              if IsPlusExpr(R, R1) then
              begin
                Evaluate(R1, First, Last, nullable);
                for I := 0 to Length(Last) - 1 do ISUnion(FPositionTable[Last[I]].FollowPos, First);
              end
              else
                if IsOptExpr(R, R1) then
                begin
                  Evaluate(R1, First, Last, nullable);
                  nullable := True;
                end
                else
                  if IsConcatExpr(R, R1, R2) then
                  begin
                    Evaluate(R1, First, Last1, nullable);
                    Evaluate(R2, First1, Last, nullable1);
                    for I := 0 to Length(Last1) - 1 do ISUnion(FPositionTable[Last1[I]].FollowPos, First1);
                    if nullable then ISUnion(First, First1);
                    if nullable1 then ISUnion(Last, Last1);
                    nullable := nullable and nullable1;
                  end
                  else
                    if IsAltExpr(R, R1, R2) then
                    begin
                      Evaluate(R1, First, Last, nullable);
                      Evaluate(R2, First1, Last1, nullable1);
                      ISUnion(First, First1);
                      ISUnion(Last, Last1);
                      nullable := nullable or nullable1;
                    end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.AddExpression(R: PRegExpr; var First: TIntegerList);

// Construct the position table, as well as first position sets.
//
// The position table stores symbol positions in regular expressions of
// the Lex grammar. It also allows to store the corresponding first
// and Follow sets. By this means, the position table represents an eps-
// free nondeterministic automaton for the regular expressions of the
// Lex grammar (FCaretFlag. Aho/Sethi/Ullman, Compilers : Principles, Techniques
// and Tools, 1986, Section 3.9, on constructing NFA'S from regular
// expressions using position tables).

var
  Last: TIntegerList;
  nullable: Boolean;

begin
  Evaluate(R, First, Last, nullable);
end;

//----------------------------------------------------------------------------------------------------------------------

// Scanner routines:

// The following routines provide access to the source line and handle
// macro substitutions. To perform macro substitution, an input buffer
// is maintained which contains the rest of the line to be parsed, plus
// any pending macro substitutions. The input buffer is organized as
// a stack onto which null-terminated replacement strings are pushed
// as macro substitutions are processed (the terminating null-character
// is used as an endmarker for macros, in order to keep track of the
// number of pending macro substitutions); characters are popped from the
// stack via calls to the FetchNextParserChar routine.
//
// In order to perform macro substitution, the scanner also has to
// maintain some state information to be able to determine when it
// is scanning quoted characters, strings or character classes
// (no macro substitution is performed in such cases).
//
// The scanner also keeps track of the current source line position in
// variable FCurrentPosition; if there are any macro substitutions on the stack,
// FCurrentPosition will point to the position of the original macro call in the
// source line. This is needed to give proper error diagnostics.

procedure TLexerGenerator.MarkError(ErrorType: TMessageType; Msg: String; Offset: Integer; Source: Boolean);

// mark error position (Offset = Offset of error position (to the left of FCurrentPosition))

begin
  if FMacroCount = 0 then Error(ErrorType, Msg, FCurrentPosition - Offset, Source)
                     else Error(ErrorType, Msg + ' in regular expression definition', FCurrentPosition, Source);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.PushString(Str: String);

// push Str onto input stack

var
  I: Integer;

begin
  Inc(FBufferPosition, Length(Str));
  if FBufferPosition > MacroBufferSize then FatalError(SMacroStackOverflow);
  for I := 1 to Length(Str) do FMacroBuffer[FBufferPosition - I + 1] := Str[I];
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.InitScanner(Source: Boolean);

// initialize the scanner

begin
  FCurrentPosition := 1;
  FBufferPosition := 0;
  FParserStringState := False;
  FParserCClassState := False;
  FParserQuoteState := False;
  FMacroCount := 0;
  if Source then PushString(FCurrentSourceLine)
            else PushString(FCurrentTemplateLine);
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.CurrentParserChar(Source: Boolean): Char;

// current character (#0 if stNone)

  //--------------- local functions -------------------------------------------

  function PushMacro: Boolean;

  // check for macro call at current position in input buffer

    //---------------------------------------------------------------------------

    function ScanMacro(var Name: String): Boolean;

    var
      I: Integer;

    begin
      if (FBufferPosition > 1) and
         (FMacroBuffer[FBufferPosition] = '') and
         (FMacroBuffer[FBufferPosition - 1] in Letters) then
      begin
        Name := '';
        I := FBufferPosition - 1;

        while (I > 0) and (FMacroBuffer[I] in Alpha) do
        begin
          Name := Name + FMacroBuffer[I];
          Dec(I);
        end;

        if (I > 0) and (FMacroBuffer[I] = '') then
        begin
          Result := True;
          Name := Name + '';
          FBufferPosition := I - 1;
        end
        else
        begin
          Result := False;
          MarkError(mtSyntaxError, SMacroEndExpected, -Length(Name), Source);
          FBufferPosition := I;
        end
      end
      else Result := False;
    end;

    //---------------------------------------------------------------------------

  var
    Name: String;

  begin
    if ScanMacro(Name) then
    begin
      Result := True;
      with FSymbolTable[Key(Name, MaxSymbols, Lookup, Entry)] do
        if SymbolType = stMacro then
        begin
          PushString(Substitute + #0);
          Inc(FMacroCount);
        end
        else MarkError(mtSemanticError, SSymbolUndefined, 0, Source);
    end
    else Result := False;
  end;

  //---------------------------------------------------------------------------

  function PopMacro: Boolean;

  // check for macro endmarker

  begin
    if (FBufferPosition > 0) and (FMacroBuffer[FBufferPosition] = #0) then
    begin
      Dec(FBufferPosition);
      Dec(FMacroCount);
      if FMacroCount = 0 then
        if Source then FCurrentPosition := Length(FCurrentSourceLine) - FBufferPosition + 1
                  else FCurrentPosition := Length(FCurrentTemplateLine) - FBufferPosition + 1;
      Result := True;
    end
    else Result := False;
  end;

  //--------------- end local functions ---------------------------------------

begin
  if not (FParserStringState or FParserCClassState or FParserQuoteState) then
    while PushMacro do
      while PopMacro do ;

  if FBufferPosition = 0 then Result := #0
                         else
  begin
    while PopMacro do ;
    if FBufferPosition = 0 then Result := #0
                           else Result := FMacroBuffer[FBufferPosition];
  end
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.FetchNextParserChar(Source: Boolean);

// get next character

begin
  if FBufferPosition > 0 then
  begin
    case FMacroBuffer[FBufferPosition] of
      '\' : FParserQuoteState := not FParserQuoteState;
      '"' : if FParserQuoteState then FParserQuoteState := False
                                 else
              if not FParserCClassState then FParserStringState := not FParserStringState;
      '[' : if FParserQuoteState then FParserQuoteState := False
                                 else
              if not FParserStringState then FParserCClassState := True;
      ']' : if FParserQuoteState then FParserQuoteState := False
                                 else
              if not FParserStringState then FParserCClassState := False;
      else  FParserQuoteState := False;
    end;
    Dec(FBufferPosition);
    if FMacroCount = 0 then
      if Source then FCurrentPosition := Length(FCurrentSourceLine) - FBufferPosition + 1
                else FCurrentPosition := Length(FCurrentTemplateLine) - FBufferPosition + 1;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

// Semantic routines

procedure TLexerGenerator.AddStartState(Symbol: String);

// add start state to FParserStartStates array

begin
  with FSymbolTable[Key(Symbol, MaxSymbols, Lookup, Entry)] do
    if SymbolType = stStartState then
    begin
      SetLength(FParserStartStates, FParserStateCount + 1);
      FParserStartStates[FParserStateCount] := StartState;
      Inc(FParserStateCount);
    end
    else MarkError(mtSemanticError, SSymbolUndefined, Length(Symbol), True);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.ParseRule(RuleNumber: Integer);

  //--------------- local functions -------------------------------------------

  procedure Rule(var Result: Boolean);

  // parse Rule according to syntax:
  //
  // Rule			: StartStatePrefix Caret
  // 				  expression [ '$' | '/' expression ]
  // 				;
  //
  // StartStatePrefix	: /* empty */
  // 				| '<' StartStateList '>'
  // 				;
  //
  // StartStateList         : Identifier { ',' Identifier }
  //                          ;
  //
  // Caret			: /* empty */
  // 				| '^'
  // 				;
  //
  // expression			: Term { '|' Term }
  // 				;
  //
  // Term			: Factor { Factor }
  // 				;
  //
  // Factor			: char
  // 				| string
  // 				| cclass
  // 				| '.'
  // 				| '(' expression ')'
  // 				| Factor '*'
  // 				| Factor '+'
  // 				| Factor '?'
  // 				| Factor '{' num [ ',' num ] '}'
  // 				;

    //-------------------------------------------------------------------------
    
    function StartStatePrefix: Boolean;

      //-----------------------------------------------------------------------

      function StartStateList: Boolean;

        //---------------------------------------------------------------------

        function Identifier: Boolean;

        var
          IDString: String;

        begin
          Result := CurrentParserChar(True) in Letters;
          if Result then
          begin
            IDString := CurrentParserChar(True);
            FetchNextParserChar(True);
            while CurrentParserChar(True) in Alpha do
            begin
              IDString := IDString + CurrentParserChar(True);
              FetchNextParserChar(True);
            end;
            AddStartState(IDString);
          end;
        end;

        //---------------------------------------------------------------------

      begin
        Result := Identifier;
        if Result then
        begin
          while CurrentParserChar(True) = ',' do
          begin
            FetchNextParserChar(True);
            Result := Identifier;
            if not Result then Break;
          end;
        end;
      end;

      //-----------------------------------------------------------------------

    begin
      FParserStateCount := 0;
      if CurrentParserChar(True) = '<' then
      begin
        FetchNextParserChar(True);
        Result := StartStateList;
        if Result then
        begin
          if CurrentParserChar(True) = '>' then
          begin
            Result := True;
            FetchNextParserChar(True);
          end
          else Result := False;
        end;
      end
      else Result := True;
    end;

    //-------------------------------------------------------------------------

    function Caret: Boolean;

    begin
      Result := True;
      FCaretFlag := CurrentParserChar(True) = '^';
      if CurrentParserChar(True) = '^' then FetchNextParserChar(True);
    end;

  //---------------------------------------------------------------------------

  function ScanCharacter(var C: Char): Boolean;

  var
    Octal: Byte;
    Count: Integer;

  begin
    Result := True;
    if CurrentParserChar(True) = '\' then
    begin
      FetchNextParserChar(True);
      case CurrentParserChar(True) of
        #0:
          Result := False;
        'n':
          begin
            C := LF;
            FetchNextParserChar(True);
          end;
        'r':
          begin
            C := CR;
            FetchNextParserChar(True);
          end;
        't':
          begin
            C := TAB;
            FetchNextParserChar(True);
          end;
        'b':
          begin
            C := BS;
            FetchNextParserChar(True);
          end;
        'f':
          begin
            C := FF;
            FetchNextParserChar(True);
          end;
        '0'..'7':
          begin
            Octal := Ord(CurrentParserChar(True)) - Ord('0');
            FetchNextParserChar(True);
            Count := 1;
            while ('0' <= CurrentParserChar(True)) and
                 (CurrentParserChar(True) <= '7') and
                  (Count < 3) do
            begin
              Inc(Count);
              Octal := Octal*8 + Ord(CurrentParserChar(True)) - Ord('0');
              FetchNextParserChar(True);
            end;
            C := Chr(Octal);
          end;
      else
        C := CurrentParserChar(True);
        FetchNextParserChar(True);
      end
    end
    else
    begin
      C := CurrentParserChar(True);
      FetchNextParserChar(True);
    end;
  end;

  //---------------------------------------------------------------------------

  function ScanString(var Str: String): Boolean;

  var
    C: Char;

  begin
    Str := '';
    FetchNextParserChar(True);
    while (CurrentParserChar(True) <> #0) and (CurrentParserChar(True) <> '"') do
    begin
      Result := ScanCharacter(C);
      if not Result then Exit;
      Str := Str + C;
    end;

    if CurrentParserChar(True) = #0 then Result := False
                                    else
    begin
      FetchNextParserChar(True);
      Result := True;
    end
  end;

  //---------------------------------------------------------------------------

  function ScanCharacterClass(var CC: TCClass): Boolean;

  // scan a character class

  var
    Caret: Boolean;
    C,
    Char1: Char;

  begin
    CC := [];
    FetchNextParserChar(True);
    if CurrentParserChar(True) = '^' then
    begin
      Caret := True;
      FetchNextParserChar(True);
    end
    else Caret := False;

    while (CurrentParserChar(True) <> #0) and (CurrentParserChar(True) <> ']') do
    begin
      Result := ScanCharacter(c);
      if not Result then Exit;
      if CurrentParserChar(True) = '-' then
      begin
        FetchNextParserChar(True);
        if (CurrentParserChar(True) <> #0) and (CurrentParserChar(True) <> ']') then
        begin
          Result := ScanCharacter(Char1);
          if not Result then Exit;
          CC := CC + [C..Char1];
        end
        else CC := CC + [C, '-'];
      end
      else CC := CC + [C];
    end;

    if CurrentParserChar(True) = #0 then Result := False
                              else
    begin
      FetchNextParserChar(True);
      Result := True;
    end;

    if Caret then CC := [#1..#255] - cc;
  end;

  //---------------------------------------------------------------------------

  function ScanNumber(var N: Integer): Boolean;

  var
    Str: String;

  begin
    if CurrentParserChar(True) in Digits then
    begin
      Str := CurrentParserChar(True);
      FetchNextParserChar(True);
      while CurrentParserChar(True) in Digits do
      begin
        Str := Str + CurrentParserChar(True);
        FetchNextParserChar(True);
      end;
      Result := IsInt(Str, N);
    end
    else Result := False;
  end;

  //---------------------------------------------------------------------------

  function Expression(var R: PRegExpr): Boolean;

    //-------------------------------------------------------------------------

    function Term(var R: PRegExpr): Boolean;

      //-----------------------------------------------------------------------

      function Factor(var R: PRegExpr): Boolean;

      var
        Str: String;
        CC: TCClass;
        C: Char;
        N, M: Integer;

      begin
        case CurrentParserChar(True) of
          '"':
            begin
              Result := ScanString(Str);
              if Result then R := MakeStringNode(Str);
            end;
          '[':
            begin
              Result := ScanCharacterClass(CC);
              if Result then R := MakeCClassNode(CC);
            end;
          '.':
            begin
              FetchNextParserChar(True);
              R := MakeCClassNode([#1..#255] - [LF]);
              Result := True;
            end;
          '(':
            begin
              FetchNextParserChar(True);
              Result := Expression(R);
              if Result and (CurrentParserChar(True) = ')') then
              begin
                FetchNextParserChar(True);
                Result := True;
              end
              else Result := False;
            end;
        else
          Result := ScanCharacter(C);
          if Result then R := MakeCharNode(C);
        end;

        while Result and (CurrentParserChar(True) in ['*', '+', '?', '{']) do
          case CurrentParserChar(True) of
            '*':
              begin
                FetchNextParserChar(True);
                R := MakeStarNode(R);
              end;
            '+':
              begin
                FetchNextParserChar(True);
                R := MakePlusNode(R);
              end;
            '?':
              begin
                FetchNextParserChar(True);
                R := MakeOptionalNode(R);
              end;
            '{':
              begin
                FetchNextParserChar(True);
                Result := ScanNumber(M);
                if Result then
                begin
                  if CurrentParserChar(True) = ',' then
                  begin
                    FetchNextParserChar(True);
                    Result := ScanNumber(N);
                    if not Result then Exit;
                    R := ExpandMNExpression(R, M, N);
                  end
                  else R := ExpandMNExpression(R, M, M);
                  if CurrentParserChar(True) = '}' then
                  begin
                    FetchNextParserChar(True);
                    Result := True;
                  end
                  else Result := False;
                end;
              end;
          end
        end;

      //-----------------------------------------------------------------------

      const
        Delimiter: TCClass = [#0, ' ', TAB, '$', '|', ')', '/'];

      var
        R1: PRegExpr;

      begin
        if not (CurrentParserChar(True) in Delimiter) then
        begin
          Result := Factor(R);
          if Result then
            while not (CurrentParserChar(True) in Delimiter) do
            begin
              Result := Factor(R1);
              if not Result then Break;
              R := MakeConcatExpression(R, R1);
            end;
        end
        else
        begin
          R := EpsEpr;
          Result := True;
        end;
      end;

    //-------------------------------------------------------------------------

    var
      R1: PRegExpr;

    begin
      Result := Term(R);
      if Result then
        while CurrentParserChar(True) = '|' do
        begin
          FetchNextParserChar(True);
          Result := Term(R1);
          if not Result then Break;
          R := MakeAlternativeExpression(R, R1);
        end
    end;

  //---------------------------------------------------------------------------

  var
    R1, R2: PRegExpr;

  begin
    Result := StartStatePrefix;
    if Result then
    begin
      Result := Caret;
      if Result then
      begin
        Result := Expression(R1);
        if Result then
        begin
          if CurrentParserChar(True) = '$' then
          begin
            FParsedExpression := MakeConcatExpression(MakeConcatExpression(R1, MakeMarkNode(RuleNumber, 1)), MakeCClassNode([LF]));
            FetchNextParserChar(True);
          end
          else
            if CurrentParserChar(True) = '/' then
            begin
              FetchNextParserChar(True);
              Result := Expression(R2);
              if not Result then Exit;
              FParsedExpression := MakeConcatExpression(MakeConcatExpression(R1, MakeMarkNode(RuleNumber, 1)), R2);
            end
          else FParsedExpression := MakeConcatExpression(R1, MakeMarkNode(RuleNumber, 1));

          FParsedExpression := MakeConcatExpression(FParsedExpression, MakeMarkNode(RuleNumber, 0));
          Result := CurrentParserChar(True) in [#0, ' ', TAB];
        end;
      end;
    end;
  end;

  //--------------- end local functions ---------------------------------------

var
  Result: Boolean;

begin
  InitScanner(True);
  Rule(Result);
  if Result then
  begin
    FCurrentExpression := Copy(FCurrentSourceLine, 1, FCurrentPosition - 1);
    FCurrentStatement := Copy(FCurrentSourceLine, FCurrentPosition, Length(FCurrentSourceLine));
  end
  else MarkError(mtSyntaxError, SErrorInRule, 0, True);
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.StripComment(const Line: String): String;

var
  I: Integer;
  InString: Boolean;

begin
  if Line = '' then Result := ' '
               else
  begin
    Result := TrimRight(Line);
    // consider string constants before stripping out the comment
    I := 1;
    InString := False;
    while I <= Length(Result) do
    begin
      if not InString and (Result[I] = '') then
      begin
        Delete(Result, I, Length(Result));
        Break;
      end;
      if Result[I] = '"' then InString := not InString;
      Inc(I);
    end;
    Result := TrimRight(Result);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.NextSourceLine: Boolean;

begin
  if FSourceIndex < Cardinal(FSource.Count) then
  begin
    Result := True;
    FCurrentSourceLine := StripComment(FSource[FSourceIndex]);
    if assigned(FOnSourceLine) then FOnSourceLine(Self, FSource[FSourceIndex], FSourceIndex + 1);
    Inc(FSourceIndex);
  end
  else Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.NextTemplateLine: Boolean;

begin
  if FTemplateIndex < Cardinal(FTemplate.Count) then
  begin
    Result := True;
    FCurrentTemplateLine := StripComment(TrimRight(FTemplate[FTemplateIndex]));
    if assigned(FOnTemplateLine) then FOnTemplateLine(Self, FTemplate[FTemplateIndex], FTemplateIndex + 1);
    Inc(FTemplateIndex);
  end
  else Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.Globals;

// read in global definitions to direct the work of the generator

var
  Symbol,
  Value: String;

begin
  while NextSourceLine do
  begin
    if FCurrentSourceLine = SectionDelimiter then Break;

    Value := Trim(FCurrentSourceLine);
    if Value > '' then
    begin                                                      
      Split(Value, 2);
      Symbol := UpperCase(Itemv(1));
      Value := Itemv(2);
      if (Symbol = '%CS') or (Symbol = '%CASESENSITIVE') then
      begin
        Value := UpperCase(Value);
        if (Value = 'T') or (Value = 'TRUE') then FCaseSensitive := True
                                             else FCaseSensitive := False;
      end
      else
      begin
        if not CheckID(Symbol) then Error(mtSyntaxError, SIdentifierExpected, FItemPos[1], True)
                               else DefineMacro(Itemv(1), Value);
      end;
    end;
  end;
  FKeywords.CaseSensitive := FCaseSensitive;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.Keywords;

var
  Keyword: String;

begin
  while NextSourceLine do
  begin
    if FCurrentSourceLine = SectionDelimiter then Break;
    Keyword := Trim(FCurrentSourceLine);
    if Keyword > '' then FKeywords.Add(Keyword);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TLexerGenerator.ReplaceInvalidChars(const S: String): String;

// converts characters not allowed in Delphi identifiers into their literal equivalent

var
  I: Integer;
  C: Char;

begin
  I := 1;
  Result := S;
  while I <= Length(Result) do
  begin
    // invalid character found?
    if Result[I] in InvalidChars then
    begin
      // remove the char in question...
      C := Result[I];
      Delete(Result, I, 1);
      // ...and insert a textual expression for this particular character
      case C of
        '#':
          Insert('Pound', Result, I);
        '$':
          Insert('Dollar', Result, I);
        '&':
          Insert('Ampersand', Result, I);
        '''':
          Insert('Apostrophe', Result, I);
        '(':
          Insert('OpenParens', Result, I);
        ')':
          Insert('CloseParens', Result, I);
        '*':
          Insert('Star', Result, I);
        '+':
          Insert('Plus', Result, I);
        ',':
          Insert('Comma', Result, I);
        '':
          Insert('Minus', Result, I);
        '.':
          Insert('Period', Result, I);
        '/':
          Insert('Slash', Result, I);
        ':':
          Insert('Colon', Result, I);
        ';':
          Insert('Semicolon', Result, I);
        '<':
          Insert('OpenAngle', Result, I);
        '=':
          Insert('Equal', Result, I);
        '>':
          Insert('CloseAngle', Result, I);
        '@':
          Insert('At', Result, I);
        '[':
          Insert('OpenBracket', Result, I);
        ']':
          Insert('CloseBracket', Result, I);
        '^':
          Insert('Caret', Result, I);
        '{':
          Insert('OpenBrace', Result, I);
        '}':
          Insert('CloseBrace', Result, I);
        '%':
          Insert('Score', Result, I);
        '!':
          Insert('ExclamMark', Result, I);
        '?':
          Insert('QuestionMark', Result, I);
        '~':
          Insert('Tilde', Result, I);
        '':
          Insert('Circle', Result, I);
      end;
    end;
    Inc(I);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.WriteKeywordSupport;

var
  I,
  Index: Integer;
  Line,
  KW: String;

begin
  Index := FInsetStartLine;

  if FKeywords.Count > 0 then
  begin
    // print declartions
    FTarget.Insert(Index, 'type // keyword table support');
    Inc(Index);
    FTarget.Insert(Index, '     TKeyword = record');
    Inc(Index);
    FTarget.Insert(Index, '       Keyword: String;');
    Inc(Index);
    FTarget.Insert(Index, '       Token: Integer;');
    Inc(Index);
    FTarget.Insert(Index, '     end;');
    Inc(Index);
    FTarget.Insert(Index, '');
    Inc(Index);
    FTarget.Insert(Index, 'const KeywordCount = ' + IntToStr(FKeywords.Count) + ';');
    Inc(Index);
    FTarget.Insert(Index, '      Keywords : array[0..KeywordCount - 1] of TKeyword = (');
    Inc(Index);

    // print keyword table
    for I := 0 to FKeywords.Count - 1 do
    begin
      KW := ReplaceInvalidChars(FKeywords[I]);
      Line := '        (Keyword: ''' + KW + '''; Token: kw'+ UpperCase(KW) + ')';
      if I < FKeywords.Count - 1 then Line := Line + ', ';
      FTarget.Insert(Index, Line);
      Inc(Index);
    end;
    FTarget.Insert(Index, ');');
    Inc(Index);

    // print keyword search routine
    FTarget.Insert(Index, '');                                                         Inc(Index);
    FTarget.Insert(Index, 'function GetKeywordToken(const S: String): Integer;');      Inc(Index);
    FTarget.Insert(Index, '');                                                         Inc(Index);
    FTarget.Insert(Index, '// internal search function for keywords');                 Inc(Index);
    FTarget.Insert(Index, '');                                                         Inc(Index);
    FTarget.Insert(Index, 'var L, R, K : Integer;');                                   Inc(Index);
    FTarget.Insert(Index, '    Res     : Integer;');                                   Inc(Index);
    FTarget.Insert(Index, '');                                                         Inc(Index);
    FTarget.Insert(Index, 'begin');                                                    Inc(Index);
    FTarget.Insert(Index, '  Result := 0;');                                           Inc(Index);
    FTarget.Insert(Index, '');                                                         Inc(Index);
    FTarget.Insert(Index, '  // binary search');                                       Inc(Index);
    FTarget.Insert(Index, '  L := 0;');                                                Inc(Index);
    FTarget.Insert(Index, '  R := KeywordCount - 1;');                                 Inc(Index);
    FTarget.Insert(Index, '  while L <= R do');                                        Inc(Index);
    FTarget.Insert(Index, '  begin');                                                  Inc(Index);
    FTarget.Insert(Index, '    K := L + (R - L) div 2;');                              Inc(Index);
    if FCaseSensitive then
    begin
      FTarget.Insert(Index, '    Res := CompareStr(S, Keywords[K].Keyword);');         Inc(Index);
    end
    else
    begin
      FTarget.Insert(Index, '    Res := CompareText(S, Keywords[K].Keyword);');        Inc(Index);
    end;
    FTarget.Insert(Index, '    if Res = 0 then');                                      Inc(Index);
    FTarget.Insert(Index, '    begin');                                                Inc(Index);
    FTarget.Insert(Index, '      Result := Keywords[K].Token;');                       Inc(Index);
    FTarget.Insert(Index, '      Break;');                                             Inc(Index);
    FTarget.Insert(Index, '    end');                                                  Inc(Index);
    FTarget.Insert(Index, '    else');                                                 Inc(Index);
    FTarget.Insert(Index, '      if Res > 0 then L := K + 1');                         Inc(Index);
    FTarget.Insert(Index, '                 else R := K - 1;');                        Inc(Index);
    FTarget.Insert(Index, '  end;');                                                   Inc(Index);
    FTarget.Insert(Index, 'end;');                                                     Inc(Index);

  end
  else
  begin
    // print dummy keyword search routine
    FTarget.Insert(Index, '');                                                         Inc(Index);
    FTarget.Insert(Index, 'function GetKeywordToken(const S: String): Integer;');      Inc(Index);
    FTarget.Insert(Index, '');                                                         Inc(Index);
    FTarget.Insert(Index, '// internal search function for keywords,');                Inc(Index);
    FTarget.Insert(Index, '// since no keywords are defined the result is always 0');  Inc(Index);
    FTarget.Insert(Index, '');                                                         Inc(Index);
    FTarget.Insert(Index, 'begin');                                                    Inc(Index);
    FTarget.Insert(Index, '  Result := 0;');                                           Inc(Index);
    FTarget.Insert(Index, 'end;');                                                     Inc(Index);
  end;

  FInsetStartLine := Index;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.KewordConstants;

var
  I: Integer;
  Line: String;

begin
  if FKeyWords.Count > 0 then
  begin
    // print keyword constant definitions
    FTarget.Add('const kwFirstKeyword = ' + IntToStr(FCurrentTokenID) + ';'); // range start
    for I := 0 to FKeywords.Count - 1 do
    begin
      Line := 'kw' + ReplaceInvalidChars(UpperCase(FKeywords[I])) + ' = ' + IntToStr(FCurrentTokenID) + ';';
      Inc(FCurrentTokenID);
      Line := '      ' + Line;
      FTarget.Add(Line);
    end;
    FTarget.Add('      kwLastKeyword = ' + IntToStr(FCurrentTokenID - 1) + ';'); // range end
    FTarget.Add('');
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TLexerGenerator.TokenConstants;

var
  I: Integer;
  Line: String;

begin
  if FTokens.Count > 0 then
  begin
    // print token constant definitions
    for I := 0 to FTokens.Count - 1 do
    begin
      Line := FTokens[I] + ' = ' + IntToStr(FCurrentTokenID) + ';';
      Inc(FCurrentTokenID);
      if I =  0 then Line := 'const ' + Line
                else Line := '      ' + Line;
      FTarget.Add(Line);
    end;
    FTarget.Add('');
    FTarget.Add('');
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

end.
