unit ParserGenerator;

// 05-DEC-99 ml: structural rework
// 20-JUN-99 ml: last change
//----------------------------------------------------------------------------------------------------------------------
// This unit contains a class used to create a parser class.

interface

uses
  Windows, Classes, DCGTools;

const MaxDepth = 1024;

type TParserAcceptance = (peNone, peAccept, peAbort, peError);

// Maximum table sizes: 
const MaxKeys           =  997;  // Length of hash symbol table(prime number!)
      MaxNonterminals   =  300;  // maximum number of nonterminals
      MaxLiterals       =  556;  // number of literals(300 + 256)
      MaxRules          =  501;  // number of rules(500 + 1)
      MaxTypeTags       =  100;  // number of type tags
      MaxPrecLevel      =   50;  // maximum precedence level
      MaxStates         = 1200;  // number of LR(0) states
      MaxItems          = 4800;  // number of items
      MaxTransitions    = 4800;  // number of transitions
      MaxReductions     = 2400;  // number of reductions

      MaxRuleLength     =  128;  // maximum length of rules
      MaxSetItems       =  128;  // maximum number of items in an item set

type TParserGenerator = class;

     // table data structures

     // Symbol table: The symbol table consists of a hash table which stores
     // print names and internal symbol numbers, and a Key table which stores,
     // for each internal symbol number, the corresponding hash Key.

     TSymbolEntry = record
       Name: String; // print name; empty entries are denoted by ''
       defined : Boolean; // flag denoting whether symbol is already defined
       Symbol: Integer;
             // internal symbol number(0 or positive: literal symbols
             //(literal characters have symbol numbers 1 thru 255);
             // negative: non-terminal symbols; 0 denotes endmarker,
             // - 1 augmented start non-terminal, 256 is reserved for
             // Error token; note that the predefined symbols except
             // the Error literal are not actually stored in the symbol
             // table; the Error symbol is entered at initialization
             // so that it always has number 256)
     end;

     TSymbolTable = array[1..MaxKeys] of TSymbolEntry;

     // hash keys for non-terminal and literal symbols
     TSymbolKeyTable = array[-MaxNonterminals..MaxLiterals - 1] of Integer;

     // Rule table: the rule table consists of an array storing the rules
     // sequentially in the order in which they appear in the source grammar;
     // a rule numbers table which is used to sort rules with regard to left-hand side
     // nonterminals (after the rule table has been constructed and sorted, all
     // references to rules are done indirectly via the RuleNumber array so that the
     // rules for each non-terminal can be accessed easily); and an offset table
     // which stores, for each non-terminal, the corresponding first and last
     // index in the rule numbers table.

     TRuleEntry = record
       lhs_sym: Integer; // lhs non-terminal
       rhs_len: Integer; // Length of rhs
       rhs_sym: array[1..MaxRuleLength] of Integer; // rhs symbols
     end;

     TRuleTable = array[1..MaxRules] of TRuleEntry;

     TRuleNumberTable = array[1..MaxRules] of Integer;

     TRuleOffsetEntry = record
       RuleLow,
       RuleHigh: Integer;
     end;

     TRuleOffsetTable = array[1..MaxNonterminals] of TRuleOffsetEntry;

     // Symbol type table: The symbol type table stores the types associated
     // with the non-terminal and terminal grammar symbols(0 if none).

     // types declared in the Definitions section
     TTypeTable = array[1..MaxTypeTags] of Integer;

     // symbol types
     TSymbolTypeTable = array[-MaxNonterminals..MaxLiterals - 1] of Integer;

     // Precedence table: The precedence table stores the type of each
     // precedence level(ptLeft, ptRight, ptNonAssoc) and, for each literal
     // symbol and grammar rule, the assigned precedence level(precedence
     // level 0 if none).

     TPrecedenceType = (ptLeft, ptRight, ptNonAssoc);

     TPrecendenceTable = array[1..MaxPrecLevel] of TPrecedenceType;

     TSymbolPrecendenceTable = array[0..MaxLiterals - 1] of Integer;

     TRulePrecendenceTable = array[1..MaxRules] of Integer;

     // Closure and first symbols table: The Closure table stores, for each
     // non-terminal X, the set of those nonterminals Y for which there is a
     // rightmost derivation X => +  Y ... . Similarly, the first set table
     // stores, for each non-terminal X, the set of literals a for which there
     // is a derivation X => +  a ... . Both WriteTables are of type TSymbolSetTable.
     //
     // The nullable table stores, for each non-terminal, a flag denoting whether
     // the non-terminal is nullable (i.e. may be derived to the empty string).
     TSymbolSetTable = array[1..MaxNonterminals] of TIntegerList;

     TNullableTable = array[1..MaxNonterminals] of Boolean;


     // State table:
     // Each state stores the first and last index of the kernel items,
     // transitions and reductions belonging to it, and a hash Key determined
     // from the kernel items which is used to speed up searches for existing
     // states.
     //
     // The items table stores the individual kernel items in the LR(0) set.
     // Each Entry consists of a rule number together with the Item position,
     // and a "Next" field indicating the associated Item in the successor state
     // (0 if there is none). The TItemSetEntry type is used to retrieve and manipulate
     // individual Item sets from the Item table.
     //
     // The transition table stores the shift and goto transitions in each state
     // (each transition is denoted by a(symbol, NextState) pair). Similarly,
     // the reductions table stores the reductions in each state, where each
     // action is denoted by a(symbolset, ruleno) pair.
     TStateEntry = record
       ItemLow,
       ItemHigh: Integer;
       TransitionLow,
       TransitionHigh: Integer;
       ReductionsLow,
       ReductionsHigh: Integer;
       Key: Integer;
     end;

     TStateTable = array[0..MaxStates - 1] of TStateEntry;

     TItemEntry = record
       RuleNumber,
       PositionNumber: Integer;
       Next: Integer;
     end;

     TItemSetEntry = record
       ItemCount: Integer;
       Item: array[1..MaxSetItems] of TItemEntry;
     end;

     TItemTable = array[1..MaxItems] of TItemEntry;

     TTransitionEntry = record
       Symbol,
       NextState: Integer;
     end;

     TTransitionTable = array[1..MaxTransitions] of TTransitionEntry;

     TReductionEntry = record
       SymbolSet: TIntegerList;
       RuleNumber: Integer;
     end;

     TReductionTable = array[1..MaxReductions] of TReductionEntry;

     // lookaheads table: This table stores, for each kernel item, the
     // corresponding LALR(1) lookahead symbol sets.
     TLookaheadTable = array[1..MaxItems] of TIntegerList;

     // The propagation table is used to keep track of how Lookaheads are
     // propagated from kernel items to other lookahead sets.
     PPropagationEntry = ^TPropagationEntry;
     TPropagationEntry = record
       SymbolSet: TIntegerList;  
       Next: PPropagationEntry;
     end;

     TPropagationTable = array[1..MaxItems] of PPropagationEntry;

     TSymbolSetArray = array[1..MaxSetItems] of TIntegerList;

     TParserGenerator = class
     private
       FSource,
       FTemplate: TStrings;
       FTarget: TStringList;
       FCurrentSourceLine,
       FCurrentTemplateLine: String;             // current input lines without parser comments
       FCharIndex: Integer;
       FSourceIndex,                             // index into FSource
       FTemplateIndex,                           // dito for FTemplate
       FRuleCount,
       FTypeCount,
       FLiteralsCount,
       FPrecedenceCount,
       FItemCount,
       FTransitionCount,
       FReductionCount,
       FErrors: Cardinal;
       FStateCount,
       FNonTerminalCount: Integer;
       FReport: Boolean;                         // denotes report mode
       FGenerating: Boolean;

       // work tables
       FStartNonTerminal: Integer;               // start non-terminal of grammar (0 if undefined)
       FSymbolTable: ^TSymbolTable;              // symbol table
       FSymbolKeyTable: ^TSymbolKeyTable;        // symbol keys
       FRuleTable: ^TRuleTable;                  // rule table
       FTypeTable: ^TTypeTable;                  // type table
       FSymbolTypeTable: ^TSymbolTypeTable;      // symbol types
       FPrecendenceTable: ^TPrecendenceTable;    // precedence table
       FSymbolPrecendenceTable: ^TSymbolPrecendenceTable; // literal symbols precedence
       FRulePrecendenceTable: ^TRulePrecendenceTable;     // rules precedence
       FRuleNumberTable: ^TRuleNumberTable;      // rule number table
       FRuleOffsetTable: ^TRuleOffsetTable;      // rule offset table
       FClosureTable: ^TSymbolSetTable;          // closure table
       FFirstSetTable: ^TSymbolSetTable;         // first set table
       FNullableTable: ^TNullableTable;          // nullable flags table
       FStateTable: ^TStateTable;                // LR(0) state table
       FItemTable: ^TItemTable;                  // LR(0) kernel Item table
       FTransitionTable: ^TTransitionTable;      // transition table
       FReductionTable: ^TReductionTable;        // reduction table
       FLookaheadTable: ^TLookaheadTable;        // LALR lookaheads table
       FPropagationTable: ^TPropagationTable;    // lookahead propagation table
       FCurrentPrecedence: Integer;              // active precedence level in token and precedence
                                                 // declarations (0 in %token declaration)
       FCurrentTypeTag: Integer;                 // active type Tag in token, precedence and type declarations
       FCurrentState: Integer;                   // state currently considered
       FCurrentTokenLen: Integer;                // Length of current token
       FCurrentRule: TRuleEntry;
       FActionCount: Integer;
       FActionMarker: Boolean;
       FTempItems: TItemSetEntry;                // used while sorting itemsq

       // data structures used in lookahead computation
       FItemSet: TItemSetEntry;
       FLookaheadSet: TSymbolSetArray;
       FKernelItemsCount: Integer;

       // data structures used during parse table generation
       FShiftReduceConflicts,
       FReduceReduceConflicts,
       FNeverReduced: Integer;                   // number of parsing conflicts and unreduced rules detected
       FGeneratorActionCount,
       FGeneratorGotoCount,
       FGeneratorStateCount: Cardinal;
       FGeneratorDefaultTable: array[0..MaxStates - 1] of Integer;
       FGeneratorActionLowTable,
       FGeneratorActionHighTable,
       FGeneratorGotoLowTable,
       FGeneratorGotoHighTable: array[0..MaxStates - 1] of Integer;

       // parser-parser variables
       FLastError: Integer;
       FAcceptance: TParserAcceptance;
       FLastValue,
       FCurrentValue: Variant;
       FParserState,
       FStackPointer,
       FNextAction: Integer;
       FStack: array[1..MaxDepth] of Integer;
       FValues: array[1..MaxDepth] of Integer;
       FLastToken: Integer;

       // events for application callbacks
       FOnSourceLine,
       FOnTemplateLine: TOnProcessLineEvent;
       FOnMessage: TOnMessageEvent;
       procedure Error(Msg: String; Pos: Integer; InSource: Boolean);
       procedure FatalError(Msg: String);
       procedure RuleToAction(Rule: Integer);
       procedure SortTypes;
       function CompareReductions(I, J: Integer): Boolean;
       procedure SwapReductions(I, J: Integer);
       function CompareRules(I, J: Integer): Boolean;
       procedure SwapRules(I, J: Integer);
       function CompareItems(I, J: Integer): Boolean;
       procedure SwapItems(I, J: Integer);
       function CompareTypes(I, J: Integer): Boolean;
       procedure SwapTypes(I, J: Integer);
       function CurrentParserChar: Char;
       procedure AddAction;
       procedure AddItem(RuleNumber, PositionNumber: Integer);
       procedure AddLiteral(Symbol: Integer);
       procedure AddNextLinks;
       procedure AddReduction(SymbolSet: TIntegerList; RuleNumber: Integer);
       procedure AddRule(R: TRuleEntry);
       procedure AddRuleAction;
       procedure AddRulePrecedence(Symbol: Integer);
       function AddState: Integer;
       procedure AddSymbol(Symbol: Integer);
       procedure AddTransition(Symbol, NextState: Integer);
       procedure AddType(K: Integer);
       procedure BuildParseTables;
       procedure Closure(var ItemSet: TItemSetEntry);
       procedure CopyAction;
       procedure CopyCode;
       procedure CopyRestOfFile;
       procedure CopySingleAction;
       procedure InitializeCounters;
       procedure DefineKey(K, Symbol: Integer);
       procedure Definitions;
       procedure EndBody;
       procedure EndReductions;
       procedure EndTransitions;
       procedure Entry(K: Integer; Symbol: String);
       function FindItem(S, RuleNumber, PositionNumber: Integer): Integer;
       procedure GenerateParser;
       procedure GetItemSet(s: Integer; var ItemSet: TItemSetEntry);
       function GetKey(Symbol: String): Integer;
       procedure GetSymbols(var ItemSet: TItemSetEntry; var SymbolSet: TIntegerList);
       procedure IncludeItemSet(var ItemSet: TItemSetEntry; RuleNumber, PositionNumber: Integer);
       procedure InitLookaheads;
       function IsDefinedKey(K: Integer; var Symbol: Integer): Boolean;
       function ItemToStr(var ItemSet: TItemSetEntry; I: Integer): String;
       function LiteralSymbol(K, N: Integer): Integer;
       function LookaheadChar: Char;
       procedure Lookaheads;
       function Lookup(K: Integer): String;
       procedure LR0Set;
       function MakeState(var ItemSet: TItemSetEntry; Symbol: Integer): Integer;
       function StateItemCount(s: Integer): Integer;
       function MakeNewLiteral: Integer;
       function MakeNewNonTerminal: Integer;
       function MakeNewPrecedenceLevel(PrecedenceType: TPrecedenceType): Integer;
       procedure MakeNewState;
       procedure NextChar;
       function NextSourceLine: Boolean;
       procedure NextTemplateSection;
       function NonTerminalSymbol(K: Integer): Integer;
       procedure ParseTable;
       function Name(Symbol: Integer): String;
       procedure Propagate;
       procedure RuleOffsets;
       function RuleToStr(I: Integer): String;
       function ScanValue: String;
       function SearchType(Symbol: String): Boolean;
       procedure SortReductions;
       procedure SortRules;
       procedure ComputeLookaheads;
       procedure StartBody;
       procedure StartReductions;
       procedure StartRule(Symbol: Integer);
       procedure StartTransition;
       function IDFromSymbol(K: Integer): Integer;
       procedure WriteTables;
       procedure SortItemSet(var ItemSet: TItemSetEntry);
       procedure DoLog(const S: String);
       procedure Closures;
       procedure FirstSets;
     protected
       procedure ClearStructures;
       function NextToken: Integer;
       function Parse: Integer;
     public
       constructor Create; virtual;
       destructor Destroy; override;

       procedure Execute;
       procedure ParserNoError;
       procedure ParserAbort;
       procedure ParserAccept;
       procedure ParserError;

       property Generating: Boolean read FGenerating;
       property Report: Boolean read FReport write FReport;
       property Source: TStrings read FSource write FSource;
       property Target: TStringList read FTarget;
       property Template: TStrings read FTemplate write FTemplate;

       property OnMessage: TOnMessageEvent read FOnMessage write FOnMessage;
       property OnSourceLine: TOnProcessLineEvent read FOnSourceLine write FOnSourceLine;
       property OnTemplateLine: TOnProcessLineEvent read FOnTemplateLine write FOnTemplateLine;
     end;

//----------------------------------------------------------------------------------------------------------------------

implementation

uses DCGStrings, SysUtils;

const ERRORTOKEN = 256;
      ID = 257;
      C_ID = 258;
      LITERAL = 259;
      LITID = 260;
      NUMBER = 261;
      PTOKEN = 262;
      PLEFT = 263;
      PRIGHT = 264;
      PNONASSOC = 265;
      PTYPE = 266;
      PSTART = 267;
      PPREC = 268;
      PP = 269;
      LCURL = 270;
      RCURL = 271;
      ILLEGAL = 272;

// parse tables
type TActionTableEntry = record
       Symbol,
       Action: Integer;
     end;

     TRuleTableEntry = record
       Len,
       Symbol: Integer;
     end;

const
  ShiftReduceActionCount = 251;
  GotoActionCount = 146;
  StateCount = 128;
  RuleCount = 102;

  ActionTable: array[1..ShiftReduceActionCount] of TActionTableEntry = (
   (Symbol: 256; Action: 12),
   (Symbol: 262; Action: 13),
   (Symbol: 263; Action: 14),
   (Symbol: 264; Action: 15),
   (Symbol: 265; Action: 16),
   (Symbol: 266; Action: 17),
   (Symbol: 267; Action: 18),
   (Symbol: 269; Action: 19),
   (Symbol: 270; Action: 20),
   (Symbol: 0; Action: 0),
   (Symbol: 256; Action: 24),
   (Symbol: 257; Action: 25),
   (Symbol: 60; Action: 28),
   (Symbol: 256; Action: -50),
   (Symbol: 257; Action: -50),
   (Symbol: 262; Action: -50),
   (Symbol: 263; Action: -50),
   (Symbol: 264; Action: -50),
   (Symbol: 265; Action: -50),
   (Symbol: 266; Action: -50),
   (Symbol: 267; Action: -50),
   (Symbol: 269; Action: -50),
   (Symbol: 270; Action: -50),
   (Symbol: 256; Action: 34),
   (Symbol: 271; Action: 35),
   (Symbol: 256; Action: 39),
   (Symbol: 270; Action: 20),
   (Symbol: 258; Action: -70),
   (Symbol: 256; Action: 43),
   (Symbol: 257; Action: 25),
   (Symbol: 262; Action: -49),
   (Symbol: 263; Action: -49),
   (Symbol: 264; Action: -49),
   (Symbol: 265; Action: -49),
   (Symbol: 266; Action: -49),
   (Symbol: 267; Action: -49),
   (Symbol: 269; Action: -49),
   (Symbol: 270; Action: -49),
   (Symbol: 257; Action: 25),
   (Symbol: 60; Action: 28),
   (Symbol: 256; Action: -50),
   (Symbol: 257; Action: -50),
   (Symbol: 259; Action: -50),
   (Symbol: 260; Action: -50),
   (Symbol: 60; Action: 28),
   (Symbol: 256; Action: -50),
   (Symbol: 257; Action: -50),
   (Symbol: 259; Action: -50),
   (Symbol: 260; Action: -50),
   (Symbol: 60; Action: 28),
   (Symbol: 256; Action: -50),
   (Symbol: 257; Action: -50),
   (Symbol: 259; Action: -50),
   (Symbol: 260; Action: -50),
   (Symbol: 60; Action: 28),
   (Symbol: 256; Action: -50),
   (Symbol: 257; Action: -50),
   (Symbol: 259; Action: -50),
   (Symbol: 260; Action: -50),
   (Symbol: 258; Action: 51),
   (Symbol: 124; Action: 56),
   (Symbol: 256; Action: 57),
   (Symbol: 258; Action: 51),
   (Symbol: 0; Action: -29),
   (Symbol: 269; Action: -29),
   (Symbol: 44; Action: 61),
   (Symbol: 256; Action: 62),
   (Symbol: 257; Action: 25),
   (Symbol: 262; Action: -48),
   (Symbol: 263; Action: -48),
   (Symbol: 264; Action: -48),
   (Symbol: 265; Action: -48),
   (Symbol: 266; Action: -48),
   (Symbol: 267; Action: -48),
   (Symbol: 269; Action: -48),
   (Symbol: 270; Action: -48),
   (Symbol: 62; Action: 64),
   (Symbol: 256; Action: 65),
   (Symbol: 256; Action: 71),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 256; Action: 71),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 256; Action: 71),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 256; Action: 71),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 269; Action: 19),
   (Symbol: 0; Action: -31),
   (Symbol: 256; Action: 34),
   (Symbol: 271; Action: 35),
   (Symbol: 256; Action: 83),
   (Symbol: 257; Action: 25),
   (Symbol: 44; Action: 61),
   (Symbol: 256; Action: 86),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 262; Action: -47),
   (Symbol: 263; Action: -47),
   (Symbol: 264; Action: -47),
   (Symbol: 265; Action: -47),
   (Symbol: 266; Action: -47),
   (Symbol: 267; Action: -47),
   (Symbol: 269; Action: -47),
   (Symbol: 270; Action: -47),
   (Symbol: 261; Action: 88),
   (Symbol: 44; Action: -59),
   (Symbol: 256; Action: -59),
   (Symbol: 257; Action: -59),
   (Symbol: 259; Action: -59),
   (Symbol: 260; Action: -59),
   (Symbol: 262; Action: -59),
   (Symbol: 263; Action: -59),
   (Symbol: 264; Action: -59),
   (Symbol: 265; Action: -59),
   (Symbol: 266; Action: -59),
   (Symbol: 267; Action: -59),
   (Symbol: 269; Action: -59),
   (Symbol: 270; Action: -59),
   (Symbol: 261; Action: 88),
   (Symbol: 44; Action: -60),
   (Symbol: 256; Action: -60),
   (Symbol: 257; Action: -60),
   (Symbol: 259; Action: -60),
   (Symbol: 260; Action: -60),
   (Symbol: 262; Action: -60),
   (Symbol: 263; Action: -60),
   (Symbol: 264; Action: -60),
   (Symbol: 265; Action: -60),
   (Symbol: 266; Action: -60),
   (Symbol: 267; Action: -60),
   (Symbol: 269; Action: -60),
   (Symbol: 270; Action: -60),
   (Symbol: 44; Action: 61),
   (Symbol: 256; Action: 86),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 262; Action: -45),
   (Symbol: 263; Action: -45),
   (Symbol: 264; Action: -45),
   (Symbol: 265; Action: -45),
   (Symbol: 266; Action: -45),
   (Symbol: 267; Action: -45),
   (Symbol: 269; Action: -45),
   (Symbol: 270; Action: -45),
   (Symbol: 44; Action: 61),
   (Symbol: 256; Action: 86),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 262; Action: -43),
   (Symbol: 263; Action: -43),
   (Symbol: 264; Action: -43),
   (Symbol: 265; Action: -43),
   (Symbol: 266; Action: -43),
   (Symbol: 267; Action: -43),
   (Symbol: 269; Action: -43),
   (Symbol: 270; Action: -43),
   (Symbol: 44; Action: 61),
   (Symbol: 256; Action: 86),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 262; Action: -41),
   (Symbol: 263; Action: -41),
   (Symbol: 264; Action: -41),
   (Symbol: 265; Action: -41),
   (Symbol: 266; Action: -41),
   (Symbol: 267; Action: -41),
   (Symbol: 269; Action: -41),
   (Symbol: 270; Action: -41),
   (Symbol: 58; Action: 91),
   (Symbol: 256; Action: 95),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 61; Action: 105),
   (Symbol: 123; Action: 106),
   (Symbol: 256; Action: 107),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 268; Action: 108),
   (Symbol: 0; Action: -93),
   (Symbol: 59; Action: -93),
   (Symbol: 124; Action: -93),
   (Symbol: 258; Action: -93),
   (Symbol: 269; Action: -93),
   (Symbol: 258; Action: 51),
   (Symbol: 59; Action: 112),
   (Symbol: 0; Action: -83),
   (Symbol: 124; Action: -83),
   (Symbol: 256; Action: -83),
   (Symbol: 258; Action: -83),
   (Symbol: 269; Action: -83),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 61; Action: 105),
   (Symbol: 123; Action: 106),
   (Symbol: 256; Action: 107),
   (Symbol: 257; Action: 25),
   (Symbol: 259; Action: 72),
   (Symbol: 260; Action: 73),
   (Symbol: 268; Action: 108),
   (Symbol: 0; Action: -93),
   (Symbol: 59; Action: -93),
   (Symbol: 124; Action: -93),
   (Symbol: 258; Action: -93),
   (Symbol: 269; Action: -93),
   (Symbol: 125; Action: 119),
   (Symbol: 256; Action: 120),
   (Symbol: 59; Action: 112),
   (Symbol: 0; Action: -80),
   (Symbol: 124; Action: -80),
   (Symbol: 256; Action: -80),
   (Symbol: 258; Action: -80),
   (Symbol: 269; Action: -80),
   (Symbol: 61; Action: 105),
   (Symbol: 123; Action: 106),
   (Symbol: 0; Action: - 101),
   (Symbol: 59; Action: - 101),
   (Symbol: 124; Action: - 101),
   (Symbol: 256; Action: - 101),
   (Symbol: 258; Action: - 101),
   (Symbol: 269; Action: - 101),
   (Symbol: 61; Action: 105),
   (Symbol: 123; Action: 106),
   (Symbol: 0; Action: - 101),
   (Symbol: 59; Action: - 101),
   (Symbol: 124; Action: - 101),
   (Symbol: 256; Action: - 101),
   (Symbol: 258; Action: - 101),
   (Symbol: 269; Action: - 101),
   (Symbol: 61; Action: 105),
   (Symbol: 123; Action: 106),
   (Symbol: 0; Action: - 101),
   (Symbol: 59; Action: - 101),
   (Symbol: 124; Action: - 101),
   (Symbol: 256; Action: - 101),
   (Symbol: 258; Action: - 101),
   (Symbol: 269; Action: - 101)
  );

  GotoTable: array[1..GotoActionCount] of TActionTableEntry = (
   (Symbol: -27; Action: 1),
   (Symbol: -2; Action: 2),
   (Symbol: -32; Action: 3),
   (Symbol: - 16; Action: 4),
   (Symbol: - 15; Action: 5),
   (Symbol: - 13; Action: 6),
   (Symbol: - 12; Action: 7),
   (Symbol: - 11; Action: 8),
   (Symbol: - 10; Action: 9),
   (Symbol: -9; Action: 10),
   (Symbol: -8; Action: 11),
   (Symbol: -33; Action: 21),
   (Symbol: -29; Action: 22),
   (Symbol: -3; Action: 23),
   (Symbol: -34; Action: 26),
   (Symbol: -24; Action: 27),
   (Symbol: -39; Action: 29),
   (Symbol: -38; Action: 30),
   (Symbol: -37; Action: 31),
   (Symbol: -35; Action: 32),
   (Symbol: - 17; Action: 33),
   (Symbol: -44; Action: 36),
   (Symbol: -28; Action: 37),
   (Symbol: - 16; Action: 38),
   (Symbol: -42; Action: 40),
   (Symbol: -40; Action: 41),
   (Symbol: -3; Action: 42),
   (Symbol: -3; Action: 44),
   (Symbol: -34; Action: 45),
   (Symbol: -24; Action: 27),
   (Symbol: -34; Action: 46),
   (Symbol: -24; Action: 27),
   (Symbol: -34; Action: 47),
   (Symbol: -24; Action: 27),
   (Symbol: -34; Action: 48),
   (Symbol: -24; Action: 27),
   (Symbol: -43; Action: 49),
   (Symbol: -4; Action: 50),
   (Symbol: -47; Action: 52),
   (Symbol: -43; Action: 53),
   (Symbol: -31; Action: 54),
   (Symbol: -21; Action: 55),
   (Symbol: -4; Action: 50),
   (Symbol: -45; Action: 58),
   (Symbol: -42; Action: 59),
   (Symbol: - 18; Action: 60),
   (Symbol: -3; Action: 42),
   (Symbol: -25; Action: 63),
   (Symbol: -41; Action: 66),
   (Symbol: -36; Action: 67),
   (Symbol: -6; Action: 68),
   (Symbol: -5; Action: 69),
   (Symbol: -3; Action: 70),
   (Symbol: -41; Action: 66),
   (Symbol: -36; Action: 74),
   (Symbol: -6; Action: 68),
   (Symbol: -5; Action: 69),
   (Symbol: -3; Action: 70),
   (Symbol: -41; Action: 66),
   (Symbol: -36; Action: 75),
   (Symbol: -6; Action: 68),
   (Symbol: -5; Action: 69),
   (Symbol: -3; Action: 70),
   (Symbol: -41; Action: 66),
   (Symbol: -36; Action: 76),
   (Symbol: -6; Action: 68),
   (Symbol: -5; Action: 69),
   (Symbol: -3; Action: 70),
   (Symbol: -48; Action: 77),
   (Symbol: -30; Action: 78),
   (Symbol: - 15; Action: 79),
   (Symbol: -52; Action: 80),
   (Symbol: - 17; Action: 81),
   (Symbol: -42; Action: 82),
   (Symbol: -3; Action: 42),
   (Symbol: -41; Action: 84),
   (Symbol: - 18; Action: 85),
   (Symbol: -6; Action: 68),
   (Symbol: -5; Action: 69),
   (Symbol: -3; Action: 70),
   (Symbol: -7; Action: 87),
   (Symbol: -7; Action: 89),
   (Symbol: -41; Action: 84),
   (Symbol: - 18; Action: 85),
   (Symbol: -6; Action: 68),
   (Symbol: -5; Action: 69),
   (Symbol: -3; Action: 70),
   (Symbol: -41; Action: 84),
   (Symbol: - 18; Action: 85),
   (Symbol: -6; Action: 68),
   (Symbol: -5; Action: 69),
   (Symbol: -3; Action: 70),
   (Symbol: -41; Action: 84),
   (Symbol: - 18; Action: 85),
   (Symbol: -6; Action: 68),
   (Symbol: -5; Action: 69),
   (Symbol: -3; Action: 70),
   (Symbol: - 19; Action: 90),
   (Symbol: -49; Action: 92),
   (Symbol: -46; Action: 93),
   (Symbol: -41; Action: 94),
   (Symbol: -6; Action: 68),
   (Symbol: -5; Action: 69),
   (Symbol: -3; Action: 70),
   (Symbol: -50; Action: 96),
   (Symbol: -53; Action: 97),
   (Symbol: -51; Action: 98),
   (Symbol: -26; Action: 99),
   (Symbol: -22; Action: 100),
   (Symbol: - 14; Action: 101),
   (Symbol: -6; Action: 102),
   (Symbol: -5; Action: 103),
   (Symbol: -3; Action: 104),
   (Symbol: -43; Action: 109),
   (Symbol: -4; Action: 50),
   (Symbol: -49; Action: 110),
   (Symbol: -20; Action: 111),
   (Symbol: -54; Action: 113),
   (Symbol: -6; Action: 114),
   (Symbol: -5; Action: 115),
   (Symbol: -3; Action: 116),
   (Symbol: -53; Action: 97),
   (Symbol: -51; Action: 117),
   (Symbol: -26; Action: 99),
   (Symbol: -22; Action: 100),
   (Symbol: - 14; Action: 101),
   (Symbol: -6; Action: 102),
   (Symbol: -5; Action: 103),
   (Symbol: -3; Action: 104),
   (Symbol: -23; Action: 118),
   (Symbol: -57; Action: 121),
   (Symbol: -56; Action: 122),
   (Symbol: -58; Action: 123),
   (Symbol: -20; Action: 111),
   (Symbol: -55; Action: 124),
   (Symbol: -53; Action: 125),
   (Symbol: -26; Action: 99),
   (Symbol: -22; Action: 100),
   (Symbol: -55; Action: 126),
   (Symbol: -53; Action: 125),
   (Symbol: -26; Action: 99),
   (Symbol: -22; Action: 100),
   (Symbol: -55; Action: 127),
   (Symbol: -53; Action: 125),
   (Symbol: -26; Action: 99),
   (Symbol: -22; Action: 100)
 );

  DefaultActionTable: array[0..StateCount - 1] of Integer = (
    -33, 0, 0, -34, -38, -28, 0, 0, -46, -44, -42, -40, -35, -6, -7, -8, -9, - 10,
    - 11, - 13, - 14, 0, 0, -36, -37, - 1, 0, 0, -24, 0, 0, 0, 0, -39, - 16, - 15, 0,
    0, -72, -76, -63, 0, -69, -66, 0, 0, 0, 0, 0, -71, -78, -2, -75, -81, 0, -82,
    -20, -77, 0, -64, 0, - 17, -67, -51, -25, -26, -52, 0, 0, -58, 0, -55, -3, -4,
    0, 0, 0, 0, -30, -32, -84, -73, -65, -68, -53, 0, -56, -61, -5, -62, -79, - 18,
    0, 0, -54, -57, -84, -88, 0, -92, -90, 0, -86, -85, -87, -27, -21, -89, - 12,
    -74, 0, - 100, - 19, 0, -96, -94, -98, 0, -91, -22, -23, 0, 0, 0, -97, - 102, -95,
    -99
  );

  ActionsLow: array[0..StateCount - 1] of Integer = (
    1, 1, 10, 11, 11, 11, 11, 13, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
    24, 24, 26, 29, 29, 29, 29, 39, 40, 40, 45, 50, 55, 60, 60, 60, 60, 61, 66, 66,
    66, 66, 77, 77, 77, 79, 83, 87, 91, 95, 95, 95, 95, 95, 95, 97, 97, 97, 97, 99,
    99, 101, 101, 101, 101, 101, 101, 101, 114, 128, 128, 142, 142, 142, 142, 155,
    168, 181, 182, 182, 182, 182, 182, 182, 182, 182, 186, 186, 186, 186, 186, 186,
    186, 198, 199, 199, 199, 199, 199, 205, 205, 205, 208, 208, 208, 208, 208, 208,
    208, 208, 208, 220, 220, 220, 222, 222, 222, 222, 228, 228, 228, 228, 236, 244,
    252, 252, 252, 252
 );

  ActionsHigh: array[0..StateCount - 1] of Integer = (
    0, 9, 10, 10, 10, 10, 12, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23,
    23, 25, 28, 28, 28, 28, 38, 39, 39, 44, 49, 54, 59, 59, 59, 59, 60, 65, 65, 65,
    65, 76, 76, 76, 78, 82, 86, 90, 94, 94, 94, 94, 94, 94, 96, 96, 96, 96, 98, 98,
    100, 100, 100, 100, 100, 100, 100, 113, 127, 127, 141, 141, 141, 141, 154, 167,
    180, 181, 181, 181, 181, 181, 181, 181, 181, 185, 185, 185, 185, 185, 185, 185,
    197, 198, 198, 198, 198, 198, 204, 204, 204, 207, 207, 207, 207, 207, 207, 207,
    207, 207, 219, 219, 219, 221, 221, 221, 221, 227, 227, 227, 227, 235, 243, 251,
    251, 251, 251, 251
 );

  GotoLow: array[0..StateCount - 1] of Integer = (
    1, 3, 12, 12, 12, 13, 14, 15, 17, 18, 19, 20, 21, 21, 21, 21, 21, 21, 21, 21,
    21, 21, 22, 25, 25, 25, 25, 28, 29, 29, 31, 33, 35, 37, 37, 37, 37, 39, 44, 45,
    45, 45, 48, 48, 48, 49, 54, 59, 64, 69, 69, 70, 70, 70, 70, 72, 73, 73, 73, 74,
    74, 76, 76, 76, 76, 76, 76, 76, 81, 82, 82, 83, 83, 83, 83, 88, 93, 98, 99, 99,
    99, 100, 101, 101, 101, 101, 105, 105, 105, 105, 105, 106, 106, 114, 116, 116,
    116, 117, 117, 118, 118, 119, 122, 122, 122, 122, 122, 122, 122, 122, 122, 130,
    130, 130, 131, 132, 133, 134, 135, 135, 135, 135, 139, 143, 147, 147, 147, 147
 );

  GotoHigh: array[0..StateCount - 1] of Integer = (
   2, 11, 11, 11, 12, 13, 14, 16, 17, 18, 19, 20, 20, 20, 20, 20, 20, 20, 20, 20,
   20, 21, 24, 24, 24, 24, 27, 28, 28, 30, 32, 34, 36, 36, 36, 36, 38, 43, 44, 44,
   44, 47, 47, 47, 48, 53, 58, 63, 68, 68, 69, 69, 69, 69, 71, 72, 72, 72, 73, 73,
   75, 75, 75, 75, 75, 75, 75, 80, 81, 81, 82, 82, 82, 82, 87, 92, 97, 98, 98, 98,
   99, 100, 100, 100, 100, 104, 104, 104, 104, 104, 105, 105, 113, 115, 115, 115,
   116, 116, 117, 117, 118, 121, 121, 121, 121, 121, 121, 121, 121, 121, 129, 129,
   129, 130, 131, 132, 133, 134, 134, 134, 134, 138, 142, 146, 146, 146, 146, 146
 );

  RuleTable: array[1..RuleCount] of TRuleTableEntry = (
   (Len: 1; Symbol: -3),
   (Len: 1; Symbol: -4),
   (Len: 1; Symbol: -5),
   (Len: 1; Symbol: -6),
   (Len: 1; Symbol: -7),
   (Len: 1; Symbol: -8),
   (Len: 1; Symbol: -9),
   (Len: 1; Symbol: - 10),
   (Len: 1; Symbol: - 11),
   (Len: 1; Symbol: - 12),
   (Len: 1; Symbol: - 13),
   (Len: 1; Symbol: - 14),
   (Len: 1; Symbol: - 15),
   (Len: 1; Symbol: - 16),
   (Len: 1; Symbol: - 17),
   (Len: 1; Symbol: - 17),
   (Len: 1; Symbol: - 18),
   (Len: 1; Symbol: - 19),
   (Len: 1; Symbol: -20),
   (Len: 1; Symbol: -21),
   (Len: 1; Symbol: -22),
   (Len: 1; Symbol: -23),
   (Len: 1; Symbol: -23),
   (Len: 1; Symbol: -24),
   (Len: 1; Symbol: -25),
   (Len: 1; Symbol: -25),
   (Len: 1; Symbol: -26),
   (Len: 0; Symbol: -29),
   (Len: 0; Symbol: -31),
   (Len: 6; Symbol: -2),
   (Len: 0; Symbol: -30),
   (Len: 1; Symbol: -30),
   (Len: 0; Symbol: -27),
   (Len: 2; Symbol: -27),
   (Len: 2; Symbol: -27),
   (Len: 2; Symbol: -32),
   (Len: 2; Symbol: -32),
   (Len: 0; Symbol: -33),
   (Len: 3; Symbol: -32),
   (Len: 0; Symbol: -35),
   (Len: 4; Symbol: -32),
   (Len: 0; Symbol: -37),
   (Len: 4; Symbol: -32),
   (Len: 0; Symbol: -38),
   (Len: 4; Symbol: -32),
   (Len: 0; Symbol: -39),
   (Len: 4; Symbol: -32),
   (Len: 3; Symbol: -32),
   (Len: 2; Symbol: -32),
   (Len: 0; Symbol: -34),
   (Len: 3; Symbol: -34),
   (Len: 1; Symbol: -36),
   (Len: 2; Symbol: -36),
   (Len: 3; Symbol: -36),
   (Len: 1; Symbol: -36),
   (Len: 2; Symbol: -36),
   (Len: 3; Symbol: -36),
   (Len: 1; Symbol: -41),
   (Len: 1; Symbol: -41),
   (Len: 1; Symbol: -41),
   (Len: 2; Symbol: -41),
   (Len: 2; Symbol: -41),
   (Len: 1; Symbol: -40),
   (Len: 2; Symbol: -40),
   (Len: 3; Symbol: -40),
   (Len: 1; Symbol: -40),
   (Len: 2; Symbol: -40),
   (Len: 3; Symbol: -40),
   (Len: 1; Symbol: -42),
   (Len: 0; Symbol: -44),
   (Len: 2; Symbol: -28),
   (Len: 0; Symbol: -45),
   (Len: 0; Symbol: -46),
   (Len: 5; Symbol: -28),
   (Len: 2; Symbol: -28),
   (Len: 1; Symbol: -28),
   (Len: 2; Symbol: -28),
   (Len: 0; Symbol: -48),
   (Len: 0; Symbol: -50),
   (Len: 6; Symbol: -43),
   (Len: 1; Symbol: -47),
   (Len: 0; Symbol: -52),
   (Len: 4; Symbol: -47),
   (Len: 0; Symbol: -49),
   (Len: 2; Symbol: -49),
   (Len: 2; Symbol: -49),
   (Len: 2; Symbol: -49),
   (Len: 2; Symbol: -49),
   (Len: 2; Symbol: -49),
   (Len: 0; Symbol: -54),
   (Len: 3; Symbol: -53),
   (Len: 1; Symbol: -53),
   (Len: 0; Symbol: -51),
   (Len: 0; Symbol: -56),
   (Len: 4; Symbol: -51),
   (Len: 0; Symbol: -57),
   (Len: 4; Symbol: -51),
   (Len: 0; Symbol: -58),
   (Len: 4; Symbol: -51),
   (Len: 2; Symbol: -51),
   (Len: 0; Symbol: -55),
   (Len: 1; Symbol: -55)
 );

//----------------------------------------------------------------------------------------------------------------------
// This implementation is based on algorithms 4.12 and 4.13 in Aho/Sethi/
// Ullman 1986 (with some optimizations added) which avoid the need to
// construct the full LR(1) set and are able to compute lookaheads from
// LR(0) kernel items only.
//
// We start off with the LR(0) state set together with corresponding (shift
// and goto) transitions already computed. We compute the LALR(1) lookahead
// sets for kernel items and do record all corresponding reduce actions
// in the reduction table, where we also have to consider non-kernel items
// with empty right-hand side. These items also require a reduction, but
// never appear in the kernel item table.
//
// This implementation uses some simple optimizations to speed up the
// algorithm. The lookahead sets are represented by TIntegerLists.
// Lookahead sets are associated with each kernel item in the item table
// and with each reduction in the reduction table. A kernel item
// requiring a reduction shares its lookahead set pointer with the
// corresponding Entry in the reduction table. The lookahead set for
// a non-kernel reduction item, with empty right-hand side, only
// appears in the reduction table.
//
// The algorithm consists of two phases:
//
// 1. Initialization:
//
//    The initialization phase consists of a single traversal of the LR(0)
//    set where we compute lookahead sets generated spontaneously (lookaheads
//    which are passed on from non-kernel items to the corresponding items in
//    successor states), initialize lookahead sets and enter them into the
//    lookahead and reduction table. Furthermore, during the initialization
//    phase we also initialize the links for the propagation of lookaheads
//    in the second phase.
//
//    To determine lookaheads and propagation links we compute the look-
//    aheads for the closures of single LR(0) sets shorthand, according
//    to the method in Aho/Sethi/Ullman 1986 (with some modifications),
//    where we associate with each kernel item I a corresponding endmarker
//    symbol #I as its lookahead symbol.
//
//    The initialization phase proceeds as follows:
//
//    1) Initialize all non-kernel Item lookahead sets to empty.
//
//       Now we pass over each state s in the LR0 set, repeating steps 2) thru
//       5) specified below:
//
//    2) Compute the closure closure(K(S)) of the states's kernel set K(s).
//
//    3) Compute the lookahead sets for closure(K(S))(described in detail
//       below) where each kernel Item I is associated with a special
//       endmarker symbol #I as lookahead.
//
//    Now the lookahead symbols, reductions and propagation links are entered
//    into the corresponding WriteTables as follows:
//
//    4) Process kernel items: Add a propagation link from the kernel item
//       to the lookahead set of the linked item in the corresponding
//       successor state (as specified by the Next field). If there is no
//       successor item (kernel item requiring a reduction), add a
//       corresponding entry into the reduction table instead.
//
//    5) Process non-kernel items: find the corresponding kernel item in the
//       successor state which is generated spontaneously from the non-kernel
//       item. Add the spontaneous lookahead symbols (except endmarker
//       symbols) of the non-kernel item determined in step 3) to the kernel
//       item. If the non-kernel item has an empty right-hand side (non-kernel
//       item requiring a reduction), add a corresponding entry into the
//       reduction table instead. Furthermore, for each endmarker symbol
//       #I in the spontaneous lookahead set of the non-kernel item add
//       a corresponding propagation link from the Ith kernel item to the
//       lookahead set of the spontaneous kernel Item.
//
//    To compute the spontaneous Lookaheads (step 3)) we proceed as follows:
//
//    3a) First compute the first sets of tail strings of all items in
//        closure(K(S)). The "tail string" of an Item [ X -> v.Yw ], where
//        Y is a non-terminal, is the symbol sequence w, whose first set
//        induces corresponding spontaneous lookaheads in the non-kernel
//        items of the state with left-hand side Y. Note that the first
//        sets of "tail strings" in items [ X -> v.yw ], where y is a
//        *terminal* symbol, are not required and hence it is not
//        necessary to compute them. We also record for each item whether
//        its tail string is "nullable", i.e. may be derived to an empty
//        string. In this case, the item also passes on its own lookaheads,
//        in addition to the first symbols of its tail string. First sets
//        and nullable flags are computed using the information in 
//        first and nullable tables.
//
//    3b) Now follows an initialization part in which each Item [ X -> v.Yw ]
//        passes on the first symbols of its tail string to the lookahead
//        sets of each corresponding non-kernel Item [ Y -> .u ].
//
//    3c) Finally, we repeatedly pass over the item set, passing on
//        lookaheads from items with nullable tail strings. Each item
//        [ X -> v.Yw ] with nullable w propagates its own lookaheads to
//        all corresponding non-kernel items [ Y -> .u]. Step 3c) terminates
//        as soon as no lookahead sets have been modified during the previous
//        pass.
//
// 2. Propagation:
//
//    The second phase of the lookahead computation algorithm now is quite
//    simple. We repeatedly pass over all kernel items, propagating lookaheads
//    according to the propagation links determined in the initialization
//    phase. The algorithm terminates as soon as no lookahead sets have been
//    modified during the previous pass.

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.ComputeLookaheads;

// compute spontaneous Lookaheads for FItemSet; negative symbols are
// used for endmarkers(-I denotes endmarker #I)

var Count, I,
    LastCount: Integer;
    FirstSymbols: TSymbolSetArray;
    nullable: array[1..MaxSetItems] of Boolean;

  //--------------- local functions -------------------------------------------

  function SymbolCount(N: Integer): Integer;

  // counts lookahead symbols

  var Count, I: Integer;

  begin
    Count := 0;
    for I := 1 to N do Inc(Count, Length(FLookaheadSet[I]));
    Result := Count;
  end;

  //---------------------------------------------------------------------------

  procedure ComputeFirstSymbols(I: Integer);

  // computes first set and nullable flag for tail string of Item number I

  var J: Integer;

  begin
    FirstSymbols[I] := nil;
    nullable[I] := True;
    with FItemSet, Item[I], FRuleTable^[RuleNumber] do
      if (PositionNumber <= rhs_len) and (rhs_sym[PositionNumber] < 0) then
      begin
        J := PositionNumber + 1;
        while (J <= rhs_len) and nullable[I] do
        begin
          if rhs_sym[J] < 0 then
          begin
            ISUnion(FirstSymbols[I], FFirstSetTable[-rhs_sym[J]]);
            nullable[I] := FNullableTable[-rhs_sym[J]];
          end
          else
          begin
            ISInclude(FirstSymbols[I], rhs_sym[J]);
            nullable[I] := False;
          end;
          Inc(J);
        end;
      end;
  end;

  //---------------------------------------------------------------------------

  procedure InitialLookaheads(I: Integer);

  // computes initial lookaheads induced by first sets of tail string of Item I

  var Symbol, J: Integer;

  begin
    with FItemSet, Item[I], FRuleTable[RuleNumber] do
      if (PositionNumber <= rhs_len) and (rhs_sym[PositionNumber] < 0) then
      begin
        Symbol := rhs_sym[PositionNumber];
        for J := FKernelItemsCount + 1 to ItemCount do
          with Item[J], FRuleTable[RuleNumber] do
            if lhs_sym = Symbol then ISUnion(FLookaheadSet[J], FirstSymbols[I]);
      end
  end;

  //---------------------------------------------------------------------------

  procedure Propagate(I: Integer);

  // propagates lookahead symbols of item I

  var Symbol, J: Integer;

  begin
    with FItemSet, Item[I], FRuleTable^[RuleNumber] do
      if (PositionNumber <= rhs_len) and (rhs_sym[PositionNumber] < 0) and nullable[I] then
      begin
        Symbol := rhs_sym[PositionNumber];
        for J := FKernelItemsCount + 1 to ItemCount do
          with Item[J], FRuleTable^[RuleNumber] do
            if lhs_sym = Symbol then ISUnion(FLookaheadSet[J], FLookaheadSet[I]);
      end
  end;

  //--------------- end local functions ---------------------------------------

begin
  with FItemSet do
  begin
    // initialize kernel lookahead sets
    for I := 1 to FKernelItemsCount do ISSingleton(FLookaheadSet[I], -I);

    // compute first sets and nullable flags
    for I := 1 to ItemCount do ComputeFirstSymbols(I);

    // initialize non-kernel lookahead sets
    for I := FKernelItemsCount + 1 to ItemCount do FLookaheadSet[I] := nil;
    for I := 1 to ItemCount do InitialLookaheads(I);

    // repeated passes until no more lookaheads have been added
    // during the previous pass
    Count := SymbolCount(ItemCount);
    repeat
      LastCount := Count;
      for I := 1 to ItemCount do Propagate(I);
      Count := SymbolCount(ItemCount);
    until LastCount = Count;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.CompareReductions(I, J: Integer): Boolean;

begin
  Result := FReductionTable[I].RuleNumber < FReductionTable[J].RuleNumber
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.SwapReductions(I, J: Integer);

var x: TReductionEntry;

begin
  x := FReductionTable[I];
  FReductionTable[I] := FReductionTable[J];
  FReductionTable[J] := x;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.SortReductions;

// sort reduction entries in FCurrentState with regard to rule numbers

begin
  with FStateTable[FCurrentState] do
    QuickSort(ReductionsLow, ReductionsHigh, CompareReductions, SwapReductions);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.InitLookaheads;

// initialization phase of lookahead computation algorithm

  //--------------- local functions -------------------------------------------

  procedure AddPropagationLink(I: Integer; SymbolSet: TIntegerList);

  // add a propagation link to kernel item I

  var Prop: PPropagationEntry;

  begin
    New(Prop);
    Prop.SymbolSet := SymbolSet;
    Prop.Next := FPropagationTable[I];
    FPropagationTable^[I] := Prop;
  end;

  //--------------- end local functions ---------------------------------------

var I, J, K: Integer;
    Lookaheads: TIntegerList;

begin
  // initialize lookahead sets and propagation links
  for I := 1 to FItemCount do FLookaheadTable[I] := nil;
  for I := 1 to FItemCount do FPropagationTable^[I] := nil;

  FCurrentState := 0;
  repeat
    with FStateTable^[FCurrentState], FItemSet do
    begin
      StartReductions;
      GetItemSet(FCurrentState, FItemSet);
      FKernelItemsCount := ItemCount;

      // compute LR(0) closure
      Closure(FItemSet);

      // compute spontaneous lookaheads
      ComputeLookaheads;

      // process kernel items
      for I := 1 to FKernelItemsCount do
        with Item[I] do
          if Next > 0 then // add propagation link
                           AddPropagationLink(ItemLow + I - 1, FLookaheadTable[Next])
                      else // enter reduce action
                           AddReduction(FLookaheadTable[ItemLow + I - 1], RuleNumber);

      // process non-kernel items

      // find successor items
      for K := TransitionLow to TransitionHigh do
        with FTransitionTable^[K] do
          for I := FKernelItemsCount + 1 to ItemCount do
            with Item[I], FRuleTable^[RuleNumber] do
              if PositionNumber > rhs_len then Next := 0
                                          else
                if rhs_sym[PositionNumber] = Symbol then
                  Next := FindItem(NextState, RuleNumber, PositionNumber + 1);

      // add spontaneous lookaheads and propagation links
      for I := FKernelItemsCount + 1 to ItemCount do
        with Item[I] do
          if Next > 0 then
          begin
            // lookaheads are generated spontaneously for successor item
            for J := 0 to High(FLookaheadSet[I]) do
              if FLookaheadSet[I][J] >= 0
                then ISInclude(FLookaheadTable[Next], FLookaheadSet[I][J])
                else AddPropagationLink(ItemLow + (-FLookaheadSet[I][J]) - 1, FLookaheadTable[Next])
          end
          else
          begin
            // non-kernel reduction item

            // the lookahead set 'Lookaheads' used here is only locally allocated and added
            // to the reduction table instead of the propagation table, hence we have to
            // store a diiferent index to mark the different table (but AFTER all additions
            // have been made)
            Lookaheads := nil;
            for J := 0 to High(FLookaheadSet[I]) do
              if FLookaheadSet[I][J] >= 0
                then ISInclude(Lookaheads, FLookaheadSet[I][J])
                else AddPropagationLink(ItemLow + (-FLookaheadSet[I][J]) - 1, Lookaheads);
            AddReduction(Lookaheads, RuleNumber);
          end;

      EndReductions;
      SortReductions;
    end;
    Inc(FCurrentState);
  until FCurrentState = FStateCount;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.Propagate;

// propagation phase of lookahead computation algorithm

var I, L: Integer;
    done: Boolean;
    Prop: PPropagationEntry;

begin
  // repeated passes over the kernel items table until no more Lookaheads
  // could be added in the previous pass:
  repeat
    done := True;
    for I := 1 to FItemCount do
    begin
      Prop := FPropagationTable^[I];
      while Prop <> nil do
        with Prop^ do
        begin
          L := Length(SymbolSet);
          ISUnion(SymbolSet, FLookaheadTable[I]);
          if Length(SymbolSet) > L then done := False;
          Prop := Next;
        end;
    end;
  until done;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.Lookaheads;

begin
  InitLookaheads;
  Propagate;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.GetSymbols(var ItemSet: TItemSetEntry; var SymbolSet: TIntegerList);

// gets the symbols for which there are transitions in ItemSet

var I: Integer;

begin
  with ItemSet do
  begin
    SymbolSet := nil;
    for I := 1 to ItemCount do
      with Item[I], FRuleTable^[RuleNumber] do
        if PositionNumber <= rhs_len then ISInclude(SymbolSet, rhs_sym[PositionNumber]);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.MakeState(var ItemSet: TItemSetEntry; Symbol: Integer): Integer;

// constructs a new state for the transitions in ItemSet on symbol Symbol
// returns: the new state number

var I: Integer;

begin
  with ItemSet do
  begin
    // add the new state:
    MakeNewState;
    for I := 1 to ItemCount do
      with Item[I], FRuleTable^[RuleNumber] do
        if (PositionNumber <= rhs_len) and (rhs_sym[PositionNumber] = Symbol) then
          AddItem(RuleNumber, PositionNumber + 1);
    Result := AddState;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddNextLinks;

// adds links to successor items for kernel items in the active state

var K, I: Integer;

begin
  with FStateTable^[FCurrentState] do
    for K := TransitionLow to TransitionHigh do
      with FTransitionTable^[K] do
        for I := ItemLow to ItemHigh do
          with FItemTable^[I], FRuleTable^[RuleNumber] do
            if (PositionNumber <= rhs_len) and (rhs_sym[PositionNumber] = Symbol) then
              Next := FindItem(NextState, RuleNumber, PositionNumber + 1);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.LR0Set;

// constructs the LR(0) state set, shift and goto transitions and
// corresponding kernel items

var Items: TItemSetEntry;
    Symbols: TIntegerList;
    I: Integer;

begin
  // initialize state 0
  MakeNewState;
  // augmented start production
  AddItem(1, 1);
  FCurrentState := AddState;

  // build the state table
  repeat
    // compute the closure of the current state
    GetItemSet(FCurrentState, Items);
    Closure(Items);
    SortItemSet(Items);

    // determine symbols used in shift and goto transitions
    GetSymbols(Items, Symbols);
    // and add transitions
    StartTransition;
    for I := 0 to High(Symbols) do
      if Symbols[I] = 0 then // accept action
                             AddTransition(0, 0)
                        else // shift/goto action
                             AddTransition(Symbols[I], MakeState(Items, Symbols[I]));
    EndTransitions;
    // add next links to kernel items
    AddNextLinks;
    // switch to next state
    Inc(FCurrentState);
  until FCurrentState = FStateCount;
end;

//----------------------------------------------------------------------------------------------------------------------

// YaccClosures.pas

procedure TParserGenerator.Closures;

// The closure set of a non-terminal X is the set of all nonterminals Y
// so that Y appears as the first symbol in a rightmost derivation from the
// non-terminal X(i.e. X => +  Y ... in a rightmost derivation). We can
// easily compute closure sets as follows:
// - Initialize the closure set for any non-terminal X to contain all
//   nonterminals Y for which there is a rule X: Y ...
// - Now repeatedly pass over the already constructed sets and, for
//   any non-terminal Y which has already been added to the closure set
//   of some non-terminal X, also include the closure elements of Y in
//   the closure set of X.
// The algorithm terminates as soon as no additional symbols have been
// added during the previous pass.

var Symbol,
    I,
    Count,
    PreviousCount: Integer;
    Symbols: TIntegerList;

begin
  // initialize closure sets
  Symbols := nil;
  PreviousCount := 0;
  Count := 0;
  for Symbol := 1 to FNonTerminalCount do
  begin
    FClosureTable[Symbol] := nil;
    with FRuleOffsetTable^[Symbol] do
      for I := RuleLow to RuleHigh do
        with FRuleTable^[FRuleNumberTable[I]] do
          if (rhs_len > 0) and (rhs_sym[1] < 0) then ISInclude(FClosureTable[Symbol], rhs_sym[1]);
    Inc(Count, Length(FClosureTable[Symbol]));
  end;

  // repeated passes until no more symbols have been added during the last pass
  while PreviousCount < Count do
  begin
    PreviousCount := Count;
    Count := 0;
    for Symbol := 1 to FNonTerminalCount do
    begin
      Symbols := FClosureTable[Symbol];
      for I := 0 to High(Symbols) do ISUnion(FClosureTable[Symbol], FClosureTable[-Symbols[I]]);
      Inc(Count, Length(FClosureTable[Symbol]));
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.FirstSets;

// The first set of a non-terminal X is the set of all literal symbols
// y so that X => +  y ... in some derivation of the non-terminal X. In
// addition X is nullable if the empty string can be derived from X.
// Using the first set construction algorithm of Aho/Sethi/Ullman,
// Section 4.4, the first sets and nullable flags are computed as
// follows:
//
// For any production X -> y1 ... yn, where the yi are grammar symbols,
// add the symbols in the first set of y1 (y1 itself if it is a literal)
// to the first set of X. If y1 is a nullable non-terminal then proceed
// with y2 etc. until either all yi have been considered or yi is non-
// nullable (or a literal symbol). If all of the yi are nullable (in
// particular if N = 0), then also set nullable[X] to True.
//
// This procedure is repeated until no more symbols have been added to any
// first set and none of the nullable flags have been changed during the
// previous pass.

var I, J, L,
    Symbol: Integer;
    N, null, done: Boolean;

begin
  // initialize tables
  for Symbol := 1 to FNonTerminalCount do
  begin
    FNullableTable[Symbol] := False;
    FFirstSetTable[Symbol] := nil;
  end;

  // repeated passes until no more symbols added and no nullable flags modified:
  repeat
    done := True;
    for I := 1 to FRuleCount do
      with FRuleTable^[I] do
      begin
        L := Length(FFirstSetTable[-lhs_sym]);
        N := FNullableTable[-lhs_sym];
        null := True;
        J := 1;

        while (J <= rhs_len) and null do
        begin
          if rhs_sym[J] < 0 then
          begin
            ISUnion(FFirstSetTable[-lhs_sym], FFirstSetTable[-rhs_sym[J]]);
            null := FNullableTable[-rhs_sym[J]];
          end
          else
          begin
            ISInclude(FFirstSetTable[-lhs_sym], rhs_sym[J]);
            null := False;
          end;
          Inc(J);
        end;

        if null then FNullableTable[-lhs_sym] := True;
        if (L < Length(FFirstSetTable[-lhs_sym])) or (N <> FNullableTable[-lhs_sym]) then done := False;
      end;
  until done;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.RuleToStr(I: Integer): String;

// returns print representation of rule number I

var J: Integer;

begin
  with FRuleTable^[I] do
  begin
    Result := Name(lhs_sym) + ':';
    for J := 1 to rhs_len do
      Result := Result + ' ' + Name(rhs_sym[J]);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.ItemToStr(var ItemSet: TItemSetEntry; I: Integer): String;

// returns print representation of Item number I in ItemSet

var J: Integer;

begin
  with ItemSet, Item[I], FRuleTable^[RuleNumber] do
  begin
    Result := Name(lhs_sym) + ':';
    for J := 1 to PositionNumber - 1 do Result := Result + ' ' + Name(rhs_sym[J]);
    Result := Result + ' _';
    for J := PositionNumber to rhs_len do Result := Result + ' ' + Name(rhs_sym[J]);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.BuildParseTables;

// builds the parse table, resolves conflicts

var I, J, K, S,
    n_errors,
    n_shifts,
    n_gotos,
    n_reductions,
    n_conflicts: Integer;

    ItemSet: TItemSetEntry;
    reduced: array[1..MaxRules] of Boolean;


begin
  // initialize
  FShiftReduceConflicts := 0;
  FReduceReduceConflicts := 0;
  FNeverReduced := 0;
  for I := 1 to FRuleCount do reduced[I] := False;

  // traverse the state table:
  for S := 0 to FStateCount - 1 do
    with FStateTable^[S] do
    begin
      // Check shift and reduce actions, resolve conflicts.
      // The number of error actions generated by nonassoc's is counted
      // in n_errors, the number of conflicts reported in n_conflicts.
      // Shift actions ruled out by disambiguating rules are flagged by
      // setting the corresponding NextState to - 1.

      n_errors := 0;
      n_conflicts := 0;

      for I := TransitionLow to TransitionHigh do
        with FTransitionTable^[I] do
          if Symbol >= 0 then
            for J := ReductionsLow to ReductionsHigh do
              with FReductionTable[J] do
                if ISMember(SymbolSet, Symbol)
                  then
                    if (FSymbolPrecendenceTable^[Symbol] > 0) and (FRulePrecendenceTable^[RuleNumber] > 0)
                      then
                        // resolve conflict using precedence
                        if FRulePrecendenceTable^[RuleNumber] = FSymbolPrecendenceTable^[Symbol]
                          then
                            case FPrecendenceTable^[FSymbolPrecendenceTable^[Symbol]] of
                              ptLeft    : // reduce
                                          NextState := -1;
                              ptRight   : // shift
                                          ISExclude(SymbolSet, Symbol);
                              ptNonAssoc: // error
                                          begin
                                            Inc(n_errors);
                                            NextState := -1;
                                            ISExclude(SymbolSet, Symbol);
                                          end;
                            end
                          else
                            if FRulePrecendenceTable^[RuleNumber] > FSymbolPrecendenceTable^[Symbol] then NextState := -1 // reduce
                      else ISExclude(SymbolSet, Symbol) // shift
                  else
                  begin
                    // shift/reduce conflict
                    if FReport then
                    begin
                      if n_conflicts = 0 then DoLog(TAB + '*** conflicts:');
                      DoLog(TAB + 'shift ' + IntToStr(NextState) + ', ' + 'reduce '  + IntToStr(RuleNumber - 1) + ' on ' + Name(Symbol));
                    end;
                    Inc(n_conflicts);
                    Inc(FShiftReduceConflicts);
                    ISExclude(SymbolSet, Symbol);
                  end;

      for I := ReductionsLow to ReductionsHigh do
        for J := I + 1 to ReductionsHigh do
          with FReductionTable[J] do
          begin
            for K := 0 to High(SymbolSet) do
              if ISMember(FReductionTable[I].SymbolSet, SymbolSet[K]) then
              begin
                // reduce/reduce conflict
                if FReport then
                begin
                  if n_conflicts = 0 then DoLog(TAB + '*** conflicts:');
                  DoLog(TAB + 'reduce ' + IntToStr(FReductionTable[I].RuleNumber - 1) + ', '  +
                        'reduce ' + IntToStr(RuleNumber - 1) +   ' on ' + Name(SymbolSet[K]));
                end;
                Inc(n_conflicts);
                Inc(FReduceReduceConflicts);
              end;
            ISMinus(SymbolSet, FReductionTable[I].SymbolSet);
          end;

      // count goto, shift and reduce actions to generate
      n_gotos := 0;
      n_shifts := 0;
      n_reductions := 0;

      for I := TransitionLow to TransitionHigh do
        with FTransitionTable^[I] do
          if NextState <> - 1 then
            if Symbol < 0 then Inc(n_gotos)
                          else Inc(n_shifts);

      for I := ReductionsLow to ReductionsHigh do
        with FReductionTable[I] do
          if Length(SymbolSet) > 0 then Inc(n_reductions);

      // determine default action
      if (n_shifts + n_errors = 0) and (n_reductions = 1)
        then
          // default action is the reduction
          with FReductionTable[ReductionsLow] do FGeneratorDefaultTable[S] := -(RuleNumber - 1)
        else
          // default action is error
          FGeneratorDefaultTable[S] := 0;

      // flag reduced rules 
      for I := ReductionsLow to ReductionsHigh do
        with FReductionTable[I] do reduced[RuleNumber] := True;

      if FReport then
      begin
        // list kernel items
        DoLog('');
        GetItemSet(S, ItemSet);
        Closure(ItemSet);
        SortItemSet(ItemSet);
        with ItemSet do
        begin
          for I := 1 to ItemCount do
            with Item[I], FRuleTable^[RuleNumber] do
              if (RuleNumber = 1) or (PositionNumber > 1) or(rhs_len = 0) then
                if PositionNumber > rhs_len then DoLog(TAB + ItemToStr(ItemSet, I) + TAB + '(' + IntToStr(RuleNumber - 1) + ')')
                                            else DoLog(TAB + ItemToStr(ItemSet, I));
        end;

        // list parse actions
        // shift, reduce and default actions
        if (n_shifts + n_errors = 0) and (n_reductions = 1) then
        begin
          with FReductionTable[ReductionsLow] do
          begin
            // default action is the reduction
            DoLog('');
            DoLog(TAB + '.' + TAB + 'reduce ' + IntToStr(RuleNumber - 1));
          end;
        end
        else
        begin
          // default action is Error
          DoLog('');

          for I := TransitionLow to TransitionHigh do
            with FTransitionTable^[I] do
              if NextState <> - 1 then
                if Symbol = 0 then DoLog(TAB + Name(Symbol) + TAB + 'accept') // accept action
              else
                if Symbol > 0 then DoLog(TAB + Name(Symbol) + TAB + 'shift ' + IntToStr(NextState)); // shift action

          for I := ReductionsLow to ReductionsHigh do
            with FReductionTable[I] do
              // reduce action
              for J := 0 to High(SymbolSet) do DoLog(TAB + Name(SymbolSet[J]) + TAB + 'reduce ' + IntToStr(RuleNumber - 1));

          // Error action
          DoLog(TAB + '.' + TAB + 'error');
        end;

        // goto actions
        if n_gotos > 0 then
        begin
          DoLog('');
          for I := TransitionLow to TransitionHigh do
            with FTransitionTable^[I] do
              if Symbol < 0 then DoLog(TAB + Name(Symbol) + TAB + 'goto ' + IntToStr(NextState));
        end;
      end;
    end;

  for I := 2 to FRuleCount do
    if not reduced[I] then Inc(FNeverReduced);

  // report rules never reduced
  if (FNeverReduced > 0) and FReport then
  begin
    DoLog('');
    DoLog('*** rules never reduced:');
    for I := 2 to FRuleCount do
      if not reduced[I] then DoLog(RuleToStr(I) + TAB + '(' + IntToStr(I - 1) + ')');
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.InitializeCounters;

// initialize counters and offsets 

var S, I: Integer;

begin
  FGeneratorStateCount := FStateCount;
  FGeneratorActionCount := 0;
  FGeneratorGotoCount := 0;

  for S := 0 to FStateCount - 1 do
    with FStateTable^[S] do
    begin
      FGeneratorActionLowTable[S] := FGeneratorActionCount + 1;
      FGeneratorGotoLowTable[S] := FGeneratorGotoCount + 1;
      if FGeneratorDefaultTable[S] = 0 then
      begin
        for I := TransitionLow to TransitionHigh do
          with FTransitionTable^[I] do
            if (Symbol >= 0) and (NextState <> - 1) then Inc(FGeneratorActionCount);

        for I := ReductionsLow to ReductionsHigh do
          with FReductionTable[I] do Inc(FGeneratorActionCount, Length(SymbolSet));
      end;

      for I := TransitionLow to TransitionHigh do
        with FTransitionTable^[I] do
          if Symbol < 0 then Inc(FGeneratorGotoCount);
          
      FGeneratorActionHighTable[S] := FGeneratorActionCount;
      FGeneratorGotoHighTable[S] := FGeneratorGotoCount;
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.WriteTables;

// write tables to target

var S, J: Integer;
    I, Count: Cardinal;
    Line: String;

begin
  FTarget.Add('type TActionTableEntry = record');
  FTarget.Add('       Symbol,');
  FTarget.Add('       Action: Integer;');
  FTarget.Add('     end;');
  FTarget.Add('');
  FTarget.Add('     TRuleTableEntry = record');
  FTarget.Add('       Len,');
  FTarget.Add('       Symbol: Integer;');
  FTarget.Add('     end;');
  FTarget.Add('');

  FTarget.Add('const');

  // counters
  FTarget.Add('  ShiftReduceActionCount = ' + IntToStr(FGeneratorActionCount) + ';');
  FTarget.Add('  GotoActionCount = ' + IntToStr(FGeneratorGotoCount) + ';');
  FTarget.Add('  StateCount = ' + IntToStr(FGeneratorStateCount) + ';');
  FTarget.Add('  RuleCount = ' + IntToStr(FRuleCount - 1) + ';');

  // shift/reduce table 
  FTarget.Add('');
  FTarget.Add('  ActionTable: array[1..ShiftReduceActionCount] of TActionTableEntry = (');
  Count := 0;
  for S := 0 to FStateCount - 1 do
    with FStateTable^[S] do
    begin
      FTarget.Add('{ ' + IntToStr(s) + ': }');
      if FGeneratorDefaultTable[S] = 0 then
      begin
        for I := TransitionLow to TransitionHigh do
          with FTransitionTable^[I] do
            if (NextState <> - 1) and (Symbol >= 0) then
            begin
              Inc(Count);
              if Symbol = 0 then Line := '   (Symbol: 0; Action: 0)'
                            else Line := '   (Symbol: ' + IntToStr(Symbol) + '; Action: ' + IntToStr(NextState) + ')';
              if Count < FGeneratorActionCount then Line := Line + ',';
              FTarget.Add(Line);
            end;

        for I := ReductionsLow to ReductionsHigh do
          with FReductionTable[I] do
            for J := 0 to High(SymbolSet) do
            begin
              Inc(Count);
              Line := '   (Symbol: ' + IntToStr(SymbolSet[J]) + '; Action: ' + IntToStr(-(RuleNumber - 1)) + ')';
              if Count < FGeneratorActionCount then Line := Line + ',';
              FTarget.Add(Line);
            end;
      end;
    end;
  FTarget.Add('  );');

  // goto table 
  FTarget.Add('');
  FTarget.Add('  GotoTable: array[1..FGeneratorGotoCount] of TActionTableEntry = (');
  Count := 0;
  for S := 0 to FStateCount - 1 do
    with FStateTable^[S] do
    begin
      for I := TransitionLow to TransitionHigh do
        with FTransitionTable^[I] do
          if Symbol < 0 then
          begin
            Inc(Count);
            Line := ' (Symbol: ' + IntToStr(Symbol) + '; Action: ' + IntToStr(NextState) + ')';
            if Count < FGeneratorGotoCount then Line := Line + ',';
            FTarget.Add(Line);
          end;
    end;
  FTarget.Add(');');

  // default action table
  FTarget.Add('');
  FTarget.Add('  DefaultActionTable: array[0..StateCount - 1] of Integer = (');
  for S := 0 to FStateCount - 1 do
  begin
    Line := '    ' + IntToStr(FGeneratorDefaultTable[S]);
    if S < FStateCount - 1 then Line := Line + ',';
    FTarget.Add(Line);
  end;
  FTarget.Add(');');

  // offset tables
  FTarget.Add('');
  FTarget.Add('  ActionsLow: array[0..StateCount - 1] of Integer = (');
  for S := 0 to FStateCount - 1 do
  begin
    Line := '    ' + IntToStr(FGeneratorActionLowTable[S]);
    if S < FStateCount - 1 then Line := Line + ',';
    FTarget.Add(Line);
  end;
  FTarget.Add(');');

  FTarget.Add('');
  FTarget.Add('  ActionsHigh: array[0..StateCount - 1] of Integer = (');
  for S := 0 to FStateCount - 1 do
  begin
    Line := '    ' + IntToStr(FGeneratorActionHighTable[S]);
    if S < FStateCount - 1 then Line := Line + ',';
    FTarget.Add(Line);
  end;
  FTarget.Add(');');

  FTarget.Add('');
  FTarget.Add('  GotoLow: array[0..StateCount - 1] of Integer = (');
  for S := 0 to FStateCount - 1 do
  begin
    Line := '    ' + IntToStr(FGeneratorGotoLowTable[S]);
    if S < FStateCount - 1 then Line := Line + ',';
    FTarget.Add(Line);
  end;
  FTarget.Add(');');

  FTarget.Add('');
  FTarget.Add('  GotoHigh: array[0..StateCount - 1] of Integer = (');
  for S := 0 to FStateCount - 1 do
  begin
    Line := '    ' + IntToStr(FGeneratorGotoHighTable[S]);
    if S < FStateCount - 1 then Line := Line + ',';
    FTarget.Add(Line);
  end;
  FTarget.Add(');');

  // rule table
  FTarget.Add('');
  FTarget.Add('  RuleTable: array[1..RuleCount] of TRuleTableEntry = (');
  for I := 2 to FRuleCount do with FRuleTable^[I] do
  begin
    Line := '    (Len: ' + IntToStr(rhs_len) + '; Symbol: ' + IntToStr(lhs_sym) + ')';
    if I < FRuleCount then Line := Line + ',';
    FTarget.Add(Line);
  end;
  FTarget.Add(');');
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.ParseTable;

// Constructs the parse table from the information in the state,
// transition and reduction table, and writes parse and rule table
// information to the target.
//
// Rules never reduced are detected and parsing conflicts resolved
// according to the usual disambiguting rules:
//
// - by default, shift/reduce conflicts are resolved in favour of
//   shift, and reduce/reduce conflicts are resolved in favour of
//   the rule appearing first in the grammar
//
// - in the presence of precedence information, shift/reduce conflicts
//   are resolved as follows:
//   - if the rule has higher precedence than the input symbol,
//     reduce
//   - if the input symbol has higher precedence than the rule,
//     shift
//   - if rule and input symbol have the same precedence, use
//     associativity to resolve the conflict: if the symbol is
//     left-associative, reduce; if right-associative, shift;
//     if nonassociative, Error.
//
// The default action for any state is Error, unless the state
// only has a single reduce action and no shift (or nonassoc-induced
// Error) actions, in which case the default action is the reduction.
// An accept action is generated for the shift-endmarker action.
//
// If the FReport option is enabled the ParseTable routine also writes
// a readable listing of the generated parser to the log, including
// descriptions of parse conflicts and rules never reduced.
//
// Parse table actions are encoded as follows:
// - positive: next state (shift or goto action)
// - negative: rule to reduce (reduce action)
// - 0: Error(in default action table) or accept (in shift/reduce action table)
//
// For a demonstration of the output look into this unit immediately after the
// implementation keyword.

begin
  BuildParseTables;
  InitializeCounters;
  WriteTables;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.CurrentParserChar: Char;

begin
  if FCharIndex > Length(FCurrentSourceLine) then
    if FSourceIndex = Cardinal(FSource.Count) then Result := #0
                                              else Result := NL
                                             else Result := FCurrentSourceLine[FCharIndex];
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.LookaheadChar: Char;

begin
  if Succ(FCharIndex) > Length(FCurrentSourceLine) then
    if FSourceIndex = Cardinal(FSource.Count) then Result := #0
                                              else Result := NL
                                                   else Result := FCurrentSourceLine[Succ(FCharIndex)];
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.NextChar;

begin
  if FCharIndex > Length(FCurrentSourceLine) then
    if FSourceIndex = Cardinal(FSource.Count) then // nop
                                              else NextSourceLine
                                             else Inc(FCharIndex);
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.IDFromSymbol(K: Integer): Integer;

// Returns internal symbol number for the symbol K.
// If K is yet undefined a new non-terminal or literal symbol is created, according to the
// appearance of symbol K (non-terminal if an ordinary identifier, literal otherwise).

var S: Integer;

begin
  if IsDefinedKey(K, S) then Result := S
                        else
    if FSymbolTable[K].Name[1] = '''' then
    begin
      S := MakeNewLiteral;
      DefineKey(K, S);
      Result := S;
    end
    else
    begin
      S := MakeNewNonTerminal;
      DefineKey(K, S);
      Result := S;
    end
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.NonTerminalSymbol(K: Integer): Integer;

// Like Symbol, but requires symbol K to be a non-terminal symbol.
// If it is already defined as literal an Error message is issued and a dummy
// non-terminal symbol is returned.

var S: Integer;

begin
  if IsDefinedKey(K, S) then
    if S < 0 then Result := S
    else
    begin
      Error(SNonterminalExpected, FCharIndex, True);
      Result := -1;
    end
                        else
    if FSymbolTable^[K].Name[1] = '''' then
    begin
      Error(SNonterminalExpected, FCharIndex, True);
      Result := -1;
    end
  else
  begin
    S := MakeNewNonTerminal;
    DefineKey(K, S);
    Result := S;
  end
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.LiteralSymbol(K: Integer; N: Integer): Integer;

// Same for literal symbols as in NonTerminalSymbol.
// If N > 0 it denotes the literal number to be assigned to the symbol.
// When a new literal identifier is defined a corresponding constant definition
// is also written to the target

var S: Integer;

begin
  if IsDefinedKey(K, S) then
    if S >= 0 then
    begin
      if N > 0 then Error(SDoubleTokenDef, FCharIndex, True);
      Result := S;
    end
    else
    begin
      Error(SLiteralExpected, FCharIndex, True);
      Result := 1;
    end
                        else
    if FSymbolTable^[K].Name[1] = '''' then
    begin
      if N > 0 then
      begin
        AddLiteral(N);
        S := N;
      end
      else S := MakeNewLiteral;
      DefineKey(K, S);
      Result := S;
    end
    else
    begin
      if N > 0 then
      begin
        AddLiteral(N);
        S := N;
      end
      else S := MakeNewLiteral;
      DefineKey(K, S);
      FTarget.Add('const ' + Name(S) + ' = ' + IntToStr(S) + ';');
      Result := S;
    end
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.NextTemplateSection;

// find next section mark (%%) in Code template

begin
  while FTemplateIndex < Cardinal(FTemplate.Count) do
  begin
    FCurrentTemplateLine := FTemplate[FTemplateIndex];
    if assigned(FOnTemplateLine) then FOnTemplateLine(Self, FTemplate[FTemplateIndex], FTemplateIndex + 1);
    Inc(FTemplateIndex);
    if FCurrentTemplateLine = '%%' then Break;
    FTarget.Add(FCurrentTemplateLine);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.Definitions;

// if necessary write out definition of the semantic value type YYSType

var I: Integer;

begin
  if FTypeCount > 0 then
  begin
    FTarget.Add('');
    FTarget.Add('type YYSType = record case Integer of');
    for I := 1 to FTypeCount do
      FTarget.Add(Format('          %d:(yy%S: %S);',
                         [I, FSymbolTable[FTypeTable[I]].Name, FSymbolTable[FTypeTable[I]].Name]));
    FTarget.Add('         end;');
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.CopyCode;

// Copy source Code section(%{ ... %}) to output file

var InString: Boolean;
    Line: String;
    
begin
  InString := False;
  Line :='';
  while CurrentParserChar <> #0 do
  begin
    if CurrentParserChar = NL then
    begin
      FTarget.Add(Line);
      Line := '';
      NextChar;
    end
    else
      if CurrentParserChar = '''' then
      begin
        Line := Line + '''';
        InString := not InString;
        NextChar;
      end
    else
      if not InString and (CurrentParserChar = '%') and (LookaheadChar = '}') then
      begin
        Break;
      end
      else
      begin
        Line := Line + CurrentParserChar;
        NextChar;
      end;
  end;
  if Line <> '' then FTarget.Add(Line);
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.ScanValue: String;

// process a $ value in an action
// (not very pretty, but it does its job)

var Tag,
    Number: String;
    I, Code: Integer;

begin
  FCurrentTokenLen := 0;
  Result := '';
  NextChar;
  if CurrentParserChar = '<' then
  begin
    // process type tag
    NextChar;
    Tag := '';
    while (CurrentParserChar <> NL) and (CurrentParserChar <> #0) and (CurrentParserChar <> '>') do
    begin
      Tag := Tag + CurrentParserChar;
      NextChar;
    end;

    if CurrentParserChar = '>' then
    begin
      if not SearchType(Tag) then
      begin
        FCurrentTokenLen := Length(Tag);
        Error(SUnknownIdentifier, FCharIndex, True);
      end;
      NextChar;
    end
    else Error(SSyntaxError, FCharIndex, True);
  end
  else Tag := '';
  
  FCurrentTokenLen := 0;
  if CurrentParserChar='$' then
  begin
    // left-hand side value
    Result := Result + 'yyval';
    // check for value type
    if (Tag = '') and (FTypeCount > 0) then
      with FCurrentRule do
        if FSymbolTypeTable[lhs_sym] > 0 then Tag := FSymbolTable[FSymbolTypeTable[lhs_sym]].Name
                                         else
        begin
          FCurrentTokenLen := 1;
          Error(STypeError, FCharIndex, True);
        end;
    if Tag <> '' then Result := Result + '.yy' + Tag;
    NextChar;
  end
  else
  begin
    // right-hand side value
    if CurrentParserChar = '-' then
    begin
      Number := '-';
      NextChar;
    end
    else Number := '';
    
    while CurrentParserChar in ['0'..'9'] do
    begin
      Number := Number + CurrentParserChar;
      NextChar;
    end;

    if Number <> '' then
    begin
      Val(Number, I, Code);
      if Code = 0 then
        if I <= FCurrentRule.rhs_len then
        begin
          Result := Result + 'yyv[yysp - ' + IntToStr(FCurrentRule.rhs_len - I) + ']';
          // check for value type
          if (Tag = '') and (FTypeCount > 0) then
            with FCurrentRule do
              if I <= 0 then
              begin
                FCurrentTokenLen := Length(Number) + 1;
                Error(STypeError, FCharIndex, True);
              end
              else
                if FSymbolTypeTable^[rhs_sym[I]] > 0 then Tag := FSymbolTable^[FSymbolTypeTable^[rhs_sym[I]]].Name
                                                     else
                begin
                  FCurrentTokenLen := Length(Number) + 1;
                  Error(STypeError, FCharIndex, True);
                end;

          if Tag <> '' then Result := Result + '.yy' + Tag;
        end
        else
        begin
          FCurrentTokenLen := Length(Number);
          Error(SRangeError, FCharIndex, True);
        end
      else Error(SSyntaxError, FCharIndex, True)
    end
    else Error(SSyntaxError, FCharIndex, True)
  end
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.CopyAction;

// Copy an action to the output file

var InString: Boolean;
    Line: String;

begin
  InString := False;
  while CurrentParserChar = ' ' do NextChar;

  Line := '         ';
  while CurrentParserChar <> #0 do
    if CurrentParserChar = NL then
    begin
      FTarget.Add(Line);
      NextChar;
      while CurrentParserChar = ' ' do NextChar;
      Line := '         ';
    end
    else
      if CurrentParserChar = '''' then
      begin
        Line := Line + '''';
        InString := not InString;
        NextChar;
      end
      else
        if not InString and (CurrentParserChar = '}') then
        begin
          Break;
        end
        else
          if not InString and (CurrentParserChar='$') then Line := Line + ScanValue
                                                      else
          begin
            Line := Line + CurrentParserChar;
            NextChar;
          end;
  if Line <> '         ' then FTarget.Add(Line);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.CopySingleAction;

// like CopyAction, but action must be single statement terminated with ';'

var InString: Boolean;
    Line: String;

begin
  InString := False;
  while CurrentParserChar = ' ' do NextChar;

  Line := '         ';
  while CurrentParserChar <> #0 do
    if CurrentParserChar = NL then
    begin
      FTarget.Add(Line);
      NextChar;
      while CurrentParserChar = ' ' do NextChar;
      Line := '         ';
    end
    else
      if CurrentParserChar = '''' then
      begin
        Line := Line + '''';
        InString := not InString;
        NextChar;
      end
      else
        if not InString and (CurrentParserChar = ';') then
        begin
          Break;
        end
        else
          if not InString and (CurrentParserChar = '$') then ScanValue
                                                        else
          begin
            Line := Line + CurrentParserChar;
            NextChar;
          end;
  if Line <> '         ' then FTarget.Add(Line + ';');
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.CopyRestOfFile;

// copies the rest of the source file to the output file

var Line: String;

begin
  Line := '';
  while CurrentParserChar <> #0 do
    if CurrentParserChar = NL then
    begin
      FTarget.Add(Line);
      Line := '';
      NextChar;
    end
    else
    begin
      Line := Line + CurrentParserChar;
      NextChar;
    end;
  if Line <> '' then FTarget.Add(Line);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.StartRule(Symbol: Integer);

// start a new rule with lhs non-terminal symbol sym

begin
  if FRuleCount = 0 then
  begin
    // fix start non-terminal of the grammar
    if FStartNonTerminal = 0 then FStartNonTerminal := Symbol;
    // add augmented start production:
    with FCurrentRule do
    begin
      lhs_sym := -1;
      rhs_len := 2;
      rhs_sym[1] := FStartNonTerminal;
      rhs_sym[2] := 0; // end marker
    end;
    AddRule(FCurrentRule);
  end;
  FCurrentRule.lhs_sym := Symbol;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.StartBody;

// start a new rule body (rhs)

begin
  FCurrentRule.rhs_len := 0;
  FActionMarker := False;
  FTarget.Add('    ' + IntToStr(FRuleCount) + ': begin');
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.EndBody;

// end a rule body

begin
  if not FActionMarker and (FCurrentRule.rhs_len > 0) then
    // add default action:
    FTarget.Add('         yyval := yyv[yysp - ' + IntToStr(FCurrentRule.rhs_len - 1) + '];');
  AddRule(FCurrentRule);
  FTarget.Add('       end;');
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddRuleAction;

// process an action inside a rule

var K: Integer;
    R: TRuleEntry;

begin
  FTarget.Add('       end;');
  Inc(FActionCount);
  K := GetKey('$$' + IntToStr(FActionCount));
  with R do
  begin
    lhs_sym := MakeNewNonTerminal;
    DefineKey(K, lhs_sym);
    rhs_len := 0;
  end;

  with FCurrentRule do
  begin
    Inc(rhs_len);
    if rhs_len > MaxRuleLength then FatalError('rule table overflow');
    rhs_sym[rhs_len] := R.lhs_sym;
  end;
  AddRule(R);
  FRulePrecendenceTable^[FRuleCount + 1] := FRulePrecendenceTable^[FRuleCount];
  FRulePrecendenceTable^[FRuleCount] := 0;
  FTarget.Add('    ' + IntToStr(FRuleCount) + ': begin');
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddSymbol(Symbol: Integer);

// add the denoted symbol to the current rule body

begin
  if FActionMarker then AddRuleAction;
  FActionMarker := False;
  with FCurrentRule do
  begin
    Inc(rhs_len);
    if rhs_len > MaxRuleLength then FatalError('rule table overflow');
    rhs_sym[rhs_len] := Symbol;
    if Symbol >= 0 then
      FRulePrecendenceTable^[FRuleCount + 1] := FSymbolPrecendenceTable^[Symbol];
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddAction;

// add an action to the current rule body

begin
  if FActionMarker then AddRuleAction;
  FActionMarker := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddRulePrecedence(Symbol: Integer);

// add the precedence of terminal symbol sym to the current rule

begin
  FRulePrecendenceTable^[FRuleCount + 1] := FSymbolPrecendenceTable^[Symbol];
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.GenerateParser;

// generate the parse table

begin
  if FStartNonTerminal = 0 then Error(SEmptyGrammar, 0, True);
  if FErrors = 0 then
  begin
    DoLog('sort ... ');
    SortRules;
    RuleOffsets;
    DoLog('closures ... ');
    Closures;
    DoLog('first sets ... ');
    FirstSets;
    DoLog('LR0 set ... ');
    LR0Set;
    DoLog('lookaheads ... ');
    Lookaheads;
    DoLog('code generation ... ');
    ParseTable;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.MakeNewNonTerminal: Integer;

// returns a new non-terminal number (< -1)

begin
  Inc(FNonTerminalCount);
  if FNonTerminalCount > MaxNonterminals then FatalError('nt table overflow');
  FSymbolTypeTable^[-FNonTerminalCount] := 0;
  Result := -FNonTerminalCount;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.MakeNewLiteral: Integer;

// returns a new literal number above 256

begin
  if FLiteralsCount >= MaxLiterals then FatalError('lit table overflow');
  FSymbolTypeTable^[FLiteralsCount] := 0;
  FSymbolPrecendenceTable^[FLiteralsCount] := 0;
  Result := FLiteralsCount;
  Inc(FLiteralsCount);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddLiteral(Symbol: Integer);

// This routine allows to add a user-defined literal symbol.
// The current literal symbols count is adjusted accordingly.

begin
  if Cardinal(Symbol) > FLiteralsCount then FLiteralsCount := Symbol;
  if FLiteralsCount > MaxLiterals then FatalError('lit table overflow');
  FSymbolTypeTable^[Symbol] := 0;
  FSymbolPrecendenceTable^[Symbol] := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.Lookup(K: Integer): String;

// get print name of symbol number K

begin
  with FSymbolTable^[K] do Result := Name
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.Entry(K: Integer; Symbol: String);

// enter symbol into table

begin
  FSymbolTable^[K].Name := Symbol;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.GetKey(Symbol: String): Integer;

// returns a hash key for symbol

begin
  Result := Key(Symbol, MaxKeys, Lookup, Entry);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.DefineKey(K: Integer; Symbol: Integer);

// defines K to be a new symbol with internal symbol number sym

begin
  FSymbolKeyTable^[Symbol] := K;
  FSymbolTable^[K].defined := True;
  FSymbolTable^[K].Symbol := Symbol;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.IsDefinedKey(K: Integer; var Symbol: Integer): Boolean;

// Checks whether symbol denoted by symbol table key K is already defined.
// If so then the corresponding symbol number is returned.

begin
  if FSymbolTable^[K].defined then
  begin
    Symbol := FSymbolTable^[K].Symbol;
    Result := True;
  end
  else Result := False
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.Name(Symbol: Integer): String;

// returns the print name of an internal symbol ('$end' for
// symbol 0, '$accept' for non-terminal -1 and a single quoted
// character for literals 1..255)

begin
  case Symbol of
    1..255: Result := SingleQuoteStr(Chr(Symbol));
    0     : Result := '$end';
    -1    : Result := '$accept';
  else
    if FSymbolTable[FSymbolKeyTable^[Symbol]].Name[1] = '''' then
      Result := SingleQuoteStr(Copy(FSymbolTable[FSymbolKeyTable^[Symbol]].Name, 2,
                               Length(FSymbolTable[FSymbolKeyTable[Symbol]].Name) - 2))
    else Result := FSymbolTable^[FSymbolKeyTable^[Symbol]].Name;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddRule(R: TRuleEntry);

// add a rule to the rule table

begin
  Inc(FRuleCount);
  if FRuleCount > MaxRules then FatalError('rule table overflow');
  FRuleTable^[FRuleCount] := R;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.CompareRules(I, J: Integer): Boolean;

begin
  if FRuleTable^[FRuleNumberTable[I]].lhs_sym = FRuleTable^[FRuleNumberTable[J]].lhs_sym
    then Result := FRuleNumberTable[I] < FRuleNumberTable[J]
    else Result := FRuleTable^[FRuleNumberTable[I]].lhs_sym > FRuleTable^[FRuleNumberTable[J]].lhs_sym;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.SwapRules(I, J: Integer);

var X: Integer;

begin
  X := FRuleNumberTable[I];
  FRuleNumberTable[I] := FRuleNumberTable[J];
  FRuleNumberTable[J] := X;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.SortRules;

// sorts rules with regard to left-hand sides into the rule no table

var I: Integer;

begin
  for I := 1 to FRuleCount do FRuleNumberTable[I] := I;
  QuickSort(1, FRuleCount, CompareRules, SwapRules);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.RuleOffsets;

// computes rule offsets after rules have been sorted

var I: Cardinal;
    Symbol: Integer;

begin
  for Symbol := 1 to FNonTerminalCount do
    with FRuleOffsetTable^[Symbol] do
    begin
      RuleLow := 1;
      RuleHigh := 0;
    end;

  I := 1;
  while (I <= FRuleCount) do
  begin
    Symbol := FRuleTable^[FRuleNumberTable[I]].lhs_sym;
    FRuleOffsetTable^[-Symbol].RuleLow := I;
    Inc(I);
    while (I <= FRuleCount) and (FRuleTable^[FRuleNumberTable[I]].lhs_sym = Symbol) do Inc(I);
    FRuleOffsetTable^[-Symbol].RuleHigh := I - 1;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddType(K: Integer);

// add a type identifier to the table

begin
  Inc(FTypeCount);
  if FTypeCount > MaxTypeTags then FatalError('type table overflow');
  FTypeTable^[FTypeCount] := K;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.CompareTypes(I, J: Integer): Boolean;

begin
  Result := FSymbolTable^[FTypeTable^[I]].Name  < FSymbolTable^[FTypeTable^[J]].Name;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.SwapTypes(I, J: Integer);

var X: Integer;

begin
  X := FTypeTable^[I];
  FTypeTable^[I] := FTypeTable^[J];
  FTypeTable^[J] := X;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.SearchType(Symbol: String): Boolean;

// search the sorted types table for the given type symbol

var L, R, K: Cardinal;

begin
  // binary search
  L := 1;
  R := FTypeCount;
  K := L  + (R-L) div 2;
  while (L < R) and (FSymbolTable^[FTypeTable^[K]].Name  <> Symbol) do
  begin
    if FSymbolTable^[FTypeTable^[K]].Name < Symbol then L := Succ(K)
                                                   else R := Pred(K);
    K := L  + (R - L) div 2;
  end;
  Result := (K <= FTypeCount) and (FSymbolTable^[FTypeTable^[K]].Name = Symbol);
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.MakeNewPrecedenceLevel(PrecedenceType: TPrecedenceType): Integer;

// adds a new precedence level of the denoted type; returns: the New level

begin
  Inc(FPrecedenceCount);
  if FPrecedenceCount > MaxPrecLevel then FatalError('prec table overflow');
  FPrecendenceTable^[FPrecedenceCount] := PrecedenceType;
  Result := FPrecedenceCount;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.MakeNewState;

// build a new state

begin
  Inc(FStateCount);
  if FStateCount > MaxStates then FatalError('state table overflow');
  FStateTable^[FStateCount - 1].ItemLow := FItemCount + 1;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddItem(RuleNumber, PositionNumber: Integer);

// add an item to the new state (initialize its next field to 0)

begin
  Inc(FItemCount);
  if FItemCount > MaxItems then FatalError('item table overflow');
  FItemTable^[FItemCount].RuleNumber := RuleNumber;
  FItemTable^[FItemCount].PositionNumber := PositionNumber;
  FItemTable^[FItemCount].Next := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.AddState: Integer;

// Adds the new state to the state table. If an equivalent state is already
// in the table, dispose the new state and return the existing state number,
// otherwise return the new state number.

  //--------------- local functions -------------------------------------------

  function HashFromState(S: Integer): Integer;

  // determines a hash key for state S

  const MaxStateKeys = 4001;

  // should be prime number so that hash keys are distributed evenly

  var I, K: Integer;

  begin
    with FStateTable^[S] do
    begin
      K := 0;
      for I := ItemLow to ItemHigh do
        with FItemTable^[I] do Inc(K, RuleNumber + PositionNumber);
      Result := K mod MaxStateKeys;
    end;
  end;

  //---------------------------------------------------------------------------

  function SearchState(S, Lo, Hi: Integer; var T: Integer): Boolean;

  // Searches the range Lo..Hi in the state table for a state with the
  // same kernel items as S. Returns True if found and then the
  // corresponding state number in T.

    //-------------------------------------------------------------------------

    function ItemAreEqual(S, T: Integer): Boolean;

    // compares kernel Item sets of states S and t

    var I, ItemS, ItemT: Integer;

    begin
      if StateItemCount(S) <> StateItemCount(T) then Result := False
                                                else
      begin
        ItemS := FStateTable^[S].ItemLow;
        ItemT := FStateTable^[t].ItemLow;
        for I := 0 to StateItemCount(S) - 1 do
          if (FItemTable^[ItemS + I].RuleNumber <> FItemTable^[ItemT + I].RuleNumber) or
             (FItemTable^[ItemS + I].PositionNumber <> FItemTable^[ItemT + I].PositionNumber) then
          begin
            Result := False;
            Exit;
          end;
        Result := True;
      end
    end;

    //-------------------------------------------------------------------------

  var I: Integer;

  begin
    with FStateTable^[S] do
      for I := Lo to Hi do
        if (Key = FStateTable^[I].Key) and ItemAreEqual(S, I) then
        begin
          Result := True;
          T := I;
          Exit;
        end;
    Result := False;
  end;

  //--------------- end local functions ---------------------------------------

var S: Integer;

begin
  with FStateTable^[FStateCount - 1] do
  begin
    ItemHigh := FItemCount;
    Key := HashFromState(FStateCount - 1);
    if SearchState(FStateCount - 1, 0, FStateCount - 2, S) then
    begin
      FItemCount := ItemLow;
      Dec(FStateCount);
      Result := S;
    end
    else Result := FStateCount - 1;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.StartTransition;

// starts building transitions of the active state

begin
  FStateTable^[FCurrentState].TransitionLow := FTransitionCount + 1;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddTransition(Symbol, NextState: Integer);

// adds a transition to the active state

begin
  Inc(FTransitionCount);
  if FTransitionCount > MaxTransitions then FatalError('trans table overflo9w');
  FTransitionTable^[FTransitionCount].Symbol := Symbol;
  FTransitionTable^[FTransitionCount].NextState := NextState;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.EndTransitions;

// ends transitions of the active state

begin
  FStateTable^[FCurrentState].TransitionHigh := FTransitionCount;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.StartReductions;

// starts building reduction actions of the active state

begin
  FStateTable^[FCurrentState].ReductionsLow := FReductionCount + 1;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.AddReduction(SymbolSet: TIntegerList; RuleNumber: Integer);

// adds a reduction to the active state

begin
  Inc(FReductionCount);
  if FReductionCount>MaxReductions then FatalError('redn table overflow');
  FReductionTable[FReductionCount].SymbolSet := SymbolSet;
  FReductionTable[FReductionCount].RuleNumber := RuleNumber;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.EndReductions;

// ends reduction actions of the active state

begin
  FStateTable^[FCurrentState].ReductionsHigh := FReductionCount;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.StateItemCount(S: Integer): Integer;

begin
  with FStateTable^[S] do Result := ItemHigh - ItemLow + 1
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.FindItem(S: Integer; RuleNumber, PositionNumber: Integer): Integer;

// finds item (RuleNumber, PositionNumber) in state S
// returns the item number

var I: Integer;

begin
  Result := 0;
  with FStateTable^[S] do
    for I := ItemLow to ItemHigh do
      if (FItemTable^[I].RuleNumber = RuleNumber) and
         (FItemTable^[I].PositionNumber=PositionNumber) then
      begin
        Result := I;
        Break;
      end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.IncludeItemSet(var ItemSet: TItemSetEntry; RuleNumber, PositionNumber: Integer);

// add the denoted Item to the given Item set

begin
  with ItemSet do
  begin
    Inc(ItemCount);
    if ItemCount > MaxSetItems then FatalError('item table overflow');
    Item[ItemCount].RuleNumber := RuleNumber;
    Item[ItemCount].PositionNumber := PositionNumber;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.GetItemSet(S: Integer; var ItemSet: TItemSetEntry);

// obtain the item set of state S from the item table

begin
  with FStateTable^[S], ItemSet do
  begin
    ItemCount := StateItemCount(S);
    Move(FItemTable^[ItemLow], Item, ItemCount * SizeOf(TItemEntry));
  end
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.Closure(var ItemSet: TItemSetEntry);

// compute the closure of ItemSet (using the closure table)

var I, J: Integer;
    NTSymbols0,
    NTSymbols: TIntegerList;

begin
  with ItemSet do
  begin
    // get the nonterminals at current positions in items
    NTSymbols0 := nil;
    for I := 1 to ItemCount do
      with Item[I], FRuleTable^[RuleNumber] do
        if (PositionNumber <= rhs_len) and (rhs_sym[PositionNumber] < 0) then ISInclude(NTSymbols0, rhs_sym[PositionNumber]);
    NTSymbols := NTSymbols0;

    // add closure symbols
    for I := 0 to High(NTSymbols0) do ISUnion(NTSymbols, FClosureTable[-NTSymbols0[I]]);

    // add the non-kernel items for the non-terminal symbols
    for I := 0 to Length(NTSymbols) - 1 do
      with FRuleOffsetTable^[-NTSymbols[I]] do
        for J := RuleLow to RuleHigh do IncludeItemSet(ItemSet, FRuleNumberTable[J], 1);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

// comparison and swap routines for SortItemSet:

function TParserGenerator.CompareItems(I, J: Integer): Boolean;

begin
  with FTempItems do
    if Item[I].PositionNumber = Item[J].PositionNumber
      then Result := Item[I].RuleNumber < Item[J].RuleNumber
      else Result := Item[I].PositionNumber > Item[J].PositionNumber;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.SwapItems(I, J: Integer);

var Entry: TItemEntry;

begin
  with FTempItems do
  begin
    Entry := Item[I];
    Item[I] := Item[J];
    Item[J] := Entry;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.SortItemSet(var ItemSet: TItemSetEntry);

// sorts an Item set with regard to position and rule numbers (higher positions,
// lower rules first)

begin
  FTempItems := ItemSet;
  QuickSort(1, FTempItems.ItemCount, CompareItems, SwapItems);
  ItemSet := FTempItems;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.SortTypes;

var I, J, Count: Cardinal;

begin
  QuickSort(1, FTypeCount, CompareTypes, SwapTypes);

  // eliminate dups
  I := 1;
  J := 1;
  Count := 0;
  while I <= FTypeCount do
  begin
    if I <> J then FTypeTable^[J] := FTypeTable^[I];
    while (I < FTypeCount) and (FTypeTable^[I + 1] = FTypeTable^[I]) do
    begin
      Inc(I);
      Inc(Count);
    end;
    Inc(I);
    Inc(J);
  end;
  Dec(FTypeCount, Count);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.ParserNoError;

begin
  FLastError := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.ParserAccept;

begin
  FAcceptance := peAccept;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.ParserAbort;

begin
  FAcceptance := peAbort;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.ParserError;

begin
  FAcceptance := peError;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.RuleToAction(Rule: Integer);

begin
  case Rule of
   1: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
   2: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
   3: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
   4: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
   5: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
   6: begin
        ParserNoError;
      end;
   7: begin
        ParserNoError;
      end;
   8: begin
        ParserNoError;
      end;
   9: begin
        ParserNoError;
      end;
  10: begin
        ParserNoError;
      end;
  11: begin
        ParserNoError;
      end;
  12: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  13: begin
        ParserNoError;
      end;
  14: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  15: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  16: begin
        Error(SRightCurlyExpected, FCharIndex, True);
      end;
  17: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  18: begin
        ParserNoError;
      end;
  19: begin
        ParserNoError;
      end;
  20: begin
        ParserNoError;
      end;
  21: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  22: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  23: begin
        Error(SRightBraceExpected, FCharIndex, True);
      end;
  24: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  25: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  26: begin
        Error(SRightAngleExpected, FCharIndex, True);
      end;
  27: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  28: begin
        SortTypes;
        Definitions;
        NextTemplateSection;
       end;
  29: begin
        NextTemplateSection;
        GenerateParser;
        NextTemplateSection;
      end;
  30: begin
        FCurrentValue := FValues[FStackPointer - 5];
      end;
  31: begin
      end;
  32: begin
        CopyRestOfFile;
      end;
  33: begin
      end;
  34: begin
        ParserNoError;
      end;
  35: begin
        Error(SErrorInDefinition, FCharIndex, True);
      end;
  36: begin
        FStartNonTerminal := NonTerminalSymbol(FValues[FStackPointer - 0]);
      end;
  37: begin
        Error(SIdentifierExpected, FCharIndex, True);
      end;
  38: begin
        CopyCode;
      end;
  39: begin
        FCurrentValue := FValues[FStackPointer - 2];
      end;
  40: begin
        FCurrentPrecedence := 0;
      end;
  41: begin
        FCurrentValue := FValues[FStackPointer - 3];
      end;
  42: begin
        FCurrentPrecedence := MakeNewPrecedenceLevel(ptLeft);
      end;
  43: begin
        FCurrentValue := FValues[FStackPointer - 3];
      end;
  44: begin
        FCurrentPrecedence := MakeNewPrecedenceLevel(ptRight);
      end;
  45: begin
        FCurrentValue := FValues[FStackPointer - 3];
      end;
  46: begin
        FCurrentPrecedence := MakeNewPrecedenceLevel(ptNonAssoc);
      end;
  47: begin
        FCurrentValue := FValues[FStackPointer - 3];
      end;
  48: begin
        FCurrentValue := FValues[FStackPointer - 2];
      end;
  49: begin
        FCurrentValue := FValues[FStackPointer - 1];
      end;
  50: begin
        FCurrentTypeTag := 0;
      end;
  51: begin
        FCurrentTypeTag := FValues[FStackPointer - 1];
        AddType(FValues[FStackPointer - 1]);
      end;
  52: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  53: begin
        ParserNoError;
      end;
  54: begin
        ParserNoError;
      end;
  55: begin
        Error(SIdentifierExpected, FCharIndex, True);
      end;
  56: begin
        Error(SErrorInDefinition, FCharIndex, True);
      end;
  57: begin
        Error(SIdentifierExpected, FCharIndex, True);
      end;
  58: begin
        if FCurrentTypeTag <> 0 then FSymbolTypeTable[FValues[FStackPointer - 0]] := FCurrentTypeTag;
        if FCurrentPrecedence <> 0 then FSymbolPrecendenceTable[FValues[FStackPointer - 0]] := FCurrentPrecedence;
      end;
  59: begin
        LiteralSymbol(FValues[FStackPointer - 0], 0);
        if FCurrentTypeTag <> 0 then FSymbolTypeTable^[LiteralSymbol(FValues[FStackPointer - 0], 0)] := FCurrentTypeTag;
        if FCurrentPrecedence <> 0 then FSymbolPrecendenceTable^[LiteralSymbol(FValues[FStackPointer - 0], 0)] := FCurrentPrecedence;
      end;
  60: begin
        LiteralSymbol(FValues[FStackPointer - 0], 0);
        if FCurrentTypeTag <> 0 then FSymbolTypeTable^[LiteralSymbol(FValues[FStackPointer - 0], 0)] := FCurrentTypeTag;
        if FCurrentPrecedence <> 0 then FSymbolPrecendenceTable^[LiteralSymbol(FValues[FStackPointer - 0], 0)] := FCurrentPrecedence;
      end;
  61: begin
        LiteralSymbol(FValues[FStackPointer - 1], 0);
        if FCurrentTypeTag <> 0 then FSymbolTypeTable^[LiteralSymbol(FValues[FStackPointer - 1], FValues[FStackPointer - 0])] := FCurrentTypeTag;
        if FCurrentPrecedence <> 0 then FSymbolPrecendenceTable^[LiteralSymbol(FValues[FStackPointer - 1], 0)] := FCurrentPrecedence;
      end;
  62: begin
        LiteralSymbol(FValues[FStackPointer - 1], 0);
        if FCurrentTypeTag <> 0 then FSymbolTypeTable^[LiteralSymbol(FValues[FStackPointer - 1], FValues[FStackPointer - 0])] := FCurrentTypeTag;
        if FCurrentPrecedence <> 0 then FSymbolPrecendenceTable^[LiteralSymbol(FValues[FStackPointer - 1], 0)] := FCurrentPrecedence;
      end;
  63: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  64: begin
        ParserNoError;
      end;
  65: begin
        ParserNoError;
      end;
  66: begin
        Error(SIdentifierExpected, FCharIndex, True);
      end;
  67: begin
        Error(SErrorInDefinition, FCharIndex, True);
      end;
  68: begin
        Error(SIdentifierExpected, FCharIndex, True);
      end;
  69: begin
        if FCurrentTypeTag <> 0 then FSymbolTypeTable^[NonTerminalSymbol(FValues[FStackPointer - 0])] := FCurrentTypeTag;
      end;
  70: begin
        NextTemplateSection;
      end;
  71: begin
        FCurrentValue := FValues[FStackPointer - 1];
      end;
  72: begin
        CopyCode;
      end;
  73: begin
        NextTemplateSection;
      end;
  74: begin
        FCurrentValue := FValues[FStackPointer - 4];
      end;
  75: begin
        ParserNoError;
      end;
  76: begin
        Error(SErrorInRule, FCharIndex, True);
      end;
  77: begin
        Error(SErrorInRule, FCharIndex, True);
      end;
  78: begin
        StartRule(NonTerminalSymbol(FValues[FStackPointer - 0]));
      end;
  79: begin
        StartBody;
      end;
  80: begin
        EndBody;
      end;
  81: begin
        FCurrentValue := FValues[FStackPointer - 0];
      end;
  82: begin
        StartBody;
      end;
  83: begin
        EndBody;
      end;
  84: begin
      end;
  85: begin
        AddSymbol(FValues[FStackPointer - 0]);
        ParserNoError;
      end;
  86: begin
        AddSymbol(IDFromSymbol(FValues[FStackPointer - 0]));
        ParserNoError;
      end;
  87: begin
        AddSymbol(IDFromSymbol(FValues[FStackPointer - 0]));
        ParserNoError;
      end;
  88: begin
        AddAction;
        ParserNoError;
      end;
  89: begin
        Error(SErrorInRule, FCharIndex, True);
      end;
  90: begin
        CopyAction;
      end;
  91: begin
        FCurrentValue := FValues[FStackPointer - 2];
      end;
  92: begin
        CopySingleAction;
      end;
  93: begin
      end;
  94: begin
        AddRulePrecedence(FValues[FStackPointer - 0]);
      end;
  95: begin
        FCurrentValue := FValues[FStackPointer - 3];
      end;
  96: begin
        AddRulePrecedence(LiteralSymbol(FValues[FStackPointer - 0], 0));
      end;
  97: begin
        FCurrentValue := FValues[FStackPointer - 3];
      end;
  98: begin
        AddRulePrecedence(LiteralSymbol(FValues[FStackPointer - 0], 0));
      end;
  99: begin
        FCurrentValue := FValues[FStackPointer - 3];
      end;
 100: begin
        FCurrentValue := FValues[FStackPointer - 1];
      end;
 101: begin
      end;
 102: begin
        AddAction;
      end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function FindAction(State, Symbol: Integer; var Action: Integer): Boolean;

// search action table

var K: Integer;

begin
  K := ActionsLow[State];
  while (K <= ActionsHigh[State]) and (ActionTable[K].Symbol <> Symbol) do Inc(K);
  if K > ActionsHigh[State] then Result := False
                            else
  begin
    Action := ActionTable[K].Action;
    Result := True;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function FindNewState(State, Symbol: Integer; var NewState: Integer): Boolean;

// search goto table

var K: Integer;

begin
  K := GotoLow[State];
  while (K <= GotoHigh[State]) and (GotoTable[K].Symbol <> Symbol) do Inc(K);
  if K > GotoHigh[State] then Result := False
                         else
  begin
    NewState := GotoTable[K].Action;
    Result := True;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.Parse: Integer;

// main parsing routine
// parses the input

label parse, next, error, errlab, shift, reduce, accept, abort;

begin
  DoLog('parsing source grammar ... ');
  // initialize
  FParserState := 0;
  FLastToken := -1;
  FErrors := 0;
  FLastError := 0;
  FStackPointer := 0;
  FLastValue := 0;

parse:

  // push state and value
  Inc(FStackPointer);
  if FStackPointer > MaxDepth then
  begin
    FatalError('Parse stack overflow');
    goto abort;
  end;
  FStack[FStackPointer] := FParserState;
  FValues[FStackPointer] := FCurrentValue;

Next:
  if (DefaultActionTable[FParserState] = 0) and (FLastToken = - 1) then
  begin
    // get next symbol
    FLastToken := NextToken;
    if FLastToken < 0 then FLastToken := 0;
  end;

  // determine parse action
  FNextAction := DefaultActionTable[FParserState];
  if FNextAction <> 0 then goto reduce; // simple state

  // no default action, search parse table
  if not FindAction(FParserState, FLastToken, FNextAction) then goto Error
                                                           else
    if FNextAction > 0 then goto shift
                       else
      if FNextAction < 0 then goto reduce
                         else goto accept;

error:
  // start error recovery
  // if FLastError=0 then yyerror('syntax Error'); syntax errror wird im Original ignoriert

errlab:
  // new error?
  if FLastError = 0 then Inc(FErrors);

  if FLastError <= 2 then                 // incomplete recovery, try again 
  begin
    FLastError := 3;
    // uncover a state with shift action on error token
    while (FStackPointer > 0) and
           not(FindAction(FStack[FStackPointer], ERRORTOKEN, FNextAction) and
           (FNextAction > 0)) do
    begin
      if FReport then
        if FStackPointer > 1
          then DoLog('Error recovery pops state ' + IntToStr(FStack[FStackPointer]) + ', uncovers ' + IntToStr(FStack[FStackPointer - 1]))
          else DoLog('Error recovery fails ... abort');
      Dec(FStackPointer);
    end;

    if FStackPointer = 0 then goto abort; // parser has fallen from stack abort
    FParserState := FNextAction;          // simulate shift on Error
    goto parse;
  end
  else                                    // no shift yet, discard symbol
  begin
    if FReport then DoLog('Error recovery discards char ' + Char(FLastToken));
    if FLastToken = 0 then goto abort;    // end of input, abort
    FLastToken := -1; goto Next;          // clear lookahead char and try again
  end;

shift:
  // go to new state, clear lookahead character
  FParserState := FNextAction;
  FLastToken := -1;
  FCurrentValue := FLastValue;
  if FLastError <> 0 then Dec(FLastError);
  goto parse;

reduce:
  // execute action, pop rule from stack and go to next state
  FAcceptance := peNone;
  RuleToAction(-FNextAction);
  Dec(FStackPointer, RuleTable[-FNextAction].Len);
  if FindNewState(FStack[FStackPointer], RuleTable[-FNextAction].Symbol, FNextAction) then FParserState := FNextAction;

  // handle action calls to ParserAccept, ParserAbort and ParserError
  case FAcceptance of
    peAccept: goto accept;
    peAbort: goto abort;
    peError: goto errlab;
  end;
  goto parse;

accept:
  Result := 0;
  Exit;

abort:
  Result := 1;
  Exit;
end;

//----------------------------------------------------------------------------------------------------------------------

// lexical analyzer

function TParserGenerator.NextToken: Integer;

  //--------------- local functions -------------------------------------------

  function EOI: Boolean;

  begin
    EOI := (FCharIndex > Length(FCurrentSourceLine)) and (FSourceIndex = Cardinal(FSource.Count));
  end;

  //---------------------------------------------------------------------------

  procedure Scan;

  // scans for nonempty character, skips comments

    //-------------------------------------------------------------------------

    procedure ScanComment;

    var I: Integer;

    begin
      I := Pos('*/', Copy(FCurrentSourceLine, FCharIndex, Length(FCurrentSourceLine)));
      if I > 0 then FCharIndex := FCharIndex + Succ(I)
               else
      begin
        while (I = 0) and NextSourceLine do I := Pos('*/', FCurrentSourceLine);

        if I = 0 then
        begin
          FCharIndex := Succ(Length(FCurrentSourceLine));
          Error(SOpenCommentAtEOI, FCharIndex, True);
        end
        else FCharIndex := I + 2;
      end
    end;

    //-------------------------------------------------------------------------

    begin
      while not EOI do
        if FCharIndex <= Length(FCurrentSourceLine) then
        begin
          case FCurrentSourceLine[FCharIndex] of
            ' ',
            TAB: Inc(FCharIndex);
            '/': if (FCharIndex < Length(FCurrentSourceLine)) and (FCurrentSourceLine[Succ(FCharIndex)] = '*') then
                 begin
                   Inc(FCharIndex, 2);
                   ScanComment
                 end
                 else Exit;
          else Exit;
          end;
        end
        else NextSourceLine;
    end;

  //---------------------------------------------------------------------------
  
  function ScanIdentifier: Integer;

  var IDString: String;

  begin
    IDString := FCurrentSourceLine[FCharIndex];
    Inc(FCharIndex);
    while (FCharIndex <= Length(FCurrentSourceLine)) and
          (FCurrentSourceLine[FCharIndex] in Alpha + ['.']) do
    begin
      IDString := IDString + FCurrentSourceLine[FCharIndex];
      Inc(FCharIndex);
    end;
    FLastValue := GetKey(IDString);
    Scan;
    if not EOI and (FCurrentSourceLine[FCharIndex] = ':') then Result := C_ID
                                                          else ScanIdentifier := ID;
  end;

  //---------------------------------------------------------------------------

  function ScanLiteral: Integer;

  // scan a literal, i.e. string
  var IDString: String;
      Octal: Byte;

  begin
    IDString := FCurrentSourceLine[FCharIndex];
    Inc(FCharIndex);
    while (FCharIndex <= Length(FCurrentSourceLine)) and (FCurrentSourceLine[FCharIndex] <> IDString[1]) do
      if FCurrentSourceLine[FCharIndex] = '\' then
        if FCharIndex < Length(FCurrentSourceLine) then
        begin
          Inc(FCharIndex);
          case FCurrentSourceLine[FCharIndex] of
            'N': begin
                   IDString := IDString + NL;
                   Inc(FCharIndex);
                 end;
            'R': begin
                   IDString := IDString + CR;
                   Inc(FCharIndex);
                 end;
            't': begin
                   IDString := IDString + TAB;
                   Inc(FCharIndex);
                 end;
            'b': begin
                   IDString := IDString + BS;
                   Inc(FCharIndex);
                 end;
            'f': begin
                   IDString := IDString + FF;
                   Inc(FCharIndex);
                 end;
            '0'..'7': begin
                        Octal := Ord(FCurrentSourceLine[FCharIndex]) - Ord('0');
                        Inc(FCharIndex);
                        while (FCharIndex <= Length(FCurrentSourceLine)) and
                            (FCurrentSourceLine[FCharIndex] in ['0'..'7']) do
                        begin
                          Octal := Octal * 8 + Ord(FCurrentSourceLine[FCharIndex]) - Ord('0');
                          Inc(FCharIndex);
                        end;
                        IDString := IDString + Chr(Octal);
                      end
            else
            begin
              IDString := IDString + FCurrentSourceLine[FCharIndex];
              Inc(FCharIndex);
            end
          end
        end
        else Inc(FCharIndex)
      else
      begin
        IDString := IDString + FCurrentSourceLine[FCharIndex];
        Inc(FCharIndex)
      end;

    if FCharIndex>Length(FCurrentSourceLine) then Error(SMissingStringTerminator, FCharIndex, True)
                                             else Inc(FCharIndex);
    if Length(IDString) = 2 then
    begin
      FLastValue := Ord(IDString[2]);
      ScanLiteral := LITERAL;
    end
    else
      if Length(IDString) > 1 then
      begin
        FLastValue := GetKey('''' + Copy(IDString, 2, Pred(Length(IDString))) + '''');
        ScanLiteral := LITID;
      end
      else ScanLiteral := ILLEGAL;
  end;

  //---------------------------------------------------------------------------

  function ScanNumber: Integer;

  // Scan an unsigned Integer

  var NumString: String;

  begin
    NumString := FCurrentSourceLine[FCharIndex];
    Inc(FCharIndex);
    while (FCharIndex <= Length(FCurrentSourceLine)) and
          (FCurrentSourceLine[FCharIndex]in Digits) do
    begin
      NumString := NumString + FCurrentSourceLine[FCharIndex];
      Inc(FCharIndex)
    end;
    try
      FLastValue := StrToInt(NumString);
      Result := NUMBER;
    except
      Result := ILLEGAL;
    end;
  end;

  //---------------------------------------------------------------------------

  function ScanKeyword: Integer;

  // scan %xy

    //-------------------------------------------------------------------------

    function Lookup(Key: String; var Token: Integer): Boolean;

    // table of parser keywords (unstropped)

    const KeywordCount = 11;
          Keywords: array[1..KeywordCount] of String = (
            '0', '2', 'binary', 'left', 'nonassoc', 'prec', 'right',
            'start', 'term', 'token', 'type');

          Tokens: array[1..KeywordCount] of Integer = (
            PTOKEN, PNONASSOC, PNONASSOC, PLEFT, PNONASSOC, PPREC, PRIGHT,
            PSTART, PTOKEN, PTOKEN, PTYPE);

    var M, N, K: Integer;

    begin
      // binary search
      M := 1;
      N := KeywordCount;
      Result := True;
      while M <= N do
      begin
        K := M + (N - M) div 2;
        if Key = Keywords[K] then
        begin
          Token := Tokens[K];
          Exit;
        end
        else
          if Key > Keywords[K] then M := K + 1
                               else N := K - 1;
      end;
      Result := False;
    end;

    //-------------------------------------------------------------------------

  var Keyword: String;
      Token: Integer;

  begin
    Inc(FCharIndex);
    if FCharIndex <= Length(FCurrentSourceLine) then
      case FCurrentSourceLine[FCharIndex] of
        '<': begin
               Result := PLEFT;
               Inc(FCharIndex);
             end;
        '>': begin
               Result := PRIGHT;
               Inc(FCharIndex);
             end;
        '=': begin
               Result := PPREC;
               Inc(FCharIndex);
             end;
        '%',
        '\': begin
               Result := PP;
               Inc(FCharIndex);
             end;
        '{': begin
               Result := LCURL;
               Inc(FCharIndex);
             end;
        '}': begin
               Result := RCURL;
               Inc(FCharIndex);
             end;
        'A'..'Z',
        'a'..'z',
        '0'..'9': begin
                    Keyword := FCurrentSourceLine[FCharIndex];
                    Inc(FCharIndex);
                    while (FCharIndex <= Length(FCurrentSourceLine)) and
                          (FCurrentSourceLine[FCharIndex]in Alpha) do
                    begin
                      Keyword := Keyword + FCurrentSourceLine[FCharIndex];
                      Inc(FCharIndex);
                    end;
                    if Lookup(Keyword, Token) then Result := Token
                                              else Result := ILLEGAL;
                  end;
      else Result := ILLEGAL
      end
    else
      Result := ILLEGAL;
  end;

  //---------------------------------------------------------------------------

  function ScanCharacter: Integer;

  // scan any single character

  begin
    Result := Ord(FCurrentSourceLine[FCharIndex]);
    Inc(FCharIndex);
  end;

  //--------------- end local functions ---------------------------------------
  
var LineNumber: Cardinal;
    CharNumber: Integer;

begin
  FCurrentTokenLen := 0;
  Scan;
  LineNumber := FSourceIndex;
  CharNumber := FCharIndex;
  if EOI then Result := 0
         else
    case FCurrentSourceLine[FCharIndex] of
      'A'..'Z',
      'a'..'z', '_': Result := ScanIdentifier;
      '''', '"': Result := ScanLiteral;
      '0'..'9': Result := ScanNumber;
      '%', '\': Result := ScanKeyword;
      '=': if (FCharIndex < Length(FCurrentSourceLine)) and (FCurrentSourceLine[Succ(FCharIndex)] = '{') then
           begin
             Inc(FCharIndex);
             Result := ScanCharacter;
           end
           else Result := ScanCharacter;
    else Result := ScanCharacter;
    end;
  if FSourceIndex = LineNumber then FCurrentTokenLen := FCharIndex - CharNumber;
end;

//----------------------------------------------------------------------------------------------------------------------

constructor TParserGenerator.Create;

begin
  FTarget := TStringList.Create;
  FReport := False;

  // allocate tables
  New(FSymbolTable);
  New(FSymbolKeyTable);
  New(FRuleTable);
  New(FRuleNumberTable);
  New(FRuleOffsetTable);
  New(FTypeTable);
  New(FSymbolTypeTable);
  New(FPrecendenceTable);
  New(FSymbolPrecendenceTable);
  New(FRulePrecendenceTable);
  New(FClosureTable);
  New(FFirstSetTable);
  New(FNullableTable);
  New(FStateTable);
  New(FItemTable);
  New(FTransitionTable);
  New(FReductionTable);
  New(FLookaheadTable);
  New(FPropagationTable);
end;

//----------------------------------------------------------------------------------------------------------------------

destructor TParserGenerator.Destroy;

begin
  Dispose(FSymbolTable);
  Dispose(FSymbolKeyTable);
  Dispose(FRuleTable);
  Dispose(FRuleNumberTable);
  Dispose(FRuleOffsetTable);
  Dispose(FTypeTable);
  Dispose(FSymbolTypeTable);
  Dispose(FPrecendenceTable);
  Dispose(FSymbolPrecendenceTable);
  Dispose(FRulePrecendenceTable);
  Dispose(FClosureTable);
  Dispose(FFirstSetTable);
  Dispose(FNullableTable);
  Dispose(FStateTable);
  Dispose(FItemTable);
  Dispose(FTransitionTable);
  Dispose(FReductionTable);
  Dispose(FLookaheadTable);
  Dispose(FPropagationTable);
  inherited;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.DoLog(const S: String);

var MI: TMessageInfo;

begin
  if assigned(FOnMessage) then
  begin
    with MI do
    begin
      Code := mtLog;
      Message := S;
      Source := '';
      Line := 0;
      Position := 0;
    end;
    FOnMessage(Self, MI);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.Error(Msg : String; Pos : Integer; InSource: Boolean);

var MI: TMessageInfo;

begin
  Inc(FErrors);
  if assigned(FOnMessage) then
  begin
    with MI do
    begin
      Code := mtError;
      Message := Msg;
      if InSource then
      begin
        Source := 'S';  // actual filenames must be inserted by the receiver
        Line := FSourceIndex;
      end
      else
      begin
        Source := 'T';
        Line := FTemplateIndex;
      end;
      Position := Pos;
    end;
    FOnMessage(Self, MI);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.FatalError(Msg : String);

var MI: TMessageInfo;

begin
  Inc(FErrors);
  if assigned(FOnMessage) then
  begin
    with MI do
    begin
      Code := mtFatalError;
      Message := Msg;
      Source := '';
      Line := 0;
      Position := 0;
    end;
    FOnMessage(Self, MI);
  end;
  Abort;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.Execute;

begin
  if FSource = nil then FatalError('source not assigned');
  if FTemplate = nil then FatalError('template not assigned');
  FGenerating := True;
  try
    ClearStructures;
    FSourceIndex := 0;
    FCharIndex := 1;
    FCurrentSourceLine := '';
    NextTemplateSection;

    // call main parsing routine
    if Parse = 0 then // done
                 else
      if FLastToken = 0 then Error(SUnexpectedEOI, FCharIndex, True)
                        else Error(SSyntaxError, FCharIndex, True);

    // print statistics
    if FErrors = 0 then 
    begin
      DoLog(IntToStr(FSourceIndex) + ' line(s),      ' +
            IntToStr(FRuleCount - 1)  + ' rule(s),      ' +
            IntToStr(FStateCount) + ' state(s),      ' +
            IntToStr(FItemCount)  + ' item(s),      ' +
            IntToStr(FTransitionCount) + ' transition(s),      ' +
            IntToStr(FReductionCount) + ' reduction(s)');
      if FShiftReduceConflicts > 0 then DoLog(IntToStr(FShiftReduceConflicts) + ' shift/reduce conflicts');
      if FReduceReduceConflicts > 0 then DoLog(IntToStr(FReduceReduceConflicts) + ' reduce/reduce conflicts');
      if FNeverReduced > 0 then DoLog(IntToStr(FNeverReduced) + ' rules never reduced');
    end;
  finally
    FGenerating := False;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TParserGenerator.NextSourceLine: Boolean;

begin
  if FSourceIndex < Cardinal(FSource.Count) then
  begin
    FCurrentSourceLine := FSource[FSourceIndex];
    if assigned(FOnSourceLine) then FOnSourceLine(Self, FSource[FSourceIndex], FSourceIndex + 1);
    Inc(FSourceIndex);
    FCharIndex := 1;
    Result := True;
  end
  else Result := False;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TParserGenerator.ClearStructures;

var I: Integer;

begin
  FErrors := 0;
  FStartNonTerminal := 0;
  FNonTerminalCount := 1;
  FLiteralsCount := 257;
  FRuleCount := 0;
  FTypeCount := 0;
  FPrecedenceCount := 0;
  FStateCount := 0;
  FItemCount := 0;
  FTransitionCount := 0;
  FReductionCount := 0;

  // initialize symbol table
  for I := 1 to MaxKeys do
    with FSymbolTable^[I] do
      begin
        Name := '';
        defined := False;
      end;

  // enter predefined error symbol into symbol table
  DefineKey(GetKey('error'), 256);

  // initialize type and precedence tables
  for I := -MaxNonterminals to MaxLiterals - 1 do FSymbolTypeTable^[I] := 0;
  for I := 0 to MaxLiterals - 1 do FSymbolPrecendenceTable^[I] := 0;
  for I := 1 to MaxRules do FRulePrecendenceTable^[I] := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

end.

