unit Project;

// 13-NOV-99 ml: docking
// 07-NOV-99 ml: structural rework, docking
// 18-JUL-99 ml: bug fixes in LoadDesktop and SaveDesktop
// 06-JUL-99 ml: Added highlighter menu entries
// 04-JUL-99 ml: Ctrl-Tab switches now from one tab sheet to the next
// 04-JUL-99 ml: implementation so far
//----------------------------------------------------------------------------------------------------------------------
// This unit contains the code for a project editor window. This form is not automatically
// created as it can appear as multiple instances. It contains mainly a page control which holds
// a syntax editor control on each page. There are functions to load and store a project and
// its associated desktop file. The project accepts dropped files and opens them as usual.
// (Note: DCG project files (*.dcg) are handled as simple text files. If you want them opend
//        as projects then drop them on the main form).

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Menus,
  ComCtrls, StdCtrls, ExtCtrls, Dialogs, Buttons,
  SyntaxEditor, Highlighter, DelphiHighlighter, DCGHighlighter,
  DCGData, LexerGenerator, ParserGenerator, DCGTools, MessageWindow;

type
  TProjectForm = class(TBaseProjectForm)
    PageControl: TPageControl;
    StatusBar: TStatusBar;
    EditorPopup: TPopupMenu;
    CutItem: TMenuItem;
    CopyItem: TMenuItem;
    PasteItem: TMenuItem;
    SelectAllItem: TMenuItem;
    PropertiesItem: TMenuItem;
    Undo1: TMenuItem;
    Redo1: TMenuItem;
    N2: TMenuItem;
    N3: TMenuItem;
    HSplitter: TSplitter;
    N1: TMenuItem;
    ClosePageItem: TMenuItem;
    DeleteItem1: TMenuItem;
    N4: TMenuItem;
    Editorproperties1: TMenuItem;
    N5: TMenuItem;
    InsertItem: TMenuItem;
    DCGCommentItem: TMenuItem;
    MacroIDItem: TMenuItem;
    N6: TMenuItem;
    HighlighterItem: TMenuItem;
    BottomDockPanel: TPanel;
    VSplitter: TSplitter;
    LeftDockPanel: TPanel;
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure ProjectChange(Sender: TObject);
    procedure LexerSourcePathEditKeyPress(Sender: TObject; var Key: Char);
    procedure EditCaretChange(Sender: TCustomSyntaxEdit; X, Y: Cardinal);
    procedure PageControlChange(Sender: TObject);
    procedure EditChange(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormDestroy(Sender: TObject);
    procedure ClosePageItemClick(Sender: TObject);
    procedure Editorproperties1Click(Sender: TObject);
    procedure PageControlDrawTab(Control: TCustomTabControl; TabIndex: Integer; const Rect: TRect; Active: Boolean);
    procedure InsertItemClick(Sender: TObject);
    procedure EditorPopupPopup(Sender: TObject);
    procedure BottomDockPanelDockOver(Sender: TObject; Source: TDragDockObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean);
    procedure LeftDockPanelDockOver(Sender: TObject; Source: TDragDockObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean);
    procedure SplitterMoved(Sender: TObject);
    procedure BottomDockPanelEndDock(Sender, Target: TObject; X, Y: Integer);
    procedure DockPanelDockDrop(Sender: TObject; Source: TDragDockObject; X, Y: Integer);
    procedure DockPanelUnDock(Sender: TObject; Client: TControl; NewTarget: TWinControl; var Allow: Boolean);
    procedure LeftDockPanelGetSiteInfo(Sender: TObject; DockClient: TControl; var InfluenceRect: TRect;
      MousePos: TPoint; var CanDock: Boolean);
  private
    FNewProject,
    FInitialShow: Boolean;
    FCurrentEdit: TSyntaxEdit;
    FProjectChanged: Boolean;
    FWindowList: Pointer; // DisableTaskWindows
    FFirstActivated: Boolean;
    FDefaultLeft,
    FDefaultTop,
    FDefaultWidth,
    FDefaultHeight: Integer;
    FErrors,
    FHints,
    FWarnings: Cardinal;
    FMessageForm: TMessageForm;
    function EditFromFilename(const FileName: String): TSyntaxEdit;
    function EditNeedsSave: Boolean;
    function SaveEdit(Edit: TSyntaxEdit): Boolean;
    function SaveEditAs(Edit: TSyntaxEdit): Boolean;
    function AskForSave(const Name: String): Word;
    function GetEdit(Index: Integer): TSyntaxEdit;
    procedure FocusEdit(Edit: TSyntaxEdit);
    procedure RemoveEdit(Edit: TSyntaxEdit);
    function GetShowControlChars: Boolean;
    procedure SetShowControlChars(const Value: Boolean);
    procedure ReindexSheets;
    function HighlighterFromFileExt(const Extension: String): TSyntaxHighlighter;
    procedure FindCoreEdits;
    procedure HighlighterClick(Sender: TObject);
    procedure WMGetMinMaxInfo(var Message: TWMGetMinMaxInfo); message WM_GETMINMAXINFO;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
    procedure WMDropFiles(var Msg: TWMDropFiles); message WM_DROPFILES;
    procedure ShowDockPanel(APanel: TPanel; MakeVisible: Boolean; Client: TControl);
  protected
    function GetNewProject: Boolean; override;
    procedure UpdateProjectSettings;
  public
    ProjectSettings: TProjectSettings;
    constructor CreateFromFile(AOwner: TComponent; const FileName: String; DoShow: Boolean);

    procedure ApplyGlobalSettings;
    function AddEdit(FileName: String; Focus: Boolean = False; ReadOnly: Boolean = False): TSyntaxEdit;
    procedure AddMessage(Info: TMessageInfo);
    function CanShow: Boolean; override;
    procedure CheckFileAges; override;
    procedure ClearMessages;
    procedure Execute; override;
    procedure GeneratingDone;
    procedure SetStatus(const Text: String); override;
    procedure Save; override;
    function SaveAll: Boolean; override;
    procedure SaveAs; override;
    function SaveProject: Boolean; override;
    function SaveProjectAs: Boolean; override;
    procedure LoadDesktop;
    procedure SaveDesktop;
    procedure ShowMarkedEdit(MI: PMessageInfo);
    procedure ValidateControls;

    property CurrentEdit: TSyntaxEdit read FCurrentEdit;
    property LastErrorCount: Cardinal read FErrors;
    property LastWarningCount: Cardinal read FWarnings;
    property LastHintCount: Cardinal read FHints;

    property ShowControlChars: Boolean read GetShowControlChars write SetShowControlChars;
  end;

var
  ProjectForm: TProjectForm;

//----------------------------------------------------------------------------------------------------------------------

implementation

uses
  Clipbrd, DCGStrings, CompileDlg, ShellAPI, EnvironmentOptionsDlg,
  SystemImageList, CommCtrl, ImgList, Main;

{$R *.DFM}

const
  BooleanToString: array[Boolean] of Char = ('F', 'T');

//----------------- TProjectForm ---------------------------------------------------------------------------------------

constructor TProjectForm.CreateFromFile(AOwner: TComponent; const FileName: String; DoShow: Boolean);

var
  Settings : TStringList;
  I: Integer;
  Item: TMenuItem;

begin
  FInitialShow := DoShow;
  FFirstActivated := True;
  Create(AOwner);
  FMessageForm := TMessageForm.Create(Self);
  DragAcceptFiles(Handle, True);
  PageControl.TabHeight := ApplicationData.SystemImages.Height + 4;
  if GlobalSettings.ShowFileIcons then PageControl.Images := ApplicationData.SystemImages
                                  else PageControl.Images := nil;
  BeginLoading;

  if (FileName = '') or
     not FileExists(FileName) then
  begin
    if FileName = '' then WorkingFile := 'new project.dcg'
                     else WorkingFIle := Filename;
    AddEdit('Lexer.l');
    FNewProject := True;
  end
  else
  begin
    WorkingFile := FileName;
    FNewProject := False;
  end;

  LoadDesktop;

  // add all highlighter references to the popup menu
  HighlighterItem.Delete(0); // remove dummy entry
  with GlobalSettings.HighlighterList do
    for I := 0 to Count - 1 do
      with TSyntaxHighlighter(Items[I]) do
      begin
        Item := NewItem(LanguageName, 0, False, True, HighlighterClick, 0, '');
        Item.RadioItem := True;
        Item.GroupIndex := 1;
        HighlighterItem.Add(Item);
      end;

  // load project if available
  if FileExists(WorkingFile) then
  begin
    Settings := TStringList.Create;
    try
      FillChar(ProjectSettings, SizeOf(ProjectSettings), 0);
      Settings.LoadFromFile(WorkingFile);
      with ProjectSettings do
      begin
        // read settings from file
        LexerSource := Settings.Values['LexicalDescription'];
        LexerTemplate := Settings.Values['LexerCodeTemplate'];
        LexerTarget := Settings.Values['LexerTargetFile'];
        GenerateLexer := StringToBoolean(Settings.Values['GenerateLexer']);

        ParserSource := Settings.Values['ParserDescription'];
        ParserTemplate := Settings.Values['ParserCodeTemplate'];
        ParserTarget := Settings.Values['ParserTargetFile'];
        GenerateParser := StringToBoolean(Settings.Values['GenerateParser']);

        CompileDelphiProject := StringToBoolean(Settings.Values['CompileDelphiProject']);
        DelphiProject := Settings.Values['DelphiProject'];
      end;
      FNewProject := False;
      ValidateControls;
    finally
      Settings.Free;
      FProjectChanged := False;
      NeedSave := False;
    end;
  end;
  EndLoading;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.SaveProject: Boolean;

var
 Settings: TStringList;

begin
  if FNewproject then Result := SaveProjectAs
                 else
    if FProjectChanged then
    begin
      Settings:=TStringList.Create;
      try
        with ProjectSettings do
        begin
          Settings.Values['LexicalDescription'] := LexerSource;
          Settings.Values['LexerCodeTemplate'] := LexerTemplate;
          Settings.Values['LexerTargetFile'] := LexerTarget;
          Settings.Values['GenerateLexer'] := BooleanToString[GenerateLexer];
          Settings.Values['ParserDescription'] := ParserSource;
          Settings.Values['ParserCodeTemplate'] := ParserTemplate;
          Settings.Values['ParserTargetFile'] := ParserTarget;
          Settings.Values['GenerateParser'] := BooleanToString[GenerateParser];
          Settings.Values['CompileDelphiProject'] := BooleanToString[CompileDelphiProject];
          Settings.Values['DelphiProject'] := DelphiProject;
        end;
        Settings.SaveToFile(WorkingFile);
        FProjectChanged := False;
        NeedSave := EditNeedsSave;
        Result := True;
      finally
        Settings.Free;
      end;
    end
    else Result := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.LoadDesktop;

var
  Settings: TStringList;
  DeskFile: String;
  I, J: Integer;
  NewEdit: TSyntaxEdit;
  Parameters: array of String;

  //---------------------------------------------------------------------------

  procedure SplitParameters(Value: String);

  // splits a list of parameter strings in Value into the Parameters array

  var
    Index: Integer;
    Delimiter: Integer;

  begin
    Parameters := nil;
    Index := 0;
    Value := Trim(Value);
    while Length(Value) > 0 do
    begin
      SetLength(Parameters, Index + 1);
      Delimiter := Pos(',', Value);
      if Delimiter = 0 then
      begin
        Parameters[Index] := Value;
        Value := '';
      end
      else
      begin
        Parameters[Index] := TrimRight(Copy(Value, 1, Delimiter - 1));
        Delete(Value, 1, Delimiter);
      end;
      Inc(Index);
    end;
  end;

  //---------------------------------------------------------------------------

var
  X, Y: Integer;
  Value: String;
  
begin
  if not FNewproject then
  begin
    DeskFile := ChangeFileExt(WorkingFile, '.dsk');
    if FileExists(DeskFile) then
    begin
      Settings := TStringList.Create;
      Settings.LoadFromFile(DeskFile);
      try
        Value := Trim(Settings.Values['Left']);
        FDefaultLeft := StrToIntDef(Value, Left);

        Value := Trim(Settings.Values['Top']);
        FDefaultTop := StrToIntDef(Value, Top);

        Value := Trim(Settings.Values['Width']);
        FDefaultWidth := StrToIntDef(Value, Width);

        Value := Trim(Settings.Values['Height']);
        FDefaultHeight := StrToIntDef(Value, Height);

        SetBounds(FDefaultLeft, FDefaultTop, FDefaultWidth, FDefaultHeight);
        if StringToBoolean(Trim(Settings.Values['Maximized'])) then WindowState := wsMaximized
                                                               else WindowState := wsNormal;

        Value := Trim(Settings.Values['DockHeight']);
        FMessageForm.TBDockHeight := StrToIntDef(Value, 150);

        Value := Trim(Settings.Values['DockWidth']);
        FMessageForm.LRDockWidth := StrToIntDef(Value, 150);

        // reopen editor files
        // read at most 50 files into the page control
        I := 0;
        while PageControl.PageCount < 50 do
        begin
          Value := Settings.Values[Format('File%d', [I])];
          if Value = '' then Break;
          Inc(I);

          SplitParameters(Value);

          // file name
          if Length(Parameters) > 0 then
          begin
            NewEdit := AddEdit(Parameters[0]);
            PageControl.ActivePage := PageControl.Pages[NewEdit.Tag];

            NewEdit.BeginUpdate;
            try
              X := 0;
              for J := 1 to High(Parameters) do
                if Parameters[J] <> '' then
                begin
                  case J of
                    1: // top line
                      NewEdit.TopLine := StrToIntDef(Parameters[J], 0);
                    2, 3: // caret position
                      if J = 2 then NewEdit.CaretX := StrToIntDef(Parameters[J], 0)
                               else NewEdit.CaretY := StrToIntDef(Parameters[J], 0);
                    4..23: // bookmarks
                      begin
                        if Odd (J) then
                        begin
                          Y := StrToIntDef(Parameters[J], 0);
                          NewEdit.SetBookMark(J div 2 - 2, X, Y);
                        end
                        else
                        begin
                          // keep X value of this bookmark and apply it in next run
                          X := StrToIntDef(Parameters[J], 0);
                        end;
                      end;
                    24: // gutter width
                      NewEdit.GutterWidth := StrToIntDef(Parameters[J], 50);
                    25: // right margin
                      NewEdit.RightMargin := StrToIntDef(Parameters[J], 120);
                  end;
                end;
            finally
              NewEdit.EndUpdate;
            end;
          end;
        end;

        Value := Trim(Settings.Values['ActivePage']);
        if Value <> '' then
        begin
          I := StrToIntDef(Value, 0);
          if I < PageControl.PageCount then PageControl.ActivePage := PageControl.Pages[I];
        end;

      finally
        Settings.Free;
      end;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.SaveDesktop;

var
  Settings: TStringList;
  Maximized: Boolean;
  I, J, X, Y: Integer;
  S: String;

begin
  if not FNewproject then
  begin
    Settings := TStringList.Create;
    try
      Maximized := IsZoomed(Handle);
      Settings.Add('[Globals]');
      Settings.Values['Maximized'] := BooleanToString[Maximized];
      Settings.Values['Left'] := IntToStr(FDefaultLeft);
      Settings.Values['Top'] := IntToStr(FDefaultTop);
      Settings.Values['Width'] := IntToStr(FDefaultWidth);
      Settings.Values['Height'] := IntToStr(FDefaultHeight);
      Settings.Values['DockHeight'] := IntToStr(FMessageForm.TBDockHeight);
      Settings.Values['DockWidth'] := IntToStr(FMessageForm.LRDockWidth);
      if Assigned(PageControl.ActivePage) then
        Settings.Values['ActivePage'] := IntToStr(PageControl.ActivePage.PageIndex);

      // editor files
      Settings.Add('');
      Settings.Add('[Editor Files]');
      // write file name and edit settings to desk file
      for I := 0 to PageControl.PageCount - 1 do
        with TSyntaxEdit(PageControl.Pages[I].Controls[0]) do
        begin
          // file name
          S := ExtractRelativePath(ExtractFilePath(WorkingFile), Info) + ',';
          // top line
          S := S + IntToStr(TopLine) + ',';
          // caret
          S := S + Format('%d,%d,', [CaretX, CaretY]);
          // bookmarks
          for J := 0 to 9 do
          begin
            if GetBookMark(J, X, Y) then S := S + Format('%d,%d,', [X, Y])
                                    else S := S + ',,';
          end;
          // gutter width
          S := S + IntToStr(GutterWidth) + ',';
          // right margin
          S := S + IntToStr(RightMargin);
          Settings.Values[Format('File%d', [I])] := S;
        end;

      Settings.SaveToFile(ExtractFilePath(WorkingFile) + PureName(WorkingFile) + '.dsk');
    finally
      Settings.Free;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.Save;

begin
  if FCurrentEdit.Modified then
  begin
    SaveEdit(FCurrentEdit);
    StatusBar.Panels[1].Text := '';
  end;
  NeedSave := EditNeedsSave;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.SaveProjectAs: Boolean;

begin
  with ApplicationData.SaveDialog do
  begin
    FileName := WorkingFile;
    DefaultExt := '*.dcg';
    Filter := SProjectFilter;
    Options := [ofHideReadOnly, ofEnableSizing];
    Title := Format(SSaveAsTitel, [PureName(WorkingFile)]);
    if Execute then
    begin
      WorkingFile := FileName;
      Result := True;
      FNewProject := False;
      SaveProject;
    end
    else Result:=False;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.SetStatus(const Text: String);

begin
  StatusBar.SimpleText := Text;
  StatusBar.SimplePanel := Text <> '';
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.AskForSave(const Name: String): Word;

begin
  Result := MessageDlg('Save changes to ''' + Name + '''?', mtConfirmation, [mbYes, mbNo, mbCancel], 0);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);

// By passing CanClose = False I mark closing the window itself, which is not the same as
// closing the entire project (in which case changes have to be written out).

var
  I: Integer;

begin
  if not CanClose then
  begin
    // project needs save?
    if FProjectChanged then
    begin
      case AskForSave(PureName(WorkingFile)) of
        mrYes: CanClose := SaveProject;
        mrNo: CanClose := True;
      else
        Show;
        CanClose := False;
      end;
    end
    else CanClose := True;

    // one of the edits needs save?
    if CanClose and EditNeedsSave then
    begin
      for I := 0 to PageControl.PageCount - 1 do
      begin
        if not CanClose then Break;
        if TSyntaxEdit(PageControl.Pages[I].Controls[0]).Modified then
        begin
          case AskForSave(ExtractFileName(TSyntaxEdit(PageControl.Pages[I].Controls[0]).Info)) of
            mrYes: CanClose := SaveEdit(TSyntaxEdit(PageControl.Pages[I].Controls[0]));
            mrCancel: CanClose := False;
          end;
        end;
      end;
    end;

    if GlobalSettings.AutoSaveDesktop then SaveDesktop;
  end
  else CanClose := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.ProjectChange(Sender: TObject);

begin
  FProjectChanged := True;
  NeedSave := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.LexerSourcePathEditKeyPress(Sender: TObject; var Key: Char);

begin
  if Key = #13 then
  begin
    Key := #0;
    ValidateControls;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.ValidateControls;

begin
  // nothing to do yet
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.EditCaretChange(Sender: TCustomSyntaxEdit; X, Y: Cardinal);

begin
  with Sender as TSyntaxEdit do
  begin
    MarkLine := -1;
    StatusBar.SimplePanel := False;
    StatusBar.Panels[0].Text := Format('%d: %d', [Y + 1, X + 1]);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.PageControlChange(Sender: TObject);

begin
  // there's only one control on each tab sheet
  FCurrentEdit := PageControl.ActivePage.Controls[0] as TSyntaxEdit;
  Caption := ExtractRelativePath(ExtractFilePath(WorkingFile), FCurrentEdit.Info);

  with FCurrentEdit do
  begin
    if Modified then StatusBar.Panels[1].Text := 'Modified'
                else StatusBar.Panels[1].Text := '';
    if eoReadOnly  in Options then StatusBar.Panels[2].Text := 'Read only'
                              else
      if eoInserting in Options then StatusBar.Panels[2].Text := 'Insert'
                                else StatusBar.Panels[2].Text := 'Overwrite';
    StatusBar.SimplePanel := False;
    StatusBar.Panels[0].Text := Format('%d: %d', [CaretY + 1, CaretX + 1]);
    FocusEdit(FCurrentEdit);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.EditChange(Sender: TObject);

begin
  NeedSave := True;
  StatusBar.Panels[1].Text := 'Modified';
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.GetNewProject: Boolean;

begin
  Result := FNewProject;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.FormActivate(Sender: TObject);

  //--------------- local function --------------------------------------------

  procedure EnsureEdit;

  // make sure at least one edit is available (create a new one if no other
  // could be found)

  var
    I: Integer;

  begin
    // try default files first
    for I := 0 to 5 do
    begin
      // load the edit
      GetEdit(I);
      // successfully loaded?
      if PageControl.PageCount > 0 then Break;
    end;

    // if still no edit available then create one
    if PageControl.PageCount = 0 then AddEdit('Lexer.l');
  end;

  //--------------- end local function ----------------------------------------

begin
  MainForm.ActiveProject := Self;
  if FFirstActivated then
  begin
    FFirstActivated := False;
    if PageControl.PageCount = 0 then EnsureEdit;
    PageControlChange(PageControl);
  end
  else
    if PageControl.PageCount = 0 then
    begin
      EnsureEdit;
      PageControlChange(PageControl);
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.EditNeedsSave: Boolean;

var
  I: Integer;

begin
  Result :=False;
  // there's only one control on each tab sheet
  for I := 0 to PageControl.PageCount - 1 do
    if TSyntaxEdit(PageControl.Pages[I].Controls[0]).Modified then
    begin
      Result := True;
      Break;
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.SaveEditAs(Edit: TSyntaxEdit): Boolean;

var
  Extension,
  FileFilter: String;
  Index: Integer;

begin
  Extension := Copy(ExtractFileExt(Edit.Info), 2, 3);
  FileFilter := SDelphiProjectFilter + '|' +
                SUnitFilter + '|' +
                SLexerFilter + '|' +
                SCodeTemplateFilter + '|' +
                SParserFilter + '|' +
                SAnyFileFilter;
  if CompareText(Extension, 'dpr') = 0 then Index := 1
                                       else
    if CompareText(Extension, 'pas') = 0 then Index := 2
                                         else
      if CompareText(Extension, 'l') = 0 then Index := 3
                                           else
        if CompareText(Extension, 'cod') = 0 then Index := 4
                                             else
          if CompareText(Extension, 'y') = 0 then Index := 5
                                               else Index := 6;

  with ApplicationData.SaveDialog do
  begin
    FileName := Edit.Info;
    DefaultExt := Extension;
    Filter := FileFilter;
    FilterIndex := Index;
    Options := [ofHideReadOnly, ofEnableSizing, ofOverwritePrompt, ofPathMustExist,
                ofEnableIncludeNotify];
    Title := SSaveTitel;
    if Execute then
    begin
      Edit.Info := FileName;
      SaveEdit(Edit);
      // update references
      PageControl.Pages[Edit.Tag].Caption := PureName(Edit.Info);
      Caption := ExtractRelativePath(ExtractFilePath(WorkingFile), Edit.Info);
      UpdateProjectSettings;
      Result := True;
    end
    else Result := False;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.SaveEdit(Edit: TSyntaxEdit): Boolean;

var
  Backup: String;
  Period: Integer;

begin
  if ExtractFilePath(Edit.Info) = '' then Result := SaveEditAs(Edit)
                                     else
  begin
    if GlobalSettings.CreateBackups then
    begin
      Backup := Edit.Info;
      Period := Pos('.', Backup);
      if Period > 0 then Insert(' (Backup)', Backup, Period)
                    else Backup := Backup + ' (Backup)';
      if FileExists(Backup) then DeleteFile(Backup);
      RenameFile(Edit.Info, Backup);
    end;
    Edit.Lines.SaveToFile(Edit.Info);
    PageControl.Pages[Edit.Tag].Tag := FileAge(Edit.Info);
    Edit.Modified := False;
    Result := True;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.SaveAs;

begin
  SaveEditAs(FCurrentEdit);
  NeedSave := EditNeedsSave or FProjectChanged;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.FindCoreEdits;

// goes through the project settings and looks for files which are loaded but don't have
// already set their reference in the settings

var
  BasePath: String;
  
begin
  BasePath := ExtractFilePath(WorkingFile);
  with ProjectSettings do
  begin
    LexerSourceEdit := EditFromFileName(BasePath + LexerSource);
    LexerTemplateEdit := EditFromFileName(BasePath + LexerTemplate);
    LexerTargetEdit := EditFromFileName(BasePath + LexerTarget);
    ParserSourceEdit := EditFromFileName(BasePath + ParserSource);
    ParserTemplateEdit := EditFromFileName(BasePath + ParserTemplate);
    ParserTargetEdit := EditFromFileName(BasePath + ParserTarget);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.Execute;

var
  CanProceed: Boolean;

begin
  if GlobalSettings.AutoSaveEditor then CanProceed := SaveAll
                                   else CanProceed := True;
  if CanProceed then
  begin
    FindCoreEdits;
    with ProjectSettings do
    begin
      // make sure the target edits exist (all others are silently loaded if necessary)
      if GenerateLexer then GetEdit(2);
      if GenerateParser then GetEdit(5);
    end;
    with FMessageForm do
    begin
      if FirstShow then ManualDock(BottomDockPanel);
      Show;
      BringToFront;
    end;
    FWindowList := DisableTaskWindows(Application.Handle);
    CompileDialog.Show;
    CompileDialog.DoGenerate(Self);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.GeneratingDone;

const
  ErrorTypes = [mtError, mtSyntaxError, mtSemanticError, mtFatalError];

var
  I: Integer;
  MI: PMessageInfo;

begin
  if Assigned(FWindowList) then
  begin
    EnableTaskWindows(FWindowList);
    FWindowList := nil;
  end;

  with FMessageForm do
  begin
    if FErrors > 0 then
    begin
      // there were errors -> mark line in editor
      for I := 0 to MessageList.Items.Count - 1 do
        if Assigned(MessageList.Items.Objects[I]) and
           (PMessageInfo(MessageList.Items.Objects[I]).Code in ErrorTypes) and
           (PMessageInfo(MessageList.Items.Objects[I]).Source <> '') then
        begin
          MI := PMessageInfo(MessageList.Items.Objects[I]);
          ShowMarkedEdit(MI);
          Break;
        end;
    end
    else FocusEdit(FCurrentEdit);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.WMGetMinMaxInfo(var Message: TWMGetMinMaxInfo);

var
  R: TRect;

begin
  inherited;
  if not (csReading in ComponentState) then
  begin
    SystemParametersInfo(SPI_GETWORKAREA, 0, @R, 0);
    Message.MinMaxInfo.ptMaxPosition.y := R.Top + MainForm.Height - 4;
    Message.MinMaxInfo.ptMaxSize.y := R.Bottom - R.Top - MainForm.Height + 4;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);

begin
  // handle some default keys
  case Key of
    VK_F5: // maximize/restore project form
      if IsZoomed(Handle) then ShowWindow(Handle, SW_SHOWNORMAL)
                          else ShowWindow(Handle, SW_MAXIMIZE);
    VK_TAB:
      if ssCtrl in Shift then PageControl.SelectNextPage(not (ssShift in Shift));
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.WMWindowPosChanged(var Message: TWMWindowPosChanged);

begin
  inherited;
  with Message.WindowPos^ do
    if not Loading and
       not IsZoomed(Handle) and
       (((flags and SWP_NOSIZE) = 0) or
       ((flags and SWP_NOMOVE) = 0)) then
    begin
      if (flags and SWP_NOMOVE) = 0 then
      begin
        FDefaultLeft := x;
        FDefaultTop := y;
      end;
      if (flags and SWP_NOSIZE) = 0 then
      begin
        FDefaultWidth := cx;
        FDefaultHeight := cy;
      end;
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.AddMessage(Info: TMessageInfo);

// Adds the given message info to the message list box.

var
  NewEntry: PMessageInfo;
  Prefix,
  Line: String;

begin
  New(NewEntry);
  NewEntry^ := Info;
  case Info.Code of
    mtError:
      begin
        Inc(FErrors);
        Prefix := SError;
      end;
    mtSyntaxError:
      begin
        Inc(FErrors);
        Prefix := SSyntaxError;
      end;
    mtSemanticError:
      begin
        Inc(FErrors);
        Prefix := SSemanticError;
      end;
    mtFatalError:
      begin
        Inc(FErrors);
        Prefix := SFatalError;
      end;
    mtWarning:
      begin
        Inc(FWarnings);
        Prefix := SWarning;
      end;
    mtHint:
      begin
        Inc(FHints);
        Prefix := SHint;
      end;
    mtLog:
      begin
        Prefix := SLog;
      end;
  end;
  if Info.Source = '' then Line := ''
                      else Line := Format(' %s(%d): ', [ExtractFileName(Info.Source), Info.Line]);
  with FMessageForm do
  begin
    MessageList.Items.AddObject(Prefix + Line + Info.Message, Pointer(NewEntry));
    MessageList.Update;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.ClearMessages;

var
  I: Integer;

begin
  with FMessageForm do
  begin
    if Assigned(MessageList) then
    begin
      for I:= 0 to MessageList.Items.Count -1 do
        if assigned(MessageList.Items.Objects[I]) then Dispose(PMessageInfo(MessageList.Items.Objects[I]));
      MessageList.Clear;
    end;
    FErrors := 0;
    FHints := 0;
    FWarnings := 0;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.FormDestroy(Sender: TObject);

begin
  ClearMessages;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.EditFromFilename(const FileName: String): TSyntaxEdit;

var
  I: Integer;

begin
  Result := nil;
  for I := 0 to PageControl.PageCount - 1 do
    if CompareText(TSyntaxEdit(PageControl.Pages[I].Controls[0]).Info, FileName) = 0 then
    begin
      Result := PageControl.Pages[I].Controls[0] as TSyntaxEdit;
      Break;
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.HighlighterFromFileExt(const Extension: String): TSyntaxHighlighter;

begin
  with ApplicationData do
    if CompareText(Extension, '.pas') = 0 then Result := DelphiHighlighter1
                                            else
      if (CompareText(Extension, '.htm') = 0) or
         (CompareText(Extension, '.html') = 0) then Result := HTMLHighlighter1
                                               else
        if (CompareText(Extension, '.l') = 0) or
           (CompareText(Extension, '.y') = 0) or
           (CompareText(Extension, '.cod') = 0) then Result := DCGHighlighter1
                                                else
          if (CompareText(Extension, '.c') = 0) or
             (CompareText(Extension, '.cpp') = 0) or
             (CompareText(Extension, '.h') = 0) then Result := CPPHighlighter1
                                                else Result := nil;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.AddEdit(FileName: String; Focus: Boolean = False; ReadOnly: Boolean = False): TSyntaxEdit;

// adds a new page control sheet with an edit window and loads the given file
// into the edit (if file name is not empty)

var
  NewSheet: TTabSheet;
  NewIndex: Integer;
  Path: String;

begin
  // absolute path?
  Path := ExtractFileDir(FileName);
  // add path of project if there's no path yet
  if Path = '' then FileName := ExtractFilePath(WorkingFile) + FileName;
               
  Result := EditFromFilename(FileName);
  if Result = nil then
  begin
    // create new sheet where the edit will be placed on
    if PageControl.PageCount > 0 then NewIndex := PageControl.ActivePage.PageIndex + 1
                                 else NewIndex := 0;
    NewSheet := TTabSheet.Create(PageControl);
    NewSheet.PageControl := PageControl;
    NewSheet.Caption := PureName(FileName);
    NewSheet.PageIndex := NewIndex;
    NewSheet.ImageIndex := ApplicationData.SystemImages.GetImageIndex(Filename, False, False, []);

    // create new editor window
    Result := TSyntaxEdit.Create(NewSheet);
    with Result do
    begin
      Parent := NewSheet;
      Keystrokes := GlobalSettings.EditorCommands;
      Cursor := crIBeam;
      Align := alClient;
      BevelEdges := [beLeft, beTop];
      BevelInner := bvNone;
      BevelKind := bkFlat;
      BorderStyle := bsNone;
      CharWidth := 8;
      Ctl3D := True;
      Font.Name := GlobalSettings.EditorFontName;
      Font.Size := GlobalSettings.EditorFontSize;
      GutterWidth := 50;
      RightMargin := 120;
      Options := GlobalSettings.EditOptions + [eoAllowMarginResize, eoWantTabs, eoReadOnlySelection];
      if ReadOnly then Options := Options + [eoReadOnly];
      TabSize := GlobalSettings.TabSize;
      IndentSize := GlobalSettings.BlockIndent;
      MaxUndo := GlobalSettings.UndoLimit;
      MarkColor.Foreground := GlobalSettings.EditorMarkForeground;
      MarkColor.Background := GlobalSettings.EditorMarkBackground;
      MaxRightChar := 256;
      PopupMenu := EditorPopup;
      ScrollbarMode := GlobalSettings.ScrollBarMode;
      Scrollbars := ssBoth;
      if FileExists(FileName) then
      begin
        Lines.LoadFromFile(FileName);
        NewSheet.Tag := FileAge(FileName);
      end
      else NewSheet.Tag := 0;
      Modified := False;
      Info := FileName;
      Highlighter := HighlighterFromFileExt(ExtractFileExt(FileName));
      OnCaretChange := EditCaretChange;
      OnChange := EditChange;
    end;
  end;

  ReindexSheets;
  if Focus then
  begin
    PageControl.ActivePage := PageControl.Pages[Result.Tag];
    PageControlChange(PageControl);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.GetEdit(Index: Integer): TSyntaxEdit;

// makes sure the specific core edit exists and returns it to the caller
// Note: The passed index is NOT the tab sheet index of the page the edits is located on,
//       but a simple running number defined to identify the core edits.

begin
  Result := nil;
  case Index of
    0:
      begin
        if ProjectSettings.LexerSourceEdit = nil then
        begin
          if ProjectSettings.LexerSource = '' then ProjectSettings.LexerSourceEdit := AddEdit('Lexer.l')
                                              else ProjectSettings.LexerSourceEdit := AddEdit(ProjectSettings.LexerSource);
        end;
        Result := ProjectSettings.LexerSourceEdit;
      end;
    1:
      begin
        if ProjectSettings.LexerTemplateEdit = nil then
        begin
          if ProjectSettings.LexerTemplate = '' then ProjectSettings.LexerTemplateEdit := AddEdit('Lexer.cod')
                                                else ProjectSettings.LexerTemplateEdit := AddEdit(ProjectSettings.LexerTemplate);
        end;
        Result := ProjectSettings.LexerTemplateEdit;
      end;
    2:
      begin
        if ProjectSettings.LexerTargetEdit = nil then
        begin
          if ProjectSettings.LexerTarget = '' then ProjectSettings.LexerTargetEdit := AddEdit('Lexer.pas')
                                              else ProjectSettings.LexerTargetEdit := AddEdit(ProjectSettings.LexerTarget);
        end;
        Result := ProjectSettings.LexerTargetEdit;
      end;
    3:
      begin
        if ProjectSettings.ParserSourceEdit = nil then
        begin
          if ProjectSettings.ParserSource = '' then ProjectSettings.ParserSourceEdit := AddEdit('Parser.y')
                                               else ProjectSettings.ParserSourceEdit := AddEdit(ProjectSettings.ParserSource);
        end;
        Result := ProjectSettings.ParserSourceEdit;
      end;
    4:
      begin
        if ProjectSettings.ParserTemplateEdit = nil then
        begin
          if ProjectSettings.ParserTemplate = '' then ProjectSettings.ParserTemplateEdit := AddEdit('Parser.cod')
                                                 else ProjectSettings.ParserTemplateEdit := AddEdit(ProjectSettings.ParserTemplate);
        end;
        Result := ProjectSettings.ParserTemplateEdit;
      end;
    5:
      begin
        if ProjectSettings.ParserTargetEdit = nil then
        begin
          if ProjectSettings.ParserTarget = '' then ProjectSettings.ParserTargetEdit := AddEdit('Parser.pas')
                                               else ProjectSettings.ParserTargetEdit := AddEdit(ProjectSettings.ParserTarget);
        end;
        Result := ProjectSettings.ParserTargetEdit;
      end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.ShowMarkedEdit(MI: PMessageInfo);

// Loads the file given in MI and positions the cursor at the provided position.
// The mark line is also place at this position.

var
  Edit: TSyntaxEdit;

begin
  if Assigned(MI) and (MI.Source <> '') then
  begin
    Edit := AddEdit(MI.Source, True);
    // the caret is zero-based while the error info is one-based
    Edit.CaretXY := Point(MI.Position, MI.Line - 1);
    Edit.MarkLine := Edit.CaretY;
    Edit.EnsureCursorPosVisible;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.ReindexSheets;

// assigns each edit an index of the tab sheet it belongs too

var
  I: Integer;

begin
  for I := 0 to PageControl.PageCount - 1 do PageControl.Pages[I].Controls[0].Tag := I;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.ClosePageItemClick(Sender: TObject);

// closes the currently active page and (if this is the last page) also the
// editor

var
  CanClose: Boolean;
  Tag: Integer;

begin
  CanClose := True;
  if FCurrentEdit.Modified then
    case AskForSave(ExtractFileName(FCurrentEdit.Info)) of
      mrYes: CanClose := SaveEdit(FCurrentEdit);
      mrCancel: CanClose := False;
    end;

  if CanClose then
  begin
    Tag := FCurrentEdit.Tag;
    RemoveEdit(FCurrentEdit);
    if PageControl.PageCount > 0 then
    begin
      if Tag > 0 then PageControl.ActivePage := PageControl.Pages[Tag - 1];
      ReindexSheets;
      PageControl.Invalidate;
      PageControlChange(PageControl);
    end
    else
    begin
      FCurrentEdit := nil;
      NeedSave := FProjectChanged;
      Close;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.FocusEdit(Edit: TSyntaxEdit);

// activates the given edit control;
// Windows.SetFocus is used since there might be some wrong settings while doing
// Disable/EnableTaskWindows.

begin
  Windows.SetFocus(Edit.Handle);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.UpdateProjectSettings;

// synchronizes the project settings (in particular file names) with the editor
// files

begin
  with ProjectSettings do
  begin
    if Assigned(LexerSourceEdit) and (LexerSource <> LexerSourceEdit.Info) then
    begin
      LexerSource := LexerSourceEdit.Info;
      FProjectChanged := True;
      NeedSave := True;
    end;
    if Assigned(LexerTemplateEdit) and (LexerTemplate <> LexerTemplateEdit.Info) then
    begin
      LexerTemplate := LexerTemplateEdit.Info;
      FProjectChanged := True;
      NeedSave := True;
    end;
    if Assigned(LexerTargetEdit) and (LexerTarget <> LexerTargetEdit.Info) then
    begin
      LexerTarget := LexerTargetEdit.Info;
      FProjectChanged := True;
      NeedSave := True;
    end;
    if Assigned(ParserSourceEdit) and (ParserSource <> ParserSourceEdit.Info) then
    begin
      ParserSource := ParserSourceEdit.Info;
      FProjectChanged := True;
      NeedSave := True;
    end;
    if Assigned(ParserTemplateEdit) and (ParserTemplate <> ParserTemplateEdit.Info) then
    begin
      ParserTemplate := ParserTemplateEdit.Info;
      FProjectChanged := True;
      NeedSave := True;
    end;
    if Assigned(ParserTargetEdit) and (ParserTarget <> ParserTargetEdit.Info) then
    begin
      ParserTarget := ParserTargetEdit.Info;
      FProjectChanged := True;
      NeedSave := True;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.RemoveEdit(Edit: TSyntaxEdit);

// removes the given edit from the page control and updates the reference in
// the project settings if necessary

var
  Sheet: TTabSheet;

begin
  with ProjectSettings do
  begin
    // remove the short hand reference from the project settings if the edit
    // is one of the core edits
    if Edit = LexerSourceEdit then LexerSourceEdit := nil
                              else
      if Edit = LexerTemplateEdit then LexerTemplateEdit := nil
                                  else
        if Edit = LexerTargetEdit then LexerTargetEdit := nil
                                  else
          if Edit = ParserSourceEdit then ParserSourceEdit := nil
                                     else
            if Edit = ParserTemplateEdit then ParserTemplateEdit := nil
                                         else
              if Edit = ParserTargetEdit then ParserTargetEdit := nil;
  end;

  // finally free the sheet where the edit is located on (and free the edit btw.)
  Sheet := PageControl.Pages[Edit.Tag];
  Sheet.PageControl := nil;
  Sheet.Free;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.WMDropFiles(var Msg: TWMDropFiles);

// the common game here if a file has been dropped on the editor

var
  Buffer: array[0..MAX_PATH] of Char;
  Count, I: Integer;

begin
  Count := DragQueryFile(Msg.Drop, DWORD(-1), nil, 0);
  for I := 0 to Count - 1 do
  begin
    DragQueryFile(Msg.Drop, I, Buffer, MAX_PATH);
    AddEdit(Buffer, True);
  end;
  DragFinish(Msg.Drop);
  Msg.Result := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.GetShowControlChars: Boolean;

begin
  Result := eoShowControlChars in FCurrentEdit.Options;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.SetShowControlChars(const Value: Boolean);

begin
  if Value then FCurrentEdit.Options := FCurrentEdit.Options + [eoShowControlChars]
           else FCurrentEdit.Options := FCurrentEdit.Options - [eoShowControlChars];
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.ApplyGlobalSettings;

// If the user has changed the short cut list or other options in the global settings
// then this method will be called.

var
  I: Integer;
  OtherOptions: TSyntaxEditOptions;

begin
  // page control settings
  PageControl.MultiLine := GlobalSettings.MultilinePageControl;
  if GlobalSettings.ShowFileIcons then PageControl.Images := ApplicationData.SystemImages
                                  else PageControl.Images := nil;

  // edit settings
  for I := 0 to PageControl.PageCount - 1 do
    with TSyntaxEdit(PageControl.Pages[I].Controls[0]) do
    begin
      Keystrokes := GlobalSettings.EditorCommands;
      // OtherOptions are either options always to be used or local to each edit
      OtherOptions := [eoWantTabs, eoReadOnlySelection] + Options * [eoShowControlChars] + Options * [eoReadOnly];
      Options := GlobalSettings.EditOptions + OtherOptions;
      TabSize := GlobalSettings.TabSize;
      IndentSize := GlobalSettings.BlockIndent;
      MaxUndo := GlobalSettings.UndoLimit;
      Font.Name := GlobalSettings.EditorFontName;
      Font.Size := GlobalSettings.EditorFontSize;
      MarkColor.Foreground := GlobalSettings.EditorMarkForeground;
      MarkColor.Background := GlobalSettings.EditorMarkBackground;
      ScrollbarMode := GlobalSettings.ScrollBarMode;
    end;

  if Assigned(FCurrentEdit) then
    if eoReadOnly  in FCurrentEdit.Options then StatusBar.Panels[2].Text := 'Read only'
                              else
      if eoInserting in FCurrentEdit.Options then StatusBar.Panels[2].Text := 'Insert'
                                             else StatusBar.Panels[2].Text := 'Overwrite';
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.Editorproperties1Click(Sender: TObject);

begin
  MainForm.ShowEnvironmentSettings(True);
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.CanShow: Boolean;

begin
  Result := PageControl.PageCount > 0;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.PageControlDrawTab(Control: TCustomTabControl; TabIndex: Integer; const Rect: TRect; Active: Boolean);

// I have to draw the tabs myself because I couldn't get the icons to draw
// transparently otherwise. I haven't found anything wrong with TSystemImageList,
// but this bloody control doesn't paint the icon transparently regardless of what
// I have tried with its properties (maybe I'm just too stupid).

var
  R: TRect;
  Sheet: TTabSheet;

begin
  R := Rect;
  with PageControl, ApplicationData.SystemImages do
  begin
    Sheet := Pages[TabIndex];
    Inc(R.Left, 2);
    if Assigned(Images) and (Sheet.ImageIndex > -1) then
    begin
      if Active then
      begin
        Inc(R.Left, 2);
        Draw(Control.Canvas, R.Left, R.Top + 4, Sheet.ImageIndex);
      end
      else Draw(Control.Canvas, R.Left, R.Top + 2, Sheet.ImageIndex);
      Inc(R.Left, Width);
    end;
    if Active then Canvas.Font.Color := clBlack
              else Canvas.Font.Color := clBtnShadow;
    DrawText(Control.Canvas.Handle, PChar(Sheet.Caption), Length(Sheet.Caption), R,
             DT_CENTER or DT_VCENTER or DT_SINGLELINE);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TProjectForm.SaveAll: Boolean;

var
  I: Integer;

begin
  Result := True;
  if FProjectChanged then Result := SaveProject;

  // One of the edits needs save?
  if Result and EditNeedsSave then
  begin
    for I := 0 to PageControl.PageCount - 1 do
    begin
      if TSyntaxEdit(PageControl.Pages[I].Controls[0]).Modified then
        Result := SaveEdit(TSyntaxEdit(PageControl.Pages[I].Controls[0]));
      if not Result then Break;
    end;
  end;
  if not Result then
  begin
    // might not have been saved (user can break the chain)
    if not FCurrentEdit.Modified then StatusBar.Panels[1].Text := '';
    NeedSave := FProjectChanged or EditNeedsSave;
  end
  else NeedSave := False;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.CheckFileAges;

var
  I: Integer;
  ThisAge: Integer;
  Edit: TSyntaxEdit;
  LastCaret: TPoint;
  LastTopLine: Integer;

begin
  for I := 0 to PageControl.PageCount - 1 do
  begin
    Edit := PageControl.Pages[I].Controls[0] as TSyntaxEdit;
    if FileExists(Edit.Info) then
    begin
      ThisAge := FileAge(Edit.Info);
      if (PageControl.Pages[I].Tag <> ThisAge) and
         (MessageDlg(Format(SFileDateTime, [Edit.Info]), mtConfirmation, [mbYes, mbNo], 0) = mrYes) then
      begin
        LastCaret := Edit.CaretXY;
        LastTopLine := Edit.TopLine;
        Edit.Lines.LoadFromFile(Edit.Info);
        Edit.Modified := False;
        Edit.CaretXY := LastCaret;
        Edit.TopLine := LastTopLine;

        if PageControl.ActivePage = PageControl.Pages[I] then StatusBar.Panels[1].Text := '';
      end;
      PageControl.Pages[I].Tag := ThisAge;
    end;
  end;
  NeedSave := FProjectChanged or EditNeedsSave;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.InsertItemClick(Sender: TObject);

begin
  case TMenuItem(Sender).MenuIndex of
    0:
      FCurrentEdit.SelText := #$BB' new comment';
    1:
      FCurrentEdit.SelText := #$A4'MacroIdentifier'#$A4;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.HighlighterClick(Sender: TObject);

begin
  FCurrentEdit.Highlighter := GlobalSettings.HighlighterList[TMenuItem(Sender).MenuIndex];
  TMenuItem(Sender).Checked := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.EditorPopupPopup(Sender: TObject);

// updates the checked highlighter item

var
  I: Integer;

begin
  if FCurrentEdit.Highlighter = nil then
  begin
    // remove check mark if there's currently no highlighter assigned
    for I := 0 to HighlighterItem.Count - 1 do
      HighlighterItem.Items[I].Checked := False;
  end
  else
    with GlobalSettings do
      for I := 0 to HighlighterList.Count - 1 do
        if HighlighterList[I] = FCurrentEdit.Highlighter then
        begin
          HighlighterItem.Items[I].Checked := True;
          Break;
        end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.BottomDockPanelDockOver(Sender: TObject; Source: TDragDockObject; X, Y: Integer; State: TDragState;
  var Accept: Boolean);

// produce feedback while docking, this method is only called if there's nothing yet
// docked to the DockPanel

var
  R: TRect;

begin
  Accept := True;
  if Accept then
  begin
    // modify the DockRect to preview dock area
    R.TopLeft := BottomDockPanel.ClientToScreen(Point(0, -FMessageForm.TBDockHeight));
    R.BottomRight := BottomDockPanel.ClientToScreen(Point(BottomDockPanel.Width, 0));
    Source.DockRect := R;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.LeftDockPanelDockOver(Sender: TObject; Source: TDragDockObject; X, Y: Integer; State: TDragState;
  var Accept: Boolean);
  
// produce feedback while docking, this method is only called if there's nothing yet
// docked to the DockPanel

var
  R: TRect;

begin
  Accept := True;
  if Accept then
  begin
    // modify the DockRect to preview dock area
    R.TopLeft := LeftDockPanel.ClientToScreen(Point(0, 0));
    R.BottomRight := LeftDockPanel.ClientToScreen(Point(FMessageForm.LRDockWidth, LeftDockPanel.Height));
    Source.DockRect := R;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.SplitterMoved(Sender: TObject);

begin
  if Sender = HSplitter then FMessageForm.TBDockHeight := ClientHeight - HSplitter.Top
                        else FMessageForm.LRDockWidth := VSplitter.Left;   
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.BottomDockPanelEndDock(Sender, Target: TObject; X, Y: Integer);

begin
  BottomDockPanel.Top := ClientHeight - FMessageForm.TBDockHeight;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.ShowDockPanel(APanel: TPanel; MakeVisible: Boolean; Client: TControl);

// Client - the docked client to show if we are re-showing the panel.
// Client is ignored if hiding the panel.

// Since docking to a non-visible docksite isn't allowed, instead of setting
// Visible for the panels we set the width/height to zero. The default InfluenceRect
// for a control extends a few pixels beyond it's boundaries, so it is possible
// to dock to zero width controls.

begin
  // don't try to hide a panel which has visible dock clients
  if not MakeVisible and (APanel.VisibleDockClientCount > 1) then Exit;

  if APanel = LeftDockPanel then VSplitter.Visible := MakeVisible
                            else HSplitter.Visible := MakeVisible;

  if MakeVisible then
    if APanel = LeftDockPanel then
    begin
      APanel.Width := ClientWidth div 3;
      VSplitter.Left := APanel.Width + VSplitter.Width;
    end
    else
    begin
      APanel.Height := ClientHeight div 3;
      HSplitter.Top := ClientHeight - APanel.Height - HSplitter.Height;
    end
  else
    if APanel = LeftDockPanel then APanel.Width := 0
                              else APanel.Height := 0;

  if MakeVisible and Assigned(Client) then
  begin
    Client.Show;
    Client.Align := alClient;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.DockPanelDockDrop(Sender: TObject; Source: TDragDockObject; X, Y: Integer);

begin
  // OnDockDrop gets called AFTER the client has actually docked,
  // so we check for DockClientCount = 1 before making the dock panel visible.
  if (Sender as TPanel).DockClientCount = 1 then ShowDockPanel(Sender as TPanel, True, nil);
  (Sender as TPanel).DockManager.ResetBounds(True);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.DockPanelUnDock(Sender: TObject; Client: TControl; NewTarget: TWinControl;
  var Allow: Boolean);

begin
  Allow := True;
  // OnUnDock gets called BEFORE the client is undocked, in order to optionally
  // disallow the undock. DockClientCount is never 0 when called from this event.
  if (Sender as TPanel).DockClientCount = 1 then ShowDockPanel(Sender as TPanel, False, nil);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TProjectForm.LeftDockPanelGetSiteInfo(Sender: TObject; DockClient: TControl; var InfluenceRect: TRect;
  MousePos: TPoint; var CanDock: Boolean);

begin
  CanDock := True;
end;

//----------------------------------------------------------------------------------------------------------------------

end.



 


 
