unit DelphiLexer;
 
// this unit contains the implementation of a lexical analyzer
// see documentation for an extensive description
 
interface
 
uses Classes;
 
const CR = #13;
      LF = #10;
 
// user defined start states
const Normal = 2;
      String1 = 4;
      String2 = 6;
 
      asmString1 = 8;
      asmString2 = 10;
      asmState = 12;
 
      OverreadState = 14;
 
      CompilerSwitch = 16;
 
      Comment = 18;
 
      CharTable = 20;
 
 
const kwABSOLUTE= 257;
      kwABSTRACT= 258;
      kwAND= 259;
      kwARRAY= 260;
      kwAS= 261;
      kwASM= 262;
      kwASSEMBLER= 263;
      kwAT= 264;
      kwAUTOMATED= 265;
      kwBEGIN= 266;
      kwCASE= 267;
      kwCDECL= 268;
      kwCLASS= 269;
      kwCONST= 270;
      kwCONSTRUCTOR= 271;
      kwCONTAINS= 272;
      kwDEFAULT= 273;
      kwDESTRUCTOR= 274;
      kwDISPID= 275;
      kwDISPINTERFACE= 276;
      kwDIV= 277;
      kwDO= 278;
      kwDOWNTO= 279;
      kwDYNAMIC= 280;
      kwELSE= 281;
      kwEND= 282;
      kwEXCEPT= 283;
      kwEXPORT= 284;
      kwEXPORTS= 285;
      kwEXTERNAL= 286;
      kwFAR= 287;
      kwFILE= 288;
      kwFINALIZATION= 289;
      kwFINALLY= 290;
      kwFOR= 291;
      kwFORWARD= 292;
      kwFUNCTION= 293;
      kwGOTO= 294;
      kwIF= 295;
      kwIMPLEMENTATION= 296;
      kwIMPLEMENTS= 297;
      kwIN= 298;
      kwINDEX= 299;
      kwINHERITED= 300;
      kwINITIALIZATION= 301;
      kwINLINE= 302;
      kwINTERFACE= 303;
      kwIS= 304;
      kwLABEL= 305;
      kwLIBRARY= 306;
      kwMESSAGE= 307;
      kwMOD= 308;
      kwNAME= 309;
      kwNEAR= 310;
      kwNIL= 311;
      kwNODEFAULT= 312;
      kwNOT= 313;
      kwOBJECT= 314;
      kwOF= 315;
      kwON= 316;
      kwOR= 317;
      kwOUT= 318;
      kwOVERLOAD= 319;
      kwOVERRIDE= 320;
      kwPACKAGE= 321;
      kwPACKED= 322;
      kwPASCAL= 323;
      kwPRIVATE= 324;
      kwPROCEDURE= 325;
      kwPROGRAM= 326;
      kwPROPERTY= 327;
      kwPROTECTED= 328;
      kwPUBLIC= 329;
      kwPUBLISHED= 330;
      kwRAISE= 331;
      kwREAD= 332;
      kwREADONLY= 333;
      kwRECORD= 334;
      kwREGISTER= 335;
      kwREINTRODUCE= 336;
      kwREPEAT= 337;
      kwREQUIRES= 338;
      kwRESIDENT= 339;
      kwRESOURCESTRING= 340;
      kwSAFECALL= 341;
      kwSET= 342;
      kwSHL= 343;
      kwSHR= 344;
      kwSTDCALL= 345;
      kwSTORED= 346;
      kwSTRING= 347;
      kwTHEN= 348;
      kwTHREADVAR= 349;
      kwTO= 350;
      kwTRY= 351;
      kwTYPE= 352;
      kwUNIT= 353;
      kwUNTIL= 354;
      kwUSES= 355;
      kwVAR= 356;
      kwVIRTUAL= 357;
      kwWHILE= 358;
      kwWITH= 359;
      kwWRITE= 360;
      kwWRITEONLY= 361;

type
  TDelphiLexer = class
  private
    FCurrentToken: Integer;            // type of current token
    FInput: PChar;                     // input string to be tokenized
    FCurrentPos: Cardinal;             // current lexer position in FInput
    FTokenPos: Cardinal;               // current token position in FInput
    FLookahead: Char;                  // next char after current position
    FCurrentState,
    FStartState,
    FLineState: Cardinal;              // lexer states
    FLastChar: Char;                   // last matched char (#0 if no match)
    FPositions: array of Cardinal;     // for each rule the last marked position, zeroed when
                                       // rule has already been considered (allocated once)
    FMatchStack: array of Cardinal;    // stack containing matched rules (growing dynamically)
    FMatchCount: Integer;              // contains current number of matches
    FRejected: Boolean;                // current match rejected?
    FDoStop: Boolean;                  // token or eof found?
    procedure RuleToToken(Rule: Integer);
    function IsKeyword: Boolean;
    function GetNextChar: Char;
    function MatchRule(var Rule: Integer): Boolean;
    function DoDefault: Boolean;
  protected
    function GetToken: String; virtual;
    procedure SetMatch(I: Integer);
    procedure SetState(Value: Cardinal); virtual;
    procedure SetToken(Token: Integer);
  public
    constructor Create; virtual;
    destructor Destroy; override;
 
    function GetEol: Boolean; virtual;
    function GetTokenInfo(var S: PChar; var Pos: Integer): Integer; virtual;
 
    procedure Next; virtual;
    procedure SetInput(const Value: String); virtual;
    procedure Reject;
    procedure Reset; virtual;
 
    property NextChar: Char read GetNextChar;
    property Lookahead: Char read FLookahead;
    property State: Cardinal read FCurrentState write SetState; // read to get current lexer state
                                                                // write to set start state for next run
    property TokenPosition: Cardinal read FTokenPos;
  end;
 
//--------------------------------------------------------------------------------
 
implementation
 
uses SysUtils;
 
// DFA table
type TTransition = record
       CharClass : set of Char;
       NextState : Integer;
     end;

const
  MarkPositionCount = 174;
  MatchCount        = 174;
  TransitionCount   = 251;
  StateCount        = 141;

  MarkPositionTable : array[0..MarkPositionCount - 1] of Integer = (
    11, 11, 0, 57, 1, 57, 47, 57, 15, 57, 28, 57, 57, 29, 57, 55, 57, 49, 57, 44, 
    57, 45, 57, 37, 38, 39, 57, 40, 57, 41, 57, 42, 57, 43, 57, 46, 57, 48, 57, 
    50, 57, 51, 57, 52, 57, 53, 57, 54, 57, 38, 57, 57, 2, 4, 4, 4, 6, 14, 13, 14, 
    9, 14, 12, 14, 7, 26, 10, 26, 26, 47, 16, 26, 17, 26, 18, 26, 51, 25, 26, 25, 
    26, 25, 26, 22, 23, 26, 50, 24, 26, 26, 26, 26, 39, 26, 40, 26, 41, 26, 42, 
    26, 43, 26, 44, 26, 45, 26, 46, 26, 48, 26, 49, 26, 52, 26, 53, 26, 54, 26, 
    55, 26, 56, 62, 58, 62, 59, 60, 62, 61, 62, 62, 66, 68, 67, 68, 68, 65, 63, 
    65, 64, 65, 69, 15, 35, 28, 27, 28, 29, 30, 36, 31, 32, 33, 34, 3, 5, 8, 16, 
    17, 18, 18, 25, 19, 25, 20, 25, 25, 58, 67, 63, 27, 27, 3, 21, 27);

  MatchTable : array [0..MatchCount - 1] of Integer = (
    11, 11, 0, 57, 1, 57, 47, 57, 15, 57, 28, 57, 57, 29, 57, 55, 57, 49, 57, 44, 
    57, 45, 57, 37, 38, 39, 57, 40, 57, 41, 57, 42, 57, 43, 57, 46, 57, 48, 57, 
    50, 57, 51, 57, 52, 57, 53, 57, 54, 57, 38, 57, 57, 2, 4, 4, 4, 6, 14, 13, 14, 
    9, 14, 12, 14, 7, 26, 10, 26, 26, 47, 16, 26, 17, 26, 18, 26, 51, 25, 26, 25, 
    26, 25, 26, 22, 23, 26, 50, 24, 26, 26, 26, 26, 39, 26, 40, 26, 41, 26, 42, 
    26, 43, 26, 44, 26, 45, 26, 46, 26, 48, 26, 49, 26, 52, 26, 53, 26, 54, 26, 
    55, 26, 56, 62, 58, 62, 59, 60, 62, 61, 62, 62, 66, 68, 67, 68, 68, 65, 63, 
    65, 64, 65, 69, 15, 35, 28, 28, 29, 30, 36, 31, 32, 33, 34, 3, 5, 8, 16, 17, 
    18, 18, 25, 19, 25, 20, 25, 25, 58, 67, 63, 27, 27, 27, 3, 21, 27);

  TransitionTable : array [0..TransitionCount - 1] of TTransition = (
    (CharClass: [#0]; NextState: 33),
    (CharClass: [#1..#9, #11..' ']; NextState: 47),
    (CharClass: [#10]; NextState: 34),
    (CharClass: ['!'..'"', '%'..'&', '?', '\', '`', '|'..#255]; NextState: 48),
    (CharClass: ['#']; NextState: 23),
    (CharClass: ['$']; NextState: 27),
    (CharClass: ['''']; NextState: 22),
    (CharClass: ['(']; NextState: 24),
    (CharClass: [')']; NextState: 41),
    (CharClass: ['*']; NextState: 37),
    (CharClass: ['+']; NextState: 35),
    (CharClass: [',']; NextState: 40),
    (CharClass: ['-']; NextState: 36),
    (CharClass: ['.']; NextState: 29),
    (CharClass: ['/']; NextState: 38),
    (CharClass: ['0'..'9']; NextState: 26),
    (CharClass: [':']; NextState: 30),
    (CharClass: [';']; NextState: 42),
    (CharClass: ['<']; NextState: 31),
    (CharClass: ['=']; NextState: 39),
    (CharClass: ['>']; NextState: 32),
    (CharClass: ['@']; NextState: 43),
    (CharClass: ['A'..'Z', '_', 'a'..'z']; NextState: 28),
    (CharClass: ['[']; NextState: 45),
    (CharClass: [']']; NextState: 46),
    (CharClass: ['^']; NextState: 44),
    (CharClass: ['{']; NextState: 25),
    (CharClass: [#0]; NextState: 33),
    (CharClass: [#1..#9, #11..' ']; NextState: 47),
    (CharClass: [#10]; NextState: 34),
    (CharClass: ['!'..'"', '%'..'&', '?', '\', '`', '|'..#255]; NextState: 48),
    (CharClass: ['#']; NextState: 23),
    (CharClass: ['$']; NextState: 27),
    (CharClass: ['''']; NextState: 22),
    (CharClass: ['(']; NextState: 24),
    (CharClass: [')']; NextState: 41),
    (CharClass: ['*']; NextState: 37),
    (CharClass: ['+']; NextState: 35),
    (CharClass: [',']; NextState: 40),
    (CharClass: ['-']; NextState: 36),
    (CharClass: ['.']; NextState: 29),
    (CharClass: ['/']; NextState: 38),
    (CharClass: ['0'..'9']; NextState: 26),
    (CharClass: [':']; NextState: 30),
    (CharClass: [';']; NextState: 42),
    (CharClass: ['<']; NextState: 31),
    (CharClass: ['=']; NextState: 39),
    (CharClass: ['>']; NextState: 32),
    (CharClass: ['@']; NextState: 43),
    (CharClass: ['A'..'Z', '_', 'a'..'z']; NextState: 28),
    (CharClass: ['[']; NextState: 45),
    (CharClass: [']']; NextState: 46),
    (CharClass: ['^']; NextState: 44),
    (CharClass: ['{']; NextState: 25),
    (CharClass: [#1..#9, #11..'"', '$'..'&', '('..#255]; NextState: 51),
    (CharClass: ['#']; NextState: 50),
    (CharClass: ['''']; NextState: 49),
    (CharClass: [#1..#9, #11..'"', '$'..'&', '('..#255]; NextState: 51),
    (CharClass: ['#']; NextState: 50),
    (CharClass: ['''']; NextState: 49),
    (CharClass: [#1..#9, #11..'&', '('..#255]; NextState: 54),
    (CharClass: [#10]; NextState: 53),
    (CharClass: ['''']; NextState: 52),
    (CharClass: [#1..#9, #11..'&', '('..#255]; NextState: 54),
    (CharClass: [#10]; NextState: 53),
    (CharClass: ['''']; NextState: 52),
    (CharClass: [#1..#9, #11..'&', '('..#255]; NextState: 54),
    (CharClass: [#10]; NextState: 53),
    (CharClass: ['''']; NextState: 55),
    (CharClass: [#1..#9, #11..'&', '('..#255]; NextState: 54),
    (CharClass: [#10]; NextState: 53),
    (CharClass: ['''']; NextState: 55),
    (CharClass: [#1..#9, #11..'!', '#'..#255]; NextState: 54),
    (CharClass: [#10]; NextState: 53),
    (CharClass: ['"']; NextState: 56),
    (CharClass: [#1..#9, #11..'!', '#'..#255]; NextState: 54),
    (CharClass: [#10]; NextState: 53),
    (CharClass: ['"']; NextState: 56),
    (CharClass: [#1..#9, #11..' ']; NextState: 68),
    (CharClass: [#10]; NextState: 66),
    (CharClass: ['!', '#', '%', '?', '\', '`', '|'..#255]; NextState: 70),
    (CharClass: ['"']; NextState: 58),
    (CharClass: ['$']; NextState: 69),
    (CharClass: ['&']; NextState: 85),
    (CharClass: ['''']; NextState: 57),
    (CharClass: ['(']; NextState: 59),
    (CharClass: [')']; NextState: 79),
    (CharClass: ['*']; NextState: 73),
    (CharClass: ['+']; NextState: 71),
    (CharClass: [',']; NextState: 78),
    (CharClass: ['-']; NextState: 72),
    (CharClass: ['.']; NextState: 84),
    (CharClass: ['/']; NextState: 74),
    (CharClass: ['0'..'1']; NextState: 63),
    (CharClass: ['2'..'7']; NextState: 64),
    (CharClass: ['8'..'9']; NextState: 65),
    (CharClass: [':']; NextState: 80),
    (CharClass: [';']; NextState: 67),
    (CharClass: ['<']; NextState: 76),
    (CharClass: ['=']; NextState: 75),
    (CharClass: ['>']; NextState: 77),
    (CharClass: ['@']; NextState: 62),
    (CharClass: ['A'..'Z', '_', 'a'..'z']; NextState: 61),
    (CharClass: ['[']; NextState: 82),
    (CharClass: [']']; NextState: 83),
    (CharClass: ['^']; NextState: 81),
    (CharClass: ['{']; NextState: 60),
    (CharClass: [#1..#9, #11..' ']; NextState: 68),
    (CharClass: [#10]; NextState: 66),
    (CharClass: ['!', '#', '%', '?', '\', '`', '|'..#255]; NextState: 70),
    (CharClass: ['"']; NextState: 58),
    (CharClass: ['$']; NextState: 69),
    (CharClass: ['&']; NextState: 85),
    (CharClass: ['''']; NextState: 57),
    (CharClass: ['(']; NextState: 59),
    (CharClass: [')']; NextState: 79),
    (CharClass: ['*']; NextState: 73),
    (CharClass: ['+']; NextState: 71),
    (CharClass: [',']; NextState: 78),
    (CharClass: ['-']; NextState: 72),
    (CharClass: ['.']; NextState: 84),
    (CharClass: ['/']; NextState: 74),
    (CharClass: ['0'..'1']; NextState: 63),
    (CharClass: ['2'..'7']; NextState: 64),
    (CharClass: ['8'..'9']; NextState: 65),
    (CharClass: [':']; NextState: 80),
    (CharClass: [';']; NextState: 67),
    (CharClass: ['<']; NextState: 76),
    (CharClass: ['=']; NextState: 75),
    (CharClass: ['>']; NextState: 77),
    (CharClass: ['@']; NextState: 62),
    (CharClass: ['A'..'Z', '_', 'a'..'z']; NextState: 61),
    (CharClass: ['[']; NextState: 82),
    (CharClass: [']']; NextState: 83),
    (CharClass: ['^']; NextState: 81),
    (CharClass: ['{']; NextState: 60),
    (CharClass: [#0]; NextState: 88),
    (CharClass: [#1..#9, #11..'!', '#'..'&', ')'..'z', '|'..#255]; NextState: 91),
    (CharClass: ['"']; NextState: 90),
    (CharClass: ['''']; NextState: 89),
    (CharClass: ['(']; NextState: 86),
    (CharClass: ['{']; NextState: 87),
    (CharClass: [#0]; NextState: 88),
    (CharClass: [#1..#9, #11..'!', '#'..'&', ')'..'z', '|'..#255]; NextState: 91),
    (CharClass: ['"']; NextState: 90),
    (CharClass: ['''']; NextState: 89),
    (CharClass: ['(']; NextState: 86),
    (CharClass: ['{']; NextState: 87),
    (CharClass: [#0]; NextState: 92),
    (CharClass: [#1..#9, #11..')', '+'..'|', '~'..#255]; NextState: 95),
    (CharClass: ['*']; NextState: 93),
    (CharClass: ['}']; NextState: 94),
    (CharClass: [#0]; NextState: 92),
    (CharClass: [#1..#9, #11..')', '+'..'|', '~'..#255]; NextState: 95),
    (CharClass: ['*']; NextState: 93),
    (CharClass: ['}']; NextState: 94),
    (CharClass: [#0]; NextState: 98),
    (CharClass: [#1..#9, #11..')', '+'..'|', '~'..#255]; NextState: 99),
    (CharClass: ['*']; NextState: 96),
    (CharClass: ['}']; NextState: 97),
    (CharClass: [#0]; NextState: 98),
    (CharClass: [#1..#9, #11..')', '+'..'|', '~'..#255]; NextState: 99),
    (CharClass: ['*']; NextState: 96),
    (CharClass: ['}']; NextState: 97),
    (CharClass: [#1..#9, #11..#255]; NextState: 100),
    (CharClass: [#1..#9, #11..#255]; NextState: 100),
    (CharClass: ['*']; NextState: 101),
    (CharClass: ['.']; NextState: 102),
    (CharClass: ['.']; NextState: 104),
    (CharClass: ['0'..'9']; NextState: 103),
    (CharClass: ['E', 'e']; NextState: 105),
    (CharClass: ['0'..'9', 'A'..'F', 'a'..'f']; NextState: 106),
    (CharClass: ['0'..'9', 'A'..'Z', '_', 'a'..'z']; NextState: 107),
    (CharClass: [')']; NextState: 109),
    (CharClass: ['.']; NextState: 108),
    (CharClass: ['=']; NextState: 110),
    (CharClass: ['=']; NextState: 111),
    (CharClass: ['>']; NextState: 112),
    (CharClass: ['=']; NextState: 113),
    (CharClass: ['$']; NextState: 115),
    (CharClass: ['0'..'9']; NextState: 114),
    (CharClass: ['''']; NextState: 116),
    (CharClass: ['''']; NextState: 117),
    (CharClass: ['*']; NextState: 118),
    (CharClass: ['0'..'9', 'A'..'Z', '_', 'a'..'z']; NextState: 119),
    (CharClass: ['0'..'9', 'A'..'Z', '_', 'a'..'z']; NextState: 121),
    (CharClass: ['@']; NextState: 120),
    (CharClass: ['0'..'1']; NextState: 122),
    (CharClass: ['2'..'7']; NextState: 124),
    (CharClass: ['8'..'9']; NextState: 127),
    (CharClass: ['A', 'C'..'F', 'a', 'c'..'f']; NextState: 126),
    (CharClass: ['B', 'b']; NextState: 123),
    (CharClass: ['O', 'o']; NextState: 125),
    (CharClass: ['0'..'7']; NextState: 124),
    (CharClass: ['8'..'9']; NextState: 127),
    (CharClass: ['A'..'F', 'a'..'f']; NextState: 126),
    (CharClass: ['O', 'o']; NextState: 125),
    (CharClass: ['0'..'9']; NextState: 127),
    (CharClass: ['A'..'F', 'a'..'f']; NextState: 126),
    (CharClass: ['0'..'9', 'A'..'F', 'a'..'f']; NextState: 128),
    (CharClass: ['*']; NextState: 129),
    (CharClass: [')']; NextState: 130),
    (CharClass: [')']; NextState: 131),
    (CharClass: ['.']; NextState: 104),
    (CharClass: ['0'..'9']; NextState: 103),
    (CharClass: ['E', 'e']; NextState: 105),
    (CharClass: [#1..'-', '/', ':'..'D', 'F'..'d', 'f'..#255]; NextState: 134),
    (CharClass: ['0'..'9']; NextState: 132),
    (CharClass: ['E', 'e']; NextState: 133),
    (CharClass: ['+', '-']; NextState: 135),
    (CharClass: ['0'..'9']; NextState: 136),
    (CharClass: ['0'..'9', 'A'..'F', 'a'..'f']; NextState: 106),
    (CharClass: ['0'..'9', 'A'..'Z', '_', 'a'..'z']; NextState: 107),
    (CharClass: ['0'..'9']; NextState: 114),
    (CharClass: ['0'..'9', 'A'..'F', 'a'..'f']; NextState: 137),
    (CharClass: ['0'..'9', 'A'..'Z', '_', 'a'..'z']; NextState: 119),
    (CharClass: ['0'..'9', 'A'..'Z', '_', 'a'..'z']; NextState: 121),
    (CharClass: ['@']; NextState: 120),
    (CharClass: ['0'..'9', '@'..'Z', '_', 'a'..'z']; NextState: 121),
    (CharClass: ['0'..'1']; NextState: 122),
    (CharClass: ['2'..'7']; NextState: 124),
    (CharClass: ['8'..'9']; NextState: 127),
    (CharClass: ['A', 'C'..'F', 'a', 'c'..'f']; NextState: 126),
    (CharClass: ['B', 'b']; NextState: 123),
    (CharClass: ['H', 'h']; NextState: 138),
    (CharClass: ['O', 'o']; NextState: 125),
    (CharClass: ['0'..'9', 'A'..'F', 'a'..'f']; NextState: 126),
    (CharClass: ['H', 'h']; NextState: 138),
    (CharClass: ['0'..'7']; NextState: 124),
    (CharClass: ['8'..'9']; NextState: 127),
    (CharClass: ['A'..'F', 'a'..'f']; NextState: 126),
    (CharClass: ['H', 'h']; NextState: 138),
    (CharClass: ['O', 'o']; NextState: 125),
    (CharClass: ['0'..'9', 'A'..'F', 'a'..'f']; NextState: 126),
    (CharClass: ['H', 'h']; NextState: 138),
    (CharClass: ['0'..'9']; NextState: 127),
    (CharClass: ['A'..'F', 'a'..'f']; NextState: 126),
    (CharClass: ['H', 'h']; NextState: 138),
    (CharClass: ['0'..'9', 'A'..'F', 'a'..'f']; NextState: 128),
    (CharClass: [#1..'-', '/', ':'..'D', 'F'..'d', 'f'..#255]; NextState: 134),
    (CharClass: ['0'..'9']; NextState: 132),
    (CharClass: ['E', 'e']; NextState: 133),
    (CharClass: ['+', '-']; NextState: 139),
    (CharClass: ['0'..'9']; NextState: 136),
    (CharClass: ['0'..'9']; NextState: 136),
    (CharClass: [#1..'-', '/', ':'..#255]; NextState: 134),
    (CharClass: ['0'..'9']; NextState: 140),
    (CharClass: ['0'..'9', 'A'..'F', 'a'..'f']; NextState: 137),
    (CharClass: ['0'..'9']; NextState: 136),
    (CharClass: [#1..'-', '/', ':'..#255]; NextState: 134),
    (CharClass: ['0'..'9']; NextState: 140)
    );

  MarksLow : array [0..StateCount-1] of Integer = (
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 4, 6, 8, 
    10, 12, 13, 15, 17, 19, 21, 23, 24, 25, 27, 29, 31, 33, 35, 37, 39, 41, 43, 
    45, 47, 49, 51, 52, 54, 55, 56, 58, 59, 60, 62, 64, 66, 68, 70, 72, 74, 77, 
    79, 81, 83, 84, 87, 89, 90, 91, 93, 95, 97, 99, 101, 103, 105, 107, 109, 111, 
    113, 115, 117, 119, 121, 122, 124, 125, 127, 129, 130, 131, 132, 134, 135, 136, 
    138, 139, 140, 141, 142, 143, 144, 145, 145, 146, 147, 148, 149, 150, 151, 152, 
    153, 154, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 164, 165, 166, 
    167, 168, 169, 170, 170, 170, 170, 171, 172, 173, 173);

  MarksHigh : array [0..StateCount-1] of Integer = (
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
    3, 5, 7, 9, 11, 12, 14, 16, 18, 20, 22, 23, 24, 26, 28, 30, 32, 34, 36, 38, 
    40, 42, 44, 46, 48, 50, 51, 53, 54, 55, 57, 58, 59, 61, 63, 65, 67, 69, 71, 
    73, 76, 78, 80, 82, 83, 86, 88, 89, 90, 92, 94, 96, 98, 100, 102, 104, 106, 
    108, 110, 112, 114, 116, 118, 120, 121, 123, 124, 126, 128, 129, 130, 131, 133, 
    134, 135, 137, 138, 139, 140, 141, 142, 143, 144, 144, 145, 146, 147, 148, 149, 
    150, 151, 152, 153, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 163, 
    164, 165, 166, 167, 168, 169, 169, 169, 169, 170, 171, 172, 172, 173);

  MatchesLow : array [0..StateCount-1] of Integer = (
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 4, 6, 8, 
    10, 12, 13, 15, 17, 19, 21, 23, 24, 25, 27, 29, 31, 33, 35, 37, 39, 41, 43, 
    45, 47, 49, 51, 52, 54, 55, 56, 58, 59, 60, 62, 64, 66, 68, 70, 72, 74, 77, 
    79, 81, 83, 84, 87, 89, 90, 91, 93, 95, 97, 99, 101, 103, 105, 107, 109, 111, 
    113, 115, 117, 119, 121, 122, 124, 125, 127, 129, 130, 131, 132, 134, 135, 136, 
    138, 139, 140, 141, 142, 143, 144, 144, 144, 145, 146, 147, 148, 149, 150, 151, 
    152, 153, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 163, 164, 165, 
    166, 167, 168, 169, 170, 171, 171, 171, 172, 173, 173);

  MatchesHigh : array [0..StateCount-1] of Integer = (
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
    3, 5, 7, 9, 11, 12, 14, 16, 18, 20, 22, 23, 24, 26, 28, 30, 32, 34, 36, 38, 
    40, 42, 44, 46, 48, 50, 51, 53, 54, 55, 57, 58, 59, 61, 63, 65, 67, 69, 71, 
    73, 76, 78, 80, 82, 83, 86, 88, 89, 90, 92, 94, 96, 98, 100, 102, 104, 106, 
    108, 110, 112, 114, 116, 118, 120, 121, 123, 124, 126, 128, 129, 130, 131, 133, 
    134, 135, 137, 138, 139, 140, 141, 142, 143, 143, 143, 144, 145, 146, 147, 148, 
    149, 150, 151, 152, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 162, 
    163, 164, 165, 166, 167, 168, 169, 170, 170, 170, 171, 172, 172, 173);

  TransitionsLow : array [0..StateCount-1] of Integer = (
    0, 0, 0, 27, 54, 57, 60, 63, 66, 69, 72, 75, 78, 107, 136, 142, 148, 152, 156, 
    160, 164, 165, 166, 166, 166, 168, 168, 171, 172, 173, 175, 176, 178, 179, 179, 
    179, 179, 179, 179, 179, 179, 179, 179, 179, 179, 179, 179, 179, 179, 179, 179, 
    181, 181, 182, 182, 182, 183, 183, 183, 183, 184, 184, 185, 187, 193, 197, 199, 
    199, 199, 199, 200, 200, 200, 200, 200, 200, 200, 200, 200, 200, 200, 200, 200, 
    200, 200, 200, 200, 201, 201, 201, 201, 201, 201, 201, 202, 202, 202, 203, 203, 
    203, 203, 203, 203, 203, 206, 209, 211, 212, 213, 213, 213, 213, 213, 213, 213, 
    214, 215, 215, 215, 215, 216, 218, 219, 226, 228, 233, 233, 235, 238, 239, 239, 
    239, 239, 242, 244, 244, 245, 247, 248, 248, 249);

  TransitionsHigh : array [0..StateCount-1] of Integer = (
    -1, -1, 26, 53, 56, 59, 62, 65, 68, 71, 74, 77, 106, 135, 141, 147, 151, 155, 
    159, 163, 164, 165, 165, 165, 167, 167, 170, 171, 172, 174, 175, 177, 178, 178, 
    178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 
    180, 180, 181, 181, 181, 182, 182, 182, 182, 183, 183, 184, 186, 192, 196, 198, 
    198, 198, 198, 199, 199, 199, 199, 199, 199, 199, 199, 199, 199, 199, 199, 199, 
    199, 199, 199, 199, 200, 200, 200, 200, 200, 200, 200, 201, 201, 201, 202, 202, 
    202, 202, 202, 202, 202, 205, 208, 210, 211, 212, 212, 212, 212, 212, 212, 212, 
    213, 214, 214, 214, 214, 215, 217, 218, 225, 227, 232, 232, 234, 237, 238, 238, 
    238, 238, 241, 243, 243, 244, 246, 247, 247, 248, 250);

type // keyword table support
     TKeyword = record
       Keyword: String;
       Token: Integer;
     end;

const KeywordCount = 105;
      Keywords : array[0..KeywordCount - 1] of TKeyword = (
        (Keyword: 'absolute'; Token: kwABSOLUTE), 
        (Keyword: 'abstract'; Token: kwABSTRACT), 
        (Keyword: 'and'; Token: kwAND), 
        (Keyword: 'array'; Token: kwARRAY), 
        (Keyword: 'as'; Token: kwAS), 
        (Keyword: 'asm'; Token: kwASM), 
        (Keyword: 'assembler'; Token: kwASSEMBLER), 
        (Keyword: 'at'; Token: kwAT), 
        (Keyword: 'automated'; Token: kwAUTOMATED), 
        (Keyword: 'begin'; Token: kwBEGIN), 
        (Keyword: 'case'; Token: kwCASE), 
        (Keyword: 'cdecl'; Token: kwCDECL), 
        (Keyword: 'class'; Token: kwCLASS), 
        (Keyword: 'const'; Token: kwCONST), 
        (Keyword: 'constructor'; Token: kwCONSTRUCTOR), 
        (Keyword: 'contains'; Token: kwCONTAINS), 
        (Keyword: 'default'; Token: kwDEFAULT), 
        (Keyword: 'destructor'; Token: kwDESTRUCTOR), 
        (Keyword: 'dispid'; Token: kwDISPID), 
        (Keyword: 'dispinterface'; Token: kwDISPINTERFACE), 
        (Keyword: 'div'; Token: kwDIV), 
        (Keyword: 'do'; Token: kwDO), 
        (Keyword: 'downto'; Token: kwDOWNTO), 
        (Keyword: 'dynamic'; Token: kwDYNAMIC), 
        (Keyword: 'else'; Token: kwELSE), 
        (Keyword: 'end'; Token: kwEND), 
        (Keyword: 'except'; Token: kwEXCEPT), 
        (Keyword: 'export'; Token: kwEXPORT), 
        (Keyword: 'exports'; Token: kwEXPORTS), 
        (Keyword: 'external'; Token: kwEXTERNAL), 
        (Keyword: 'far'; Token: kwFAR), 
        (Keyword: 'file'; Token: kwFILE), 
        (Keyword: 'finalization'; Token: kwFINALIZATION), 
        (Keyword: 'finally'; Token: kwFINALLY), 
        (Keyword: 'for'; Token: kwFOR), 
        (Keyword: 'forward'; Token: kwFORWARD), 
        (Keyword: 'function'; Token: kwFUNCTION), 
        (Keyword: 'goto'; Token: kwGOTO), 
        (Keyword: 'if'; Token: kwIF), 
        (Keyword: 'implementation'; Token: kwIMPLEMENTATION), 
        (Keyword: 'implements'; Token: kwIMPLEMENTS), 
        (Keyword: 'in'; Token: kwIN), 
        (Keyword: 'index'; Token: kwINDEX), 
        (Keyword: 'inherited'; Token: kwINHERITED), 
        (Keyword: 'initialization'; Token: kwINITIALIZATION), 
        (Keyword: 'inline'; Token: kwINLINE), 
        (Keyword: 'interface'; Token: kwINTERFACE), 
        (Keyword: 'is'; Token: kwIS), 
        (Keyword: 'label'; Token: kwLABEL), 
        (Keyword: 'library'; Token: kwLIBRARY), 
        (Keyword: 'message'; Token: kwMESSAGE), 
        (Keyword: 'mod'; Token: kwMOD), 
        (Keyword: 'name'; Token: kwNAME), 
        (Keyword: 'near'; Token: kwNEAR), 
        (Keyword: 'nil'; Token: kwNIL), 
        (Keyword: 'nodefault'; Token: kwNODEFAULT), 
        (Keyword: 'not'; Token: kwNOT), 
        (Keyword: 'object'; Token: kwOBJECT), 
        (Keyword: 'of'; Token: kwOF), 
        (Keyword: 'on'; Token: kwON), 
        (Keyword: 'or'; Token: kwOR), 
        (Keyword: 'out'; Token: kwOUT), 
        (Keyword: 'overload'; Token: kwOVERLOAD), 
        (Keyword: 'override'; Token: kwOVERRIDE), 
        (Keyword: 'package'; Token: kwPACKAGE), 
        (Keyword: 'packed'; Token: kwPACKED), 
        (Keyword: 'pascal'; Token: kwPASCAL), 
        (Keyword: 'private'; Token: kwPRIVATE), 
        (Keyword: 'procedure'; Token: kwPROCEDURE), 
        (Keyword: 'program'; Token: kwPROGRAM), 
        (Keyword: 'property'; Token: kwPROPERTY), 
        (Keyword: 'protected'; Token: kwPROTECTED), 
        (Keyword: 'public'; Token: kwPUBLIC), 
        (Keyword: 'published'; Token: kwPUBLISHED), 
        (Keyword: 'raise'; Token: kwRAISE), 
        (Keyword: 'read'; Token: kwREAD), 
        (Keyword: 'readonly'; Token: kwREADONLY), 
        (Keyword: 'record'; Token: kwRECORD), 
        (Keyword: 'register'; Token: kwREGISTER), 
        (Keyword: 'reintroduce'; Token: kwREINTRODUCE), 
        (Keyword: 'repeat'; Token: kwREPEAT), 
        (Keyword: 'requires'; Token: kwREQUIRES), 
        (Keyword: 'resident'; Token: kwRESIDENT), 
        (Keyword: 'resourcestring'; Token: kwRESOURCESTRING), 
        (Keyword: 'safecall'; Token: kwSAFECALL), 
        (Keyword: 'set'; Token: kwSET), 
        (Keyword: 'shl'; Token: kwSHL), 
        (Keyword: 'shr'; Token: kwSHR), 
        (Keyword: 'stdcall'; Token: kwSTDCALL), 
        (Keyword: 'stored'; Token: kwSTORED), 
        (Keyword: 'string'; Token: kwSTRING), 
        (Keyword: 'then'; Token: kwTHEN), 
        (Keyword: 'threadvar'; Token: kwTHREADVAR), 
        (Keyword: 'to'; Token: kwTO), 
        (Keyword: 'try'; Token: kwTRY), 
        (Keyword: 'type'; Token: kwTYPE), 
        (Keyword: 'unit'; Token: kwUNIT), 
        (Keyword: 'until'; Token: kwUNTIL), 
        (Keyword: 'uses'; Token: kwUSES), 
        (Keyword: 'var'; Token: kwVAR), 
        (Keyword: 'virtual'; Token: kwVIRTUAL), 
        (Keyword: 'while'; Token: kwWHILE), 
        (Keyword: 'with'; Token: kwWITH), 
        (Keyword: 'write'; Token: kwWRITE), 
        (Keyword: 'writeonly'; Token: kwWRITEONLY)
);

function GetKeywordToken(const S: String): Integer;

// checks whether the current token is a keyword,
// if so then the current token type is set accordingly

var L, R, K : Integer;
    Res     : Integer;

begin
  Result := 0;

  // binary search
  L := 0;
  R := KeywordCount - 1;
  while L <= R do
  begin
    K := L + (R - L) div 2;
    Res := CompareText(S, Keywords[K].Keyword);
    if Res = 0 then
    begin
      Result := Keywords[K].Token;
      Break;
    end
    else
      if Res > 0 then L := K + 1
                 else R := K - 1;
  end;
end;
 
//--------------------------------------------------------------------------------
 
constructor TDelphiLexer.Create;
 
begin
  // allocate positions array (length is highest possible mark position)
  SetLength(FPositions, MarksHigh[StateCount - 1]);
end;
 
//-----------------------------------------------------------------------------
 
destructor TDelphiLexer.Destroy;
 
begin
  FPositions := nil;
  FMatchStack := nil;
  inherited;
end;
 
//-----------------------------------------------------------------------------
 
function TDelphiLexer.IsKeyword: Boolean;
 
// checks whether the current token is a keyword,
// if so then the current token type is set accordingly
 
var Word: String;
 
begin
  SetString(Word, FInput + FTokenPos, FCurrentPos - FTokenPos);
  FCurrentToken := GetKeywordToken(Word);
  if FCurrentToken > 0 then Result :=True
                       else Result := False;
end;
 
//-----------------------------------------------------------------------------
 
function TDelphiLexer.GetEol: Boolean;
 
begin
  Result := FInput[FCurrentPos] = #0;
end;
 
//-----------------------------------------------------------------------------
 
function TDelphiLexer.GetToken: String;
 
var Len: Cardinal;
 
begin
  Len := FCurrentPos - FTokenPos;
  SetString(Result, FInput + FTokenPos, Len);
end;
 
//-----------------------------------------------------------------------------
 
procedure TDelphiLexer.Reset;
 
begin
  FTokenPos := 0;
  FCurrentPos := 0;
  FCurrentToken := 0;
  FStartState := 0;
  FLineState := 1;
  FLastChar := #0;
  FMatchCount := 0;
end;
 
//-----------------------------------------------------------------------------
 
procedure TDelphiLexer.SetToken(Token: Integer);
 
begin
  FCurrentToken := Token;
  FDoStop := True;
end;
 
//-----------------------------------------------------------------------------
 
procedure TDelphiLexer.RuleToToken(Rule: Integer);
 
var S : String;
 
begin
  SetString(S, FInput + FTokenPos, FCurrentPos - FTokenPos);
  // actions
  case Rule of
    // <Normal>'
         0 : 
          begin
            s := '';
            Start(String2);
          end;
    // <Normal>"#"
         1 : 
          begin
            s := '';
            Start (String1);
            Unget_Char (yyText [1]);
          end;
    // <String1>'
         2 : 
          begin
            Start (String2);
          end;
    // <String1>#integer
         3 : 
          begin   { #123 oder #$12 im String enthalten }
            Delete (yyText, 1, 1);
            Val (yyText, i, Result);
            if Result = 0 then
              s := s + Char (i)
            else begin
              SetToken (UNSIGNED_INTEGER);
              RaiseScannerError (IDS_ErrorInIntegerConst, ['.']);
            end;
          end;
    // <String1>.
         4 : 
          begin
            Start (Normal);
            SetToken (STRING_CONST);
            UnGet_Char (yytext [1]);
            Input2Output (''''+s+'''');
            {RaiseScannerError (IDS_ErrorInIntegerConst, ['.']);}
          end;
    // <String2>"''"
         5 : 
          s := s+'''';
    // <String2>'
         6 : 
          Start (String1);
    // <asmState>'
         7 : 
          begin
            Start (asmString1);
            s := '';
          end;
    // <asmString1>"''"
         8 : 
          s := s+'''';
    // <asmString1>'
         9 : 
          begin
            Start (asmState);
            SetToken (STRING_CONST);
            Input2Output (''''+s+'''');
          end;
    // <asmState>\"
        10 : 
          begin
            Start (asmString2);
            s := '';
          end;
    // <asmString2>""""
        11 : 
          s := s+'"';
    // <asmString2>\"
        12 : 
          begin
            Start (asmState);
            SetToken (STRING_CONST);
            Input2Output ('"'+s+'"');
          end;
    // <String2,asmString1,asmString2>\n
        13 : 
          begin
            RaiseScannerError (IDS_StringExceeedsLine, ['.']);  Halt;
          end;
    // <String2,asmString1,asmString2>.
        14 : 
          s := s + yyText;
    // <Normal>"(*"|"{"
        15 : 
          begin {State: Normal}
            EvalCommentStart(Normal);
          end;
    // <asmState>"(*"|"{"
        16 : 
          begin
            EvalCommentStart(asmState);
          end;
    // <asmState>identifier
        17 : 
          begin
            Input2Output (yyText);
            {if isAssemblerKeyword (yyText, KW) then
              SetToken (KW)
            else if isAssemblerRegister (yyText, KW) then
              SetToken (KW)
            else }
              SetToken (IDENTIFIER);
            if Upper (yyText) = 'end' then begin
              Return2OldState; {Could only be Normal}
              SetToken (_END_);
            end;
          end;
    // <asmState>asmLabel
        18 : 
          begin
            SetToken (ASM_LABEL);
            Input2Output (yytext);
          end;
    // <asmState>asmBinary
        19 : 
          begin
            Input2Output (yyText);
            Dec (Byte (yyText [0]));
            MakeInt (yyText, BinBase);
          end;
    // <asmState>asmOctal
        20 : 
          begin
            Input2Output (yyText);
            Dec (Byte (yyText [0]));
            MakeInt (yyText, OctBase);
          end;
    // <asmState>asmHex
        21 : 
          begin
            Input2Output (yyText);
            Dec (Byte (yyText [0]));
            MakeInt (yyText, HexBase);
          end;
    // <asmState>\n
        22 : 
          begin
            Input2Output( NEWLINE );
            SetToken (ASM_SEPERATOR);
          end;
    // <asmState>";"
        23 : 
          begin
            Input2Output (';');
            SetToken (ASM_SEPERATOR);
          end;
    // <asmState>asmWhite
        24 : 
          Input2Output (yyText);
    // <asmState>integer
        25 : 
          begin
            Val (yyText, yylVal.yyLongint, Result);
            Input2Output (yyText);
            if Result=0 then
              SetToken (UNSIGNED_INTEGER)
            else
              SetToken (ILLEGAL);
          end;
    // <asmState>.
        26 : 
          begin
            RaiseScannerError (IDS_IllegalChar, [yyText, '.']
                               {'Illegal character: ', yyText});
            SetToken (ILLEGAL);
          end;
    // <Normal>realinteger/[^\.]
        27 : 
          begin {State: Normal}
            if (yyactchar = '.') then
              {hngt jetzt bei [1..2]; Verwechslung mit 1.2e4 abfangen}
              if (yylastchar = '.') then begin
                {hngt jetzt bei [1..2]; Verwechslung mit 1.2e4 abfangen}
                Val (copy(yyText,1,length(yytext)-1), yylVal.yyLongint, Result);
                Input2Output (copy(yyText,1,length(yytext)-1));
                if Result=0 then begin
                  SetToken (UNSIGNED_INTEGER);
                  unget_char( '.' );
                  exit;
                end else
                  SetToken (ILLEGAL);
              end else {z.B. '1.3e1..' oder '1.3..' }
                SetToken (ILLEGAL);
            {Normaler Real}
            Val (yyText, yylVal.yyExtended, Result);
            Input2Output (yyText);
            if Result=0 then begin
              SetToken (UNSIGNED_REAL);
            end else
              SetToken (ILLEGAL);
          end;
    // <Normal>integer
        28 : 
          begin {State: Normal}
            Val (yyText, yylVal.yyLongint, Result);
            Input2Output (yyText);
            if Result=0 then
              SetToken (UNSIGNED_INTEGER)
            else
              SetToken (ILLEGAL);
          end;
    // <Normal>identifier
        29 : 
          begin {State: Normal}
            Input2Output (yyText);
            if isKeyword(yyText, KW) then begin
              SetToken (KW);
              if KW = _ASM_ then
                SwitchState (Normal,asmState);
            end else if  isStdInstruction(yyText, KW) then
              if WouldAcceptChar(KW) then
                SetToken(KW)
              else
                SetToken (IDENTIFIER)
            else
              SetToken (IDENTIFIER);
          end;
    // <Normal>".."
        30 : 
          begin
            Input2Output (yyText);
            SetToken (DOTDOT);
          end;
    // <Normal>":="
        31 : 
          begin
            Input2Output (yyText);
            SetToken (_ASSIGNMENT_);
          end;
    // <Normal>"<="
        32 : 
          begin
            Input2Output (yyText);
            SetToken (LE);
          end;
    // <Normal>"<>"
        33 : 
          begin
            Input2Output (yyText);
            SetToken (NOTEQUAL);
          end;
    // <Normal>">="
        34 : 
          begin
            Input2Output (yyText);
            SetToken (GE);
          end;
    // <Normal>"(."
        35 : 
          begin
            Input2Output (yyText);
            yyText := '[';
            SetToken (LBRAC);
          end;
    // <Normal>".)"
        36 : 
          begin {State: Normal}
            Input2Output (yyText);
            yyText := ']';
            SetToken (RBRAC);
          end;
    // <Normal>eof
        37 : 
          begin {State: Normal}
            Input2Output (yyText);
            SetToken (0);
          end;
    // <Normal>white
        38 : 
          begin {State: Normal}
            if yyText = #10 then
              Input2Output ( NEWLINE )
            else
              Input2Output (yyText);
          end;
    // <asmState,Normal>"+"
        39 : 
          SetToken(PLUS);
    // <asmState,Normal>"-"
        40 : 
          SetToken(MINUS);
    // <asmState,Normal>"*"
        41 : 
          SetToken(STAR);
    // <asmState,Normal>"/"
        42 : 
          SetToken(SLASH);
    // <asmState,Normal>"="
        43 : 
          SetToken(EQUAL);
    // <asmState,Normal>"<"
        44 : 
          SetToken(LT);
    // <asmState,Normal>">"
        45 : 
          SetToken(GT);
    // <asmState,Normal>","
        46 : 
          SetToken(COMMA);
    // <asmState,Normal>"("
        47 : 
          SetToken(LPAREN);
    // <asmState,Normal>")"
        48 : 
          SetToken(RPAREN);
    // <asmState,Normal>":"
        49 : 
          SetToken(COLON);
    // <asmState,Normal>";"
        50 : 
          SetToken(SEMICOLON);
    // <asmState,Normal>"@"
        51 : 
          SetToken(APE_SYMBOL);
    // <asmState,Normal>"^"
        52 : 
          SetToken(UPARROW);
    // <asmState,Normal>"["
        53 : 
          SetToken(LBRAC);
    // <asmState,Normal>"]"
        54 : 
          SetToken(RBRAC);
    // <asmState,Normal>"."
        55 : 
          SetToken(DOT);
    // <asmState>"&"
        56 : 
          SetToken(ET);
    // <Normal>.
        57 : 
          begin {State: Normal}
            RaiseScannerError (IDS_IllegalChar, [yyText, '.']);
            SetToken (ILLEGAL);
          end;
    // <OverreadState>"(*"|"{"
        58 : 
          begin
            EvalCommentStart(OverreadState);
          end;
    // <OverreadState>\0
        59 : 
          begin
            Return2OldState; {Could be Normal or AsmState}
            RaiseScannerError( IDS_CS_expected, ['.'] );
          end;
    // <OverreadState>'
        60 : 
          begin
            s := '';
            if FormerState = Normal then
              SwitchState (OverreadState, String2)
            else if FormerState = asmState then
              SwitchState (OverreadState,asmString1)
            else
              RaiseScannerError( IDS_InternalScannerError, ['.'] );
          end;
    // <OverreadState>\"
        61 : 
          begin
            if FormerState = Normal then
            else if FormerState = asmState then begin
              SwitchState (OverreadState,asmString2); s := '';
            end else
              RaiseScannerError( IDS_InternalScannerError, ['.'] );
          end;
    // <OverreadState>.
        62 : 
          begin
            Input2Output (yyText);
          end;
    // <Comment>"*)"|"}"
        63 : 
          begin
            if ((yyText = '}') AND (cs = '{')) OR
               ((yyText = '*)') AND (cs = '(*')) then begin
              Input2Output (yyText);
              Return2OldState; {Could be Normal or AsmState}
            end;
          end;
    // <Comment>\0
        64 : 
          begin
            Return2OldState; {Could be Normal or AsmState or OverreadState}
            CommentEof;
          end;
    // <Comment>.
        65 : 
          begin
            Input2Output (yyText);
          end;
    // <CompilerSwitch>\0
        66 : 
          begin
            Return2OldState; {Could be Normal or AsmState or OverreadState}
            RaiseScannerError( IDS_CS_expected, ['.'] );
          end;
    // <CompilerSwitch>"*)"|"}"
        67 : 
          begin
            if ((yyText = '}') AND (cs = '{')) OR
               ((yyText = '*)') AND (cs = '(*')) then begin
              Input2Output (yyText);
              {EvalCompilerSwitch( CompSwitch );}
              Return2OldState; {Could be Normal or AsmState or OverreadState}
            end;
          end;
    // <CompilerSwitch>.
        68 : 
          begin
            Input2Output (yyText);
            CompSwitch := CompSwitch + yyText ;
          end;
    // <CharTable>.
        69 : 
          begin
            Input2Output (yyText);
            s := yyText;
            SetToken (IMMCHAR_ID);
            Return2OldState;
          end;
  end;
end;
 
//-----------------------------------------------------------------------------
 
function TDelphiLexer.MatchRule(var Rule : Integer ) : Boolean;
 
// finds the last match and the corresponding marked position and adjusts
// the matched string accordingly;
// returns:
// - True if a rule has been matched, False otherwise
// - Rule: the number of the matched rule
 
begin
  FRejected := False;
  while (FMatchCount > 0) and (FPositions[FMatchStack[FMatchCount - 1]] = 0) do Dec(FMatchCount);
 
  if FMatchCount > 0 then
  begin
    Rule := FMatchStack[FMatchCount - 1];
    FCurrentPos := FTokenPos + FPositions[Rule];
    FPositions[Rule] := 0;
    if (FCurrentPos - FTokenPos) > 0 then FLastChar := FInput[FCurrentPos]
                                     else FLastChar := #0;
    Result := True;
  end
  else
  begin
    // return position to starting point
    FCurrentPos := FTokenPos;
    FLastChar := #0;
    Result := False;
  end
end;
 
//-----------------------------------------------------------------------------
 
function TDelphiLexer.DoDefault : Boolean;
 
begin
  FRejected := False;
  if NextChar <> #0 then Result := True
                    else
  begin
    FLineState := 1;
    Result := False;
  end;
  FLastChar := FInput[FCurrentPos];
end;
 
//-----------------------------------------------------------------------------
 
procedure TDelphiLexer.Next;
 
// scans for the next token and determines its kind (FCurrentToken and FTokenPos
// are set appropriately)
 
var I : Integer;
    Len : Cardinal;
    Matched : Boolean;
 
begin
  FDoStop := False;
 
  repeat
    // initialize run
    FTokenPos := FCurrentPos;
 
    if FLastChar <> #0 then
      if FLastChar = LF then FLineState := 1
                        else FLineState := 0;
    FCurrentState := FStartState + FLineState;
 
    repeat // character scan loop
      // mark positions and matches
      Len := FCurrentPos - FTokenPos;
      for I := MarksLow[FCurrentState] to MarksHigh[FCurrentState] do FPositions[MarkPositionTable[I]] := Len;
      for I := MatchesHigh[FCurrentState] downto MatchesLow[FCurrentState] do SetMatch(MatchTable[I]);
 
      if TransitionsLow[FCurrentState] > TransitionsHigh[FCurrentState] then Break;  // dead state, do action
 
      // determine action
      I := TransitionsLow[FCurrentState];
      while (I <= TransitionsHigh[FCurrentState]) and not (FInput[FCurrentPos] in TransitionTable[I].CharClass) do Inc(I);
 
      // get next character
      NextChar;
 
      if I > TransitionsHigh[FCurrentState] then Break; // no transition on current char in this state
 
      // switch to new state
      FCurrentState := TransitionTable[I].NextState;
 
    until False;
 
    repeat
      Matched := MatchRule(I);
      if Matched then RuleToToken(I);
      // FRejected can be set in RuleToToken by calling Reject
    until Matched and not FRejected;
 
    if not (Matched or DoDefault) then
    begin
      FStartState := 0;
      FLineState := 1;
      FLastChar := #0;
    end;
  until FDoStop;
end;
 
//-----------------------------------------------------------------------------
 
procedure TDelphiLexer.SetInput(const Value: String);
 
begin
  FInput := PChar(Value);
  Reset;
end;
 
//-----------------------------------------------------------------------------
 
function TDelphiLexer.GetNextChar: Char;
 
begin
  if FInput[FCurrentPos] <> #0 then
  begin
    Inc(FCurrentPos);
    FLookahead := FInput[FCurrentPos + 1];
  end
  else FLookahead := #0;
  Result := FInput[FCurrentPos];
end;
 
//-----------------------------------------------------------------------------
 
procedure TDelphiLexer.SetState(Value: Cardinal);
 
begin
  FStartState := Value;
end;
 
//-----------------------------------------------------------------------------
 
procedure TDelphiLexer.SetMatch(I: Integer);
 
begin
  if FMatchCount >= Length(FMatchStack) then SetLength(FMatchStack, FMatchCount + 1);
  FMatchStack[FMatchCount] := I;
  Inc(FMatchCount);
end;
 
//-----------------------------------------------------------------------------
 
procedure TDelphiLexer.Reject;
 
begin
  FRejected := True;
end;
 
//-----------------------------------------------------------------------------
 
function TDelphiLexer.GetTokenInfo(var S: PChar; var Pos: Integer): Integer;
 
// returns token string, token position relative to entire input string, and token length in one call
 
begin
  S := FInput + FTokenPos;
  Pos := FTokenPos;
  Result := FCurrentPos - FTokenPos;
end;
 
//-----------------------------------------------------------------------------
 
end.
 
