unit SyntaxEditor;

// Version: 1.5
//----------------------------------------------------------------------------------------------------------------------
// This unit contains an Unicode editor control with syntax highlighting capabilities.
// The control needs Delphi 4 or up to compile.
//
//----------------------------------------------------------------------------------------------------------------------
// Open issues:
//   - Because the the control depends on Unicode.pas, which is still beta, also the syntax editor must be considered
//     as being beta. In particular not all search (and replace) functionality is fully implemented and tested.
//   - Although the edit works with wide strings, the highlighters do not yet, because new tools must be created
//     to build new highlighter classes which fully support wide strings (see also my homepage for DCG, the
//     Delphi Compiler Generator, which will be extended to create Unicode highlighters). The impact on the current
//     implementation can be seen where the edit needs to act on the text like looking for word boundaries etc.
//
// This unit is freeware for non-commercial software.
//
// 25-JUN-2000 ml:
//   bookmarks improvments
// 24-JUN-2000 ml:
//   - IME support for Windows NT does now work
//   - gutter and right margin resize with mouse
//   - OnCutText, LoadFromStream for the text checks now for too long lines
// 12-JUN-2000 ml:
//   Unicode.pas has been renamed to JclUnicode.pas and is now part of the JEDI VCL. This required, though, a few
//   changes to some names here.
// 13-MAY-2000 ml:
//   bug fixes (triple clicks, select all)
// 30-APR-2000 ml:
//   bug fixes
// JAN-MAR-2000 ml: Version 1.3
//   - search and replace, bug fixes
// NOV-DEC-99 ml: Version 1.2
//   - switched entirely to Unicode (some code already was using Unicode, e.g. output)
//   - basic IME handling
//   - mouse wheel support, undo/redo, bookmarks, painting optimization, scroll hint window, bug fixes
// JUN-JUL-99 ml: Version 1.1
//   - line numbers, structural rework, clean up, bug fixes
// 10-MAR-99 ml: Version 1.0
//   First implementation from a freeware lib (mwEdit), significant changes due to speed and needed functionality
//   (removed lots of redundancies in string handling, total rework of output, undo / redo, tab support, control char
//   output, first Unicode support etc.)
//
// (c) 1999, 2000 Dipl. Ing. Mike Lischke (public@lischke-online.de)
// Homepage: www.lischke-online.de

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Graphics, Math,
  ExtCtrls, Forms, StdCtrls, Clipbrd, EditorKeyCommands, Highlighter, JclUnicode;

const
  // these self defined cursors are added, the VCL does not a good job on cursors as there's no
  // way to determine which constants already have been occupied by other code
  crDragMove = 2910;
  crDragCopy = 3110;

type
  // eoSmartTabs and eoUseTabs are mutual exclusive (avoid activating both, as this confuses the input),
  // to use any of the tab options you must also have eoWantTabs enabled, otherwise the TAB char
  // will never even reach the edit.
  TSyntaxEditOption = (
    eoAllowGutterResize,       // allows to resize the gutter with the mouse
    eoAllowMarginResize,       // allows to resize (reposition) the right margin with the mouse 
    eoAutoIndent,              // do auto indentation when inserting a new line
    eoAutoUnindent,            // do auto undindentation when deleting a character which
                               // is only preceded by whitespace characters
    eoCursorThroughTabs,       // move cursor as if tabulators were (TabSize) spaces
    eoGroupUndo,               // while executing undo/redo handle all continous changes of
                               // the same kind in one call instead undoing/redoing each command
                               // separately
    eoHideSelection,           // hide selection when unfocused
    eoInserting,               // denotes whether the control is in insert or in overwrite mode
                               // (this option in superseeded by eoReadOnly)
    eoKeepTrailingBlanks,      // don't remove automatically white spaces at the end of a line
    eoLineNumbers,             // show line numbers in gutter
    eoOptimalFill,             // not yet implemented
    eoReadOnly,                // prevent the content of the control from being changed
    eoReadOnlySelection,       // show selection if in read only mode (useful for simulation static text)
    eoScrollPastEOL,           // allow the cursor to go past the end of the line
    eoShowControlChars,        // show tabulators, spaces and line breaks (only if not eoScrollPastEOL) as glyphs
    eoShowCursorWhileReadOnly, // don't hide the cursor if control is in read only mode
    eoShowScrollHint,          // show a hint window with the current top line while scrolling with the thumb
    eoSmartTabs,               // automatically put in enough spaces when TAB has been pressed to move
                               // the cursor to the next non-white space character of the previous line
                               // (actually only the cursor is moved, the spaces are inserted later)
    eoTripleClicks,            // allow selecting an entire line with a triple click
    eoUndoAfterSave,           // don't clear the undo/redo stack after the control has been saved
    eoUseTabs,                 // don't covert TABs to spaces but use TAB with fixed distances (see TabSize)
    eoUseSyntaxHighlighting,   // switch on syntax highlighting (ignored if no highlighter is assigned)
    eoWantTabs);               // use TABs for input rather than moving the focus to the next control
  TSyntaxEditOptions = set of TSyntaxEditOption;

const                              
  DefaultTabGlyph = WideChar($2192);
  DefaultLineBreakGlyph = WideChar('');
  DefaultSpaceGlyph = WideChar($2219); // Note: $B7 should not be used here because it might be interpreted
                                       //       as valid identifier character (e.g. in Catalan)

  DefaultOptions = [eoAutoIndent, eoAutoUnindent, eoCursorThroughTabs, eoGroupUndo, eoHideSelection,
                    eoInserting, eoScrollPastEOL, eoSmartTabs, eoTripleClicks, eoUseSyntaxHighlighting, eoWantTabs];

type
  TReplaceAction = (raCancel, raSkip, raReplace, raReplaceAll);

  TCaretType = (ctVerticalLine, ctHorizontalLine, ctHalfBlock, ctBlock);
  
  // undo / redo
  TChangeReason = (
    ecrNone,
    ecrInsert,
    ecrDelete,
    ecrReplace,
    ecrDragMove,
    ecrDragCopy,
    ecrCursorMove,
    ecrOverwrite,
    ecrIndentation
  );

  PChange = ^TChange;
  TChange = record
    Reason: TChangeReason;
    OldText: WideString;
    OldCaret,
    OldSelStart,
    OldSelEnd: TPoint;
    NewText: WideString;
    NewCaret,
    NewSelStart,
    NewSelEnd: TPoint;
  end;

  // used to record the current edit state for registering an undo/redo action
  TEditState = record
    Text: WideString;
    Caret,
    SelStart,
    SelEnd: TPoint;
  end;

  TCustomSyntaxEdit = class;

  TBookMarkEvent = procedure(Sender: TCustomSyntaxEdit; Index, X, Y: Integer; var Allowed: Boolean) of object;

  // undo and redo in one list
  TUndoList = class
  private
    FList: TList;
    FCurrent: Integer;
    FMaxUndo: Integer;
    FOwner: TCustomSyntaxEdit;
    function GetCanRedo: Boolean;
    function GetCanUndo: Boolean;
    procedure SetMaxUndo(Value: Integer);
  protected
    procedure RemoveChange(Index: Integer);
  public
    constructor Create(AOwner: TCustomSyntaxEdit);
    destructor Destroy; override;

    procedure AddChange(AReason: TChangeReason; const AOldText: WideString; AOldCaret, AOldSelStart, AOldSelEnd: PPoint;
      const ANewText: WideString; ANewCaret, ANewSelStart, ANewSelEnd: PPoint); overload;
    procedure AddChange(AReason: TChangeReason; const AOldText: WideString; AOldCaret, AOldSelStart, AOldSelEnd: PPoint); overload;
    function GetUndoChange(var Change: PChange): TChangeReason;
    function GetRedoChange(var Change: PChange): TChangeReason;
    function GetCurrentRedoReason: TChangeReason;
    function GetCurrentUndoReason: TChangeReason;
    procedure ClearList;

    property CanRedo: Boolean read GetCanRedo;
    property CanUndo: Boolean read GetCanUndo;
    property MaxUndo: Integer read FMaxUndo write SetMaxUndo;
  end;

  TSyntaxEditorStrings = class(TWideStringList)
  private
    FOwner: TCustomSyntaxEdit;
    FModified: Boolean;
    FOnLoaded: TNotifyEvent;
    FAutoResetModified: Boolean;
  protected
    procedure Put(Index: Integer; const S: WideString); override;
  public
    constructor Create(AOwner: TCustomSyntaxEdit); virtual;

    function Add(const S: WideString): Integer; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: WideString); override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
    procedure SetText(const Value: WideString); override;

    property AutoResetModified: Boolean read FAutoResetModified write FAutoResetModified default True;
    property Modified: Boolean read FModified write FModified;
    property OnLoaded: TNotifyEvent read FOnLoaded write FOnLoaded;
  end;

  TColors = class(TPersistent)
  private
    FBackground: TColor;
    FForeground: TColor;
    FOwner: TCustomSyntaxEdit;
    procedure SetBackground(Value: TColor);
    procedure SetForeground(Value: TColor);
  public
    constructor Create(AOwner: TCustomSyntaxEdit);
  published
    property Background: TColor read FBackground write SetBackground default clHighLight;
    property Foreground: TColor read FForeground write SetForeground default clHighLightText;
  end;

  TBookmark = record
    Visible: Boolean;
    X, Y: Integer;
  end;

  TBookmarkOptions = class(TPersistent)
  private
    FEnableKeys: Boolean;
    FGlyphsVisible: Boolean;
    FLeftMargin: Integer;
    FOwner: TCustomSyntaxEdit;
    procedure SetGlyphsVisible(Value: Boolean);
    procedure SetLeftMargin(Value: Integer);
  public
    constructor Create(AOwner: TCustomSyntaxEdit); 
  published
    property EnableKeys: Boolean read FEnableKeys write FEnableKeys default True;
    property GlyphsVisible: Boolean read FGlyphsVisible write SetGlyphsVisible default True;
    property LeftMargin: Integer read FLeftMargin write SetLeftMargin default 2;
  end;

  TScrollHintWindow = class(THintWindow)
  protected
    procedure Paint; override;
  end;

  TDistanceArray = array of Cardinal;

  TScrollBarMode = (
    sbmRegular,
    sbmFlat,
    sbm3D
  );

  TSearchOption = (
    soBackwards,             // search backwards instead of forward
    soEntireScope,           // search in entire text
    soIgnoreNonSpacing,      // ignore non-spacing characters in search
    soMatchCase,             // case sensitive search
    soPrompt,                // ask user for each replace action
    soRegularExpression,     // search using regular expressions
    soReplace,               // do a replace
    soReplaceAll,            // replace all occurences
    soSelectedOnly,          // search in selected text only
    soSpaceCompress,         // handle several consecutive white spaces as one white space
                             // so "ab   cd" will match "ab cd" and "ab        cd"
    soWholeWord);            // match whole words only
  TSearchOptions = set of TSearchOption;

  // states the syntax edit can enter due to various events triggered by the user
  TSyntaxEditStates = set of (
    esGutterResizing,        // gutter resizing is in progress
    esRightMarginResizing    // right margin resizing is in progress
  );
  
  // event types
  TPaintEvent = procedure(Sender: TCustomSyntaxEdit; ACanvas: TCanvas) of object;
  TCaretEvent = procedure(Sender: TCustomSyntaxEdit; X, Y: Cardinal) of object;
  TReplaceTextEvent = procedure(Sender: TCustomSyntaxEdit; const Search, Replace: WideString; Line, Start, Stop: Integer;
    var Action: TReplaceAction) of object;
  TProcessCommandEvent = procedure(Sender: TCustomSyntaxEdit; var Command: TEditorCommand; var AChar: WideChar; Data: Pointer) of object;
  TCutTextEvent = procedure(Sender: TCustomSyntaxEdit; var Text: WideString; var DoCut: Boolean) of object;

  TCustomSyntaxEdit = class(TCustomControl)
  private
    FBlockBegin: TPoint;
    FBlockEnd: TPoint;
    FCaretX: Integer;
    FCaretY: Integer;
    FCaretVisible,
    FDragWasStarted: Boolean;
    FCharsInWindow: Integer;
    FCharWidth: Integer;
    FMultiClicked: Boolean;
    FScrollTimer: TTimer;
    FFontDummy,
    FLineNumberFont: TFont;
    FGutterColor: TColor;
    FGutterRect: TRect;
    FGutterWidth: Integer;
    FLines: TWideStrings;
    FLinesInWindow: Integer;
    FOffsetX,
    FOffsetY: Integer;
    FMaxRightPos: Integer;
    FUpdateCount: Integer;
    FRightMargin: Integer;
    FScrollBars: TScrollStyle;
    FTextHeight: Integer;
    FHighLighter: TSyntaxHighlighter;
    FMarkColor,
    FSelectedColor,
    FScrollHintColor: TColors;
    FUndoList : TUndoList;
    FBookMarks: array[0..9] of TBookMark;
    FLastCaret: TPoint;
    FBookmarkOptions: TBookmarkOptions;
    FOnPaint: TPaintEvent;
    FBorderStyle: TBorderStyle;
    FOverwriteCaret: TCaretType;
    FInsertCaret: TCaretType;
    FCaretOffset: TPoint;
    FKeyStrokes: TKeyStrokes;
    FModified: Boolean;
    FCharWidthArray: TDistanceArray;
    FInternalBMList: TImageList;
    FExtraLineSpacing: Integer;
    FOptions: TSyntaxEditOptions;
    FTabSize: Integer;
    FMarginColor: TColor;
    FScrollFactor: Single;
    FScrollBarMode: TScrollBarMode;
    FLastValidLine: Integer;
    FMarkLine: Integer;
    FLastCalcIndex: Integer;
    FIndentSize: Cardinal;
    FKeyHandled,
    FDropTarget: Boolean;              // needed to know in the scroll timer event whether to change
                                       // the selection block end or just the cursor position
    FDoubleClickTime: UINT;
    FLastDblClick: UINT;
    FScrollHint: TScrollHintWindow;
    FStates: TSyntaxEditStates;
    FInfo: WideString;                 // this is a placeholder for application defined data
                                       // like a file name or something similar

    FOnCaretChange: TCaretEvent;
    FOnSettingChange: TNotifyEvent;
    FOnProcessCommand: TProcessCommandEvent;
    FOnProcessUserCommand: TProcessCommandEvent;
    FOnBookmark: TBookmarkEvent;
    FOnChange: TNotifyEvent;
    FOnReplaceText: TReplaceTextEvent;
    FOnCutText: TCutTextEvent;
    function CaretXPix: Integer;
    function CaretYPix: Integer;
    procedure ComputeCaret(X, Y: Integer);
    function GetBlockBegin: TPoint;
    function GetBlockEnd: TPoint;
    function GetFont: TFont;
    function GetLineCount: Integer;
    function GetLineText: WideString;
    function GetSelAvail: Boolean;
    function GetText: WideString;
    function GetTopLine: Integer;
    procedure FontChanged(Sender: TObject);
    procedure LineNumberFontChanged(Sender: TObject);
    procedure LinesChanged(Sender: TObject);
    procedure SetBlockBegin(Value: TPoint);
    procedure SetBlockEnd(Value: TPoint);
    procedure SetWordBlock(Value: TPoint);
    procedure SetCaretX(Value: Integer);
    procedure SetCaretY(Value: Integer);
    procedure SetGutterWidth(Value: Integer);
    procedure SetFont(const Value: TFont);
    procedure SetOffsetX(Value: Integer);
    procedure SetOffsetY(Value: Integer);
    procedure SetLines(Value: TWideStrings);
    procedure SetLineText(Value: WideString);
    procedure SetScrollBars(const Value: TScrollStyle);
    procedure SetText(const Value: WideString);
    procedure SetTopLine(Value: Integer);
    procedure OnScrollTimer(Sender: TObject);
    procedure UpdateCaret;
    procedure UpdateScrollBars;
    function GetSelText: WideString;
    procedure SetSelText(const Value: WideString);
    procedure ScanFrom(Index: Integer);
    procedure PaintHighlighted(TextCanvas: TCanvas);
    function GetCanUndo: Boolean;
    function GetCanRedo: Boolean;
    procedure SetGutterColor(Value: TColor);
    procedure SetRightMargin(Value: Integer);
    procedure SetMaxUndo(const Value: Integer);
    function GetMaxUndo: Integer;
    procedure SetHighlighter(const Value: TSyntaxHighlighter);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetInsertCaret(const Value: TCaretType);
    procedure SetOverwriteCaret(const Value: TCaretType);
    function GetCaretXY: TPoint;
    procedure SetCaretXY(const Value: TPoint);
    procedure SetKeystrokes(const Value: TKeyStrokes);
    procedure SetMaxRightChar(const Value: Integer);
    procedure SetCharWidth(const Value: Integer);
    procedure SetExtraLineSpacing(const Value: Integer);
    procedure SetSelTextExternal(const Value: WideString);
    procedure SetOptions(const Value: TSyntaxEditOptions);
    procedure SetTabSize(Value: Integer);
    function GetMaxRightChar: Integer;
    procedure SetMarginColor(const Value: TColor);
    function NextTabPos(Index: Integer): Integer;
    function Unindent(X, Y: Cardinal): Cardinal;
    procedure SetScrollBarMode(const Value: TScrollBarMode);
    procedure CalcCharWidths(Index: Integer); overload;
    procedure CalcCharWidths(const S: WideString; TabSize, CharWidth: Cardinal); overload;
    function GetModified: Boolean;
    procedure SetModified(const Value: Boolean);
    procedure SetMarkLine(Value: Integer);
    procedure SetMarkColor(const Value: TColors);
    function NextCharPos(ThisPosition: Cardinal; ForceNonDefault: Boolean = False): Cardinal; overload;
    function NextCharPos(ThisPosition, Line: Cardinal; ForceNonDefault: Boolean = False): Cardinal; overload;
    function PreviousCharPos(ThisPosition: Cardinal): Cardinal; overload;
    function PreviousCharPos(ThisPosition, Line: Cardinal): Cardinal; overload;
    function GetLineEnd: Cardinal; overload;
    function GetLineEnd(Index: Cardinal): Cardinal; overload;
    function ColumnToCharIndex(X: Cardinal): Cardinal; overload;
    function ColumnToCharIndex(Pos: TPoint): Cardinal; overload;
    function CharIndexToColumn(X: Cardinal): Cardinal; overload;
    function CharIndexToColumn(X: Cardinal; const S: WideString; TabSize, CharWidth: Cardinal): Cardinal; overload;
    function CharIndexToColumn(Pos: TPoint): Cardinal; overload;
    procedure SetIndentSize(Value: Cardinal);
    procedure ReplaceLeftTabs(var S: WideString);
    function LeftSpaces(const S: WideString): Cardinal;
    function RecordState(IncludeText: Boolean): TEditState;
    function PosInSelection(Pos: TPoint): Boolean;
    procedure TripleClick(X, Y: Integer);
    procedure SetUpdateState(Updating: Boolean);
    function CopyOnDrop(Source: TObject): Boolean;
    procedure SetLineNumberFont(const Value: TFont);
    procedure SetScrollHintColor(const Value: TColors);
    procedure SetSelectedColor(const Value: TColors);

    procedure CMMouseWheel(var Message: TCMMouseWheel); message CM_MOUSEWHEEL;
    procedure CMSysColorChange(var Message: TMessage); message CM_SYSCOLORCHANGE;
    procedure WMEraseBkgnd(var Message: TMessage); message WM_ERASEBKGND;
    procedure WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMHScroll(var Message: TWMScroll); message WM_HSCROLL;
    procedure WMImeComposition(var Message: TMessage); message WM_IME_COMPOSITION;
    procedure WMImeNotify(var Message: TMessage); message WM_IME_NOTIFY;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMVScroll(var Message: TWMScroll); message WM_VSCROLL;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    function DoCutText(var Text: WideString): Boolean; virtual;
    procedure DeleteSelection(NeedRescan: Boolean);
    procedure DoChange; virtual;
    procedure DoSettingChanged; virtual;
    function NextSmartTab(X, Y: Integer; SkipNonWhite: Boolean): Integer;
    procedure HideCaret;
    procedure InsertText(const Value: WideString);
    function IsIdentChar(const AChar: WideChar; IdChars: TIdentChars): Boolean;
    procedure ListAdded(Sender: TObject);
    procedure ListDeleted(Index: Integer);
    procedure ListInserted(Index: Integer);
    procedure ListLoaded(Sender: TObject);
    procedure ListPutted(Index: Integer);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure DblClick; override;
    procedure KeyPress(var Key: Char); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure PaintGutter(TextCanvas: TCanvas); virtual;
    procedure PaintText(TextCanvas: TCanvas); virtual;
    procedure Paint; override;
    procedure ShowCaret;
    procedure Loaded; override;
    procedure DoCaretChange; virtual;
    procedure DoStartDrag(var DragObject: TDragObject); override;
    procedure DragCanceled; override;
    procedure DragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean); override;
    procedure InitializeCaret;
    procedure InvalidateLine(Index: Integer);
    procedure InvalidateLines(Start, Stop: Integer);
    procedure InvalidateToBottom(Index: Integer);
    procedure MarkListChange(Sender: TObject);
    procedure ProcessCommand(var Command: TEditorCommand; var AChar: WideChar; Data: Pointer); virtual;
    // If the translations requires Data, memory will be allocated for it via a
    // GetMem call. The client must call FreeMem on Data if it is not nil.
    function TranslateKeyCode(Code: Word; Shift: TShiftState; var Data: Pointer): TEditorCommand;

    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property BookMarkOptions: TBookmarkOptions read FBookmarkOptions write FBookmarkOptions;
    property CharWidth: Integer read FCharWidth write SetCharWidth;
    property ExtraLineSpacing: Integer read FExtraLineSpacing write SetExtraLineSpacing default 0;
    property Font: TFont read GetFont write SetFont;
    property GutterColor: TColor read FGutterColor write SetGutterColor;
    property GutterWidth: Integer read FGutterWidth write SetGutterWidth;
    property HighLighter: TSyntaxHighlighter read FHighLighter write SetHighlighter;
    property IndentSize: Cardinal read FIndentSize write SetIndentSize default 2;
    property InsertCaret: TCaretType read FInsertCaret write SetInsertCaret default ctVerticalLine;
    property Keystrokes: TKeyStrokes read FKeyStrokes write SetKeystrokes;
    property LineNumberFont: TFont read FLineNumberFont write SetLineNumberFont;
    property Lines: TWideStrings read FLines write SetLines;
    property LineCount: Integer read GetLineCount;
    property MarginColor: TColor read FMarginColor write SetMarginColor default clSilver;
    property MarkColor: TColors read FMarkColor write SetMarkColor;
    property MarkLine: Integer read FMarkLine write SetMarkLine default -1;
    property MaxRightChar: Integer read GetMaxRightChar write SetMaxRightChar;
    property MaxUndo: Integer read GetMaxUndo write SetMaxUndo;
    property Options: TSyntaxEditOptions read FOptions write SetOptions default DefaultOptions;
    property OverwriteCaret: TCaretType read FOverwriteCaret write SetOverwriteCaret default ctBlock;
    property UpdateLock: Integer read FUpdateCount;
    property RightMargin: Integer read FRightMargin write SetRightMargin default 80;
    property ScrollBarMode: TScrollBarMode read FScrollBarMode write SetScrollBarMode;
    property ScrollBars: TScrollStyle read FScrollBars write SetScrollBars default ssBoth;
    property ScrollHintColor: TColors read FScrollHintColor write SetScrollHintColor;
    property SelectedColor: TColors read FSelectedColor write SetSelectedColor;
    property TabSize: Integer read FTabSize write SetTabSize default 8;
    property TabStop default True;

    property OnCaretChange: TCaretEvent read FOnCaretChange write FOnCaretChange;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnCutText: TCutTextEvent read FOnCutText write FOnCutText; 
    property OnPaint: TPaintEvent read FOnPaint write FOnPaint;
    property OnBookmark: TBookmarkEvent read FOnBookmark write FOnBookmark;
    property OnProcessCommand: TProcessCommandEvent read FOnProcessCommand write FOnProcessCommand;
    property OnProcessUserCommand: TProcessCommandEvent read FOnProcessUserCommand write FOnProcessUserCommand;
    property OnReplaceText: TReplaceTextEvent read FOnReplaceText write FOnReplaceText;
    property OnSettingChange: TNotifyEvent read FOnSettingChange write FOnSettingChange;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function ActivateKeyboard(const C: WideChar): Boolean;
    function CharPositionToRowColumn(Position: Cardinal): TPoint;
    procedure ClearAll(RegisterUndo, KeepUndoList: Boolean);
    procedure ClearBookMark(BookMark: Integer);
    procedure ClearUndo;
    procedure CommandProcessor(Command: TEditorCommand; AChar: WideChar; Data: Pointer); virtual;
    procedure CopyToClipboard;
    procedure CutToClipboard;
    procedure DragDrop(Source: TObject; X, Y: Integer); override;
    procedure EnsureCursorPosVisible;
    function GetBookMark(BookMark: Integer; var X, Y: Integer): Boolean;
    function GetSelEnd: Integer;
    function GetSelStart: Integer;
    function GetWordAndBounds(Pos: TPoint; var BB, BE: TPoint): WideString;
    procedure GotoBookMark(BookMark: Integer);
    procedure BlockIndent;
    procedure BlockUnindent;
    function IsBookmark(BookMark: Integer): Boolean;
    function LastWordPos: TPoint; virtual;
    function LineFromPos(Pos: TPoint): Integer;
    function NextWordPos: TPoint; virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure PasteFromClipboard;
    function PositionFromPoint(X, Y: Integer): TPoint;
    procedure Redo;
    function RowColumnToCharPosition(P: TPoint): Cardinal;
    function SearchReplace(const SearchText, ReplaceText: WideString; Options: TSearchOptions): Integer;
    procedure SelectAll;
    procedure SetBookMark(BookMark: Integer; X, Y: Integer);
    procedure SetDefaultKeystrokes; virtual;
    procedure SetSelEnd(const Value: Integer);
    procedure SetSelStart(Value: Integer);
    procedure Undo;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure WndProc(var Msg: TMessage); override;
    function WordAtPos(X, Y: Integer): WideString;

    property BlockBegin: TPoint read GetBlockBegin write SetBlockBegin;
    property BlockEnd: TPoint read GetBlockEnd write SetBlockEnd;
    property CanUndo: Boolean read GetCanUndo;
    property CanRedo: Boolean read GetCanRedo;
    property CaretX: Integer read FCaretX write SetCaretX;
    property CaretY: Integer read FCaretY write SetCaretY;
    property CaretXY: TPoint read GetCaretXY write SetCaretXY;
    property CharsInWindow: Integer read FCharsInWindow;
    property Info: WideString read FInfo write FInfo;
    property LineText: WideString read GetLineText write SetLineText;
    property LinesInWindow: Integer read FLinesInWindow;
    property Modified: Boolean read GetModified write SetModified;
    property SelAvail: Boolean read GetSelAvail;
    property SelText: WideString read GetSelText write SetSelTextExternal;
    property States: TSyntaxEditStates read FStates;
    property Text: WideString read GetText write SetText;
    property TopLine: Integer read GetTopLine write SetTopLine;
    property UndoList: TUndoList read FUndoList;
  end;

  TSyntaxEdit = class(TCustomSyntaxEdit)
  published
    property Align;
    property Anchors;
    property Constraints;
    property BevelEdges;
    property BevelInner;
    property BevelOuter;
    property BevelKind;
    property BevelWidth;
    property BorderWidth;
    property BorderStyle;
    property BookMarkOptions;
    property CharWidth;
    property Color;
    property Ctl3D;
    property Enabled;
    property ExtraLineSpacing;
    property Font;
    property GutterColor;
    property GutterWidth;
    property Height;
    property HighLighter;
    property IndentSize;
    property InsertCaret;
    property Keystrokes;
    property LineNumberFont;
    property Lines;
    property LineCount;
    property MarginColor;
    property MarkColor;
    property MarkLine;
    property MaxRightChar;
    property MaxUndo;
    property Name;
    property Options;
    property OverwriteCaret;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RightMargin;
    property ShowHint;
    property ScrollBarMode;
    property ScrollBars;
    property ScrollHintColor;
    property SelectedColor;
    property TabOrder;
    property TabSize;
    property TabStop default True;
    property Tag;
    property Visible;
    property Width;

    property OnBookmark;
    property OnCaretChange;
    property OnChange;
    property OnClick;
    property OnCutText;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnPaint;
    property OnProcessCommand;
    property OnProcessUserCommand;
    property OnReplaceText;
    property OnSettingChange;
    property OnStartDrag;
  end;

//----------------------------------------------------------------------------------------------------------------------

implementation

uses
  CommCtrl, IMM;

{$R SyntaxEdit.res}

//----------------------------------------------------------------------------------------------------------------------

procedure ConvertAndAddImages(IL: TImageList);

// loads the internal bookmark image list

var
  OneImage: TBitmap;
  I: Integer;
  MaskColor: TColor;
  Name: String;

begin
  // Since we want our bookmark images looking in the correct system colors,
  // we have to remap their colors. I really would like to use
  // IL.GetInstRes(HInstance, rtBitmap, Resname, 16, [lrMap3DColors], clWhite) for this task,
  // but could not get it working under NT. It seems that image lists are not very well
  // supported under WinNT 4.
  OneImage := TBitmap.Create;
  try
    // it is assumed that the image height determines also the width of
    // one entry in the image list
    IL.Clear;
    IL.Height := 16;
    IL.Width := 16;
    for I := 0 to 9 do
    begin
      Name := 'MARK_' + IntToStr(I);
      OneImage.Handle := CreateGrayMappedRes(HInstance, PChar(Name));
      MaskColor := OneImage.Canvas.Pixels[0, 0];
      IL.AddMasked(OneImage, MaskColor);
    end;
  finally
    OneImage.Free;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function Min(X, Y: Integer): Integer;

begin
  if X < Y then Result := X
           else Result := Y;
end;

//----------------------------------------------------------------------------------------------------------------------

function MaxPoint(P1, P2: TPoint): TPoint;

begin
  if P1.Y > P2.Y then Result := P1
                 else
    if P1.Y < P2.Y then Result := P2
                   else
      if P1.X > P2.X then Result := P1
                     else Result := P2;
end;

//----------------------------------------------------------------------------------------------------------------------

function MaxPointAsReference(var P1, P2: TPoint): PPoint;

begin
  if P1.Y > P2.Y then Result := @P1
                 else
    if P1.Y < P2.Y then Result := @P2
                   else
      if P1.X > P2.X then Result := @P1
                     else Result := @P2;
end;

//----------------------------------------------------------------------------------------------------------------------

function MinPoint(P1, P2: TPoint): TPoint;

begin
  if P1.Y < P2.Y then Result := P1
                 else
    if P1.Y > P2.Y then Result := P2
                   else
      if P1.X < P2.X then Result := P1
                     else Result := P2;
end;

//----------------------------------------------------------------------------------------------------------------------

function MinPointAsReference(var P1, P2: TPoint): PPoint;

begin
  if P1.Y < P2.Y then Result := @P1
                 else
    if P1.Y > P2.Y then Result := @P2
                   else
      if P1.X < P2.X then Result := @P1
                     else Result := @P2;
end;

//----------------------------------------------------------------------------------------------------------------------

function PointsAreEqual(P1, P2: TPoint): Boolean;

begin
  Result := (P1.X = P2.X) and (P1.Y = P2.Y);
end;

//----------------- TUndoList ------------------------------------------------------------------------------------------

constructor TUndoList.Create(AOwner: TCustomSyntaxEdit);

begin
  inherited Create;
  FOwner := AOwner;
  FList := TList.Create;
  FMaxUndo := 10;
  FCurrent := -1;
end;

//----------------------------------------------------------------------------------------------------------------------

destructor TUndoList.Destroy;

begin
  ClearList;
  FList.Free;
  inherited Destroy;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TUndoList.AddChange(AReason: TChangeReason; const AOldText: WideString; AOldCaret, AOldSelStart,
  AOldSelEnd: PPoint; const ANewText: WideString; ANewCaret, ANewSelStart, ANewSelEnd: PPoint);

// adds a new change into the undo list;
// The parameter list looks rather messy, but should serve all possible combination
// which can appear. By using pointers instead of just the records we can pass
// nil values where we don't need a value in the undo list. This is particularly useful
// to avoid creating dummy records just to fill the paramter list.
// Note: There's an overloaded funtion variant to avoid passing in the current
//       edit state (text, caret, selection).

var
  NewChange: PChange;

begin
  if FMaxUndo > 0 then
  begin
    if FList.Count >= FMaxUndo then RemoveChange(0);

    // a new undo record deletes all redo records if there are any
    while FCurrent < FList.Count - 1 do RemoveChange(FList.Count - 1);

    // add a new entry
    New(NewChange);
    ZeroMemory(NewChange, SizeOf(NewChange^));
    with NewChange^ do
    begin
      Reason := AReason;
      OldText := AOldText;
      if AOldCaret <> nil then OldCaret := AOldCaret^;
      if AOldSelStart <> nil then OldSelStart := AOldSelStart^;
      if AOldSelEnd <> nil then OldSelEnd := AOldSelEnd^;
      NewText := ANewText;
      if ANewCaret <> nil then NewCaret := ANewCaret^;
      if ANewSelStart <> nil then NewSelStart := ANewSelStart^;
      if ANewSelEnd <> nil then NewSelEnd := ANewSelEnd^;
    end;
    FList.Add(NewChange);
    // advance current undo index
    Inc(FCurrent);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TUndoList.AddChange(AReason: TChangeReason; const AOldText: WideString; AOldCaret, AOldSelStart, AOldSelEnd: PPoint);

// adds a new change into the undo list, all new values are taken from current state
// Note: There's an overloaded function variant to allow passing other New* values
//       than the ones of the editor.

var
  NewChange: PChange;

begin
  if FMaxUndo > 0 then
  begin
    if FList.Count >= FMaxUndo then RemoveChange(0);

    // a new undo record deletes all redo records if there are any
    while FCurrent < FList.Count - 1 do RemoveChange(FList.Count - 1);

    // add a new entry
    New(NewChange);
    ZeroMemory(NewChange, SizeOf(NewChange^));
    with NewChange^ do
    begin
      Reason := AReason;
      OldText := AOldText;
      if AOldCaret <> nil then OldCaret := AOldCaret^;
      if AOldSelStart <> nil then OldSelStart := AOldSelStart^;
      if AOldSelEnd <> nil then OldSelEnd := AOldSelEnd^;
      with FOwner do
      begin
        NewText := GetSelText;
        NewCaret := CaretXY;
        NewSelStart := FBlockBegin;
        NewSelEnd := FBlockEnd;
      end;
    end;
    FList.Add(NewChange);
    // advance current undo index
    Inc(FCurrent);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TUndoList.GetUndoChange(var Change: PChange): TChangeReason;

begin
  if FCurrent = -1 then Result := ecrNone
                   else
  begin
    Change := FList.Items[FCurrent];
    Result := Change.Reason;
    Dec(FCurrent);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TUndoList.GetRedoChange(var Change: PChange): TChangeReason;

begin
  if FCurrent = FList.Count - 1 then Result := ecrNone
                                else
  begin
    Inc(FCurrent);
    Change := FList.Items[FCurrent];
    Result := Change.Reason;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TUndoList.GetCanRedo: Boolean;

begin
  Result := FCurrent < FList.Count - 1;
end;

//----------------------------------------------------------------------------------------------------------------------

function TUndoList.GetCanUndo: Boolean;

begin
  Result := FCurrent > -1;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TUndoList.SetMaxUndo(Value: Integer);

begin
  if Value > 32767 then Value := 32767;
  if FMaxUndo <> Value then
  begin
    FMaxUndo := Value;
    // FCurrent ist automatically updated
    while FMaxUndo < FList.Count - 1 do RemoveChange(0);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TUndoList.RemoveChange(Index: Integer);

var
  AChange: PChange;

begin
  if (Index > -1) and (Index < FList.Count) then
  begin
    AChange := FList.Items[Index];
    Dispose(AChange);
    FList.Delete(Index);
  end;
  if FCurrent > FList.Count - 1 then FCurrent := FList.Count - 1;
end;

//----------------------------------------------------------------------------------------------------------------------

function TUndoList.GetCurrentUndoReason: TChangeReason;

begin
  if FCurrent = -1 then Result := ecrNone
                   else Result := PChange(FList.Items[FCurrent]).Reason;
end;

//----------------------------------------------------------------------------------------------------------------------

function TUndoList.GetCurrentRedoReason: TChangeReason;

begin
  if FCurrent = FList.Count - 1 then Result := ecrNone
                                else Result := PChange(FList.Items[FCurrent + 1]).Reason;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TUndoList.ClearList;

var
  I: Integer;

begin
  for I := FList.Count - 1 downto 0 do RemoveChange(I);
end;

//----------------- TBookmarkOptions -----------------------------------------------------------------------------------

constructor TBookmarkOptions.Create(AOwner: TCustomSyntaxEdit);

begin
  inherited Create;
  FOwner := AOwner;
  FEnableKeys := True;
  FGlyphsVisible := True;
  FLeftMargin := 2;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TBookmarkOptions.SetGlyphsVisible(Value: Boolean);

begin
  if FGlyphsVisible <> Value then
  begin
    FGlyphsVisible := Value;
    FOwner.Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TBookmarkOptions.SetLeftMargin(Value: Integer);

begin
  if FLeftMargin <> Value then
  begin
    FLeftMargin := Value;
    FOwner.Invalidate;
  end;
end;

//----------------- TSyntaxEditorStrings -------------------------------------------------------------------------------

constructor TSyntaxEditorStrings.Create(AOwner: TCustomSyntaxEdit);

begin
  inherited Create;
  FOwner := AOwner;
  FAutoResetModified := True;
end;

//----------------------------------------------------------------------------------------------------------------------

function TSyntaxEditorStrings.Add(const S: WideString): Integer;

begin
  if eoKeepTrailingBlanks in FOwner.FOptions then Result := inherited Add(S)
                                             else Result := inherited Add(WideTrimRight(S));
  FModified := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSyntaxEditorStrings.Delete(Index: Integer);

begin
  inherited Delete(Index);
  if Index = FOwner.FLastCalcIndex then FOwner.FLastCalcIndex := -1;
  FModified := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSyntaxEditorStrings.Insert(Index: Integer; const S: WideString);

begin
  inherited Insert(Index, S);
  if Index = FOwner.FLastCalcIndex then FOwner.FLastCalcIndex := -1;
  FModified := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSyntaxEditorStrings.Put(Index: Integer; const S: WideString);

begin
  if Get(Index) <> S then
  begin
    inherited Put(Index, S);
    FModified := True;
    FOwner.InvalidateLine(Index);
    if Index = FOwner.FLastCalcIndex then FOwner.FLastCalcIndex := -1;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSyntaxEditorStrings.LoadFromStream(Stream: TStream);

var
  I: Integer;
  S: WideString;
  MaxChar: Integer;
  
begin
  // clear entire text buffer plus some other variables and the undo stack,
  // do not register an undo action for it
  FOwner.ClearAll(False, False);
  inherited;
  FOwner.FLastCalcIndex := -1;
  FOwner.FUndoList.ClearList;

  // check for lines which are too long
  MaxChar := FOwner.FMaxRightPos div FOwner.FCharWidth;
  for I := 0 to Count - 1 do
  begin
    if Length(Self[I]) > MaxChar then
    begin
      S := Self[I];
      if FOwner.DoCutText(S) then S := Copy(S, 1, MaxChar);
      Self[I] := S;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSyntaxEditorStrings.SaveToStream(Stream: TStream);

begin
  inherited;
  if Saved and AutoResetModified then
  begin
    FModified := False;
    FOwner.FModified := False;
  end;
  if not (eoUndoAfterSave in FOwner.FOptions) then FOwner.FUndoList.ClearList;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TSyntaxEditorStrings.SetText(const Value: WideString);

begin
  inherited SetText(Value);
  FOwner.FLastCalcIndex := -1;
  FOwner.Invalidate;
  FModified := True;
  if Assigned(FOnLoaded) then FOnLoaded(Self);
end;

//----------------- TColors --------------------------------------------------------------------------------------------

constructor TColors.Create(AOwner: TCustomSyntaxEdit);

begin
  inherited Create;
  FOwner := AOwner;
  FBackground := clHighLight;
  FForeground := clHighLightText;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TColors.SetBackground(Value: TColor);

begin
  if FBackground <> Value then
  begin
    FBackground := Value;
    if FOwner <> nil then FOwner.Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TColors.SetForeground(Value: TColor);

begin
  if FForeground <> Value then
  begin
    FForeground := Value;
    if FOwner <> nil then FOwner.Invalidate;
  end;
end;

//----------------- TScrollHintWindow ----------------------------------------------------------------------------------

procedure TScrollHintWindow.Paint;

var
  R: TRect;
  
begin
  R := ClientRect;
  Inc(R.Left, 2);
  Inc(R.Top, 2);
  Canvas.Font := Font;
  DrawText(Canvas.Handle, PChar(Caption), -1, R, DT_LEFT or DT_NOPREFIX or
    DT_WORDBREAK or DrawTextBiDiModeFlagsReadingOnly);
end;

//----------------- TCustomSyntaxEdit ----------------------------------------------------------------------------------

constructor TCustomSyntaxEdit.Create(AOwner: TComponent);

var
  I: Integer;

begin
  inherited Create(AOwner);
  DoubleBuffered := False;
  FOptions := DefaultOptions;
  FLines := TSyntaxEditorStrings.Create(Self);

  FFontDummy := TFont.Create;
  FLineNumberFont := TFont.Create;
  FLineNumberFont.Name := 'Terminal';
  FLineNumberFont.Size := 6;
  FLineNumberFont.Color := clBlack;
  FLineNumberFont.Style := [];
  FLineNumberFont.OnChange := LineNumberFontChanged;

  FUndoList := TUndoList.Create(Self);
  TSyntaxEditorStrings(FLines).OnLoaded := ListLoaded;
  FSelectedColor := TColors.Create(Self);
  FMarkColor := TColors.Create(Self);
  FScrollHintColor := TColors.Create(Self);
  FScrollHintColor.FBackground := clAppWorkSpace;
  FScrollHintColor.FForeground := clInfoText;
  FBookmarkOptions := TBookmarkOptions.Create(Self);

  FScrollTimer := TTimer.Create(Self);
  FScrollTimer.enabled := False;
  FScrollTimer.Interval := 20;
  FScrollTimer.OnTimer := OnScrollTimer;

  FCharsInWindow := 10;
  // FRightMargin, FCharWidth and MaxRight have to be set before FontChanged
  // is called for the first time.
  FRightMargin := 80;
  // this char width value is only a dummy and will be updated later
  FCharWidth := 8;
  MaxRightChar := 80;
  FMarginColor := clSilver;
  FGutterWidth := 30;
  ControlStyle := ControlStyle + [csOpaque, csSetCaption, csReflector];
  Height := 150;
  Width := 200;
  Cursor := crIBeam;
  Color := clWindow;
  FCaretX := 0;
  FCaretY := 0;
  FCaretOffset := Point(0, 0);
  FGutterColor := clBtnFace;
  Font.OnChange := FontChanged;
  FFontDummy.Name := 'Courier New';
  FFontDummy.Size := 10;
  Font.Assign(FFontDummy);
  ParentFont := False;
  ParentColor := False;
  TabStop := True;
  TWideStringList(FLines).OnChange := LinesChanged;
  with FGutterRect do
  begin
    Left := 0;
    Top := 0;
    Right := FGutterWidth - 1;
    Bottom := Height;
  end;
  FTabSize := 8;
  MaxRightChar := 128;
  FUpdateCount := 0;
  FIndentSize := 2;
  FScrollBars := ssBoth;
  FBlockBegin := Point(0, 0);
  FBlockEnd := Point(0, 0);
  FInternalBMList := TImageList.Create(Self);
  ConvertAndAddImages(FInternalBMList);
  for I := 0 to 9 do FBookmarks[I].Y := -1;
  FBorderStyle := bsSingle;
  FInsertCaret := ctVerticalLine;
  FOverwriteCaret := ctBlock;
  FKeyStrokes := TKeyStrokes.Create(Self);
  FScrollFactor := 1;
  FMarkLine := -1;
  FLastCalcIndex := -1;
  // retrive double click time to be used in triple clicks,
  FDoubleClickTime := GetDoubleClickTime;
  SetDefaultKeystrokes;

  FScrollHint := TScrollHintWindow.Create(Self);
  FScrollHint.Name := 'Scroll';
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.CreateParams(var Params: TCreateParams);

const
  ScrollBar: array[TScrollStyle] of DWORD = (0, WS_HSCROLL, WS_VSCROLL, WS_HSCROLL or WS_VSCROLL);
  BorderStyles: array[TBorderStyle] of DWORD = (0, WS_BORDER);

  CS_OFF = CS_HREDRAW or CS_VREDRAW;

begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or ScrollBar[FScrollBars] or BorderStyles[FBorderStyle] or WS_CLIPCHILDREN or WS_CLIPSIBLINGS;
    WindowClass.Style := WindowClass.Style and not CS_OFF;
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

destructor TCustomSyntaxEdit.Destroy;

begin
  FKeyStrokes.Free;
  FLineNumberFont.Free;
  FFontDummy.Free;
  FLines.Free;
  FSelectedColor.Free;
  FScrollHintColor.Free;
  FBookmarkOptions.Free;
  FScrollTimer.Free;
  FUndoList.Free;
  FCharWidthArray := nil;
  // scroll hint window is automatically destroyed
  inherited Destroy;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.RecordState(IncludeText: Boolean): TEditState;

// prepare the result structure so it can be used for undo-regstration
begin
  with Result do
  begin
    if IncludeText then Text := GetSelText
                   else Text := '';
    Caret := CaretXY;
    SelStart := FBlockBegin;
    SelEnd := FBlockEnd;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.PositionFromPoint(X, Y: Integer): TPoint;

// calculates the column/line position from the given pixel coordinates (which must
// be in client coordinates)

begin
  // note: for the column value a real division with round is used (instead of
  // an integer div) to automatically switch to the proper column depending on
  // whether the position is before the half width of the column or after
  Result := Point(Round((X - FOffsetX - FGutterRect.Right + 1) / FCharWidth), Y div FTextHeight - FOffsetY);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ComputeCaret(X, Y: Integer);

// sets the caret to the line/column indicated by the given pixel position

var
  EditState: TEditState;
  NewCaret: TPoint;

begin
  EditState := RecordState(False);
  CaretXY := PositionFromPoint(X, Y);
  with EditState do
    if not PointsAreEqual(CaretXY, Caret) then
    begin
      NewCaret := CaretXY;
      FUndoList.AddChange(ecrCursorMove, '', @Caret, @SelStart, @SelEnd,
                                         '', @NewCaret, nil, nil);
    end;
  UpdateWindow(Handle);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.PosInSelection(Pos: TPoint): Boolean;

// determines whether the given position (line/column reference) is in the currrent
// selection block

var
  BB, BE: PPoint;

begin
  Result := False;
  if SelAvail then
  begin
    BB := MinPointAsReference(FBlockBegin, FBlockEnd);
    BE := MaxPointAsReference(FBlockBegin, FBlockEnd);
    if BB.Y = BE.Y then Result := (Pos.Y = BB.Y) and (Pos.X >= BB.X) and (Pos.X <= BE.X)
                   else
      Result := ((Pos.Y > BB.Y) and (Pos.Y < BE.Y)) or
                ((Pos.Y = BB.Y) and (Pos.X >= BB.X)) or
                ((Pos.Y = BE.Y) and (Pos.X <= BE.X));
    // so far so good, let's also take into account when
    // the block end is beyond a line end
    if Result and (Pos.Y = BE.Y) then
      Result := Cardinal(Pos.X) < GetLineEnd(Pos.Y);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.CopyToClipboard;

var
  Data: THandle;
  DataPtr: Pointer;
  Size: Cardinal;
  W: WideString;

begin
  // Delphi's TClipboard does not support Unicode (D5 and lower) hence we
  // need to handle the stuff manually. Add an ANSI version of the string too
  // using the system's default locale for Win9x. Note: non ANSI characters will
  // be converted to question marks. Under WinNT/Win2K where the clipboard is
  // Unicode aware you should of course use the Unicode clipboard format (CF_UNICODETEXT);
  if SelAvail then
  begin
    W := SelText;
    Size := Length(W);

    // ANSI version
    Clipboard.AsText := W;

    // Unicode version
    Data := GlobalAlloc(GMEM_MOVEABLE + GMEM_DDESHARE, 2 * Size + 2);
    try
      DataPtr := GlobalLock(Data);
      try
        Move(PWideChar(W)^, DataPtr^, 2 * Size + 2);
        Clipboard.SetAsHandle(CF_UNICODETEXT, Data);
      finally
        GlobalUnlock(Data);
      end;
    except
      GlobalFree(Data);
      raise;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.CutToClipboard;

var
  State: TEditState;

begin
  if SelAvail then
  begin
    State := RecordState(True);
    CopyToClipboard;
    SetSelText('');
    with State do
      FUndoList.AddChange(ecrDelete, Text, @Caret, @SelStart, @SelEnd);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.PasteFromClipboard;

var
  State: TEditState;
  BB, BE: TPoint;
  Data: THandle;

begin
  if Clipboard.HasFormat(CF_UNICODETEXT) or
     Clipboard.HasFormat(CF_TEXT) then
  begin
    State := RecordState(True);
    // keep the position at which the new text will be inserted
    BB := MinPoint(FBlockBegin, FBlockEnd);

    // prefer Unicode text over normal text
    if Clipboard.HasFormat(CF_UNICODETEXT) then
    begin
      Data := Clipboard.GetAsHandle(CF_UNICODETEXT);
      try
        if Data <> 0 then SetSelText(PWideChar(GlobalLock(Data)));
      finally
        if Data <> 0 then GlobalUnlock(Data);
      end;
    end
    else
      SetSelText(Clipboard.AsText);
      
    // keep end of text block we just inserted
    BE := CaretXY;
    with State do
      FUndoList.AddChange(ecrReplace, Text, @Caret, @SelStart, @SelEnd, Clipboard.AsText, @BE, @BB, @BE);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetBlockBegin: TPoint;

begin
  Result := MinPoint(FBlockBegin, FBlockEnd);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetBlockEnd: TPoint;

begin
  Result := MaxPoint(FBlockBegin, FBlockEnd);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.CaretXPix: Integer;

begin
  Result := FCaretX * FCharWidth + FGutterRect.Right + 1 + FOffsetX;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.CaretYPix: Integer;

begin
  Result := (FCaretY + FOffsetY) * FTextHeight;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.FontChanged(Sender: TObject);

var
  DC: HDC;
  Save: THandle;
  Metrics: TTextMetric;
  TMFont: HFONT;
  TMLogFont: TLogFont;

begin
  GetObject(Font.Handle, SizeOf(TLogFont), @TMLogFont);

  // We need to get the font with bold and italics set so we have enough room
  // for the widest character it will draw.  This is because fixed pitch fonts
  // always have the same character width, but ONLY when the same attributes
  // are being used.  That is, a bold character can be wider than a non-bold
  // character in a fixed pitch font.
  TMLogFont.lfWeight := FW_BOLD;
  TMLogFont.lfItalic := 1;
  TMFont := CreateFontIndirect(TMLogFont);
  try
    DC := GetDC(0);
    try
      Save := SelectObject(DC, TMFont);
      GetTextMetrics(DC, Metrics);
      SelectObject(DC, Save);
    finally
      ReleaseDC(0, DC);
    end;
  finally
    DeleteObject(TMFont);
  end;

  with Metrics do
  begin
    // Note:  Through trial-and-error I found that tmAveCharWidth should be used
    // instead of tmMaxCharWidth as you would think. I'm basing this behavior
    // on the Delphi IDE editor behavior. If tmMaxCharWidth is used, we end up
    // with chars being much farther apart than the same font in the IDE.
    CharWidth := tmAveCharWidth;

    FTextHeight := tmHeight + tmExternalLeading + FExtraLineSpacing;
    if Parent <> nil then
    begin
      FLinesInWindow := ClientHeight div FTextHeight;
      FCharsInWindow := (ClientWidth - FGutterRect.Right + 1) div FCharWidth;
    end;
    if HandleAllocated then
    begin
      if Focused then
      begin
        HideCaret;
        InitializeCaret;
      end;
      UpdateScrollBars;
    end;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetFont: TFont;

begin
  Result := inherited Font;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetLineCount: Integer;

begin
  Result := FLines.Count;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetLineText: WideString;

// returns the text of the current line, makes sure there is a line by
// creating one if necessary

var
  WasModified: Boolean;

begin
  if FLines.Count = 0 then
  begin
    WasModified := Modified;
    FLines.Add('');
    // It's a dummy line we just have inserted hence we should restore the
    // previous modification state.
    Modified := WasModified;
    if FHighlighter <> nil then
    begin
      FHighlighter.ResetRange;
      FLines.Objects[0] := FHighlighter.GetRange;
    end;
  end;                                                 
  Result := FLines[FCaretY];
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetSelAvail: Boolean;

begin
  Result := ((FBlockBegin.X <> FBlockEnd.X) or (FBlockBegin.Y <> FBlockEnd.Y)) and (FLines.Count > 0);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetSelText: WideString;

var
  BB, BE: TPoint;
  BeginX,
  EndX: Integer;
  First, Last: Integer;
  Helper: TWideStringList;

begin
  Result := '';
  if SelAvail then
  begin
    Helper := TWideStringList.Create;
    BB := BlockBegin;
    BeginX := ColumnToCharIndex(BB);
    BE := BlockEnd;
    EndX := ColumnToCharIndex(BE);
    if BB.Y = BE.Y then Result := Copy(FLines[BB.Y], BeginX + 1, EndX - BeginX)
                   else
    begin
      First := BB.Y;
      Last := BE.Y;
      Helper.Add(Copy(FLines[First], BeginX + 1, Length(FLines[First])));
      Inc(First);
      while First < Last do
      begin
        Helper.Add(FLines[First]);
        Inc(First);
      end;
      Result := Helper.Text;
      if First < FLines.Count then Result := Result + Copy(FLines[First], 1, EndX);
    end;
    Helper.Free;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetText: WideString;

begin
  Result := FLines.Text;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetTopLine: Integer;

begin
  Result := -FOffsetY + 1;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.HideCaret;

begin
  if FCaretVisible then FCaretVisible := not Windows.HideCaret(Handle);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.KeyDown(var Key: Word; Shift: TShiftState);

var
  Data: Pointer;
  Cmd: TEditorCommand;

begin
  inherited;

  Data := nil;
  try
    Cmd := TranslateKeyCode(Key, Shift, Data);
    if Cmd <> ecNone then
    begin
      Key := 0;
      FKeyHandled := True;
      CommandProcessor(Cmd, WideNull, Data);
    end;
  finally
    if Data <> nil then FreeMem(Data);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.KeyPress(var Key: Char);

var
  CW: WideChar;

begin
  inherited;
  if not FKeyHandled then
  begin
    CW := KeyUnicode(Key);
    CommandProcessor(ecChar, CW, nil);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.KeyUp(var Key: Word; Shift: TShiftState);

begin
  inherited;
  FKeyHandled := False;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.Loaded;

begin
  inherited Loaded;
  UpdateScrollBars;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.LinesChanged(Sender: TObject);

begin
  if HandleAllocated then
  begin
    UpdateScrollBars;
    if FHighlighter <> nil then ScanFrom(FCaretY);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.TripleClick(X, Y: Integer);

// mark entire line if the user clicked three times within the double click interval

var
  NewCaret: TPoint;

begin
  FMultiClicked := True;

  // set new selection only when the mouse is still on the same line as when the first click occured
  NewCaret := PositionFromPoint(X, Y);
  if NewCaret.Y = FCaretY then
  begin
    SetBlockBegin(Point(0, CaretY));
    SetBlockEnd(Point(0, CaretY + 1));
  end;
  FLastDblClick := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.DblClick;

begin
  if FLines.Count > 0 then SetWordBlock(CaretXY);
  inherited;
  FMultiClicked := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

var
  ColPos: TPoint;
  
begin
  inherited;

  if Button = mbLeft then
  begin
    if not Focused then
    begin
      SetFocus;
      // don't delete selection if the edit wasn't focused previously
      if SelAvail then Exit;
    end;

    if not PtInRect(FGutterRect, Point(X, Y)) then
    begin
      // MouseDown will also be called when the mouse has been clicked twice (double click),
      // hence we can handle the triple click stuff herein too.
      if ssDouble in Shift then FLastDblClick := GetTickCount
                           else
      begin
        if (eoTripleClicks in FOptions) and (Shift = [ssLeft]) and (FLastDblClick > 0) then
        begin
          if (GetTickCount - FLastDblClick) < FDoubleClickTime then
          begin
            TripleClick(X, Y);
            Exit;
          end;
          FLastDblClick := 0;
        end;

        SetCapture(Handle);
        // check for gutter or right margin resize
        ColPos := PositionFromPoint(X, Y);
        if (eoAllowMarginResize in FOptions) and (ColPos.X = FRightMargin) then
        begin
          Include(FStates, esRightMarginResizing);
        end
        else
          if (eoAllowGutterResize in FOptions) and (Abs(X - FGutterWidth) < 3) then
          begin
            Include(FStates, esGutterResizing);
          end
          else
          begin
            // keep last cursor position to restore it in case of cancelled drag operation
            FLastCaret := CaretXY;
            ComputeCaret(X, Y);
            if ((Shift - [ssCtrl]) = [ssLeft]) and
               PosInSelection(PositionFromPoint(X, Y)) then BeginDrag(False)
                                                       else
            begin
              if ssShift in Shift then SetBlockEnd(CaretXY)
                                  else SetBlockBegin(CaretXY);
            end;
          end;
      end;
    end;
  end
  else FScrollTimer.enabled := False;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.MouseMove(Shift: TShiftState; X, Y: Integer);

var
  P: TPoint;

begin
  inherited;

  P := PositionFromPoint(X, Y);
  if esGutterResizing in FStates then
  begin
    GutterWidth := X;
  end
  else
    if esRightMarginResizing in FStates then
    begin
      RightMargin := P.X;
    end
    else
    begin
      // don't use ssLeft in Shift as this parameter is not always reliable (mouse wheel!)
      if (csLButtonDown in ControlState) and not FMultiClicked then
      begin
        if not FScrollTimer.Enabled
           and ((FLastCaret.X - P.X <> 0) or (FLastCaret.Y - P.Y <> 0)) then FScrollTimer.Enabled := True;
      end
      else
        FScrollTimer.enabled := False;
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

begin
  FMultiClicked := False;
  FScrollTimer.enabled := False;
  FStates := FStates - [esGutterResizing, esRightMarginResizing];
  ReleaseCapture;
  inherited;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.Paint;

var
  WasVisible: Boolean;

begin
  WasVisible := FCaretVisible;
  if WasVisible then HideCaret;
  PaintGutter(Canvas);
  Canvas.Font := Font;
  if FLines.Count > 0 then
  begin
    if (FHighLighter <> nil) and
       (eoUseSyntaxHighlighting in FOptions) then PaintHighlighted(Canvas)
                                             else PaintText(Canvas);
  end
  else
    with Canvas do
    begin
      Brush.Color := Self.Color;
      FillRect(ClipRect);

      // draw the right margin marker
      Pen.Color := FMarginColor;
      Pen.Width := 1;
      MoveTo(FRightMargin * FCharWidth + FGutterRect.Right + 1 + FOffsetX, 0);
      LineTo(FRightMargin * FCharWidth + FGutterRect.Right + 1 + FOffsetX, ClientHeight);
    end;

  if Assigned(FOnPaint) then FOnPaint(Self, Canvas);
  if WasVisible then ShowCaret;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.PaintGutter(TextCanvas: TCanvas);

var
  I: Integer;
  UpdateRect, R: TRect;
  StartLine,
  EndLine,
  LineIndex,
  YOffset: Integer;
  Number: String;

begin
  // draw only if (part of) gutter rect is invalid
  if IntersectRect(UpdateRect, TextCanvas.ClipRect, FGutterRect) then
  begin
    with TextCanvas do
    begin
      Brush.Color := FGutterColor;
      UpdateRect.Left := FGutterRect.Right - 5;
      FillRect(UpdateRect);

      UpdateRect.Right := FGutterRect.Right - 3;
      Pen.Color := clBtnHighlight;
      Pen.Style := psSolid;
      Pen.Width := 1;
      MoveTo(UpdateRect.Right, 0);
      LineTo(UpdateRect.Right, UpdateRect.Bottom);

      Inc(UpdateRect.Right);
      Pen.Color := clBtnShadow;
      MoveTo(UpdateRect.Right, 0);
      LineTo(UpdateRect.Right, UpdateRect.Bottom);

      Inc(UpdateRect.Right, 2);
      Pen.Color := Self.Color;
      Pen.Width := 2;
      MoveTo(UpdateRect.Right, 0);
      LineTo(UpdateRect.Right, UpdateRect.Bottom);

      StartLine := (UpdateRect.Top div FTextHeight) - FOffsetY;
      EndLine := (UpdateRect.Bottom div FTextHeight) - FOffsetY + 1;
      if EndLine > FLines.Count - 1 then EndLine := FLines.Count - 1;
      YOffset := (UpdateRect.Top div FTextHeight) * FTextHeight;
      R := Rect(0, YOffset, FGutterRect.Right - 5, YOffset + FTextHeight);

      if eoLineNumbers in FOptions then
      begin
        Font.Assign(FLineNumberFont);
        // make the line numbers showing one-based
        for LineIndex := StartLine + 1 to EndLine + 1 do
        begin
          Number := IntToStr(LineIndex);
          FillRect(R);
          DrawText(Handle, PChar(Number), Length(Number), R, DT_VCENTER or DT_RIGHT or DT_SINGLELINE);
          OffsetRect(R, 0, FTextHeight);
        end;
      end;

      // fill rest of gutter rect if necessary
      if R.Top < UpdateRect.Bottom then
      begin
        R.Bottom := UpdateRect.Bottom;
        FillRect(R);
      end;

      if BookMarkOptions.GlyphsVisible then
      begin
        for I := 0 to 9 do
          with FBookMarks[I] do
            if Visible and (Y >= StartLine) and (Y <= EndLine) then
            begin
              FInternalBMList.Draw(TextCanvas, FBookmarkOptions.LeftMargin, (Y + FOffsetY) * FTextHeight, I);
            end;
      end;
      ExcludeClipRect(Handle, 0, 0, FGutterRect.Right, FGutterRect.Bottom);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.PaintHighlighted(TextCanvas: TCanvas);

// main drawing routine to paint the control if syntax highlighting is used

var
  YOffset,
  LineIndex,
  PaintOffset,
  OldOffset,
  StartIndex,
  StopIndex,
  TempIndex: Integer;
  BB, BE: PPoint;
  TokenData: TTokenData;

  I,
  StartLine,
  EndLine: Integer;
  UpdateRect,
  Dummy,
  R: TRect;
  SelStart,
  SelEnd: Cardinal;
  CurrentStyle: Cardinal; // to optimize drawing selected/unselected text
  ShowSelection,
  ShowLineBreak: Boolean;
  LineBase: PWideChar;

  //--------------- local functions -------------------------------------------

  procedure PaintToken(Start, Len: Cardinal);

  // paints the given string at current paint offset, no control chars are painted,
  // text attributes must be set up already on call;
  // side effects: current paint offset will be updated

  var
    Index,
    Counter: Cardinal;
    P: PWideChar;

  begin
    // replace non-printable characters,
    // calculate update rect for this token (-part) btw.
    P := LineBase + Start;
    OldOffset := PaintOffset;
    Index := Start;
    Counter := Len;
    while Counter > 0 do
    begin
      if P^ < ' ' then P^ := ' ';
      Inc(P);
      Inc(PaintOffset, FCharWidthArray[Index]);
      Inc(Index);
      Dec(Counter);
    end;

    R := Rect(OldOffset, YOffset, PaintOffset, YOffset + FTextHeight);
    // finally paint text
    ExtTextOutW(TextCanvas.Handle, OldOffset, YOffset, ETO_OPAQUE, @R, LineBase + Start, Len, @FCharWidthArray[Start]);
  end;

  //---------------------------------------------------------------------------

  procedure PaintTokenWithControlChars(Start, Len: Cardinal);

  // paints the given WideString at current paint offset, considers control characters,
  // text attributes must be set up already on call;
  // side effects: current paint offset will be updated

  var
    I,
    Counter: Cardinal;
    Run,
    TabPos: PWideChar;

  begin
    // replace space characters
    Run := LineBase + Start;
    Counter := Len; 
    while Counter > 0 do
    begin
      if Run^ = ' ' then Run^ := DefaultSpaceGlyph;
      Inc(Run);
      Dec(Counter);
    end;
    Run := LineBase + Start;

    Counter := Len;
    with TextCanvas do
      while Counter > 0 do
      begin
        // find TAB character
        TabPos := StrScanW(Run, WideTabulator);
        // something to draw before tab?
        if TabPos <> nil then I := Min(Counter, TabPos - Run)
                         else I := Counter;
        if I > 0 then
        begin
          OldOffset := PaintOffset;
          Inc(PaintOffset, I * Cardinal(FCharWidth));
          R := Rect(OldOffset, YOffset, PaintOffset, YOffset + FTextHeight);
          ExtTextOutW(Handle, OldOffset, YOffset, ETO_OPAQUE, @R, Run, I, @FCharWidthArray[Start]);
          Inc(Start, I);
          Dec(Counter, I);
          Inc(Run, I);
        end;

        // tab to draw?
        if TabPos <> nil then
        begin
          while (TabPos^ = WideTabulator) and (Counter > 0) do
          begin
            OldOffset := PaintOffset;
            Inc(PaintOffset, FCharWidthArray[Start]);
            R := Rect(OldOffset, YOffset, PaintOffset, YOffset + FTextHeight);

            SetTextAlign(Handle, TA_CENTER);
            ExtTextOutW(Handle, (PaintOffset + OldOffset) div 2, YOffset, ETO_OPAQUE, @R, DefaultTabGlyph, 1, nil);
            SetTextAlign(Handle, TA_LEFT);
            Inc(Start);
            Dec(Counter);
            Inc(TabPos);
          end;
          Run := TabPos;
        end;
    end;
  end;

  //---------------------------------------------------------------------------

  procedure PaintLineBreak;

  begin
    R := Rect(PaintOffset, YOffset, PaintOffset + FCharWidth, YOffset + FTextHeight);
    if IntersectRect(Dummy, R, UpdateRect) then
      ExtTextOutW(TextCanvas.Handle, PaintOffset, YOffset, ETO_OPAQUE, @R, DefaultLineBreakGlyph, 1, Pointer(FCharWidthArray));
    Inc(PaintOffset, FCharWidth);
  end;

  //--------------- end local functions ---------------------------------------

begin
  with TextCanvas do
  begin
    UpdateRect := ClipRect;
    // optimize drawing by using the update area
    StartLine := (UpdateRect.Top div FTextHeight) - FOffsetY;
    EndLine := (UpdateRect.Bottom div FTextHeight) - FOffsetY + 1;
    if EndLine > FLines.Count - 1 then EndLine := FLines.Count - 1;

    // The vertical position we will paint at. Start at top and increment by
    // FTextHeight for each iteration of the loop (see bottom of for loop).
    YOffset := (UpdateRect.Top div FTextHeight) * FTextHeight;

    // we will update the highlighter start states for each line we touch to prevent
    // an extra run somewhere, hence unplug the change handler
    TWideStringList(FLines).OnChange := nil;

    // tell highlighter which state it should use from now on
    if StartLine < FLines.Count - 1 then
      // already sanned?
      if StartLine <= FLastValidLine then FHighLighter.SetRange(FLines.Objects[StartLine])
                                     else
      begin
        // not yet scanned, so do it now
        I := FLastValidLine;
        // scan all lines up to the start line
        FHighLighter.SetRange(FLines.Objects[I]);
        repeat
          FHighLighter.SetLine(FLines[I]);
          while not FHighLighter.EOL do FHighLighter.Next;
          Inc(I);
          FLines.Objects[I] := FHighLighter.GetRange;
        until I = StartLine;
        // on exit we have already the correct start state for our painting
      end;

    // determine correct start and end line of selection
    BB := MinPointAsReference(FBlockBegin, FBlockEnd);
    BE := MaxPointAsReference(FBlockBegin, FBlockEnd);

    ShowLineBreak := (eoShowControlChars in FOptions) and not (eoScrollPastEOL in FOptions);
  
    // Determine in advance whether to show selection or not.
    // Note: We have four concurrent options determining when to show selection (apart
    //       from SelAvail). These are: focus state, eoReadOnly, eoReadOnlySelection and
    //       eoHideSelection. Given these states one can derive a logic table with 16 results
    //       and from this table get a logical equation. Optimizing this eq. leads to
    //       the four conditions written and described below. If we name the mentioned
    //       states (in the same order as above) with A, B, C and D, respectively, then
    //       the resulting eq. is: Sel = AB or AC or BD or CD.
    // show selection if...
    ShowSelection := // selection is available and
                     SelAvail and (
                       // edit is focused and not read only (then both selection opts have no influence) or
                       (Focused and not (eoReadOnly in FOptions)) or
                       // edit is focused and selection in r/o mode is allowed (then read only and
                       // hide selection when unfocused don't matter) or
                       (Focused and (eoReadOnlySelection in FOptions)) or
                       // edit is not read only and selection will not be hidden when unfocused (then
                       // being focused doesn't matter and r/o selection has no influence) or
                       ((FOptions * [eoReadOnly, eoHideSelection]) = []) or
                       // r/o selection is enabled and selection will not be hidden when unfocused
                       // (then being focused or in r/o mode don't matter)
                       ((eoReadOnlySelection in FOptions) and not (eoHideSelection in FOptions))
                     );

    // loop through the lines which needs repainting
    for LineIndex := StartLine to EndLine do
    begin
      // Tell highlighter what line we are working on.  This will get the first
      // token set up.
      FHighLighter.SetLine(FLines[LineIndex]);
      LineBase := PWideChar(FLines[LineIndex]);

      PaintOffset := FOffsetX + FGutterRect.Right + 1;
      CalcCharWidths(LineIndex);

      // line marker to draw (overrides selection)?
      if FMarkLine = LineIndex then
      begin
        Font.Color := FMarkColor.Foreground;
        Brush.Color := FMarkColor.Background;
        while not FHighLighter.EOL do
        begin
          TokenData := FHighLighter.GetTokenInfo;
          Font.Style := TokenData.Style;
          with TokenData do
            if eoShowControlChars in FOptions then PaintTokenWithControlChars(Position, Length)
                                              else PaintToken(Position, Length);
          FHighLighter.Next;
        end;

        Brush.Color := FMarkColor.Background;
        if ShowLineBreak then
        begin
          Font.Color := FMarkColor.Foreground;
          PaintLineBreak;
        end;
      end
      else
        // if no selection is to draw then go straight ahead
        if (LineIndex < BB.Y) or (LineIndex > BE.Y) or not ShowSelection then
        begin
          while not FHighLighter.EOL do
          begin
            TokenData := FHighLighter.GetTokenInfo;
            with TokenData do
            begin
              Font.Style := TokenData.Style;
              Font.Color := Foreground;
              Brush.Color := Background;

              if eoShowControlChars in FOptions then PaintTokenWithControlChars(Position, Length)
                                                else PaintToken(Position, Length);
            end;
            // early out if possible
            if PaintOffset >= UpdateRect.Right then Break;
            FHighLighter.Next;
          end;


          Brush.Color := Self.Color;
          if ShowLineBreak then
          begin
            Font.Color := Font.Color;
            PaintLineBreak;
          end;
        end
        else
        begin
          // selection parts are to be considered, so determine start and end position of
          // the selection on this particular line
          if BB.Y < LineIndex then SelStart := 0
                              else SelStart := ColumnToCharIndex(Point(BB.X, LineIndex));
          if BE.Y > LineIndex then SelEnd := FMaxRightPos // any large number makes it here
                              else SelEnd := ColumnToCharIndex(Point(BE.X, LineIndex));

          CurrentStyle := 0;
          StopIndex := 0;
          while not FHighLighter.EOL do
          begin
            // Get the current token (string) to be painted.
            TokenData := FHighLighter.GetTokenInfo;
            // set styles which do not change between selected and non-selected parts
            Font.Style := TokenData.Style;
            // start with the beginning of the token
            StartIndex := 0;
            // stop at the end of the token
            StopIndex := TokenData.Length;
            // Paint up to selection start. That length will be the StopIndex
            // or the start of the selection mark, whichever is less.
            TempIndex := Min(StopIndex + Integer(TokenData.Position), SelStart) - StartIndex - Integer(TokenData.Position);
            if TempIndex > 0 then
            begin
              // set needed styles
              with TokenData do
              begin
                Font.Color := Foreground;
                Brush.Color := Background;
                CurrentStyle := 1;
                if eoShowControlChars in FOptions then PaintTokenWithControlChars(Position, TempIndex)
                                                  else PaintToken(Position, TempIndex);
              end;
              // update the start index into the line text to skip past the
              // text we just painted.
              Inc(StartIndex, TempIndex);
            end;

            // Paint the selection text. That length will be the StopIndex or the
            // end of the selection mark, whichever is less.
            TempIndex := Min(StopIndex + Integer(TokenData.Position), SelEnd) - StartIndex - Integer(TokenData.Position);
            if TempIndex > 0 then // Have anything to paint?
            begin
              if CurrentStyle <> 2 then // other than selected style?
              begin
                // set the selection highlight colors
                with FSelectedColor do
                begin
                  Font.Color := Foreground;
                  Brush.Color := Background;
                end;
                CurrentStyle := 2;
              end;
              // Paint the selection text
              with TokenData do
                if eoShowControlChars in FOptions then PaintTokenWithControlChars(Position + Cardinal(StartIndex), TempIndex)
                                                  else PaintToken(Position + Cardinal(StartIndex), TempIndex);
              // Update the start index into the line text to skip past the
              // text we just painted.
              Inc(StartIndex, TempIndex);
            end;

            // Paint the post-selection text, the length is whatever is left over.
            TempIndex := StopIndex - StartIndex;
            if TempIndex > 0 then
            begin
              if CurrentStyle <> 1 then // other than unselected style?
              begin
                // set needed styles
                with TokenData do
                begin
                  Font.Color := Foreground;
                  Brush.Color := Background;
                end;
                CurrentStyle := 1;
              end;
              with TokenData do
                if eoShowControlChars in FOptions then PaintTokenWithControlChars(Position + Cardinal(StartIndex), TempIndex)
                                                  else PaintToken(Position + Cardinal(StartIndex), TempIndex);
            end;

            if CurrentStyle <> 2 then CurrentStyle := 0;
            
            // early out if possible
            if PaintOffset >= UpdateRect.Right then Break;

            FHighLighter.Next;
          end;

          // prepare drawing of the rest of the line
          if LineIndex < BE.Y then Brush.Color := FSelectedColor.Background
                              else Brush.Color := Self.Color;
          if ShowLineBreak then
          begin
            if StopIndex < Integer(SelEnd) then Font.Color := FSelectedColor.Foreground
                                           else Font.Color := Font.Color;
            PaintLineBreak;
          end;
        end;

      // clear background of the rest of the line
      FillRect(Rect(PaintOffset, YOffset, UpdateRect.Right, YOffset + FTextHeight));

      // draw the right margin marker
      Pen.Color := FMarginColor;
      Pen.Width := 1;
      MoveTo(FRightMargin * FCharWidth + FOffsetX + FGutterRect.Right + 1, YOffset);
      LineTo(FRightMargin * FCharWidth + FOffsetX + FGutterRect.Right + 1, YOffset + FTextHeight);

      // This line has been painted, increment the vertical offset and loop back
      // for the next line of text.
      Inc(YOffset, FTextHeight);

      // store state value to avoid later rescans
      if (LineIndex + 1) < FLines.Count then FLines.Objects[LineIndex + 1] := FHighLighter.GetRange;
    end;

    // finally erase all the space not covered by lines
    if YOffset < UpdateRect.Bottom then
    begin
      Brush.Color := Self.Color;
      FillRect(Rect(UpdateRect.Left, YOffset, UpdateRect.Right, UpdateRect.Bottom));

      // draw the right margin marker
      Pen.Color := FMarginColor;
      Pen.Width := 1;
      MoveTo(FRightMargin * FCharWidth + FOffsetX + FGutterRect.Right + 1, YOffset);
      LineTo(FRightMargin * FCharWidth + FOffsetX + FGutterRect.Right + 1, UpdateRect.Bottom);
    end;

    if EndLine > FLastValidLine then FLastValidLine := EndLine;
    TWideStringList(FLines).OnChange := LinesChanged;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.PaintText(TextCanvas: TCanvas);

// main drawing routine to paint the control if no syntax highlighting is used

var
  YOffset,
  LineIndex,
  PaintOffset,
  OldOffset,
  StartIndex,
  StopIndex,
  TempIndex: Integer;
  BB, BE: PPoint;
  TokenLen : Integer;

  StartLine,
  EndLine: Integer;
  UpdateRect,
  Dummy,
  R: TRect;
  SelStart,
  SelEnd: Integer;
  ShowSelection: Boolean;
  ShowLineBreak: Boolean;
  LineBase: PWideChar;

  //--------------- local functions -------------------------------------------

  procedure PaintToken(Start, Len: Cardinal);

  // paints the given WideString at current paint offset, no control chars are painted,
  // text attributes must be set up already on call;
  // side effects: current paint offset will be updated

  var
    Index,
    Counter: Cardinal;
    P: PWideChar;

  begin
    // replace non-printable characters,
    // calculate update rect for this token (-part) btw.
    P := LineBase + Start;
    OldOffset := PaintOffset;
    Index := Start;
    Counter := Len;
    while Counter > 0 do
    begin
      if P^ < ' ' then P^ := ' ';
      Inc(P);
      Inc(PaintOffset, FCharWidthArray[Index]);
      Inc(Index);
      Dec(Counter);
    end;

    R := Rect(OldOffset, YOffset, PaintOffset, YOffset + FTextHeight);
    // finally paint text
    ExtTextOutW(TextCanvas.Handle, OldOffset, YOffset, ETO_OPAQUE, @R, LineBase + Start, Len, @FCharWidthArray[Start]);
  end;

  //---------------------------------------------------------------------------

  procedure PaintTokenWithControlChars(Start, Len: Cardinal);

  // paints the given string at current paint offset, considers control characters,
  // text attributes must be set up already on call;
  // side effects: current paint offset will be updated

  var
    I,
    Counter: Cardinal;
    Run,
    TabPos: PWideChar;

  begin
    // replace space characters
    Run := LineBase + Start;
    Counter := Len;
    while Counter > 0 do
    begin
      if Run^ = ' ' then Run^ := DefaultSpaceGlyph;
      Inc(Run);
      Dec(Counter);
    end;

    Run := LineBase + Start;
    Counter := Len;
    with TextCanvas do
      while Counter > 0 do
      begin
        // find TAB character
        TabPos := StrScanW(Run, WideTabulator);
        // something to draw before tab?
        if TabPos <> nil then I := Min(Counter, TabPos - Run)
                         else I := Counter;
        if I > 0 then
        begin
          OldOffset := PaintOffset;
          Inc(PaintOffset, I * Cardinal(FCharWidth));
          R := Rect(OldOffset, YOffset, PaintOffset, YOffset + FTextHeight);
          ExtTextOutW(Handle, OldOffset, YOffset, ETO_OPAQUE, @R, Run, I, @FCharWidthArray[Start]);
          Inc(Start, I);
          Dec(Counter, I);
          Inc(Run, I);
        end;

        // tab to draw?
        if TabPos <> nil then
        begin
          while (TabPos^ = WideTabulator) and (Counter > 0) do
          begin
            OldOffset := PaintOffset;
            Inc(PaintOffset, FCharWidthArray[Start]);
            R := Rect(OldOffset, YOffset, PaintOffset, YOffset + FTextHeight);

            SetTextAlign(Handle, TA_CENTER);
            ExtTextOutW(Handle, (PaintOffset + OldOffset) div 2, YOffset, ETO_OPAQUE, @R, DefaultTabGlyph, 1, nil);
            SetTextAlign(Handle, TA_LEFT);
            Inc(Start);
            Dec(Counter);
            Inc(TabPos);
          end;
          Run := TabPos;
        end;
      end;
  end;

  //---------------------------------------------------------------------------

  procedure PaintLineBreak;

  begin
    R := Rect(PaintOffset, YOffset, PaintOffset + FCharWidth, YOffset + FTextHeight);
    if IntersectRect(Dummy, R, UpdateRect) then
      ExtTextOutW(TextCanvas.Handle, PaintOffset, YOffset, ETO_OPAQUE, @R, DefaultLineBreakGlyph, 1, Pointer(FCharWidthArray));
    Inc(PaintOffset, FCharWidth);
  end;

  //--------------- end local functions ---------------------------------------

begin
  with TextCanvas do
  begin
    Font := Self.Font;
    UpdateRect := ClipRect;
    // optimize drawing by using the update area
    StartLine := (UpdateRect.Top div FTextHeight) - FOffsetY;
    EndLine := (UpdateRect.Bottom div FTextHeight) - FOffsetY + 1;
    if EndLine > FLines.Count - 1 then EndLine := FLines.Count - 1;

    // The vertical position we will paint at. Start at top and increment by
    // FTextHeight for each iteration of the loop (see bottom of for loop).
    YOffset := (UpdateRect.Top div FTextHeight) * FTextHeight;

    // determine correct start and end line of selection
    BB := MinPointAsReference(FBlockBegin, FBlockEnd);
    BE := MaxPointAsReference(FBlockBegin, FBlockEnd);

    ShowLineBreak := (eoShowControlChars in FOptions) and not (eoScrollPastEOL in FOptions);

    // Determine in advance whether to show selection or not.
    // Note: We have four concurrent options determining when to show selection (apart
    //       from SelAvail). These are: focus state, eoReadOnly, eoReadOnlySelection and
    //       eoHideSelection. Given these states one can derive a logic table with 16 results
    //       and from this table get a logical equation. Optimizing this eq. leads to
    //       the four conditions written and described below. If we name the mentioned
    //       states (in the same order as above) with A, B, C and D, respectively, then
    //       the resulting eq. is: Sel = AB or AC or BD or CD.
    // show selection if...
    ShowSelection := // selection is available and
                     SelAvail and (
                       // edit is focused and not read only (then both selection opts have no influence) or
                       (Focused and not (eoReadOnly in FOptions)) or
                       // edit is focused and selection in r/o mode is allowed (then read only and
                       // hide selection when unfocused don't matter) or
                       (Focused and (eoReadOnlySelection in FOptions)) or
                       // edit is not read only and selection will not be hidden when unfocused (then
                       // being focused doesn't matter and r/o selection has no influence) or
                       ((FOptions * [eoReadOnly, eoHideSelection]) = []) or
                       // r/o selection is enabled and selection will not be hidden when unfocused
                       // (then being focused or in r/o mode don't matter)
                       ((eoReadOnlySelection in FOptions) and not (eoHideSelection in FOptions))
                     );

    // Loop from the top line in the window to the last line in the window.
    for LineIndex := StartLine to EndLine do
    begin
      PaintOffset := FOffsetX + FGutterRect.Right + 1;
      LineBase := PWideChar(FLines[LineIndex]);
      TokenLen := Length(FLines[LineIndex]);

      CalcCharWidths(LineIndex);

      // line marker to draw (overrides selection)?
      if (FMarkLine = LineIndex) or
         ((LineIndex < BB.Y) or (LineIndex > BE.Y) or not ShowSelection) then
      begin
        if FMarkLine = LineIndex then
        begin
          Font.Color := FMarkColor.Foreground;
          Brush.Color := FMarkColor.Background;
        end
        else
        begin
          Font.Color := Self.Font.Color;
          Brush.Color := Self.Color;
        end;

        if TokenLen > 0 then
          if eoShowControlChars in FOptions then PaintTokenWithControlChars(0, TokenLen)
                                            else PaintToken(0, TokenLen);

        if ShowLineBreak then PaintLineBreak;
      end
      else
      begin
        // selection parts are to be considered, so determine start and end position of
        // the selection on this particular line
        if BB.Y < LineIndex then SelStart := 0
                            else SelStart := ColumnToCharIndex(Point(BB.X, LineIndex));
        if BE.Y > LineIndex then SelEnd := FMaxRightPos // any large number makes it here
                            else SelEnd := ColumnToCharIndex(Point(BE.X, LineIndex));

        Font.Color := Font.Color;
        Brush.Color := Self.Color;
        // start with the beginning of the token
        StartIndex := 0;
        StopIndex := TokenLen;

        // Paint up to selection start. That length will be the StopIndex
        // or the start of the selection mark, whichever is less.
        TempIndex := Min(TokenLen, SelStart);
        if TempIndex > 0 then
        begin
          // set needed styles
          Font.Color := Self.Font.Color;
          Brush.Color := Self.Color;
          if eoShowControlChars in FOptions then PaintTokenWithControlChars(0, TempIndex)
                                            else PaintToken(0, TempIndex);
          // update the start index into the line text to skip past the
          // text we just painted.
          Inc(StartIndex, TempIndex);
        end;

        // Paint the selection text. That length will be the StopIndex or the
        // end of the selection mark, whichever is less.
        TempIndex := Min(StopIndex, SelEnd) - StartIndex;
        if TempIndex > 0 then // Have anything to paint?
        begin
          // set the selection highlight colors
          with FSelectedColor do
          begin
            Font.Color := Foreground;
            Brush.Color := Background;
          end;
          // paint the selection text
          if eoShowControlChars in FOptions then PaintTokenWithControlChars(StartIndex, TempIndex)
                                            else PaintToken(StartIndex, TempIndex);
          // Update the start index into the line text to skip past the
          // text we just painted.
          Inc(StartIndex, TempIndex);
        end;

        // Paint the post-selection text, the length is whatever is left over.
        TempIndex := StopIndex - StartIndex;
        if TempIndex > 0 then
        begin
            // set needed styles
          Font.Color := Self.Font.Color;
          Brush.Color := Self.Color;
          if eoShowControlChars in FOptions then PaintTokenWithControlChars(StartIndex, TempIndex)
                                            else PaintToken(StartIndex, TempIndex);
        end;

        // prepare drawing of the rest of the line
        if LineIndex < BE.Y then Brush.Color := FSelectedColor.Background
                            else Brush.Color := Self.Color;
        if ShowLineBreak then
        begin
          if StopIndex < Integer(SelEnd) then Font.Color := FSelectedColor.Foreground
                                         else Font.Color := Font.Color;
          PaintLineBreak;
        end;
      end;

      // clear background of the rest of the line
      FillRect(Rect(PaintOffset, YOffset, UpdateRect.Right, YOffset + FTextHeight));

      // draw the right margin marker
      Pen.Color := FMarginColor;
      Pen.Width := 1;
      MoveTo(FRightMargin * FCharWidth + FOffsetX + FGutterRect.Right + 1, YOffset);
      LineTo(FRightMargin * FCharWidth + FOffsetX + FGutterRect.Right + 1, YOffset + FTextHeight);

      // This line has been painted, increment the vertical offset and loop back
      // for the next line of text.
      Inc(YOffset, FTextHeight);
    end;

    // finally erase all the space not covered by lines
    if YOffset < UpdateRect.Bottom then
    begin
      Brush.Color := Self.Color;
      FillRect(Rect(UpdateRect.Left, YOffset, UpdateRect.Right, UpdateRect.Bottom));

      // draw the right margin marker
      Pen.Color := FMarginColor;
      Pen.Width := 1;
      MoveTo(FRightMargin * FCharWidth + FOffsetX + FGutterRect.Right + 1, YOffset);
      LineTo(FRightMargin * FCharWidth + FOffsetX + FGutterRect.Right + 1, UpdateRect.Bottom);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SelectAll;

var
  State: TEditState;
  Caret: TPoint;
  CX: Integer;

begin
  if FLines.Count > 0 then
  begin
    State := RecordState(False);
    SetBlockBegin(Point(0, 0));
    CX := GetLineEnd(FLines.Count - 1);
    SetBlockEnd(Point(CX, FLines.Count - 1));
    Caret := Point(CX, FLines.Count - 1);
    CaretXY := Caret;
    Invalidate;

    FUndoList.AddChange(ecrCursorMove, '', @State.Caret, @State.SelStart, @State.SelEnd,
                                       '', @Caret, @FBlockBegin, @FBlockEnd);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetBlockBegin(Value: TPoint);

begin
  if Value.X < 0 then Value.X := 0
                 else
    if Value.X > MaxRightChar then Value.X := MaxRightChar;
  if Value.Y < 0 then Value.Y := 0
                 else
    if Value.Y > FLines.Count then Value.Y := FLines.Count;
  if (FUpdateCount = 0) and SelAvail then InvalidateLines(FBlockBegin.Y, FBlockEnd.Y);
  FBlockBegin := Value;
  FBlockEnd := Value;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetBlockEnd(Value: TPoint);

begin
  if Value.X < 0 then Value.X := 0
                 else
    if Value.X > MaxRightChar then Value.X := MaxRightChar;
  if Value.Y < 0 then Value.Y := 0
                 else
    if Value.Y > FLines.Count then Value.Y := FLines.Count;

  if (FBlockEnd.X <> Value.X) or (FBlockEnd.Y <> Value.Y) then
  begin
    if FUpdateCount = 0 then InvalidateLines(Value.Y, FBlockEnd.Y);
    FBlockEnd := Value;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetCaretX(Value: Integer);

var
  I: Integer;

begin
  if Value < 0 then Value := 0;

  if FCaretX <> Value then
  begin
    I := GetLineEnd + 1;
    if eoScrollPastEOL in FOptions then
    begin
      if Value > MaxRightChar then Value := MaxRightChar;
    end
    else
    begin
      if Value > I then Value := I;
    end;
    FCaretX := Value;
    EnsureCursorPosVisible;
    if FUpdateCount = 0 then
    begin
      UpdateCaret;
      DoCaretChange;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetCaretY(Value: Integer);

var
  Len: Integer;

begin
  if (Value < 0) or (FLines.Count = 0) then Value := 0;
  if (FLines.Count > 0) and (Value > FLines.Count - 1) then Value := FLines.Count - 1;
  if FCaretY <> Value then
  begin
    // remove trailing blanks in line we just leaving
    if (FCaretY > -1) and
       (FCaretY < FLines.Count) and
       not (eoKeepTrailingBlanks in FOptions) then FLines[FCaretY] := WideTrimRight(FLines[FCaretY]);
    FCaretY := Value;

    if not (eoScrollPastEOL in FOptions) then
    begin
      Len := Length(LineText);
      if FCaretX > Len then
      begin
        CaretX := Len;
        if FCaretX < -FOffsetX * FCharWidth then SetOffsetX(FCaretX * FCharWidth);
        Exit;
      end;
    end;
    EnsureCursorPosVisible;
    if FUpdateCount = 0 then
    begin
      UpdateCaret;
      DoCaretChange;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetFont(const Value: TFont);

var
  SavePitch: TFontPitch;

begin
  SavePitch := Value.Pitch;
  Value.Pitch := fpFixed;
  inherited Font := Value;
  Value.Pitch := SavePitch;
  DoSettingChanged;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetGutterWidth(Value: Integer);

begin
  if Value <> FGutterWidth then
  begin
    if Value < 0 then Value := 0;
    FGutterWidth := Value;
    FGutterRect.Right := Value - 1;
    FCharsInWindow := (ClientWidth - FGutterRect.Right + 1) div FCharWidth;
    DoSettingChanged;
    UpdateCaret;
    UpdateScrollBars;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetOffsetX(Value: Integer);

var
  dX: Integer;
  R: TRect;

begin
  if -Value > (FMaxRightPos - ClientWidth + FGutterRect.Right) then Value := -(FMaxRightPos - ClientWidth + FGutterRect.Right);
  if Value > 0 then Value := 0;
  if FOffsetX <> Value then
  begin
    dX := Value - FOffsetX;
    FOffsetX := Value;
    R := Rect(FGutterRect.Right + 1, 0, ClientWidth, ClientHeight);
    ScrollWindow(Handle, dX, 0, nil, @R);
    if (CaretXPix + FCaretOffset.X) < (FGutterRect.Right + 1) then HideCaret
                                                              else ShowCaret;
    if FUpdateCount = 0 then
    begin
      UpdateWindow(Handle);
      UpdateScrollbars;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetOffsetY(Value: Integer);

// this offset is measured in lines not pixels

var
  dY: Integer;

begin
  // range check, order is important here
  if Value < (FLinesInWindow - FLines.Count) then Value := FLinesInWindow - FLines.Count;
  if Value > 0 then Value := 0;
  if FOffsetY <> Value then
  begin
    dY := FTextHeight * (Value - FOffsetY);
    FOffsetY := Value;
    ScrollWindow(Handle, 0, dY, nil, nil);
    if FUpdateCount = 0 then
    begin
      UpdateWindow(Handle);
      UpdateScrollbars;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetLines(Value: TWideStrings);

begin
  if HandleAllocated then
  begin
    FLines.Assign(Value);
    Invalidate;
    Update;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetLineText(Value: WideString);

begin
  if FLines.Count = 0 then
  begin
    FLines.Add('');
    if FHighlighter <> nil then
    begin
      FHighlighter.ResetRange;
      FLines.Objects[0] := FHighlighter.GetRange;
    end;
  end;
  if FCaretY < FLines.Count then FLines[FCaretY] := Value;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetScrollBars(const Value: TScrollStyle);

begin
  if (FScrollBars <> Value) then
  begin
    FScrollBars := Value;
    RecreateWnd;
    UpdateScrollBars;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.DoCutText(var Text: WideString): Boolean;

begin
  Result := True;
  if Assigned(FOnCutText) then FOnCutText(Self, Text, Result);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.DeleteSelection(NeedRescan: Boolean);

// Deletes all characters in the current selection and causes the highlighter
// to rescan modified lines if NeedRescan is True. Set this to False if you are going
// to insert new text at the current cursor position afterwards and start the rescan then.

var
  I,
  Count,
  Start: Integer;
  TempString,
  TempString2: WideString;
  BB, BE: TPoint;
  BeginX,
  EndX: Integer;

begin
  BB := MinPoint(FBlockBegin, FBlockEnd);
  BeginX := ColumnToCharIndex(BB);
  BE := MaxPoint(FBlockBegin, FBlockEnd);
  EndX := ColumnToCharIndex(BE);

  CaretX := BB.X;
  if BB.Y = BE.Y then
  begin
    TempString := Lines[BB.Y];
    Delete(TempString, BeginX + 1, EndX - BeginX);
    FLines[BB.Y] := TempString;
    TempString := '';
  end
  else
  begin
    CaretY := BB.Y;
    TempString := FLines[BB.Y];
    if Length(TempString) < BeginX then TempString := TempString + WideStringOfChar(' ', BeginX - Length(TempString))
                                   else Delete(TempString, BeginX + 1, Length(TempString));
    if BE.Y < FLines.Count then
    begin
      TempString2 := FLines[BE.Y];
      Delete(TempString2, 1, EndX);
      TempString := TempString + TempString2;
    end;
    Start := BB.Y;
    Count := BE.Y - BB.Y;
    for I := 0 to 0 do
      if FBookmarks[I].Y >= BE.Y then Dec(FBookmarks[I].Y, Count)
                                 else
        if FBookmarks[I].Y > BB.Y then FBookmarks[I].Y := BB.Y;

    try
      for I := Count + Start - 1 downto Start do FLines.Delete(I);
    finally
      // don't rescan if strings are empty now or something is to be inserted afterwards
      if (FHighlighter <> nil) and (Start > 0) and NeedRescan then ScanFrom(Start - 1);
    end;
    // if all lines have been removed then TempString can only be an empty WideString
    if Start < FLines.Count then FLines[Start] := TempString;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.InsertText(const Value: WideString);

var
  I,
  Len: Integer;
  TempString,
  TempString2: WideString;
  Helper: WideString;
  MaxChars: Integer;
  CX: Integer;
  Temp: TWideStringList;
  TempY: Integer;

begin
  if Value <> '' then
  begin
    Temp := TWideStringList.Create;
    try
      Temp.Text := Value + WideCR + WideLF;
      CX := ColumnToCharIndex(FCaretX);

      TempString := LineText;
      Len := CX - Length(TempString);

      if Temp.Count = 1 then
      begin
        if Len > 0 then TempString := TempString + WideStringOfChar(' ', Len);
        Insert(Temp[0], TempString, CX + 1);
        LineText := TempString;
        CaretX := CharIndexToColumn(CX + Length(Temp[0]));
      end
      else
      begin
        if Len > 0 then
        begin
          Helper := WideStringOfChar(' ', Len);
          TempString := TempString + Helper;
        end;
        TempString := Copy(TempString, 1, CX) + Temp[0];
        TempString2 := Copy(LineText, CX + 1, Length(LineText));
        MaxChars := FMaxRightPos div FCharWidth;
        if Length(TempString) > MaxChars then
        begin
          if DoCutText(TempString) then
          TempString := Copy(TempString, 1, MaxChars);
        end;
        FLines[FCaretY] := TempString;
        TempY := FCaretY + 1;
        for I := 1 to Temp.Count - 2 do
        begin
          TempString := Temp[I];
          if Length(TempString) > MaxChars then
          begin
            if DoCutText(TempString) then TempString := Copy(TempString, 1, MaxChars);
          end;
          FLines.Insert(TempY, TempString);
          Inc(TempY);
        end;
        // relocate bookmarks
        for I := 0 to 9 do
          if FBookmarks[I].Y >= CaretY then Inc(FBookmarks[I].Y, Temp.Count - 1);

        FLines.Insert(TempY, Temp[Temp.Count - 1] + TempString2);
        if FHighlighter <> nil then ScanFrom(FCaretY);
        CaretX := Length(Temp[Temp.Count - 1]);
        CaretY := TempY;
      end;
    finally
      Temp.Free;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetSelText(const Value: WideString);

begin
  FLines.BeginUpdate;
  InvalidateToBottom(Min(FBlockBegin.Y, FBlockEnd.Y));
  if SelAvail then DeleteSelection(Value = '');
  InsertText(Value);
  BlockBegin := CaretXY;
  FLines.EndUpdate;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetText(const Value: WideString);

begin
  FLines.Text := Value;
  Invalidate;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetTopLine(Value: Integer);

begin
  if Value < 1 then Value := 1;
  if TopLine <> Value then
  begin
    if Value > FLines.Count then Value := FLines.Count;
    SetOffsetY(1 - Value);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ShowCaret;

begin
  if not FCaretVisible then FCaretVisible := Windows.ShowCaret(Handle);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.OnScrollTimer(Sender: TObject);

// do autoscrolling while mouse down and set selection appropriately

var
  SP,
  CP: TPoint;
  NewCaret: TPoint;
  R: TRect;

begin
  Inc(FUpdateCount);
  // get current caret postion
  GetCursorPos(SP);
  GetWindowRect(Handle, R);
  InflateRect(R, -FTextHeight, -FTextHeight);
  CP := ScreenToClient(SP);
  if not PtInRect(R, SP) then
  begin
    // scroll only if mouse is outside the specified rectangle (which is a bit smaller
    // than the client area in order to allow to stop scrolling if the mouse is moved
    // over an other edit to drop text to)
    if SP.X < R.Left then SetOffsetX(FOffsetX + (R.Left - SP.X));
    if SP.X > R.Right then SetOffsetX(FOffsetX + (R.Right - SP.X));
    if SP.Y < R.Top then SetOffsetY(FOffsetY + (R.Top - SP.Y));
    if SP.Y > R.Bottom then SetOffsetY(FOffsetY + (R.Bottom - SP.Y));
  end;
  NewCaret := PositionFromPoint(CP.X, CP.Y);
  Dec(FUpdateCount);
  if not PointsAreEqual(CaretXY, NewCaret) then
  begin
    if not FDropTarget then SetBlockEnd(NewCaret);
    CaretXY := NewCaret;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.UpdateCaret;

var
  CX, CY: Integer;

begin
  CX := CaretXPix + FCaretOffset.X;
  CY := CaretYPix + FCaretOffset.Y;
  SetCaretPos(CX - 1, CY);
  if (CX < FGutterRect.Right + 1) or (CX > ClientWidth) or
     (CY < 0) or (CY > ClientHeight - FTextHeight) then HideCaret
                                                   else ShowCaret;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.UpdateScrollBars;

var
  ScrollInfo: TScrollInfo;
  Page: Integer;

begin
  if (FUpdateCount = 0) and (ClientWidth > 0) then
  begin
    ZeroMemory(@ScrollInfo, SizeOf(ScrollInfo));
    ScrollInfo.cbSize := SizeOf(ScrollInfo);
    ScrollInfo.fMask := SIF_ALL;
    ScrollInfo.nMin := 0;
    ScrollInfo.nTrackPos := 0;
    // vertical scrollbar
    if FScrollBars in [ssBoth, ssVertical] then
    begin
      if (FLinesInWindow > 0) and (FLines.Count > FLinesInWindow) then
      begin
        ScrollInfo.nPage := FLinesInWindow + 1;

        ScrollInfo.nMax := FLines.Count;
        // WM_VSCROLL can only handle values up to 65535,
        // consider this and leave some reserve
        if ScrollInfo.nMax > 65000 then
        begin
          FScrollFactor := ScrollInfo.nMax / 65000;
          ScrollInfo.nMax := 65000;
        end
        else FScrollFactor := 1;

        ScrollInfo.nPos := Round(-FOffsetY / FScrollFactor);
      end;
      FlatSB_SetScrollInfo(Handle, SB_VERT, ScrollInfo, True);
    end;

    if FScrollBars in [ssBoth, ssHorizontal] then
    begin
      ScrollInfo.nMax := FMaxRightPos;
      Page := ClientWidth - FGutterRect.Right + 1;
      if Page >= 0 then ScrollInfo.nPage := Page
                   else ScrollInfo.nPage := 0;
      ScrollInfo.nPos := -FOffsetX;
      FlatSB_SetScrollInfo(Handle, SB_HORZ, ScrollInfo, True);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WMEraseBkgnd(var Message: TMessage);

begin
  Message.Result := 1;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WMGetDlgCode(var Msg: TWMGetDlgCode);

begin
  Msg.Result := DLGC_WANTARROWS or DLGC_WANTCHARS;
  if eoWantTabs in FOptions then Msg.Result := Msg.Result or DLGC_WANTTAB;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WMHScroll(var Message: TWMScroll);

begin
  case Message.ScrollCode of
    SB_BOTTOM:
      SetOffsetX(-FMaxRightPos);
    SB_ENDSCROLL:
      ;
    SB_LINELEFT:
      if FOffsetX < 0 then SetOffsetX(FOffsetX + FCharWidth);
    SB_LINERIGHT:
      if -FOffsetX < (FMaxRightPos - ClientWidth + FGutterRect.Right) then SetOffsetX(FOffsetX - FCharWidth);
    SB_PAGELEFT:
      SetOffsetX(FOffsetX + ClientWidth - FGutterRect.Right + 1);
    SB_PAGERIGHT:
      SetOffsetX(FOffsetX - ClientWidth + FGutterRect.Right + 1);
    SB_THUMBPOSITION,
    SB_THUMBTRACK:
      SetOffsetX(-Message.Pos);
    SB_TOP:
      SetOffsetX(0);
  end;
  Message.Result := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WMKillFocus(var Message: TWMKillFocus);

begin
  inherited;
  HideCaret;
  DestroyCaret;
  if (eoHideSelection in FOptions) and SelAvail then Invalidate;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WMSetCursor(var Message: TWMSetCursor);

var
  P, ColPos: TPoint;

begin
  if csDesigning in ComponentState then inherited
                                   else
  begin
    GetCursorPos(P);
    P := ScreenToClient(P);
    if (Message.HitTest <> HTCLIENT) or
       PtInRect(FGutterRect, P) then Windows.SetCursor(Screen.Cursors[crArrow])
                                else
    begin
      ColPos := PositionFromPoint(P.X, P.Y);
      if ((eoAllowMarginResize in FOptions) and (ColPos.X = FRightMargin)) or
         ((eoAllowGutterResize in FOptions) and (Abs(P.X - FGutterWidth) < 3))
        then Windows.SetCursor(Screen.Cursors[crHSplit])
        else
          if SelAvail and PosInSelection(ColPos) then Windows.SetCursor(Screen.Cursors[crArrow])
                                                 else Windows.SetCursor(Screen.Cursors[Cursor]);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WMSetFocus(var Message: TWMSetFocus);

begin
  inherited;
  InitializeCaret;
  UpdateScrollBars;
  if (eoHideSelection in FOptions) and SelAvail then Invalidate;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WMSize(var Message: TWMSize);

begin
  inherited;
  if HandleAllocated then
  begin
    FCharsInWindow := (ClientWidth - FGutterRect.Right + 1) div FCharWidth;
    FLinesInWindow := ClientHeight  div FTextHeight;
    FGutterRect.Bottom := ClientHeight;
    if (FLines.Count - TopLine + 1) < FLinesInWindow then TopLine := FLines.Count - FLinesInWindow + 1;
    UpdateScrollBars;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WMVScroll(var Message: TWMScroll);

var
  LineString: WideString;
  HintRect: TRect;
  P: TPoint;
  ButtonH: Integer;

begin
  case Message.ScrollCode of
    SB_BOTTOM:
      SetOffsetY(-FLines.Count);
    SB_ENDSCROLL:
      FScrollHint.ReleaseHandle;
    SB_LINEUP:
      SetOffsetY(FOffsetY + 1);
    SB_LINEDOWN:
      SetOffsetY(FOffsetY - 1);
    SB_PAGELEFT:
      SetOffsetY(FOffsetY + FLinesInWindow);
    SB_PAGERIGHT:
      SetOffsetY(FOffsetY - FLinesInWindow);
    SB_THUMBPOSITION: // indicates end of thumbtrack operation, no need to handle this
      ;
    SB_THUMBTRACK:
      begin
        SetOffsetY(-Round(Word(Message.Pos) * FScrollFactor));
        if eoShowScrollHint in FOptions then
        begin
          LineString := IntToStr(-FOffsetY + 1);
          // calculate optimal size for scroll hint window
          HintRect := FScrollHint.CalcHintRect(2000, LineString, nil);
          // place it parallel to the thumb
          P := ClientToScreen(Point(0, 0));
          ButtonH := GetSystemMetrics(SM_CYVSCROLL)+ 8;
          OffsetRect(HintRect,
                     P.X + ClientWidth - HintRect.Right - 2,
                     ButtonH div 2 + P.Y + Round((ClientHeight - HintRect.Bottom - ButtonH) *
                                                 Word(Message.Pos) * FScrollFactor / FLines.Count));
          with FScrollHint do
          begin
            Font.Color := FScrollHintColor.FForeground;
            Color := FScrollHintColor.FBackground;
            ActivateHint(HintRect, LineString);
            UpdateWindow(Handle);
          end;
        end;
      end;
    SB_TOP:
      SetOffsetY(0);
  end;
  Message.Result := 0;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ScanFrom(Index: Integer);

var
  OldChange: TNotifyEvent;

begin
  if Index < FLines.Count then
  begin
    OldChange := TWideStringList(FLines).OnChange;
    TWideStringList(FLines).OnChange := nil;
    try
      if (Index = 0) and (FLines.Count > 0) then
      begin
        FHighLighter.ResetRange;
        FLines.Objects[0] := FHighlighter.GetRange;
      end;
      FLastValidLine := Index;
      InvalidateToBottom(Index);

    finally
      TWideStringList(FLines).OnChange := OldChange;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ListAdded(Sender: TObject);

begin
  if FHighLighter <> nil then
  begin
    if FLines.Count > 1 then
    begin
      FLines.Objects[FLines.Count - 1] := Pointer(-1);
    end
    else
    begin
      FHighLighter.ReSetRange;
      FLines.Objects[0] := FHighLighter.GetRange;
    end;
  end
  else
    // the highlighter may not be assigned yet
    FLines.Objects[FLines.Count - 1] := Pointer(-1);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ListDeleted(Index: Integer);

begin
  if FHighLighter <> nil then
  begin
    if FLines.Count > 1 then ScanFrom(Index)
                       else
      if FLines.Count = 1 then
      begin
        FHighLighter.ResetRange;
        FLines.Objects[0] := FHighLighter.GetRange;
      end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ListInserted(Index: Integer);

begin
  if FHighLighter <> nil then
  begin
    if FLines.Count > 1 then ScanFrom(Index)
                       else
      if FLines.Count = 1 then
      begin
        FHighLighter.ResetRange;
        FLines.Objects[0] := FHighLighter.GetRange;
      end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ListLoaded(Sender: TObject);

begin
  CaretXY := Point(0, 0);
  SetOffsetX(0);
  SetOffsetY(0);
  FUndoList.ClearList;
  if FHighlighter <> nil then ScanFrom(0);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ListPutted(Index: Integer);

begin
  if FHighLighter <> nil then ScanFrom(Index);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetWordBlock(Value: TPoint);

var
  BB, BE: TPoint;

begin
  GetWordAndBounds(Value, BB, BE);
  SetBlockBegin(BB);
  SetBlockEnd(BE);
  CaretXY := BE;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetCanUndo: Boolean;

begin
  Result := FUndoList.CanUndo;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetCanRedo: Boolean;

begin
  Result := FUndoList.CanRedo;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.Redo;

var
  Change: PChange;
  LastReason: TChangeReason;

  //--------------- local functions -------------------------------------------

  procedure RedoAction;

  // redo the action currently recorded in Change

  begin
    with Change^ do
      case Reason of
        ecrInsert:
          begin
            CaretXY := OldCaret;
            BlockBegin := CaretXY;
            SetSelText(NewText);
            CaretXY := NewCaret;
            BlockEnd := NewCaret;
          end;
        ecrDelete:
          begin
            SetBlockBegin(OldSelStart);
            SetBlockEnd(OldSelEnd);
            SetSelText('');
            CaretXY := NewCaret;
          end;
        ecrCursorMove:
          begin
            SetBlockBegin(NewSelStart);
            SetBlockEnd(NewSelEnd);
            CaretXY := NewCaret;
          end;
        ecrReplace:
          begin
            CaretXY := OldCaret;
            SetBlockBegin(OldSelStart);
            SetBlockEnd(OldSelEnd);
            SetSelText(NewText);
            CaretXY := NewCaret;
          end;
        ecrDragCopy,
        ecrDragMove:
          begin
            if Reason = ecrDragMove then
            begin
              SetBlockBegin(OldSelStart);
              SetBlockEnd(OldSelEnd);
              SetSelText('');
            end;
            
            CaretXY := MinPoint(NewSelStart, NewSelEnd);
            SetBlockBegin(NewSelStart);
            SetSelText(NewText);

            CaretXY := NewCaret;
            SetBlockBegin(NewSelStart);
            SetBlockEnd(NewSelEnd);
          end;
        ecrOverwrite:
          begin
            // overwriting text is always replacing it with one char 
            CaretXY := OldCaret;
            SetBlockBegin(OldCaret);
            SetBlockEnd(Point(OldCaret.X + 1, OldCaret.Y));
            SetSelText(NewText);
            SetBlockBegin(NewCaret);
            CaretXY := NewCaret;
          end;
        ecrIndentation:
          begin
            SetBlockBegin(Point(0, Min(OldSelStart.Y, OldSelEnd.Y)));
            SetBlockEnd(Point(0, Max(OldSelStart.Y, OldSelEnd.Y) + 1));
            SetSelText(NewText);
            CaretXY := NewCaret;
            SetBlockBegin(NewSelStart);
            SetBlockEnd(NewSelEnd);
          end;
      end;
  end;

  //--------------- end local functions ---------------------------------------

begin
  if CanRedo then
  begin
    BeginUpdate;
    repeat
      // pop last entry from undo stack and get its values
      LastReason := FUndoList.GetRedoChange(Change);
      RedoAction;
    until not (eoGroupUndo in FOptions) or not CanRedo or (LastReason <> FUndoList.GetCurrentRedoReason);
    EndUpdate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.Undo;

var
  Change: PChange;
  LastReason: TChangeReason;

  //--------------- local functions -------------------------------------------

  procedure UndoAction;

  // undo the action currently recorded in Change

  var I, J: Cardinal;
      T: TWideStringList;

  begin
    with Change^ do
      case Reason of
        ecrInsert:
          begin
            SetBlockBegin(NewSelStart);
            SetBlockEnd(NewSelEnd);
            SetSelText('');
            CaretXY := OldCaret;
          end;
        ecrDelete:
          begin
            CaretXY := OldCaret;
            SetBlockBegin(MinPoint(OldSelStart, OldSelEnd));
            SetSelText(OldText);
            SetBlockBegin(OldSelStart);
            SetBlockEnd(OldSelEnd);
          end;
        ecrCursorMove:
          begin
            CaretXY := OldCaret;
            BlockBegin := OldSelStart;
            BlockEnd := OldSelEnd;
          end;
        ecrReplace:
          begin
            CaretXY := OldCaret;
            SetBlockBegin(NewSelStart);
            SetBlockEnd(NewSelEnd);
            SetSelText(OldText);
            SetBlockBegin(OldSelStart);
            SetBlockEnd(OldSelEnd);
          end;
        ecrDragCopy,
        ecrDragMove:
          begin
            SetBlockBegin(NewSelStart);
            SetBlockEnd(NewSelEnd);
            SetSelText('');

            if Reason = ecrDragMove then
            begin
              CaretXY := MinPoint(OldSelStart, OldSelEnd);
              SetSelText(OldText);
            end;

            CaretXY := OldCaret;
            SetBlockBegin(OldSelStart);
            SetBlockEnd(OldSelEnd);
          end;
        ecrOverwrite:
          begin
            // overwriting text is always replacing one char by another
            CaretXY := OldCaret;
            SetBlockBegin(OldCaret);
            SetBlockEnd(Point(OldCaret.X + 1, OldCaret.Y));
            SetSelText(OldText);
            SetBlockBegin(OldCaret);
          end;
        ecrIndentation:
          begin
            // This is a special kind of replacement, as we set back entire lines.
            // With the consideration of some background information, it becomes
            // reasonably fast (no highlighter rescan necessary as we are modifying spaces
            // only, no partial line modifications).
            T := TWideStringList.Create;
            try
              T.Text := OldText;
              J := Min(NewSelStart.Y, NewSelEnd.Y);
              // restore old lines
              for I := 0 to T.Count - 1 do
                FLines[I + J] := T[I];
            finally
              T.Free;
            end;
            CaretXY := OldCaret;
            SetBlockBegin(OldSelStart);
            SetBlockEnd(OldSelEnd);
          end;
      end;
  end;

  //--------------- end local functions ---------------------------------------

begin
  if CanUndo then
  begin
    BeginUpdate;
    repeat
      // pop last entry from undo stack and get its values
      LastReason := FUndoList.GetUndoChange(Change);
      UndoAction;
    until not (eoGroupUndo in FOptions) or not CanUndo or (LastReason <> FUndoList.GetCurrentUndoReason);
    EndUpdate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ClearBookMark(BookMark: Integer);

var
  Allowed: Boolean;

begin
  if (BookMark in [0..9]) and FBookMarks[BookMark].Visible then
    with FBookMarks[BookMark] do
    begin
      Allowed := True;
      if Assigned(FOnBookmark) then FOnBookmark(Self, Bookmark, X, Y, Allowed);
      Visible := not Allowed;
      if Allowed then  
      begin
        InvalidateLine(Y);
        Y := -1;
      end;
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.GotoBookMark(BookMark: Integer);

begin
  if BookMark in [0..9] then
  begin
    CaretXY := Point(FBookMarks[BookMark].X, FBookMarks[BookMark].Y);
    EnsureCursorPosVisible;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetBookMark(BookMark: Integer; X, Y: Integer);

var
  Allowed: Boolean;

begin
  if (BookMark in [0..9]) and (Y <= FLines.Count) then
  begin
    Allowed := True;
    if Assigned(FOnBookmark) then FOnBookmark(Self, Bookmark, X, Y, Allowed);
    if Allowed then
    begin
      FBookMarks[BookMark].X := X;
      FBookMarks[BookMark].Y := Y;
      FBookMarks[BookMark].Visible := True;
      InvalidateLine(Y); 
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WndProc(var Msg: TMessage);

// prevent Alt-Backspace from beeping

begin
  if (Msg.Msg = WM_SYSCHAR) and
     (Msg.wParam = VK_BACK) and
     ((Msg.lParam and $20000000) <> 0) then
  begin
    Msg.Msg := 0;
  end
  else inherited;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.IsIdentChar(const AChar: WideChar; IdChars: TIdentChars): Boolean;

// Compatibility routine to check whether a given character is in the provided (SBCS) characters list.
// Works only if AChar is in the ANSI code page (value is < $100).
// Consider this function as being temporary until highlighters are ready which can deal with Unicode.

begin
  Result := Char(AChar) in IdChars;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetWordAndBounds(Pos: TPoint; var BB, BE: TPoint): WideString;

// Returns the word and its boundaries at position Pos. All positions are given
// as column/row values.
// Note: Compare Pos with BB and BE to determine whether the word is really at the
//       given position or has been automatically choosen because there was no word
//       at this position.

var
  CX: Integer;
  Line: WideString;
  IdChars: TIdentChars;
  Len: Integer;

begin
  // make sure the given position is valid
  if Pos.Y < 0 then Pos.Y := 0;
  if Pos.Y >= FLines.Count then Pos.Y := FLines.Count - 1;
  if Pos.X < 0 then Pos.X := 0;

  // determine WideString to be searched through
  if Pos.Y > - 1 then Line := FLines[Pos.Y]
                 else Line := '';
  Len := Length(Line);
  BB.Y := Pos.Y;
  BE.Y := Pos.Y;

  if Len > 0 then
  begin
    // setup initial states
    CX := ColumnToCharIndex(Pos);
    if CX > Len - 1 then CX := Len - 1;
    if FHighLighter <> nil then IdChars := FHighLighter.IdentChars
                           else IDchars := [#33..#255];

    // four cases are to be considered:
    // 1. IdChar at current position
    if IsIdentChar(Line[CX + 1], IdChars) then
    begin
      // find start of word
      BB.X := CX;
      while (BB.X > 0) and IsIdentChar(Line[BB.X], IdChars) do Dec(BB.X);
      // find end of word
      BE.X := CX + 1;
      while (BE.X < Len) and IsIdentChar(Line[BE.X + 1], IdChars) do Inc(BE.X);
      // copy word
      Result := Copy(Line, BB.X + 1, BE.X - BB.X);
    end
    else
    begin
      // 2. no ID char at current position, so search to the left
      BE.X := CX;
      while (BE.X > 0) and not IsIdentChar(Line[BE.X], IdChars) do Dec(BE.X);
      if BE.X > 0 then
      begin
        CX := BE.X;
        // find start of word
        while (CX > 0) and IsIdentChar(Line[CX], IdChars) do Dec(CX);
        BB.X := CX;
        Result := Copy(Line, BB.X + 1, BE.X - BB.X);
      end
      else
      begin
        // 3. no ID char found to the left, so search to the right
        BB.X := CX + 1;
        while (BB.X < Len) and not IsIdentChar(Line[BB.X + 1], IdChars) do Inc(BB.X);
        if BB.X < Len then
        begin
          // find end of word
          BE.X := BB.X + 1;
          while (BE.X < Len) and IsIdentChar(Line[BE.X + 1], IdChars) do Inc(BE.X);
          Result := Copy(Line, BB.X + 1, BE.X - BB.X);
        end
        else
        begin
          // 4. nothing found, return all we have
          BB.X := 0;
          BE.X := Len;
          Result := Line;
        end;
      end;
    end;
	
    // finally turn char indices into columns
    BB.X := CharIndexToColumn(BB);
    BE.X := CharIndexToColumn(BE);
  end
  else
  begin
    Result := '';
    BB.X := 0;
    BE.X := 0;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.WordAtPos(X, Y: Integer): WideString;

// returns the word at the given position, X and Y must be given in client coordinates,

var
  BB, BE: TPoint;

begin
  Result := GetWordAndBounds(PositionFromPoint(X, Y), BB, BE);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.DoStartDrag(var DragObject: TDragObject);

begin
  FDragWasStarted := False;
  inherited;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.DragCanceled;

begin
  FScrollTimer.Enabled := False;
  CaretXY := FLastCaret; // restore previous caret position
  if not FDragWasStarted then SetBlockBegin(FLastCaret);
  inherited;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.CopyOnDrop(Source: TObject): Boolean;

// determines wether a drag'n drop operation is a move or a copy

var
  Shift: TShiftState;
  
begin
  // determine move or copy operation
  Shift := KeyDataToShiftState(0);
  // w/o modifier keys the operation is determined source and target edit being the same or not
  if Shift = [] then Result := Source <> Self
                else Result := (ssCtrl in Shift) or (Source <> Self) and not (ssShift in Shift);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.DragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);

// drag operations are also allowed between different edits

begin
  FDragWasStarted := True;
  inherited;
  if Accept and (Source is TCustomSyntaxEdit) then
  begin
    case State of
      dsDragLeave:
        begin
          FDropTarget := False;
          FScrollTimer.Enabled := False;
          // give the timer time to process its pending message
          Application.ProcessMessages;
        end;
      dsDragEnter:
        begin
          // remember that we are drag target currently as the originator of the
          // drag operation might be another syntax edit instance
          FDropTarget := True;
          SetFocus;
          FScrollTimer.Enabled := True;
        end;
    end;

    // If source and target are identical then we cannot drop on selected text.
    // If this edit is not the source then we accept the operation and will
    // replace the selected text by the incoming one.
    if Source = Self then Accept := not PosInSelection(PositionFromPoint(X, Y))
                     else Accept := True;

    if Accept then
    begin
      // if Ctrl is pressed or Source is not this control then change
      // cursor to indicate copy instead of move
      if CopyOnDrop(Source) then TCustomSyntaxEdit(Source).DragCursor := crDragCopy
                            else TCustomSyntaxEdit(Source).DragCursor := crDragMove;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.DragDrop(Source: TObject; X, Y: Integer);

var
  DoMove,
  DoReplace: Boolean;
  NewCaret: TPoint;
  OldBB,
  OldBE: TPoint;
  OldText,
  NewText: WideString;
  ThisEditState,
  OtherEditState: TEditState;

begin
  if not (eoReadOnly in FOptions) then
  begin
    inherited;

    // determine move or copy operation
    DoMove := not CopyOnDrop(Source);
    ThisEditState := RecordState(True);

    if Source = Self then
    begin
      NewText := GetSelText;
      NewCaret := CaretXY;

      OldText := GetSelText;
      OldBB := MinPoint(FBlockBegin, FBlockEnd);
      OldBE := MaxPoint(FBlockBegin, FBlockEnd);

      if DoMove then
      begin
        if PosInSelection(NewCaret) then NewCaret := OldBB
                                    else
        begin
          // continue calculation with char indices
          NewCaret.X := ColumnToCharIndex(NewCaret);
          if (NewCaret.Y >= OldBB.Y) and (NewCaret.X >= OldBB.X) then
          begin
            // correct cursor position if the target pos is located after the selection start
            Dec(NewCaret.Y, OldBE.Y - OldBB.Y);
            if NewCaret.Y = OldBB.Y then Inc(NewCaret.X, OldBB.X);
            if NewCaret.Y = OldBE.Y then Dec(NewCaret.X, OldBE.X);
          end;
          // finally go back to columns
          NewCaret.X := CharIndexToColumn(NewCaret);
        end;
        // delete old text
        SetSelText('');
      end;

      // insert the text at new position...
      CaretXY := NewCaret;
      SetBlockBegin(NewCaret);
      SetSelText(NewText);

      // ...and mark it again as selected
      CaretXY := NewCaret;
      SetBlockBegin(NewCaret);
      SetBlockEnd(Point(NewCaret.X + OldBE.X - OldBB.X, NewCaret.Y + OldBE.Y - OldBB.Y));

      // finally register undo change
      if DoMove then FUndoList.AddChange(ecrDragMove, OldText, @FLastCaret, @OldBB, @OldBE)
                else FUndoList.AddChange(ecrDragCopy, NewText, @NewCaret, @OldBB, @OldBE);
      DoChange;
    end
    else
      if Source is TCustomSyntaxEdit then
      begin
        // drag'n drop operation between different edits
        NewText := TCustomSyntaxEdit(Source).GetSelText;
        NewCaret := CaretXY;
        DoReplace := PosInSelection(PositionFromPoint(X, Y));

        OldText := GetSelText;
        OldBB := MinPoint(FBlockBegin, FBlockEnd);
        OldBE := MaxPoint(FBlockBegin, FBlockEnd);

        // replace selected text here with selected text from other edit if the drag cursor
        // points to selected text, otherwise insert the new text
        if DoReplace then CaretXY := OldBB
                     else BlockBegin := CaretXY; // make sure nothing is selected anymore
        SetSelText(NewText);
        // if the user wants to move text then additionally delete the selected text in the other edit
        with Source as TCustomSyntaxEdit do
        begin
          if DoMove then
          begin
            // retrieve text, selection bounds and cursor position from other edit for undo registration
            OtherEditState := RecordState(True);
            CaretXY := MinPoint(FBlockBegin, FBlockEnd);
            SetSelText('');
            // register undo action in other edit
            with OtherEditState do
              FUndoList.AddChange(ecrDelete, Text, @FLastCaret, @SelStart, @SelEnd);

            DoChange;
          end;
        end;

        // mark the new text as selected
        if DoReplace then BlockBegin := OldBB
                     else BlockBegin := NewCaret;
        BlockEnd := CaretXY;
        // If there's a move operation then there was also a caret update in the code above which
        // set the caret for the source edit. We need now to tell Windows to set back the caret (which
        // is a shared ressource) to the correct coordinates in this edit.
        if DoMove then UpdateCaret;

        // finally register undo change in this edit
        with ThisEditState do
          FUndoList.AddChange(ecrInsert, '', @Caret, @SelStart, @SelEnd,
                                         NewText, @FBlockBegin, @FBlockBegin, @FBlockEnd);
        DoChange;
      end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetGutterColor(Value: TColor);

begin
  if FGutterColor <> Value then
  begin
    FGutterColor := Value;
    DoSettingChanged;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetRightMargin(Value: Integer);

begin
  if (FRightMargin <> Value) then
  begin
    FRightMargin := Value;
    DoSettingChanged;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetMaxUndo: Integer;

begin
  Result := FUndoList.MaxUndo;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetMaxUndo(const Value: Integer);

begin
  if Value > -1 then
  begin
    FUndoList.MaxUndo := Value;
    DoSettingChanged;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.Notification(AComponent: TComponent; Operation: TOperation);

begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FHighLighter) then
  begin
    FHighLighter := nil;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetHighlighter(const Value: TSyntaxHighlighter);

begin
  if Value <> FHighLighter then
  begin
    if FHighLighter <> nil then FHighLighter.WindowList.RemoveInstance(Self);
    FHighLighter := Value;
    if FHighLighter <> nil then
    begin
      Value.WindowList.AddInstance(Self);
      Value.FreeNotification(Self);
      ScanFrom(0);
    end
    else Invalidate;
    DoSettingChanged;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetBorderStyle(Value: TBorderStyle);

begin
  if FBorderStyle <> Value then
  begin
    FBorderStyle := Value;
    RecreateWnd;
    DoSettingChanged;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.InitializeCaret;

var
  ct: TCaretType;
  cw, ch: Integer;

begin
  if not (eoReadOnly in FOptions) or
     (eoShowCursorWhileReadOnly in FOptions) then
  begin
    // CreateCaret automatically destroys the previous one, so we don't have to
    // worry about cleaning up the old one here with DestroyCaret.
    // Ideally, we will have properties that control what these two carets look like.
    HideCaret;
    if eoInserting in FOptions  then ct := FInsertCaret
                                else ct := FOverwriteCaret;
    case ct of
      ctHorizontalLine:
        begin
          cw := FCharWidth;
          ch := 2;
          FCaretOffset := Point(0, FTextHeight-2);
        end;
      ctHalfBlock:
        begin
          cw := FCharWidth;
          ch := (FTextHeight - 2) div 2;
          FCaretOffset := Point(0, ch);
        end;
      ctBlock:
        begin
          cw := FCharWidth;
          ch := FTextHeight - 2;
          FCaretOffset := Point(0, 0);
        end;
    else // ctVerticalLine
      cw := 2;
      ch := FTextHeight - 2;
      FCaretOffset := Point(0, 0);
    end;
    CreateCaret(Handle, 0, cw, ch);
    UpdateCaret;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetInsertCaret(const Value: TCaretType);

begin
  if FInsertCaret <> Value then
  begin
    FInsertCaret := Value;
    InitializeCaret;
    DoSettingChanged;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetOverwriteCaret(const Value: TCaretType);

begin
  if FOverwriteCaret <> Value then
  begin
    FOverwriteCaret := Value;
    InitializeCaret;
    DoSettingChanged;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetCaretXY: TPoint;

begin
  Result := Point(FCaretX, FCaretY);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetCaretXY(const Value: TPoint);

begin
  // set Y first because ScrollPastEOL can effect the X pos, prevent SetCaretX/Y
  // from sending OnChange events
  Inc(FUpdateCount);
  CaretY := Value.Y;
  CaretX := Value.X;
  Dec(FUpdateCount);
  if FUpdateCount = 0 then
  begin
    UpdateCaret;
    UpdateScrollBars;
    UpdateWindow(Handle);
    DoCaretChange;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.DoCaretChange;

begin
  {// try setting the correct keyboard layout
  if not (csLButtonDown in ControlState) and
     (FCaretY < FLines.Count) and
     (FCaretX > 0) and
     (FCaretX < Length(FLines[FCaretY])) then
    ActivateKeyboard(FLines[FCaretY][FCaretX]);}
  if Assigned(FOnCaretChange) then FOnCaretChange(Self, FCaretX, FCaretY);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetMaxRightChar: Integer;

begin
  Result := FMaxRightPos div FCharWidth;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetMarginColor(const Value: TColor);

begin
  if FMarginColor <> Value then
  begin
    FMarginColor := Value;
    DoSettingChanged;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetMaxRightChar(const Value: Integer);

var
  I: Integer;

begin
  if FMaxRightPos <> (Value  * FCharWidth) then
  begin
    FMaxRightPos := Value * FCharWidth;
    SetLength(FCharWidthArray, Value);
    for I := 0 to Value - 1 do FCharWidthArray[I] := FCharWidth;
    if HandleAllocated then
    begin
      DoSettingChanged;
      Invalidate;
      UpdateScrollBars;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetCharWidth(const Value: Integer);

var
  I,
  RightChar: Integer;

begin
  if FCharWidth <> Value then
  begin
    RightChar := FMaxRightPos div FCharWidth;
    FCharWidth := Value;
    MaxRightChar := RightChar;
    for I := 0 to (FMaxRightPos div FCharWidth) - 1 do FCharWidthArray[I] := FCharWidth;
    DoSettingChanged;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.EnsureCursorPosVisible;

begin
  // make sure X is visible
  if FCaretX < (-FOffsetX div FCharWidth) then SetOffsetX(-FCaretX * FCharWidth)
                                          else
    if FCaretX >= (FCharsInWindow - FOffsetX div FCharWidth) then SetOffsetX(-(FCaretX - FCharsInWindow) * FCharWidth);

  // make sure Y is visible
  if FCaretY < (TopLine - 1) then TopLine := FCaretY + 1
                             else
    if FCaretY > (TopLine + (LinesInWindow - 2)) then TopLine := FCaretY - (LinesInWindow - 2);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetKeystrokes(const Value: TKeyStrokes);

begin
  if Value = nil then FKeyStrokes.Clear
                 else FKeyStrokes.Assign(Value);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetDefaultKeystrokes;

begin
  FKeyStrokes.ResetDefaults;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.TranslateKeyCode(Code: word; Shift: TShiftState; var Data: Pointer): TEditorCommand;

// If the translations requires Data, memory will be allocated for it via a
// GetMem call. The client must call FreeMem on Data if it is not nil.

var
  I: Integer;

begin
  I := Keystrokes.FindKeycode(Code, Shift);
  if I >= 0 then Result := Keystrokes[I].Command
            else Result := ecNone;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.CommandProcessor(Command: TEditorCommand; AChar: WideChar; Data: Pointer);

var
  CX: Integer;
  Len: Integer;
  Temp: WideString;
  Helper: WideString;
  OldOptions: TSyntaxEditOptions;
  Action: Boolean;
  WP: TPoint;
  Caret: TPoint;
  I, J: Integer;
  OldChar: WideChar;
  BB, BE: TPoint;
  LastState: TEditState;

  //--------------- local functions -------------------------------------------

  procedure RegisterCursorMoveChange;

  // helper function to ease the repetitive registration of the cursor move undo change

  var
    Caret: TPoint;

  begin
    if not PointsAreEqual(CaretXY, LastState.Caret) then
    begin
      Caret := CaretXY;
      FUndoList.AddChange(ecrCursorMove, '', @LastState.Caret, @LastState.SelStart, @LastState.SelEnd,
                                         '', @Caret, @FBlockBegin, @FBlockEnd);
    end;
  end;

  //---------------------------------------------------------------------------

  procedure RegisterDeleteChange;

  // helper function to ease the repetitive registration of a delete undo change

  begin
    with LastState do
      FUndoList.AddChange(ecrDelete, Text, @Caret, @SelStart, @SelEnd);
  end;

  //---------------------------------------------------------------------------

  procedure RegisterReplaceChange(NewText: WideString; StartBound, EndBound: TPoint);

  // helper function to ease the repetitive registration of a replace undo change
  // StartBound and EndBound are just to ease making undo and do not correspond
  // to a selection

  var
    Caret: TPoint;

  begin
    Caret := CaretXY;
    FUndoList.AddChange(ecrReplace, LastState.Text, @LastState.Caret, @LastState.SelStart, @LastState.SelEnd,
                                    NewText, @Caret, @StartBound, @EndBound);
  end;

  //---------------------------------------------------------------------------

  procedure RegisterInsertChange(NewText: WideString; StartBound, EndBound: TPoint);

  // helper function to ease the repetitive registration of an insert undo change,
  // StartBound and EndBound are just to ease making undo and do not correspond
  // to a selection

  var
    Caret: TPoint;

  begin
    Caret := CaretXY;
    FUndoList.AddChange(ecrInsert, '', @LastState.Caret, @LastState.SelStart, @LastState.SelEnd,
                                   NewText, @Caret, @StartBound, @EndBound);
  end;

  //---------------------------------------------------------------------------

  procedure RegisterOverwriteChange(NewChar: WideChar);

  var
    Caret: TPoint;

  begin
    Caret := CaretXY;
    FUndoList.AddChange(ecrOverwrite, OldChar, @LastState.Caret, nil, nil,
                                      NewChar, @Caret, nil, nil);
  end;

  //--------------- end local functions ---------------------------------------

begin
  ProcessCommand(Command, AChar, Data);
  if (Command = ecNone) or (Command >= ecUserFirst) then Exit;

  case Command of
    ecLeft,
    ecSelLeft:
      begin
        LastState := RecordState(False);
        CX := PreviousCharPos(FCaretX);

        if not (eoScrollPastEOL in FOptions) and (FCaretX = 0) then
        begin
          if FCaretY > 0 then
          begin
            CaretY := FCaretY - 1;
            CaretX := GetLineEnd;
          end;
        end
        else CaretX := CX;

        if Command = ecSelLeft then
        begin
          if not SelAvail then SetBlockBegin(LastState.Caret);
          SetBlockEnd(CaretXY);
        end
        else SetBlockBegin(CaretXY);

        RegisterCursorMoveChange;
      end;

    ecRight,
    ecSelRight:
      begin
        LastState := RecordState(False);
        CX := NextCharPos(FCaretX);

        if not (eoScrollPastEOL in FOptions) and (Cardinal(CX) > GetLineEnd) then
        begin
          if FCaretY < FLines.Count then
          begin
            CaretX := 0;
            CaretY := FCaretY + 1;
          end;
        end
        else CaretX := CX;

        if Command = ecSelRight then
        begin
          if not SelAvail then SetBlockBegin(LastState.Caret);
          SetBlockEnd(CaretXY);
        end
        else SetBlockBegin(CaretXY);

        RegisterCursorMoveChange;
      end;

    ecUp,
    ecSelUp:
      begin
        LastState := RecordState(False);
        CaretY := FCaretY - 1;
        if (Command = ecSelUp) then
        begin
          if not SelAvail then SetBlockBegin(Point(FCaretX, FCaretY + 1));
          SetBlockEnd(CaretXY);
        end
        else SetBlockBegin(CaretXY);

        RegisterCursorMoveChange;
      end;

    ecDown,
    ecSelDown:
      begin
        LastState := RecordState(False);
        CaretY := FCaretY + 1;
        if Command = ecSelDown then
        begin
          if not SelAvail then SetBlockBegin(Point(FCaretX, FCaretY - 1));
          SetBlockEnd(CaretXY);
        end
        else SetBlockBegin(CaretXY);

        RegisterCursorMoveChange;
      end;

    ecWordLeft,
    ecSelWordLeft:
      begin
        LastState := RecordState(False);
        if (Command = ecSelWordLeft) and (not SelAvail) then SetBlockBegin(CaretXY);
        CaretXY := LastWordPos;
        if Command = ecSelWordLeft then SetBlockEnd(CaretXY)
                                   else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecWordRight,
    ecSelWordRight:
      begin
        LastState := RecordState(False);
        if (Command = ecSelWordRight) and (not SelAvail) then SetBlockBegin(CaretXY);
        CaretXY := NextWordPos;
        if Command = ecSelWordRight then SetBlockEnd(CaretXY)
                                    else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecLineStart,
    ecSelLineStart:
      begin
        LastState := RecordState(False);
        if (Command = ecSelLineStart) and (not SelAvail) then SetBlockBegin(CaretXY);
        CaretX := 0;
        if Command = ecSelLineStart then SetBlockEnd(CaretXY)
                                    else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecLineEnd,
    ecSelLineEnd:
     begin
        LastState := RecordState(False);
        if (Command = ecSelLineEnd) and (not SelAvail) then SetBlockBegin(CaretXY);
        if eoKeepTrailingBlanks in FOptions then Helper := LineText
                                            else Helper := WideTrimRight(LineText);
        if Helper <> LineText then LineText := Helper;
        CaretX := GetLineEnd(FCaretY);
        if Command = ecSelLineEnd then SetBlockEnd(CaretXY)
                                  else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecPageUp,
    ecSelPageUp:
      begin
        LastState := RecordState(False);
        if (Command = ecSelPageUp) and (not SelAvail) then SetBlockBegin(CaretXY);
        Caret := Point(FCaretX, FCaretY - FLinesInWindow);
        if Command = ecSelPageUp then SetBlockEnd(Caret)
                                 else SetBlockBegin(Caret);
        TopLine := TopLine - LinesInWindow;
        CaretXY := Caret;
        RegisterCursorMoveChange;
      end;

    ecPageDown,
    ecSelPageDown:
      begin
        LastState := RecordState(False);
        if (Command = ecSelPageDown) and (not SelAvail) then SetBlockBegin(CaretXY);
        Caret := Point(FCaretX, FCaretY + FLinesInWindow);
        if (Command = ecSelPageDown) then SetBlockEnd(Caret)
                                     else SetBlockBegin(Caret);
        TopLine := TopLine + LinesInWindow;
        CaretXY := Caret;
        RegisterCursorMoveChange;
      end;

    ecPageLeft,
    ecSelPageLeft:
      begin
        LastState := RecordState(False);
        if (Command = ecSelPageLeft) and (not SelAvail) then SetBlockBegin(CaretXY);
        SetOffsetX(FOffsetX + CharsInWindow * FCharWidth);
        CaretX := FCaretX - CharsInWindow;
        if (Command = ecSelPageLeft) then SetBlockEnd(CaretXY)
                                     else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecPageRight,
    ecSelPageRight:
      begin
        LastState := RecordState(False);
        if (Command = ecSelPageRight) and (not SelAvail) then SetBlockBegin(CaretXY);
        SetOffsetX(FOffsetX - CharsInWindow * FCharWidth);
        CaretX := FCaretX + CharsInWindow;
        if (Command = ecSelPageRight) then SetBlockEnd(CaretXY)
                                      else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecPageTop,
    ecSelPageTop:
      begin
        LastState := RecordState(False);
        if (Command = ecSelPageTop) and (not SelAvail) then SetBlockBegin(CaretXY);
        CaretY := TopLine - 1;
        if Command = ecSelPageTop then SetBlockEnd(CaretXY)
                                  else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecPageBottom,
    ecSelPageBottom:
      begin
        LastState := RecordState(False);
        if (Command = ecSelPageBottom) and (not SelAvail) then SetBlockBegin(CaretXY);
        CaretY := TopLine + LinesInWindow - 2;
        if (Command = ecSelPageBottom) then SetBlockEnd(CaretXY)
                                       else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecEditorTop,
    ecSelEditorTop:
      begin
        LastState := RecordState(False);
        if (Command = ecSelEditorTop) and (not SelAvail) then SetBlockBegin(CaretXY);
        CaretX := 0;
        CaretY := 0;
        if Command = ecSelEditorTop then SetBlockEnd(CaretXY)
                                    else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecEditorBottom,
    ecSelEditorBottom:
      begin
        LastState := RecordState(False);
        if (Command = ecSelEditorBottom) and (not SelAvail) then SetBlockBegin(CaretXY);
        CaretY := FLines.Count - 1;
        SetOffsetY(FLinesInWindow - FLines.Count);
        CaretX := Length(FLines[FLines.Count - 1]);
        SetOffsetX(0);
        if Command = ecSelEditorBottom then SetBlockEnd(CaretXY)
                                       else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecGotoXY,
    ecSelGotoXY:
      if Data <> nil then
      begin
        LastState := RecordState(False);
        if (Command = ecSelGotoXY) and (not SelAvail) then SetBlockBegin(CaretXY);
        CaretXY := PPoint(Data)^;
        if Command = ecSelGotoXY then SetBlockEnd(CaretXY)
                                 else SetBlockBegin(CaretXY);
        RegisterCursorMoveChange;
      end;

    ecSelectAll:
      SelectAll;

    ecDeleteLastChar:
      if not (eoReadOnly in FOptions) then
      begin
        LastState := RecordState(True);
        if SelAvail then
        begin
          SetSelText('');
          RegisterDeleteChange;
        end
        else
        begin
          Temp := LineText;
          Len := Length(Temp);
          // get char index from column
          CX := ColumnToCharIndex(FCaretX);
          if CX <= Len then
          begin
            if CX > 0 then
            begin
              // current index is somewhere on the line, first check whether we
              // can do auto unindentation
              I := CX - 1;
              while (I > -1) and UnicodeIsWhiteSpace(Word(Temp[I + 1])) do Dec(I);
              if (eoAutoUnindent in FOptions) and (I = -1) then
              begin
                // fine, all conditions met to do auto unindentation
                I := Unindent(FCaretX, FCaretY);
                J := ColumnToCharIndex(I);
                Helper := Copy(Temp, J + 1, CX - J);
                Delete(Temp, J + 1, CX - J);
                LineText := Temp;
                CaretX := I;
                LastState.Text := Helper;
                RegisterDeleteChange;
              end
              else
              begin
                // no unindentation, just delete last character;
                // need to test for non-default width character though
                if CharIndexToColumn(CX) = Cardinal(FCaretX) then
                begin
                  Helper := Temp[CX];
                  Delete(Temp, CX, 1);
                  CaretX := PreviousCharPos(FCaretX, FCaretY);
                end
                else
                begin
                  Helper := Temp[CX + 1];
                  Delete(Temp, CX + 1, 1);
                  CaretX := CharIndexToColumn(CX);
                end;
                LineText := Temp;
                LastState.Text := Helper;
                RegisterDeleteChange;
              end;
            end
            else
            begin
              // character index is at line start, but the first char might be a non-default
              // width char and the current column is somewhere inbetween -> check this
              if FCaretX > 0 then
              begin
                // yes, the first char is of non-default width, so delete this char
                // instead of moving the line one step up
                LastState.Text := Copy(Temp, 1, 1);
                Delete(Temp, 1, 1);
                LineText := Temp;
                CaretX := 0;
                RegisterDeleteChange;
              end
              else
                // no, first char is just an ordinary one, so move this line one step up
                if FCaretY > 0 then
                begin
                  FLines.Delete(FCaretY);
                  // need to relocate bookmarks
                  for I := 0 to 9 do
                    if FBookmarks[I].Y >= FCaretY then Dec(FBookmarks[I].Y);
                  CaretY := CaretY - 1;
                  CaretX := GetLineEnd(FCaretY);
                  LineText := LineText + Temp;
                  LastState.Text := WideCRLF;
                  RegisterDeleteChange;
                  InvalidateToBottom(FCaretY);
                end;
            end;
            DoChange;
          end
          else
          begin
            // caret is beyond the end of the current line -> just do a cursor move
            CaretX := CharIndexToColumn(CX - 1);
            RegisterCursorMoveChange;
          end;
        end;
      end;

    ecDeleteChar:
      if not (eoReadOnly in FOptions) then
      begin
        LastState := RecordState(True);
        if SelAvail then
        begin
          SetSelText('');
          SetBlockBegin(CaretXY);
          RegisterDeleteChange;
        end
        else
        begin
          Temp := LineText;
          Len := Length(Temp);
          CX := ColumnToCharIndex(FCaretX);
          if Len > CX then
          begin
            LastState.Text := Temp[CX + 1];
            Delete(Temp, CX + 1, 1);
            LineText := Temp;
            // need a caret change here if it was place on a non-default width char
            CaretX := CharIndexToColumn(CX);
            RegisterDeleteChange;
          end
          else
          begin
            if FCaretY < FLines.Count - 1 then
            begin
              Helper := WideStringOfChar(' ', CX - Len);
              LineText := Temp + Helper + FLines[FCaretY + 1];
              FLines.Delete(FCaretY + 1);
              // need to relocate bookmarks
              for I := 0 to 9 do
                if FBookmarks[I].Y >= FCaretY then Dec(FBookmarks[I].Y);
              InvalidateToBottom(FCaretY);
              LastState.Text := WideCRLF;
              RegisterDeleteChange;
            end;
          end;
        end;
        DoChange;
      end;

    ecDeleteWord:
      if not (eoReadOnly in FOptions) then
      begin
        LastState := RecordState(True);
        WP := NextWordPos;
        if (FCaretX <> WP.X) or (FCaretY <> WP.Y) then
        begin
          SetBlockBegin(CaretXY);
          SetBlockEnd(WP);
          LastState.Text := SelText;
          SetSelText('');
          CaretXY := CaretXY; // fix up cursor position
          RegisterDeleteChange;
          DoChange;
        end;
      end;

    ecDeleteLastWord:
      if not (eoReadOnly in FOptions) then
      begin
        LastState := RecordState(True);
        WP := LastWordPos;
        if (FCaretX <> WP.X) or (CaretY <> WP.Y) then
        begin
          SetBlockBegin(WP);
          SetBlockEnd(CaretXY);
          LastState.Text := SelText;
          SetSelText('');
          CaretXY := CaretXY; // fix up cursor position
          RegisterDeleteChange;
          DoChange;
        end;
      end;

    ecDeleteBOL:
      if not (eoReadOnly in FOptions) then
      begin
        LastState := RecordState(True);
        LastState.Text := Copy(LineText, 1, ColumnToCharIndex(FCaretX));
        SetBlockBegin(Point(0, CaretY));
        SetBlockEnd(CaretXY);
        SetSelText('');
        CaretXY := Point(0, CaretY); // fix up cursor position
        RegisterDeleteChange;
        DoChange;
      end;

    ecDeleteEOL:
      if not (eoReadOnly in FOptions) then
      begin
        LastState := RecordState(True);
        CX := ColumnToCharIndex(FCaretX);
        LastState.Text := Copy(LineText, CX + 1, Length(LineText) - CX);
        SetBlockBegin(CaretXY);
        SetBlockEnd(Point(GetLineEnd, CaretY));
        SetSelText('');
        CaretXY := CaretXY; // fix up cursor position
        RegisterDeleteChange;
        DoChange;
      end;

    ecDeleteLine:
      if not (eoReadOnly in FOptions) then
      begin
        LastState := RecordState(True);
        CaretX := 0;
        if SelAvail then SetBlockBegin(CaretXY);
        LastState.Text := PWideChar(LineText + WideCRLF);
        FLines.Delete(FCaretY);
        RegisterDeleteChange;
        DoChange;
      end;

    ecClearAll:
      if not (eoReadOnly in FOptions) then
      begin
        ClearAll(True, True);
        DoChange;
      end;

    ecInsertLine,
    ecLineBreak:
      if not (eoReadOnly in FOptions) then
      begin
        LastState := RecordState(True);
        if SelAvail then
        begin
          BB := MinPoint(FBlockBegin, FBlockEnd);
          Caret := BB;
          Temp := '';
          if not (eoKeepTrailingBlanks in FOptions) then
          begin
            // include trailing blanks (which would otherwise appear and automatically
            // deleted after the line break has been inserted) into the text being
            // replaced, to be able to restore them while doing undo
            CX := ColumnToCharIndex(BB);
            // keep char index, we need it later
            I := CX;
            Helper := FLines[BB.Y];
            // extent the helper WideString to match the caret position
            if CX > Length(Helper) then Helper := Helper + WideStringOfChar(' ', CX - Length(Helper));
            Dec(CX);
            while (CX > -1) and UnicodeIsWhiteSpace(Word(Helper[CX + 1])) do Dec(CX);
            Inc(CX);
            BB.X := CharIndexToColumn(Point(CX, BB.Y));
            // copy white spaces potentially to be restored
            if CX < I then Temp := Copy(Helper, CX + 1, I - CX);
          end;
          SetSelText(WideCRLF);
          SetBlockBegin(CaretXY);
          // save additional text to be restored
          LastState.Text := Temp + LastState.Text;
          BE := CaretXY;
          // restore caret postion if we deal with a line insertion
          if Command = ecInsertLine then CaretXY := Caret;
          RegisterReplaceChange(WideCRLF, BB, BE);
        end
        else
        begin
          TWideStringList(FLines).OnChange := nil;
          Temp := LineText;
          CX := ColumnToCharIndex(FCaretX);
          LineText := Copy(Temp, 1, CX);
          BB := Point(FCaretX, FCaretY);
          Delete(Temp, 1, CX);
          InvalidateToBottom(FCaretY);
          if eoAutoIndent in FOptions then Caret := Point(NextSmartTab(0, FCaretY + 1, False), FCaretY + 1)
                                      else Caret := Point(0, FCaretY + 1);
          // the just calculated column corresponds directly to the characters we
          // have to insert, hence no conversion to a char index is needed
          Helper := WideStringOfChar(' ', Caret.X);
          Insert(Helper, Temp, 1);
          FLines.Insert(Caret.Y, Temp);
          if Command = ecLineBreak then CaretXY := Caret;

          RegisterInsertChange(WideCRLF + Helper, BB, Caret);
          TWideStringList(FLines).OnChange := LinesChanged;

          // need to relocate bookmarks
          for I := 0 to 9 do
            if FBookmarks[I].Y >= FCaretY then Inc(FBookmarks[I].Y);
        end;
        DoChange;
      end;

    ecChar:
      // #127 is Ctrl + Backspace
      if ((AChar = WideTabulator) or (AChar >= WideSpace)) and (AChar <> #127) and not (eoReadOnly in FOptions) then
      begin
        LastState := RecordState(True);
        if SelAvail then
        begin
          // Special case TAB -> it does not delete the selection, but enhances
          // it to the next smart tab pos and moves following characters accordingly
          // if the editor is using smart tabs currently
          // (that's the way the Delphi IDE does it, so I'll do it too).
          if (AChar = WideTabulator) and (eoSmartTabs in FOptions)  then
          begin
            CX := NextSmartTab(FCaretX, FCaretY, True);
            I := ColumnToCharIndex(Point(FCaretX, FCaretY));
            Temp := LineText;
            Insert(WideStringOfChar(' ', CX - FCaretX), Temp, I + 1);
            LineText := Temp;
            // adjust selection
            BB := MinPoint(FBlockBegin, FBLockEnd);
            BE := MaxPoint(FBlockBegin, FBLockEnd);
            if (FCaretY = BB.Y) and (FCaretX <= BB.X) then
            begin
              SetBlockBegin(Point(BB.X + CX - FCaretX, FCaretY));
              if FCaretY = BE.Y then SetBlockEnd(Point(BE.X + CX - FCaretX, FCaretY))
                                else SetBlockEnd(BE);
            end
            else
              if (FCaretY = BE.Y) and (FCaretX >= BE.X) then SetBlockEnd(Point(CX, FCaretY));
            CaretX := CX;
            RegisterReplaceChange(SelText, FBlockBegin, FBlockEnd);
            DoChange;
          end
          else
          begin
            BB := MinPoint(FBlockBegin, FBlockEnd);
            if (AChar = WideTabulator) and not (eoUseTabs in FOptions) then Temp := WideStringOfChar(' ', FTabSize)
                                                                       else Temp := AChar;
            SetSelText(Temp);
            CaretX := NextCharPos(BB.X, BB.Y);
            SetBlockBegin(CaretXY);
            RegisterReplaceChange(Temp, BB, CaretXY);
            DoChange;
          end;
        end
        else
        begin
          Temp := LineText;
          Len := Length(Temp);
          CX := ColumnToCharIndex(FCaretX);
          if Len < CX then Temp := Temp + WideStringOfChar(' ', CX - Len);
          OldOptions := FOptions;
          Include(FOptions, eoScrollPastEOL);
          try
            if (eoInserting in FOptions) or (Len <= CX) then
            begin
              if (AChar = WideTabulator) and not (eoUseTabs in FOptions) then
              begin
                if eoSmartTabs in FOptions then I := NextSmartTab(FCaretX, FCaretY, True)
                                           else I := FCaretX + FTabSize;
                // if I is still at FCaretX then do nothing (can happen with smart
                // tabs under certain conditions)
                if I > FCaretX then
                begin
                  // insert as much spaces as the size of the column difference
                  Helper := WideStringOfChar(' ', I - FCaretX);
                  Insert(Helper, Temp, CX + 1);
                  Action := True;
                end
                else Action := False;
              end
              else
              begin
                // insert the new character (plus some spaces if the current column is
                // within a non-default width character and the new character is not a tab)
                if AChar = WideTabulator then
                begin
                  Helper := AChar;
                  Insert(Helper, Temp, CX + 1);
                  // need to set the line text at this point, although it is redundant (as it will
                  // be set in the action branch below), because the new caret position
                  // needs also to be calculated and this computation is based on the
                  // text
                  LineText := Temp;
                  I := NextCharPos(FCaretX, True);
                end
                else
                begin
                  Helper := WideStringOfChar(' ', Cardinal(FCaretX) - CharIndexToColumn(CX)) + AChar;
                  Insert(Helper, Temp, CX + 1);
                  I := FCaretX + 1;
                end;
                Action := True;
              end;

              if Action then
              begin
                LineText := Temp;
                CaretX := I;
                RegisterInsertChange(Helper, LastState.Caret, CaretXY);
                if FCaretX >= -FOffsetX * FCharWidth + FCharsInWindow then SetOffsetX((FOffsetX - Min(25, FCharsInWindow - 1)) * FCharWidth);
                DoChange;
              end;
            end
            else
            begin
              // in overwrite mode does a tab char only advance the cursor
              if AChar = WideTabulator then
              begin
                CaretX := NextTabPos(FCaretX);
                RegisterCursorMoveChange;
              end
              else
              begin
                OldChar := Temp[CX + 1];
                Temp[CX + 1] := AChar;
                CaretX := NextCharPos(FCaretX);
                LineText := Temp;
                RegisterOverwriteChange(AChar);
                if CaretX >= -FOffsetX * FCharWidth + FCharsInWindow then SetOffsetX((FOffsetX + Min(25, FCharsInWindow - 1)) * FCharWidth);
                DoChange;
              end;
            end;
          finally
            FOptions := OldOptions;
          end;
        end;
      end;

    ecUndo:
      Undo;

    ecRedo:
      Redo;

    ecGotoMarker0..ecGotoMarker9:
        if BookMarkOptions.EnableKeys then GotoBookMark(Command - ecGotoMarker0);

    ecToggleMarker0..ecToggleMarker9:
      begin
        if BookMarkOptions.EnableKeys then
        begin
          CX := Command - ecToggleMarker0;
          Action := FBookMarks[CX].Y <> CaretY;
          ClearBookMark(CX);
          if Action then SetBookMark(CX, FCaretX, CaretY);
        end;
      end;

    ecCut:
      if not (eoReadOnly in FOptions) and SelAvail then
      begin
        CutToClipboard;
        DoChange;
      end;

    ecCopy:
      CopyToClipboard;

    ecPaste:
      if not (eoReadOnly in FOptions) then
      begin
        PasteFromClipboard;
        DoChange;
      end;

    ecScrollUp:
      begin
        TopLine := TopLine - 1;
        if CaretY > TopLine + FLinesInWindow - 2 then CaretY := TopLine + LinesInWindow - 2;
        RegisterCursorMoveChange;
      end;

    ecScrollDown:
      begin
        TopLine := TopLine + 1;
        if CaretY < TopLine - 1 then CaretY := TopLine - 1;
        RegisterCursorMoveChange;
      end;

    ecScrollLeft:
      begin
        Dec(FOffsetX, FCharWidth);
        if FCaretX > -FOffsetX * FCharWidth + CharsInWindow then CaretX := -FOffsetX * FCharWidth + CharsInWindow;
        RegisterCursorMoveChange;
      end;

    ecScrollRight:
      begin
        Inc(FOffsetX, FCharWidth);
        if FCaretX < -FOffsetX * FCharWidth then CaretX := -FOffsetX * FCharWidth;
        RegisterCursorMoveChange;
      end;

    ecInsertMode:
      begin
        Include(FOptions, eoInserting);
        InitializeCaret;
        DoSettingChanged;
      end;

    ecOverwriteMode:
      begin
        Exclude(FOptions, eoInserting);
        InitializeCaret;
        DoSettingChanged;
      end;

    ecToggleMode:
      begin
        if eoInserting in FOptions then Exclude(FOptions, eoInserting)
                                   else Include(FOptions, eoInserting);
        InitializeCaret;
        DoSettingChanged;
      end;

    ecBlockIndent:
      if not (eoReadOnly in FOptions) then
      begin
        BlockIndent;
        DoChange;
      end;

    ecBlockUnindent:
      if not (eoReadOnly in FOptions) then
      begin
        BlockUnindent;
        DoChange;
      end;
  end;

  // Add macro recorder stuff here?
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.DoSettingChanged;

begin
  if Assigned(FOnSettingChange) then FOnSettingChange(Self);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ProcessCommand(var Command: TEditorCommand; var AChar: WideChar; Data: Pointer);

begin
  if Command < ecUserFirst then
  begin
    if assigned(FOnProcessCommand) then FOnProcessCommand(Self, Command, AChar, Data);
  end
  else
  begin
    if assigned(FOnProcessUserCommand) then FOnProcessUserCommand(Self, Command, AChar, Data);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ClearAll(RegisterUndo, KeepUndoList: Boolean);

// Deletes the entire text and (in case KeepUndoList is not True) clears also the
// undo list. If KeepUndoList is True then RegisterUndo has no effect otherwise
// a new undo record is added to make the clear operation reversable.

var
  State: TEditState;

begin
  if FLines.Count > 0 then
  begin
    BeginUpdate;
    RegisterUndo := RegisterUndo and KeepUndoList;
    try
      if RegisterUndo then
      begin
        SelectAll;
        State := RecordState(True);
      end;
      FLines.Clear;
      FBlockBegin := Point(0, 0);
      FBlockEnd := Point(0, 0);
      CaretXY := Point(0, 0);
      if RegisterUndo then
        with State do
          FUndoList.AddChange(ecrDelete, Text, @Caret, @SelStart, @SelEnd);
    finally
      EndUpdate;
    end;
  end;
  if not KeepUndoList then FUndoList.ClearList;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.LineFromPos(Pos: TPoint): Integer;

// returns the index of the line at position Pos (which must be given in client coordinates)

begin
  Result := -FOffsetY + (Pos.Y div FTextHeight) + 1;
  if Result < 0 then Result := 0;
  if Result > FLines.Count - 1 then Result := FLines.Count -1;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.NextWordPos: TPoint;

// Calculates the caret position of the next word start or the line end if there's
// no further word on the current line and the caret is not already on the line's end.

var
  CX, CY: Integer;
  X: Integer;
  FoundWhiteSpace: Boolean;
  Len: Integer;
  Temp: WideString;
  IdChars: TIdentChars;

begin
  Result := CaretXY;
  // ensure at least an empty line is there
  Temp := LineText;
  CX := ColumnToCharIndex(FCaretX) + 1;
  if CX < Length(FLines[FCaretY]) then CY := FCaretY
                                  else
  begin
    CY := FCaretY + 1;
    if CY > FLines.Count - 1 then Exit;
    if WideTrim(FLines[CY]) = '' then
    begin
      Result := Point(0, CY);
      Exit;
    end;
    CX := 1;
  end;

  if FHighLighter <> nil then IdChars := FHighLighter.IdentChars
                         else IDchars := [#33..#255];

  Temp := FLines[CY];
  Len := Length(Temp);
  FoundWhiteSpace := (not IsIdentChar(Temp[CX], IdChars)) or (CY <> FCaretY);
  if not FoundWhiteSpace then
    for X := CX to Len do
      if not IsIdentChar(Temp[X], IdChars) then
      begin
        CX := X;
        FoundWhiteSpace := True;
        Break;
      end;

  if FoundWhiteSpace then
  begin
    FoundWhiteSpace := False;
    for X := CX to Len do
      if IsIdentChar(Temp[X], IdChars) then
      begin
        Result := Point(X - 1, CY);
        FoundWhiteSpace := True;
        Break;
      end;
  end;

  if not FoundWhiteSpace then Result := Point(Len, CY);
  Result.X := CharIndexToColumn(Result);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.LastWordPos: TPoint;

var
  CX, CY: Integer;
  X: Integer;
  FoundNonWhiteSpace: Boolean;
  Len: Integer;
  Temp: WideString;
  IdChars: TIdentChars;

begin
  Result := CaretXY;
  Len := Length(LineText);
  CX := ColumnToCharIndex(FCaretX);
  if CX > Len then CX := Len;
  if CX > 0 then CY := FCaretY
            else
  begin
    if FCaretY < 1 then
    begin
      Result := Point(0, 0);
      Exit;
    end
    else
    begin
      Result := Point(GetLineEnd(FCaretY - 1), FCaretY - 1);
      Exit;
    end;
  end;

  if FHighLighter <> nil then IdChars := FHighLighter.IdentChars
                         else IDchars := [#33..#255];

  Temp := FLines[CY];
  FoundNonWhiteSpace := IsIdentChar(Temp[CX], IdChars);
  if not FoundNonWhiteSpace then
    for X := CX downto 2 do
      if IsIdentChar(Temp[X - 1], IdChars) then
      begin
        CX := X - 1;
        FoundNonWhiteSpace := True;
        Break;
      end;

  if FoundNonWhiteSpace then
  begin
    FoundNonWhiteSpace := False;
    for X := CX downto 2 do
      if not IsIdentChar(Temp[X - 1], IdChars) then
      begin
        Result := Point(X - 1, CY);
        FoundNonWhiteSpace := True;
        Break;
      end;

    if not FoundNonWhiteSpace then Result := Point(0, CY);
  end
  else
  begin
    Dec(CY);
    if CY = -1 then Result := Point(0, 0)
               else Result := Point(length(FLines[CY]), CY);
  end;
  Result.X := CharIndexToColumn(Result);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetUpdateState(Updating: Boolean);

begin
  SendMessage(Handle, WM_SETREDRAW, Ord(not Updating), 0);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.BeginUpdate;

begin
  if FUpdateCount = 0 then SetUpdateState(True);
  Inc(FUpdateCount);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.EndUpdate;

begin
  if FUpdateCount > 0 then Dec(FUpdateCount);
  if FUpdateCount = 0 then
  begin
    SetUpdateState(False);
    UpdateScrollBars;
    UpdateCaret;
    DoCaretChange;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.InvalidateLine(Index: Integer);

var
  R : TRect;

begin
  R := Rect(0, (Index + FOffsetY) * FTextHeight, ClientWidth, (Index + FOffsetY + 1) * FTextHeight);
  InvalidateRect(Handle, @R, False);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.InvalidateLines(Start, Stop: Integer);

var
  R : TRect;

begin
  if Start <= Stop then
  begin
    Start := Start + FOffsetY;
    if Start < 0 then Start := 0;
    Stop := Stop + FOffsetY + 1;
    if Stop > FLinesInWindow then Stop := FLinesInWindow;
    // consider partially visible line
    if (ClientHeight mod FLinesInWindow) <> 0 then Inc(Stop);
    R := Rect(0, Start * FTextHeight, ClientWidth, Stop * FTextHeight);
  end
  else
  begin
    Stop := Stop + FOffsetY;
    if Stop < 0 then Stop := 0;
    Start := Start + FOffsetY + 1;
    if Start > FLinesInWindow then Start := FLinesInWindow;
    // consider partially visible line
    if (ClientHeight mod FLinesInWindow) <> 0 then Inc(Start);
    R := Rect(0, Stop* FTextHeight, ClientWidth, Start * FTextHeight);
  end;
  InvalidateRect(Handle, @R, False);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.InvalidateToBottom(Index: Integer);

var
  R : TRect;

begin
  R := Rect(0, (Index + FOffsetY) * FTextHeight, ClientWidth, ClientHeight);
  InvalidateRect(Handle, @R, False);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.MarkListChange(Sender: TObject);

begin
  Invalidate;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetSelStart: Integer;

var
  Loop: Integer;
  X, Y: Integer;

begin
  X := FCaretx;
  Y := FCaretY;

  Result := 0;
  Loop := 0;
  while Loop < (Y-1) do
  begin
    Result := Result+Length(FLines[Loop])+2;
    Inc(Loop);
  end;

  Result := Result+Min(X, Length(FLines[Loop])+2);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetSelStart(Value: Integer);

var
  Loop: Integer;
  Count: Integer;

begin
  Loop := 0;
  Count := 0;
  if Value < 0 then Value := 0;
  while (Count + Length(FLines[Loop]) + 2) < Value do
  begin
    Count := Count+Length(FLines[Loop])+2;
    Inc(loop);
  end;
  CaretX := Value - Count;
  CaretY := Loop + 1;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetSelEnd: Integer;

var
  Loop: Integer;
  X, Y: Integer;

begin
  X := BlockEnd.X;
  Y := BlockEnd.Y;

  Result := 0;
  Loop := 0;
  while Loop < (Y-1) do
  begin
    Result := Result+Length(FLines[Loop])+2;
    Inc(Loop);
  end;
  Result := Result+X;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetSelEnd(const Value: Integer);

var
  P: TPoint;
  Loop: Integer;
  Count: Integer;

begin
  Loop := 0;
  Count := 0;
  while (Count+Length(FLines[Loop])+2) < value do
  begin
    Count := Count+Length(FLines[Loop])+2;
    Inc(loop);
  end;
  P.X := Value-Count;
  P.Y := Loop+1;
  Blockend := P;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetExtraLineSpacing(const Value: Integer);

begin
  FExtraLineSpacing := Value;
  FontChanged(Self);
  DoSettingChanged;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.RowColumnToCharPosition(P: TPoint): Cardinal;

// Calculates the total position of the column/row reference point as would all lines be
// one continous WideString.

var
  I: Integer;
  
begin
  Result := 0;
  P.X := ColumnToCharIndex(P);
  for I := 0 to Min(P.Y, FLines.Count) - 1 do Inc(Result, Length(FLines[I]) + 2); // add 2 for CRLF
  if P.Y < FLines.Count then Inc(Result, P.X);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.CharPositionToRowColumn(Position: Cardinal): TPoint;

// Calculates the total position of the column/row reference point as would all lines be
// one continous WideString.

var
  I: Integer;
  Len: Cardinal;

begin
  Result.Y := 0;
  for I := 0 to FLines.Count - 1 do
  begin
    Len := Length(FLines[I]) + 2; // sub 2 for CRLF
    if Position < Len then Break;
    Inc(Result.Y);
    Dec(Position, Len);
  end;
  Result.X := Position;
  Result.X := ColumnToCharIndex(Result);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.SearchReplace(const SearchText, ReplaceText: WideString; Options: TSearchOptions): Integer;

var
  StartPoint,
  EndPoint: TPoint; // search range
  Current: TPoint;  // current search position
  DoBackward,
  DoFromCursor: Boolean;
  DoPrompt: Boolean;
  DoReplace,
  DoReplaceAll,
  Changed: Boolean;
  Action: TReplaceAction;
  SearchEngine: TSearchEngine;

  //--------------- local functions -------------------------------------------

  function IsInValidSearchRange(First, Last: Integer): Boolean;

  begin
    Result := not (((Current.Y = StartPoint.Y) and (First < StartPoint.X)) or
                  ((Current.Y = EndPoint.Y) and (Last > EndPoint.X)));
  end;

  //---------------------------------------------------------------------------

  procedure DoSearchReplace;

  var
    SearchLen,
    ReplaceLen,
    N: Integer;
    Start,
    Stop,
    ColStart,
    ColStop: Integer;
    Found,
    DoStop: Boolean;
    SearchOptions: TSearchFlags;
    Offset: Integer;

  begin
    // initialize the search engine
    SearchOptions := [];
    if soMatchCase in Options then Include(SearchOptions, sfCaseSensitive);
    if soIgnoreNonSpacing in Options then Include(SearchOptions, sfIgnoreNonSpacing);
    if soSpaceCompress in Options then Include(SearchOptions, sfSpaceCompress);
    if soWholeWord in Options then Include(SearchOptions, sfWholeWordOnly);
    //soWholeWord in Options;
    SearchEngine.FindPrepare(SearchText, SearchOptions);
    // search while the current search position is inside of the search range
    SearchLen := Length(SearchText);
    ReplaceLen := Length(ReplaceText);
    DoStop := False;

    while not DoStop and (Current.Y >= StartPoint.Y) and (Current.Y <= EndPoint.Y) do
    begin
      // need a running offset because further search results are determined with regard
      // to the unchanged string but when replacing the string changes all the time
      Offset := 0;
      Found := SearchEngine.FindAll(FLines[Current.Y]);
      if DoBackward then N := Pred(SearchEngine.Count)
                    else N := 0;
      // Operate on all results in this line.
      while not DoStop and Found do
      begin
        if (N < 0) or (N = SearchEngine.Count) then Break;
        SearchEngine.GetResult(N, Start, Stop);
        Inc(Start, Offset);
        Inc(Stop, Offset);
        // convert character positions to column values,
        // need to invalidate the char width array in every run because there might be non-single characters
        // like tabulators in text or replace string
        FLastCalcIndex := -1;
        ColStart := CharIndexToColumn(Point(Start, Current.Y));
        ColStop := CharIndexToColumn(Point(Stop, Current.Y));
        if DoBackward then Dec(N)
                      else Inc(N);
        // Is the search result entirely in the search range?
        if not IsInValidSearchRange(Start, Stop) then Continue;
        Inc(Result);
        // Select the text, so the user can see it in the OnReplaceText event
        // handler or as the search result.
        Current.X := ColStart;
        BlockBegin := Current;
        if DoBackward then CaretXY := Current;
        Current.X := ColStop;
        BlockEnd := Current;
        if not DoBackward then CaretXY := Current;
        // if it's a search only we can leave the procedure now
        if not (DoReplace or DoReplaceAll) then
        begin
          DoStop := True;
          Break;
        end;

        // Prompt and replace or replace all.  If user chooses to replace
        // all after prompting, turn off prompting.
        if DoPrompt then
        begin
          Action := raCancel;
          Application.ProcessMessages;
          FOnReplaceText(Self, SearchText, ReplaceText, Current.Y, ColStart, ColStop, Action);
          if Action = raCancel then
          begin
            DoStop := True;
            Break;
          end;
        end
        else Action := raReplace;
        if Action <> raSkip then
        begin
          // user has been prompted and has requested to silently replace all
          // so turn off prompting
          if Action = raReplaceAll then
          begin
            if not DoReplaceAll then
            begin
              DoReplaceAll := True;
              BeginUpdate;
            end;
            DoPrompt := False;
          end;
          SetSelTextExternal(ReplaceText);
          Changed := True;
        end;
        // calculate position offset (this offset is character index not column based)
        if not DoBackward then Inc(Offset, ReplaceLen - SearchLen);
        if not DoReplaceAll then DoStop := True;
      end;
      // search next / previous line
      if DoBackward then Dec(Current.Y)
                    else Inc(Current.Y);
    end;
  end;

  //--------------- end local functions ---------------------------------------

begin
  Result := 0;
  // can't search for or replace an empty WideString
  if (Length(SearchText) > 0) and (FLines.Count > 0) then
  begin
    Changed := False;
    // get the text range to search in, ignore search in selection if nothing is selected
    DoBackward := soBackwards in Options;
    DoPrompt := (soPrompt in Options) and Assigned(FOnReplaceText);
    DoReplace := soReplace in Options;
    DoReplaceAll := soReplaceAll in Options;
    DoFromCursor := not (soEntireScope in Options);
    if SelAvail and (soSelectedOnly in Options) then
    begin
      StartPoint := BlockBegin;
      EndPoint := BlockEnd;
      // ignore the cursor position when searching in the selection
      if DoBackward and not (soRegularExpression in Options) then Current := EndPoint
                                                             else Current := StartPoint;
    end
    else
    begin
      StartPoint := Point(0, 0);
      EndPoint.Y := FLines.Count - 1;
      EndPoint.X := Length(FLines[EndPoint.Y]) + 1;
      if DoFromCursor then
        if DoBackward then EndPoint := CaretXY
                      else StartPoint := CaretXY;
      if DoBackward then Current := EndPoint
                    else Current := StartPoint;
    end;
    if DoReplaceAll and not DoPrompt then BeginUpdate;

    if soRegularExpression in Options then SearchEngine := TURESearch.Create(FLines)
                                      else SearchEngine := TUTBMSearch.Create(FLines);
    try
      DoSearchReplace;
      if Changed then DoChange;
    finally
      SearchEngine.Free;
      if DoReplaceAll and not DoPrompt then EndUpdate;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetBookMark(BookMark: Integer; var X, Y: Integer): Boolean;

begin
  Result := (Bookmark in [0..9]) and FBookmarks[Bookmark].Visible;
  if Result then
  begin
    X := FBookmarks[Bookmark].X;
    Y := FBookmarks[Bookmark].Y;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.IsBookmark(BookMark: Integer): Boolean;

var
  X, Y: Integer;

begin
  Result := GetBookMark(BookMark, X, Y);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ClearUndo;

begin
  FUndoList.ClearList;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetSelTextExternal(const Value: WideString);

var
  State: TEditState;
  BB, BE: TPoint;

begin
  State := RecordState(True);
  // keep the position at which the new text will be inserted
  BB := MinPoint(FBlockBegin, FBlockEnd);
  SetSelText(Value);
  SetBlockBegin(CaretXY);
  // keep end of text block just inserted
  BE := CaretXY;
  with State do
    FUndoList.AddChange(ecrReplace, Text, @Caret, @SelStart, @SelEnd, Value, @BE, @BB, @BE);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetOptions(const Value: TSyntaxEditOptions);

var
  ToBeSet,
  ToBeCleared,
  ToChange: TSyntaxEditOptions;

begin
  if FOptions <> Value then
  begin
    ToBeSet := Value - FOptions;
    ToBeCleared := FOptions - Value;
    FOptions := Value;
    ToChange := ToBeSet + ToBeCleared;

    if eoReadOnly in FOptions then
      if eoShowCursorWhileReadOnly in ToBeSet then InitializeCaret
                                              else
      begin
        HideCaret;
        DestroyCaret;
      end;
    if (eoHideSelection in ToChange) or
       (eoShowControlChars in ToChange) or
       (eoUseSyntaxHighlighting in ToChange) then Invalidate;

    if eoLineNumbers in ToChange then
    begin
      FGutterRect.Right := FGutterWidth - 1;
      FCharsInWindow := (ClientWidth - FGutterRect.Right + 1) div FCharWidth;
      Invalidate;
    end;

    // reset pos in case it's past EOL currently
    if (eoScrollPastEOL in ToBeCleared) and (FCaretX > MaxRightChar) then CaretX := MaxRightChar;
    if (eoLineNumbers in ToChange) and Focused then UpdateCaret;
    DoSettingChanged;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetTabSize(Value: Integer);

begin
  if Value < 1 then Value := 1;
  if FTabSize <> Value then
  begin
    FTabSize := Value;
    DoSettingChanged;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.NextSmartTab(X, Y: Integer; SkipNonWhite: Boolean): Integer;

// Calculates the indent position starting with X and depending on the first
// line above Y which contains a white space char before X.
// X and Result are column values!

var
  I : Integer;
  Len : Integer;
  Line: WideString;

begin
  // cursor at line 0?
  if Y = 0 then Result := X // no next tab position available
           else
  begin
    // find a line above the current line which size is greater than X
    repeat
      Dec(Y);
      I := GetLineEnd(Y);
      if I > X then Break;
    until Y = 0;

    // found a line?
    if I > X then
    begin
      Line := FLines[Y];
      Len := Length(Line);
      I := ColumnToCharIndex(Point(X, Y));
      // skip non-white spaces if required
      if SkipNonWhite then
        while (I < Len) and not UnicodeIsWhiteSpace(Word(Line[I + 1])) do Inc(I);

      // now skip any white space character
      while (I < Len) and UnicodeIsWhiteSpace(Word(Line[I + 1])) do Inc(I);
      Result := CharIndexToColumn(Point(I, Y));
    end
    else Result := X;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.Unindent(X, Y: Cardinal): Cardinal;

// Calculates the next unindent position starting with X and depending on the
// first line above Y which contains a white space char before X.
// X must be a column value!

  //--------------- local functions -------------------------------------------

  function FindNonWhiteSpaceColumn(X, Y: Cardinal; var NewX: Cardinal): Boolean;

  var
    Index: Integer;
    Line: WideString;

  begin
    NewX := X;
    Result := False;
    Line := FLines[Y];
    while NewX > 0 do
    begin
      Dec(NewX);
      // turn column into its corresponding character index...
      Index := ColumnToCharIndex(Point(NewX, Y));
      // ... and turn it back into the start column of this char
      NewX := CharIndexToColumn(Point(Index, Y));
      // fork out if we found a non-white space char
      if (Index < Length(Line)) and not UnicodeIsWhiteSpace(Word(Line[Index + 1])) then
      begin
        Result := True;
        Break;
      end;
    end;
  end;

  //---------------------------------------------------------------------------

  function FindWhiteSpaceColumn(X, Y: Cardinal; var NewX: Cardinal): Boolean;

  var
    Index: Integer;
    Line: WideString;

  begin
    NewX := X;
    Result := False;
    Line := FLines[Y];
    while NewX > 0 do
    begin
      Dec(NewX);
      // turn column into its corresponding character index...
      Index := ColumnToCharIndex(Point(NewX, Y));
      // ... and turn it back into the start column of this char
      NewX := CharIndexToColumn(Point(Index, Y));
      // fork out if we found a white space
      if (Index < Length(Line)) and UnicodeIsWhiteSpace(Word(Line[Index + 1])) then
      begin
        Result := True;
        Break;
      end;
    end;
  end;

  //--------------- end local functions ---------------------------------------

begin
  // cursor at line 0?
  if (X = 0) or (Y = 0) then Result := 0 // no previous tab position available
                        else
  begin
    // find a line above the current line, which is not empty and whose first
    // non-white character is before X
    repeat
      // go one line up (picturally speaking, actually we use the previous line)
      Dec(Y);
      // find last non-white space before current column
      if FindNonWhiteSpaceColumn(X, Y, Result) then
      begin
        // we found a non-white space, now search the first white
        // space column before this column
        if FindWhiteSpaceColumn(Result, Y, Result) then
        begin
          // if there's such a column then advance to next column (which is then
          // a non-white space column)
          Result := NextCharPos(Result, Y);
        end;
        // leave loop, since we either found what we were looking for or
        // the line in question has no white chars at the beginning (which means
        // unindentation target column is 0)
        Break;
      end;
      // turn around if there was no unindent column by going up one line
    until Y = 0;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.NextTabPos(Index: Integer): Integer;

// calculates next tab stop position

begin
  Result := (Index div FTabSize + 1) * FTabSize;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.CMSysColorChange(var Message: TMessage);

begin
  inherited;
  ConvertAndAddImages(FInternalBMList);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.DoChange;

begin
  FModified := True;
  if Assigned(FOnChange) then FOnChange(Self);
end;

//----------------------------------------------------------------------------------------------------------------------

const
  ScrollBarProp: array[TScrollBarMode] of Integer = (FSB_REGULAR_MODE, FSB_FLAT_MODE, FSB_ENCARTA_MODE);

procedure TCustomSyntaxEdit.CreateWnd;

begin
  inherited;
  InitializeFlatSB(Handle);
  FlatSB_EnableScrollBar(Handle, SB_BOTH, ESB_ENABLE_BOTH);
  FlatSB_SetScrollProp(Handle, WSB_PROP_HSTYLE, ScrollBarProp[FScrollBarMode], True);
  FlatSB_SetScrollProp(Handle, WSB_PROP_VSTYLE, ScrollBarProp[FScrollBarMode], True);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetScrollBarMode(const Value: TScrollBarMode);

begin
  if FScrollBarMode <> Value then
  begin
    FScrollBarMode := Value;
    FlatSB_SetScrollProp(Handle, WSB_PROP_HSTYLE, ScrollBarProp[FScrollBarMode], True);
    FlatSB_SetScrollProp(Handle, WSB_PROP_VSTYLE, ScrollBarProp[FScrollBarMode], True);
    DoSettingChanged;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.CalcCharWidths(Index: Integer);

// Creates an array of integer values describing the width of each character
// of a given line in pixels.

var
  I, Run,
  NewPos: Integer;
  P: PWideChar;
  Multiplyer: Integer;

begin
  if Index > FLines.Count - 1 then Index := FLines.Count - 1;
  if FLastCalcIndex <> Index then
  begin
    FLastCalcIndex := Index;
    if Index = -1 then FCharWidthArray := nil
                  else
    begin
      P := PWideChar(FLines[Index]);
      SetLength(FCharWidthArray, Length(FLines[Index]));
      Multiplyer := FTabSize * FCharWidth;
      // running term for tabulator calculation
      Run := 0;
      I := 0;
      while P^ <> WideNull do
      begin
        if P^ = WideTabulator then
        begin
          // calculate new absolute position
          NewPos := (Run div Multiplyer + 1) * Multiplyer;
          // make this a relative distance
          FCharWidthArray[I] := NewPos - Run;
        end
        else
        begin
          FCharWidthArray[I] := FCharWidth;
        end;

        Inc(Run, FCharWidthArray[I]);
        Inc(I);
        Inc(P);
      end;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.CalcCharWidths(const S: WideString; TabSize, CharWidth: Cardinal);

// Creates an array of integer values describing the width of each character
// of the given WideString in pixels.

var
  I: Integer;
  Run,
  NewPos: Cardinal;

begin
  // always mark last calculated line as invalid as the given line might be from
  // somewhere else but not FLines
  FLastCalcIndex := -1;

  SetLength(FCharWidthArray, Length(S));
  // running term for tabulator calculation
  Run := 0;
  for I := 0 to Length(S) - 1 do
  begin
    if S[I + 1] = WideTabulator then
    begin
      // calculate new absolute position
      NewPos := (Run div (TabSize * CharWidth) + 1) * TabSize * CharWidth;
      // make this a relative distance
      FCharWidthArray[I] := NewPos - Run;
    end
    else FCharWidthArray[I] := FCharWidth;
    Inc(Run, FCharWidthArray[I]);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetModified: Boolean;

begin
  Result := FModified or TSyntaxEditorStrings(FLines).Modified;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetModified(const Value: Boolean);

begin
  FModified := Value;
  TSyntaxEditorStrings(FLines).Modified := Value;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetMarkLine(Value: Integer);

begin
  if Value < -1 then Value := -1;
  if FMarkLine <> Value then
  begin
    if FMarkLine > -1 then InvalidateLine(FMarkLine);
    FMarkLine := Value;
    if FMarkLine > -1 then InvalidateLine(FMarkLine);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetMarkColor(const Value: TColors);

begin
  FMarkColor.FBackground := Value.FBackground;
  FMarkColor.FForeground := Value.FForeground;
  if FMarkLine > -1 then InvalidateLine(FMarkLine);
  DoSettingChanged;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.NextCharPos(ThisPosition: Cardinal; ForceNonDefault: Boolean): Cardinal;

begin
  Result := NextCharPos(ThisPosition, FCaretY, ForceNonDefault);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.NextCharPos(ThisPosition, Line: Cardinal; ForceNonDefault: Boolean): Cardinal;

// Determines the next character position for the given line depending on the
// column given by ThisPosition. The returned index can directly be used to set the caret.

var
  PixelPos, Run: Cardinal;
  I: Cardinal;

begin
  CalcCharWidths(Line);
  // turn column into pixel position
  PixelPos := ThisPosition * Cardinal(FCharWidth);
  if ((eoCursorThroughTabs in FOptions) and not ForceNonDefault) or
     (ThisPosition >= GetLineEnd(Line)) then Result := ThisPosition + 1
                                        else
  begin
    // The task is to find the first column which corresponds directly to a
    // character (which is not always the case, eg. tabs) and is larger than
    // the given position.
    Run := 0;
    // sum up character widths until we find a pixel position larger than the
    // given one
    for I := 0 to High(FCharWidthArray) do
    begin
      // values in the char width array always correspond to an actual character index
      if PixelPos < Run then Break
                        else Inc(Run, FCharWidthArray[I]);
    end;
    Result := Run div Cardinal(FCharWidth);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.PreviousCharPos(ThisPosition: Cardinal): Cardinal;

begin
  Result := PreviousCharPos(ThisPosition, FCaretY);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.PreviousCharPos(ThisPosition, Line: Cardinal): Cardinal;

// Determines the previous character position for the given line depending on the
// column given by ThisPosition. The returned index can directly be used to set the caret.

var
  PixelPos, Run: Cardinal;
  I: Cardinal;

begin
  if ThisPosition = 0 then Result := 0
                      else
  begin
    CalcCharWidths(Line);
    // turn column into pixel position
    PixelPos := ThisPosition * Cardinal(FCharWidth);
    if (eoCursorThroughTabs in FOptions) or
       (ThisPosition > GetLineEnd(Line)) then Result := ThisPosition - 1
                                         else
    begin
      // The task is to find the last column which corresponds directly to a
      // character (which is not always the case, eg. tabs) and is smaller than
      // the given position.
      Run := 0;
      // sum up character widths until we find the largest pixel position smaller
      // than the given one
      for I := 0 to High(FCharWidthArray) do
      begin
        // values in the char width array always correspond to an actual character index
        if PixelPos <= (Run + FCharWidthArray[I]) then Break
                                                  else Inc(Run, FCharWidthArray[I]);
      end;
      Result := Run div Cardinal(FCharWidth);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetLineEnd: Cardinal;

begin
  Result := GetLineEnd(FCaretY);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.GetLineEnd(Index: Cardinal): Cardinal;

// calculates the first unused column of the line given by Index

var
  I: Integer;

begin
  CalcCharWidths(Index);
  Result := 0;
  for I := 0 to High(FCharWidthArray) do Inc(Result, FCharWidthArray[I]);
  Result := Result div Cardinal(FCharWidth);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.ColumnToCharIndex(X: Cardinal): Cardinal;

begin
  Result := ColumnToCharIndex(Point(X, FCaretY));
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.ColumnToCharIndex(Pos: TPoint): Cardinal;

// Calculates the character index into the line given by Y. X and Y are usually a
// caret position. Because of non-default width characters (like tabulators) the
// CaretX position is merely a column value and does not directly correspond to a
// particular character. The result can be used to address a character in the line
// given by Y or is a generic char index if X is beyond any char in Y.

var
  Run: Integer;

begin
  CalcCharWidths(Pos.Y);
  Run := 0;
  Pos.X := Pos.X * FCharWidth;
  Result := 0;
  while Result < Cardinal(Length(FCharWidthArray)) do
  begin
    Inc(Run, FCharWidthArray[Result]);
    if Run > Pos.X then Break;
    Inc(Result);
  end;
  // If the run is still smaller than the given column then add "virtual" chars
  // to be able to add a character (far) beyond the current end of the line.
  if Run < Pos.X then Inc(Result, (Pos.X - Run) div FCharWidth);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.CharIndexToColumn(X: Cardinal): Cardinal;

begin
  Result := CharIndexToColumn(Point(X, FCaretY));
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.CharIndexToColumn(X: Cardinal; const S: WideString; TabSize, CharWidth: Cardinal): Cardinal;

// calculates the corresponding column from the given character position using the
// text stored in S

var
  I: Cardinal;

begin
  CalcCharWidths(S, TabSize, CharWidth);
  I := 0;
  Result := 0;
  while I < X do
  begin
    if I < Cardinal(Length(FCharWidthArray)) then Inc(Result, FCharWidthArray[I])
                                             else Inc(Result, CharWidth);
    Inc(I);
  end;
  Result := Result div Cardinal(CharWidth);
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.CharIndexToColumn(Pos: TPoint): Cardinal;

// calculates the corresponding column from the given character position using the
// text stored in FLines at position Pos.Y

var
  I: Integer;

begin
  CalcCharWidths(Pos.Y);
  I := 0;
  Result := 0;
  while I < Pos.X do
  begin
    if I < Length(FCharWidthArray) then Inc(Result, FCharWidthArray[I])
                                   else Inc(Result, FCharWidth);
    Inc(I);
  end;
  Result := Result div Cardinal(FCharWidth);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.ReplaceLeftTabs(var S: WideString);

// replaces all leading tab characters in the given WideString by a number of spaces given
// by the current tab size

var
  Count: Integer;

begin
  Count := LeftSpaces(S);
  while Count > 0 do
  begin
    if S[Count] = WideTabulator then
    begin
      Delete(S, Count, 1);
      Insert(WideStringOfChar(' ', FTabSize), S, Count);
    end;
    Dec(Count);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TCustomSyntaxEdit.LeftSpaces(const S: WideString): Cardinal;

// counts the number of white spaces at the beginning of the WideString;
// a tab is considered as one space here

var
  Run: PWideChar;

begin
  Run := PWideChar(S);
  while UnicodeIsWhiteSpace(Word(Run^)) do Inc(Run);
  Result := Run - PWideChar(S);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.BlockIndent;

// Inserts FIndentSize space characters before every line in the current selection.
// Note: A designed side effect is that all leading tab characters in the concerned
//       lines will be replaced by the proper number of spaces!

var
  I, J,
  Last: Integer;
  Line: WideString;
  BB, BE: PPoint;

  SelStart,
  SelEnd,
  OldCaret,
  NewCaret: TPoint;
  T1, T2: TWideStringList;

begin
  if SelAvail then
  begin
    BeginUpdate;
    // By using references instead of a copy we can directly work on the selection
    // variables and have not to worry about whether the block begin is really
    // before the block end.
    BB := MinPointAsReference(FBlockBegin, FBLockEnd);
    BE := MaxPointAsReference(FBlockBegin, FBLockEnd);

    // exclude last selection line if it does not contain any selected chars
    if BE.X > 0 then Last := BE.Y
                else Last := BE.Y - 1;

    // save state for undo, this needs special care as we are going to replace
    // characters not in the current selection
    OldCaret := CaretXY;
    SelStart := BB^;
    SelEnd := BE^;

    T1 := TWideStringList.Create;
    T1.Capacity := Last - BB.Y + 1;
    T2 := TWideStringList.Create;
    T2.Capacity := Last - BB.Y + 1;

    for I := BB.Y to Last do
    begin
      Line := FLines[I];
      // copy all lines (entirely) which are somehow touched by the current selection
      T1.Add(Line);
      ReplaceLeftTabs(Line);
      Insert(WideStringOfChar(' ', FIndentSize), Line, 1);
      FLines[I] := Line;
      // copy also all lines after they have been changed
      T2.Add(Line);

      // adjust cursor if it is at this particular line, this needs to be done only
      // once, hence I put it herein
      if FCaretY = I then
      begin
        J := GetLineEnd(I);
        // cursor beyond end of line?
        if FCaretX >= J then CaretX := J
                        else
        begin
          // cursor in leading white space area?
          J := LeftSpaces(Line); // don't need TrimRightW here, as it has already been
                                 // verified that the cursor is before a non-white space char
          if FCaretX < J then CaretX := J
                         else CaretX := Cardinal(FCaretX) + FIndentSize;
        end;
      end
    end;

    // adjust selection appropriately (the Delphi IDE moves the selection start and
    // end to the first non-whitespace if there's no non-whitespace before the
    // particular X position, hence I'll do it too)
    Inc(BB.X, FIndentSize);
    Line := FLines[BB.Y];
    I := LeftSpaces(Line);
    if I >= BB.X then BB.X := I;
    I := GetLineEnd(BB.Y);
    if I < BB.X then BB.X := I;

    if BE.X > 0 then
    begin
      Inc(BE.X, FIndentSize);
      Line := FLines[BE.Y];
      I := LeftSpaces(Line);
      if I >= BE.X then BE.X := I;
      I := GetLineEnd(BE.Y);
      if I < BE.X then BE.X := I;
    end;

    // no further cursor adjustion, if the cusror is, by any means, not on any of
    // the lines in the current selection then it won't be updated at all
    NewCaret := CaretXY;
    FUndoList.AddChange(ecrIndentation, T1.Text, @OldCaret, @SelStart, @SelEnd,
                                        T2.Text, @NewCaret, @FBlockBegin, @FBlockEnd);

    T1.Free;
    T2.Free;
    EndUpdate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.BlockUnindent;

// Removes first FIndentSize space characters of every line in the current selection
// or all leading white spaces if their count is less than FIndentSize.
// Note: A designed side effect is that all leading tab characters in the concerned
//       lines will be replaced by the proper number of spaces!

var
  I, J,
  Last: Integer;
  Line: WideString;
  BB, BE: PPoint;

  SelStart,
  SelEnd,
  OldCaret,
  NewCaret: TPoint;
  OldText,
  NewText: WideString;

begin
  if SelAvail then
  begin
    BeginUpdate;
                               
    // By using references instead of a copy we can directly work on the selection
    // variables and have not to worry about whether the block begin is really
    // before the block end.
    BB := MinPointAsReference(FBlockBegin, FBLockEnd);
    BE := MaxPointAsReference(FBlockBegin, FBLockEnd);

    // exclude last selection line if it does not contain any selected chars
    if BE.X > 0 then Last := BE.Y
                else Last := BE.Y - 1;

    // save state for undo, this needs special care as we are going to replace
    // characters not in the current selection
    OldText := '';
    NewText := '';
    OldCaret := CaretXY;
    SelStart := BB^;
    SelEnd := BE^;

    for I := BB.Y to Last do
    begin
      Line := FLines[I];
      // copy all lines (entirely) which are somehow touched by the current selection
      OldText := OldText + Line + WideCRLF;
      ReplaceLeftTabs(Line);
      J := LeftSpaces(Line);
      if Cardinal(J) >= FIndentSize then Delete(Line, 1, FIndentSize)
                                    else Delete(Line, 1, J);
      FLines[I] := Line;
      // copy also all lines after they have been changed
      NewText := NewText + Line + WideCRLF;

      // adjust cursor if it is at this particular line, this needs to be done only
      // once, hence I put it herein
      if FCaretY = I then
      begin
        J := GetLineEnd(I);
        // cursor beyond end of line?
        if FCaretX >= J then CaretX := J
                        else
        begin
          // cursor in leading white space area?
          J := LeftSpaces(Line); // don't need TrimRightW here, as it has already been
                                 // verified that the cursor is before a non-white space char
          if FCaretX < J then CaretX := J
                         else CaretX := FCaretX - Integer(FIndentSize);
        end;
      end
    end;
    
    // adjust selection appropriately (the Delphi IDE moves the selection start and
    // end to the first non-whitespace if there's no non-whitespace before the
    // particular X position, hence I'll do it too)
    Dec(BB.X, FIndentSize);
    Line := FLines[BB.Y];
    I := LeftSpaces(Line);
    if I >= BB.X then BB.X := I;
    I := GetLineEnd(BB.Y);
    if I < BB.X then BB.X := I;

    if BE.X > 0 then
    begin
      Dec(BE.X, FIndentSize);
      Line := FLines[BE.Y];
      I := LeftSpaces(Line);
      if I >= BE.X then BE.X := I;
      I := GetLineEnd(BE.Y);
      if I < BE.X then BE.X := I;
    end;
    
    // no further cursor adjustion, if the cursor is, by any mean, not on any of
    // the lines in the current selection then it won't be updated at all
    NewCaret := CaretXY;
    FUndoList.AddChange(ecrIndentation, OldText, @OldCaret, @SelStart, @SelEnd,
                                        NewText, @NewCaret, @FBlockBegin, @FBlockEnd);

    EndUpdate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetIndentSize(Value: Cardinal);

begin
  if Value < 1 then Value := 1;
  if FIndentSize <> Value then
  begin
    FIndentSize := Value;
    DoSettingChanged;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.CMMouseWheel(var Message: TCMMouseWheel);

var
  ScrollCount: Integer;

begin
  inherited;
  if Message.Result = 0 then
    with Message do
    begin
      Result := 1;
      if ssCtrl in ShiftState then ScrollCount := FLinesInWindow * (WheelDelta div WHEEL_DELTA)
                              else ScrollCount := Mouse.WheelScrollLines * (WheelDelta div WHEEL_DELTA);
      TopLine := TopLine - ScrollCount;
      Update;
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetLineNumberFont(const Value: TFont);

begin
  FLineNumberFont.Assign(Value);
  LineNumberFontChanged(nil);
  DoSettingChanged;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.LineNumberFontChanged(Sender: TObject);

begin
  if HandleAllocated then
  begin
    InvalidateRect(Handle, @FGutterRect, False);
    UpdateWindow(Handle);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetScrollHintColor(const Value: TColors);

begin
  FScrollHintColor.FBackground := Value.FBackground;
  FScrollHintColor.FForeground := Value.FForeground;
  DoSettingChanged;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.SetSelectedColor(const Value: TColors);

begin
  FSelectedColor.FBackground := Value.FBackground;
  FSelectedColor.FForeground := Value.FForeground;
  DoSettingChanged;
  Invalidate;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WMImeComposition(var Message: TMessage);

var
  IMEContext: HIMC;
  P: PWideChar;
  Size: Integer;

begin
  if (Message.LParam and GCS_RESULTSTR) <> 0 then
  begin
    IMEContext := ImmGetContext(Handle);
    try
      // We have two choices here to retrive the composition string.
      // One is the ANSI versin of ImmGetCompositionString which will return
      // a multibyte version of the result string. This function, however, requires
      // the correct system local to be activated (which depends on the currently active
      // IME). Thus this requires remarkable extra effort. The other option (which has
      // been choosen here) is the wide/Unicode version. This directly returns the
      // composition result but does not work on Win9x. 
      Size := ImmGetCompositionStringW(IMEContext, GCS_RESULTSTR, nil, 0);
      // add space for 0 terminator
      P := AllocMem(Size + SizeOf(WideChar));
      try
        ImmGetCompositionStringW(IMEContext, GCS_RESULTSTR, P, Size);
        P[Size div 2] := #0;
        InsertText(P);
      finally
        FreeMem(P);
      end;
    finally
      ImmReleaseContext(Handle, IMEContext);
    end;
    Message.Result := 0;
  end
  else inherited;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TCustomSyntaxEdit.WMImeNotify(var Message: TMessage);

var
  IMC: HIMC;
  LogFont: TLogFont;
  
begin
  with Message do
  begin
    case WParam of
      IMN_SETOPENSTATUS:
        begin
          IMC := ImmGetContext(Handle);
          if IMC <> 0 then
          begin
            // need to tell the composition window what font we are using currently
            GetObject(Font.Handle, SizeOf(TLogFont), @LogFont);
            ImmSetCompositionFont(IMC, @LogFont);
            ImmReleaseContext(Handle, IMC);
          end;
          Message.Result := 0;
        end;
    else
      inherited;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

const
  // associated languages to particular Unicode code block,
  CodeBlockToLanguage: array[0..67] of LCID =
    (
    LANG_NEUTRAL, // Basic Latin
    LANG_NEUTRAL, // Latin-1 Supplement
    LANG_NEUTRAL, // Latin Extended-A
    LANG_NEUTRAL, // Latin Extended-B
    LANG_NEUTRAL, // IPA Extensions
    LANG_NEUTRAL, // Spacing Modifier Letters
    LANG_NEUTRAL, // Combining Diacritical Marks
    LANG_GREEK,   // Greek
    LANG_RUSSIAN, // Cyrillic
    LANG_NEUTRAL, // Armenian
    LANG_HEBREW,  // Hebrew
    LANG_ARABIC,  // Arabic
    LANG_NEUTRAL, // Devanagari
    LANG_NEUTRAL, // Bengali
    LANG_NEUTRAL, // Gurmukhi
    LANG_NEUTRAL, // Gujarati
    LANG_NEUTRAL, // Oriya
    LANG_NEUTRAL, // Tamil
    LANG_NEUTRAL, // Telugu
    LANG_NEUTRAL, // Kannada
    LANG_NEUTRAL, // Malayalam
    LANG_THAI,    // Thai
    LANG_NEUTRAL, // Lao
    LANG_NEUTRAL, // Tibetan
    LANG_NEUTRAL, // Georgian
    LANG_NEUTRAL, // Hangul Jamo
    LANG_NEUTRAL, // Latin Extended Additional
    LANG_GREEK,   // Greek Extended
    LANG_NEUTRAL, // General Punctuation
    LANG_NEUTRAL, // Superscripts and Subscripts
    LANG_NEUTRAL, // Currency Symbols
    LANG_NEUTRAL, // Combining Marks for Symbols
    LANG_NEUTRAL, // Letterlike Symbols
    LANG_NEUTRAL, // Number Forms
    LANG_NEUTRAL, // Arrows
    LANG_NEUTRAL, // Mathematical Operators
    LANG_NEUTRAL, // Miscellaneous Technical
    LANG_NEUTRAL, // Control Pictures
    LANG_NEUTRAL, // Optical Character Recognition
    LANG_NEUTRAL, // Enclosed Alphanumerics
    LANG_NEUTRAL, // Box Drawing
    LANG_NEUTRAL, // Block Elements
    LANG_NEUTRAL, // Geometric Shapes
    LANG_NEUTRAL, // Miscellaneous Symbols
    LANG_NEUTRAL, // Dingbats
    LANG_NEUTRAL, // CJK Symbols and Punctuation
    LANG_JAPANESE, // Hiragana
    LANG_JAPANESE, // Katakana
    LANG_NEUTRAL, // Bopomofo
    LANG_NEUTRAL, // Hangul Compatibility Jamo
    LANG_NEUTRAL, // Kanbun
    LANG_CHINESE, // Enclosed CJK Letters and Months
    LANG_CHINESE, // CJK Compatibility
    LANG_CHINESE, // CJK Unified Ideographs
    LANG_NEUTRAL, // Hangul Syllables
    LANG_NEUTRAL, // High Surrogates
    LANG_NEUTRAL, // High Private Use Surrogates
    LANG_NEUTRAL, // Low Surrogates
    LANG_NEUTRAL, // Private Use
    LANG_ARABIC,  // CJK Compatibility Ideographs
    LANG_NEUTRAL, // Alphabetic Presentation Forms
    LANG_ARABIC,  // Arabic Presentation Forms-A
    LANG_NEUTRAL, // Combining Half Marks
    LANG_CHINESE, // CJK Compatibility Forms
    LANG_NEUTRAL, // Small Form Variants
    LANG_ARABIC,  // Arabic Presentation Forms-B
    LANG_NEUTRAL, // Halfwidth and Fullwidth Forms
    LANG_NEUTRAL  // Specials
    );

function TCustomSyntaxEdit.ActivateKeyboard(const C: WideChar): Boolean;

// Tries to activate a keyboard layout based on the code block which contains C.
// If the layout can be activated (if needed then it is loaded) then Result is True
// otherwise False.
// Consider this function as being experimental as the character to language conversion
// does not work very well. Need to investigate this further...

var
  CodeBlock: Cardinal;
  KBID: String;

begin
  CodeBlock := CodeBlockFromChar(C);
  Result := ActivateKeyboardLayout(CodeBlockToLanguage[CodeBlock], 0) <> 0;
  if not Result then
  begin
    KBID := Format('%x', [CodeBlockToLanguage[CodeBlock]]);
    Result := LoadKeyboardLayout(PChar(KBID), KLF_SUBSTITUTE_OK) <> 0;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

initialization
  Screen.Cursors[crDragMove] := LoadCursor(HInstance, 'DRAGMOVE');
  Screen.Cursors[crDragCopy] := LoadCursor(HInstance, 'DRAGCOPY');
finalization
end.
